.TH MESSAGES M
.SH Name
messages \- Description of system console messages.
.SH Description
This section describes the various system messages 
which may appear on the system console.  The messages are 
categorized as follows:
.TP
.I Fatal
Recovery is impossible.
.TP
.I "System inconsistency"
A contradictory situation exists in the kernel.
.TP
.I Abnormal
A probably legitimate but extreme situation exists.
.TP
.I Hardware
Indicates a hardware problem.
.PP
Fatal system messages begin with ``panic:'' and indicate 
hardware problems or kernel inconsistencies that are too 
severe for continued operation.  After displaying a fatal 
message, the system will stop.  Rebooting is required. 
.PP
System inconsistency messages indicate problems usually
traceable to hardware malfunction, such as memory failure.
These messages rarely occur since associated
hardware problems are generally detected before such
an inconsistency can occur.  
.PP
Abnormal messages represent kernel operation problems, 
such as the overflow of critical tables.  It takes extreme 
situations to bring these problems about, so they 
should never occur in normal system use.
.PP
Hardware messages normally specify the device,
.I dev ,
that caused the error.  Each message gives a 
device specification of the form
.I nn/mm
where
.I nn
is the major number of the device, and
.I mm
is its minor number.  The command pipeline
.PP
.RS
ls -l /dev | grep
.I nn
| grep
.I mm
.RE
.PP
may be used to list the name of the device
associated with the given major and minor numbers.
.br
.ne 1i
.SS "System Messages"
.HP 5
** ABNORMAL System Shutdown **
.br
This message appears when errors occur during 
system shutdown.  It is usually accompanied by other
system messages.
.I "System inconsistency, fatal."
.HP 5
bad block on dev \fInn\fP/\fImm\fP
.br
A nonexistent disk block was found on,
or is being inserted in, the structure's free list.
.I System inconsistency.
.HP 5
bad count on dev \fInn\fP/\fImm\fP
.br
A structural inconsistency in the superblock of a file system.  
The system attempts a repair, but this message will probably be 
followed by more complaints about this file system.
.I System inconsistency.
.HP 5
Bad free count on dev \fInn\fP/\fImm\fP
.br
A structural inconsistency in the superblock of a file system.  
The system attempts a repair, but this message will probably 
be followed by more complaints about this file system.
.I System inconsistency.
.\"	.HP 5
.\"	error on dev \fIname\fP (\fInn\fP/\fImm\fP)
.\"	.br
.\"	This is the way that most device driver diagnostic messages start.
.\"	The message will indicate the specific driver and complaint.  The 
.\"	.I name
.\"	is a word identifying the device.
.HP 5
iaddress > 2^24
.br
This indicates an attempted reference to an illegal block number,
one so large that it could only occur on a file system larger 
than 8 billion bytes.
.I "Abnormal."
.HP 5
Inode table overflow
.br
Each open file requires an inode entry to be kept in memory.  When
this table overflows the specific request (usually
.IR open (S)
or
.IR creat (S))
is refused.  Although not fatal to the system, this event may 
damage the operation of various spoolers, daemons, the mailer, 
and other important utilities.  Anomalous results and missing 
data files are a common result.  
.I Abnormal.
.HP 5
interrupt from unknown device, vec=\fId\fP
.sp -1
.HP 5
panic: unknown interrupt
.br
This message indicates that an unknown, unused interrupt (number d)
occured on the 68000. This usually indicates a 68000, or, more rarely, a
Z80 hardware problem. \f2Hardware fatal\f1.
.\"	
.\"	The CPU received an interrupt via a supposedly unused vector.
.\"	This message is followed by ``panic: unknown interrupt.''
.\"	Typically this event comes about when a hardware failure 
.\"	miscomputes the vector of a valid interrupt.
.\"	.I Hardware.
.HP 5
no file
.br
There are too many open files, the system has run out of 
entries in its ``open file'' table.  The warnings given for 
the message ``inode\ table\ overflow'' apply here.
.I Abnormal.
.HP 5
no space on dev \fInn\fP/\fImm\fP
.br
This message means that the specified file system
has run out of free blocks.  Although not normally as 
serious, the warnings discussed for ``inode table overflow'' 
apply: often user programs are written casually and ignore the 
error code returned when they tried to write to the disk; 
this results in missing data and ``holes'' in data files.  
The system administrator should keep close watch on the 
amount of free disk space and take steps to avoid this situation.
.I Abnormal.
.HP 5
**  Normal System Shutdown  **
.sp -1
.HP 5
\&...
.sp -1
.HP 5
[Z80 Control System Halted]
.br
This message appears when the system has been shutdown properly.
It indicates that the machine may now be rebooted or powered down.
.HP 5
Out of inodes on dev \fInn\fP/\fImm\fP
.br
The indicated file system has run out of free inodes.  
The number of inodes available on a file system is 
determined when the file system is created (using
.IR mkfs (C)).
The default number is quite generous, this message 
should be very rare.  The only recourse is to remove some 
worthless files from that file system, or dump the entire 
system to a backup device, run
.IR mkfs (C)
with more inodes specified, and restore the files from backup.
.I Abnormal.
.HP 5
out of text
.br
When programs linked with the 
.I ld
.B \-i 
or 
.B \-n 
switch are run, a table entry is made so that only one copy 
of the pure text will be in memory even if there are multiple 
copies of the program running.  This message appears when this 
table is full.  The system refuses to run the program
which caused the overflow.  Note that there is only one entry 
in this table for each different pure text program.  Multiple 
copies of one program will not require multiple table entries.
Each ``sticky'' program (see 
.IR chmod (C))
requires a permanent entry in this table; nonsticky
pure text programs require an entry only when there is at 
least one copy being executed.
.I Abnormal.
.\"	.TP 5
.\"	panic: bad 287 int
.\"	.br
.\"	Attempted execution of a real mode 287 instruction.
.\"	.I System inconsistency, fatal. 
.HP 5
panic: blkdev
.br
An internal disk I/O request, already verified as valid, is 
discovered to be referring to a nonexistent disk.
.I "System inconsistency, fatal."
.HP 5
panic: devtab
.br
An internal disk I/O request, already verified as valid, is discovered
to be referring to a nonexistent disk.
.I "System inconsistency, fatal."
.HP 5
panic: iinit
.br
The super-block of the root file system could not be read.  This
message occurs only at boot time.
.I "Hardware, fatal."
.HP 5
panic: IO err in swap
.br
A fatal I/O error occurred while reading or writing the swap area.
.I "Hardware, fatal."
.HP 5
panic: memory parity
.br
A hardware memory failure trap has been taken.
.I "Hardware or system inconsistency, fatal."
.TP 5
panic: mmusub: chk
.br
An error occurred during memory management operations.
.I System inconsistency, fatal.
.HP 5
panic: no fs
.br
A file system descriptor has disappeared from its table.
.I "System inconsistency, fatal."
.HP 5
panic: no imt
.br
A mounted file system has disappeared from the mount table.
.I "System inconsistency, fatal."
.HP 5
panic: no procs
.br
Each user is limited in the amount of simultaneous processes 
he can have; an attempt to create a new process when none is 
available or when the user's limit is exceeded is refused.  
That is an occasional event and produces no console messages; 
this panic occurs when the kernel has certified
that a free process table entry is available and 
yet can't find one when it goes to get it.
.I "System inconsistency, fatal."
.HP 5
panic: Out of swap 
.br
There is insufficient space on the swap disk to hold a task.
The system refuses to create tasks when it feels there is 
insufficient disk space, but it is possible to create situations 
to fool this mechanism.
.I Abnormal, fatal.
.\"	.TP 5
.\"	panic: general protection trap
.\"	.br
.\"	General protection trap taken in kernel.
.\"	.I System inconsistency, fatal. 
.\"	.TP 5
.\"	panic: segment not present
.\"	.br
.\"	An attempt has been made to access an invalid
.\"	segment.  It may also indicate the segment-not-present 
.\"	trap has been taken in the kernel.
.\"	.I System inconsistency, fatal.
.HP 5
panic: Timeout table overflow
.br
The timeout table is full.  Timeout requests are generated by device
drivers, there should usually be room for one entry per system serial
line plus ten more for other usages.  
.HP 5
Supervisor trap \f2d\f1
.sp -1
.HP 5
panic: trap in sys
.br
The 68000 CPU has generated a TRAP number \f2d\f1 while executing
kernel or device driver code.  This message is preceded with
an information dump describing the trap.
.I "Hardware or system inconsistency, fatal."
.\"	.TP 5
.\"	panic: Invalid TSS
.\"	.br
.\"	Internal tables have become corrupted.
.\"	.I System inconsistency, fatal.
.\"	.HP 5
.\"	panic: unknown interrupt
.\"	.br
.\"	The CPU received an interrupt via a supposedly unused vector.
.\"	Typically this event comes about when a hardware failure miscomputes
.\"	the vector of a valid interrupt.
.\"	.I "Hardware, fatal."
.HP 5
panic: proc on q
.br
The system attempts to queue a process already on
the ready-to-run process queue. 
.I "System inconsistency, fatal."
.\"	.HP 5
.\"	Trap \fItype\fP 
.\"	.br
.\"	This message precedes a ``panic:'' message.  The 
.\"	.I type
.\"	is the trap number given by the processor.  
.\"	The message is followed by a dump of registers.
.\"	.I System inconsistency, fatal.
.\"	.SH Notes
.\"	Not all messages appear on all machines.  Some messages are
.\"	processor dependent.
.HP 5
\f3Warning: \f1bad configuration, resetting to defaults.
.br
The parameters specified with the "cfg" command require an unreasonable
amount of memory be used for \*(x1, and would leave the available user
memory area too small to be usable. The system ignores table size
parameters and uses the defaults. This message only occurs at boot
time. This message may be caused by running a system with less memory
than it ordinarily has. Abnormal.
.br
.ne 6
.SH "Device Driver Messages"
.TP
{Hard, Floppy, Cartridge} Drive n: <specific message>
These messages indicate a specific error, and include the following:
.RS
.TP
bad format
This means the user tried to read or write an unformatted or foreign
disk.
.TP
Further I/O aborted until device closed
.sp -1
.TP
Device closed; error cleared
This means an error occurred, and the driver refused to do any more I/O
until the program closed the device.
.TP
active disk changed
This message means that the user changed a floppy while \*(x1 was doing
I/O on it.
.TP
drive not ready
.sp -1
.TP
active drive not ready
This means that the user tried to use a drive, and it was not ready.
This can happen, for example, if the user tries to mount a floppy drive
which has no disk in the drive.
.TP
drive write protected
.sp -1
.TP
active drive write protected
The user tried to write to a drive which was write-protected.
.RE
.TP
{Hard, Floppy, Cartridge} Drive N: {hard, soft} error HHHH while
.sp -.8
.TP
\h'5.5P'{reading, writing, formatting} PLACE
An I/O error of some sort occurred. "hard" means the data was lost and
"soft" means the data was written or read correctly.
.sp
PLACE will indicate where the error occured on the media and will
include `Cylinder', `Head' or `Side', and `Sector'; or `Block' numbers.
.sp
`HHHH' will be four digits of status information from the hardware that
may be useful to technicians in case of hardware failure.
