/*
 *  "errd" -  Transcribe /dev/error somewhere
 */

#include  <signal.h>
#include  <errno.h>
#include  <fcntl.h>
#include  <sys/ioctl.h>
#include  <sys/types.h>
#include  <sys/stat.h>
#include  <sys/param.h>
#include  <sys/z80msgbuf.h>

#define NULL	0
#define EOF	(-1)
#define STAMPTIME	(2 * 60)


/*
 *  Our string dictionary
 */
char	ErrorDev[] = "/dev/error";
char	Console[] = "/dev/console";
char	MsgBuf[] = "/usr/adm/msgbuf";

char	version[] = "@(#) errd v1.1(01)\n";
char	copyright[] =
	"Copyright 1988 Tandy Corporation.  All Rights Reserved.\n";


int	fd, lostchars = 0;
long	nowtime, lasttime = 0L;
char	buf[256], lastbuf[256], tmpbuf[60];

char	*myname;
int	foreground = 0;
int	savebuffer = 0;


/*
 *  Misc. Declarations
 */
extern int	errno;
extern char	*sys_errlist[];

extern char	*ctime();
extern long	time();

int		bye();


main(argc, argv)
  int	argc;
  char	**argv;
{
	register char	*sp;
	int		c, n, timeout, repeats, same;

	myname = *argv++; --argc;
	while (argc > 0) 
	{
		int	ok = 1;

		if (argv[0][0] == '-') {
			if (argv[0][1] == 'f')
				++foreground;
			else if (argv[0][1] == 's')
				++savebuffer;
			else
				ok = 0;
			if (!ok || argv[0][2] != '\0')
				goto bad;
		}
		else {
		   bad:
			sprintf(buf, "errd: Unknown argument \"%s\"\n",
							argv[0]);
			write(2, buf, strlen(buf));
		}

		--argc, ++argv;
	}

#ifndef DEBUG
	/* We run in background automagically(tm) */
	if (!foreground)
	{
		int	pid;

		do
		    pid = fork();
		while (pid < 0);
		if (pid > 0)
			exit(0);
	}
#endif /*DEBUG*/

	/* Prepare to listen, Houston... */
	fd = open(ErrorDev, O_RDONLY);
	if (fd < 0) {
		sprintf(buf, "errd: Can't open \"%s\"; %s\n",
				ErrorDev,
				sys_errlist[errno]);
		write(2, buf, strlen(buf));
		exit(1);
	}

	acatch();
	signal(SIGTERM, bye);

	repeats = timeout = 0;
	for ( ; ; ) {
		n = lgets(buf, sizeof (buf), timeout);
		nowtime = time((long *)0);
		same = (strcmp(buf, lastbuf) == 0);
		if (!lostchars && same && nowtime - lasttime < STAMPTIME && n != EOF) {
			++repeats;
			timeout = STAMPTIME - (nowtime - lasttime) - 5;
			if (timeout > 1)
				continue;
		}

		if (repeats > 1)
			sprintf(lastbuf, "  [last message repeats %d times]\n", repeats);
		if (repeats > 0) {
			output(lastbuf);
			lastbuf[0] = '\0';
		}

		if (lostchars) {
			sprintf(tmpbuf, "  [errd lost %d characters]\n", lostchars);
			output(tmpbuf);
			lostchars = 0;
		}

		if (n != EOF)
			strcpy(lastbuf, buf);

		repeats = timeout = 0;

		if (n != EOF)
			output(buf);
	}
	/*NOTREACHED*/
}

char	lbuf[1024];
char	*fp = lbuf, *ep = lbuf;
lgetc(waittime)
  int	waittime;
{
	int	rval;

	if (ep == fp || rdchk(fd) > 0 && fp < &lbuf[sizeof (lbuf)]) {
		if (ep == fp)
			fp = ep = lbuf;
		do {
			if (waittime > 0)
				alarm(waittime);
			rval = read(fd, fp, &lbuf[sizeof (lbuf)] - fp);
			alarm(0);
		} while (rval == 0);

		if (rval < 0)
			return (EOF);
		fp += rval;
	}

	return (*ep++);
}

lgets(inbuf, max, timeout)
  register char	*inbuf;
  register int	max, timeout;
{
	register int	c, nchars;

	--max;
	nchars = 0;
	while (nchars < max && (c = lgetc(timeout)) != EOF) {
		if (c == '\0')
			++lostchars;
		else {
			if (c != '\r') {
				*inbuf++ = (char)c;
				++nchars;
				timeout = 3;
			}
			if (c == '\n')
				break;
		}
	}
	*inbuf = '\0';

	return ((c == EOF)? EOF : nchars);
}

output(buf)
  char	*buf;
{
	register char	*bp;
	char		zeroed, *tptr, outbuf[512];
	static struct stat	sbuf;
	static char		is_reg = 0, beenhere = 0;

	zeroed = 0;
	if (is_reg || !beenhere) {
		fstat(1, &sbuf);
		if (!beenhere) {
			is_reg = ((sbuf.st_mode & S_IFMT) == S_IFREG);
			beenhere = 1;
			if (!is_reg)
				goto skipit;		/* Speed hack */
		}
		if (sbuf.st_size < lseek(1, 0L, 1)) {
			zeroed = 1;
			lseek(1, sbuf.st_size, 0);
		}
	}

   skipit:
	tptr = ctime(&nowtime);
	tptr[19] = '\0';
	tptr += 4;

	outbuf[0] = '\0';
	if (nowtime - lasttime > STAMPTIME || zeroed) {
		sprintf(outbuf, "%s\n%s%s\n", 
				is_reg? "":"\r",
				tptr,
				is_reg? "":"\r");
		lasttime = nowtime;
	}
	lstrcat(outbuf, buf, is_reg);
	write(1, outbuf, strlen(outbuf));

	if (savebuffer)
		do_buffer();
}

lstrcat(buf, s, is_reg)
  register char	*buf, *s;
  register char	is_reg;
{
	while (*buf)
		++buf;
	do {
		if (!is_reg && *s == '\n')
			*buf++ = '\r';
		*buf++ = *s;
	} while (*s++);
}

acatch()
{
	signal(SIGALRM, acatch);
}

bye(code)
  int	code;
{
	close(fd);
	exit(code);
}

do_buffer()
{
	struct zm_iopb  msgbuf;
	static int	mbfd = -1;

	if (mbfd < 0) {
		mbfd = open(MsgBuf, O_WRONLY|O_CREAT, 0644);
		if (mbfd < 0) {
			sprintf(tmpbuf, "errd: Can't open \"%s\"; %s\n",
					MsgBuf,
					sys_errlist[errno]);
			write(2, tmpbuf, strlen(tmpbuf));
		}
	}
	if (mbfd >= 0 && ioctl(fd, EIOBGET, &msgbuf) >= 0) {
		lseek(mbfd, 0L, 0);
		write(mbfd, (char *)&msgbuf, sizeof (msgbuf));
	}
}
