/*
 *  "kinetic.c"  -  demo of /dev/graphics
 *
 *	George Moore 	08/28/1986
 *	Tandy System Software
 */

#include  <stdio.h>
#include  <signal.h>
#include  <sys/ioctl.h>
#include  <sys/graphics.h>

#define	BARS	15
#define	STEP	4
#define	ADD	4
/* #define	RAWGR		/* define for raw graphics mode */

extern int	errno;

char	screen[] = "/dev/graphics";
char	cantopen[] = "Can't open ";

struct screen {
	unsigned char	b[240][80];
};
struct screen	buff[2];

#ifdef	RAWGR
char	cmds[700];
#endif 	RAWGR

int	screen_fd;
char	*myname;

int	quit();

main(argc, argv)
  int	argc;
  char	**argv;
{
	register int	rval;
	int		i, x, xi, xj, yi, yj, x1, y, y1, e = 2;
	int		loop;
	int		a[BARS+1][5];

	myname = *argv;

	screen_fd = open(screen, 2);
	if (screen_fd < 0) {
		write(2, cantopen, sizeof(cantopen));
		perror(screen);
		exit(1);
	}

	errno = 0;

	/* Turn graphics off while we get ready */
	i = GR_ENABLE;
	if (ioctl(screen_fd, GIOCCBIC, &i) < 0) {
		fprintf(stderr, "%s: Can't GIOCCBIC; ", myname);
		perror("");
		exit(1);
	}

	/* Trap for exit */
	signal(SIGINT, quit);

	/* Clear real screen & turn off cursor */
	printf("Rc");
	/* clear graphics screen */
	rval = write(screen_fd, (char *)&buff[0], sizeof (buff[0]));
	if (rval != sizeof (buff[0])) {
		fprintf(stderr, "%s write error: wanted %d, got %d; ", 
					myname,
					sizeof (buff[0]),
					rval);
		perror("");
	}
	/* Set some crunchy options */
	i = (GR_ENABLE|GR_W_ASYNC);
	if (ioctl(screen_fd, GIOCCBIS, &i) < 0) {
		fprintf(stderr, "%s: Can't GIOCCBIS; ", myname);
		perror("");
		exit(1);
	}

	srand(getpid());

	xi = (rand()*STEP)/32767 + ADD;
	xj = (rand()*STEP)/32767 + ADD;
	yi = (rand()*STEP)/32767 + ADD;
	yj = (rand()*STEP)/32767 + ADD;
	x = (rand()*638)/32767;
	x1 = (rand()*638)/32767;
	y = (rand()*238)/32767;
	y1 = (rand()*238)/32767;
	i = 1;

	/* draw the box around the screen */
	line(0, 0, 639, 0, 1);	   /* top */
	line(0, 239, 0, 0, 1);	   /* left side */
	line(639, 0, 639, 239, 1); /* right side */
	line(639, 239, 0, 239, 1); /* bottom */
	refresh();

	for (loop = 0; ; loop++) {
		line(a[e][1], a[e][2], a[e][3], a[e][4], 0);
		line(x, y, x1, y1, 1);

		if (loop % 2)
			refresh();

		a[i][1] = x;
		a[i][2] = y;
		a[i][3] = x1;
		a[i][4] = y1;

		if (++i > BARS)
			i = 1;
		if (++e > BARS)
			e = 1;

		if ((x+xi < 1) || (x+xi > 638))
			xi = (-1 * sgn(xi) * (((rand() * STEP)/32767)+ADD));
		if ((x1+xj < 1) || (x1+xj > 638))
			xj = (-1 * sgn(xj) * (((rand() * STEP)/32767)+ADD));
		if ((y+yi < 1) || (y+yi > 238))
			yi = (-1 * sgn(yi) * (((rand() * STEP)/32767)+ADD));
		if ((y1+yj < 1) || (y1+yj > 238))
			yj = (-1 * sgn(yj) * (((rand() * STEP)/32767)+ADD));
		x += xi;
		y += yi;
		x1 += xj;
		y1 += yj;
	}
}

quit()
{
	int	arg;

	printf("RC");
	arg = (GR_ENABLE|GR_W_ASYNC);
	if (ioctl(screen_fd, GIOCCBIC, &arg) < 0) {
		fprintf(stderr, "%s: Can't GIOCCBIC; ", myname);
		perror("");
		exit(1);
	}
	exit(0);
}

sgn(i)
int	i;
{
	if (i > 0)
		return(1);
	if (i < 0)
		return(-1);
	if (i == 0)
		return(i);
}

line(x0, y0, x1, y1, c)
int	x0, y0, x1, y1, c;
{
	int	dx, dy, xinc, yinc, threshold;

	xinc = yinc = 1;
	dx = x1 - x0;
	dy = y1 - y0;
	if (dx < 0) {
		dx = -dx;
		xinc = -xinc;
	}
	if (dy < 0) {
		dy = -dy;
		yinc = -yinc;
	}
	threshold = 0;
	write_pixel(x0, y0, c);
	if (dx >= dy) {
		while (x0 != x1) {
			x0 += xinc;
			threshold += dy;
			if (threshold > dx) {
				threshold -= dx;
				y0 += yinc;
			}
			write_pixel(x0, y0, c);
		}
	}
	else {
		while (y0 != y1) {
			y0 += yinc;
			threshold += dx;
			if (threshold > dy) {
				threshold -= dy;
				x0 += xinc;
			}
			write_pixel(x0, y0, c);
		}
	}
}

struct screen	*cbp = &buff[0];
short	cbuff = 0;
long	lo = 80 * 240, hi = 0L;
write_pixel(x, y, c)
int	x, y, c;
{
	char	mask;
	long	seek;

	mask = 1 << (7 - (x&7));
	if (c == 1)
		cbp->b[y][x/8] |= mask;
	else if (c == 0)
		cbp->b[y][x/8] &= ~mask;

	seek = (y * 80) + (x / 8);
	if (seek < lo)
		lo = seek;
	if (seek > hi)
		hi = seek;

	/*lseek(screen_fd, seek, 0);*/
	/*write(screen_fd, &cbp->b[y][x], 1);*/
}

refresh()
{
	if (hi - lo > 0) {
		lseek(screen_fd, lo, 0);
		write(screen_fd, &cbp->b[lo/80][lo%80], (int)(hi - lo + 1));
		lo = 80 * 240;
		hi = 0L;
	}

	cbuff = 1 - cbuff;
	copybuf(&buff[cbuff], &buff[1 - cbuff], sizeof (buff[cbuff]));
	cbp = &buff[cbuff];
}

copybuf(to, from, bytes)
  register char	*from, *to;
  register int	bytes;
{
	while (bytes--)
		*to++ = *from++;
}
