/*Itst.c --	A program to test input and output on a DT-100.
	  	Author: Patty Armington

		This program must be run on a DT-100.  It continually
		sends a sequence which asks what the cursor position
		is, and the DT-100 responds.  This program is quite
		a tough test of serial output and can divide the good
		serial drivers from the mediocre ones.  It is a good
		test of how the serial driver handles quick bursts of
		input characters.

		The program appears to simply move the cursor around
		on the screen when it's working properly and will print
		a message when an error occurs.

		If itst is given any argument, it will sound the bell
		when an error occurs.
		
			
 */

#include <stdio.h>
#include <signal.h>

char	buf1[9] = "\33[1;1H", buf2[100];
long	count = 0;
long	avg = 0, acnt = 0;
int	bell;

main(argc,argv)
int argc;
char *argv[];
{
	register int i, j;
	int timout();
	int endjob();

	if (argc > 1)
		bell = 1;
	else
		bell = 0;
	signal(SIGALRM, timout);
	signal(SIGTERM, endjob);
	save_tty();
	set_tty();
	write(1, "\33[H\33[J", 6);
	write(1, "\33[25;1H", 7);
	write(1,"START: ",7);
	system("exec date");
	for (;;) {
		for (i='1'; i<='9'; i++) {
			buf1[2] = i;
			for (j='1'; j<='9'; j++) {
				buf1[4] = j;
				write(1, buf1, 6);
				write(1, "\33[6n", 4);
				count++;
				getansr();
			}
		}
	}
	endjob();
}

getansr()
{
	register char *p = buf1, *q = buf2;
	register int c, done = 0;
	static long	lasttime = 0;

	while (*p) {
		alarm(8);	
		c = read(0, q, 1);
		alarm(0);
		if (c != 1)
			break;
		p++;
		if ((c = *q++) == 'q')
			endjob();
		else if (c == 'R') {
			done = 1;
			break;
		}
	}
	*q = '\0';
	p = buf1;
	q = buf2;
	while (*p) {
		c = *p++;
		if (c == 'H')
			c = 'R';
		if ((*q++ & 0x7F) != c) {
			if (!done)
				eatinput();
			if (bell)
				printf("\7");
			printf("\33[24;1H\n%ld[%ld]: sent ",
						count,
						count - lasttime);
			avg += count - lasttime;
			++acnt;
			lasttime = count;

			fflush(stdout);
			show(buf1);
			write(1, ", got ", 6);
			show(buf2);
			return;
		}
	}
}

timout()
{
	int timeout();
	signal(SIGALRM, timout);
}

show(p)
register char *p;
{
	char cc[2];

	for ( ; *p; p++) {
		*p &= 0x7F;
		if (*p == '\33')
			write(1, "\\E", 3);
		else if (*p < ' ') {
			cc[0] = '^';
			cc[1] = *p + '@';
			write(1, cc, 2);
		} else if (*p == 0x7f)
			write(1, "^?", 2);
		else
			write(1, p, 1);
	}
}

eatinput()
{
	register int c;
	register char *p = buf2;

	while (*p)
		p++;
	do {
		alarm(2);
		c = read(0, p, 1);
		alarm(0);
		if (*p++ == 'R') {
			p++;
			break;
		}
	} while (c == 1);
	*--p = '\0';
}

endjob()
{
	long	average;

	average = (acnt > 0)? avg/acnt : count;
	eatinput();
	printf("\33[24;1H\n%ld: done; avg = %ld; number of failures = %ld\n", 
	count, average, acnt);
	restore_tty();
	exit(0);
}

/*.
*************************************************************************
*									*
* save_tty(), set_tty(), restore_tty() -  Manipulate the terminal	*
*									*
*  No input parameters or return values.				*
*									*
*************************************************************************
*/

#include  <sys/types.h>
#include  <sys/ioctl.h>
#include  <termio.h>
static struct  termio	old, new;

save_tty()
{
	ioctl(fileno(stdin), TCGETA, &old);

	new = old;
	new.c_lflag &= ~(ISIG|ICANON|ECHO);
	new.c_cc[VMIN] = 1;
	new.c_cc[VTIME] = 0;
}

set_tty()
{
	ioctl(fileno(stdin), TCSETA, &new);
}

restore_tty()
{
	ioctl(fileno(stdin), TCSETAF, &old);
}
