/*
 *	Tandy 68000/XENIX disk driver's defines
 */

/* Misc unrelated defines */
#define	RETRIES	1
#define SPL()   spl5() 



/*
 * minor number:
 *		XDDUUPPP
 *		 X  - unused
 *		 DD - device number, where 0 is floppy, 1 is hard, and 2 is cart
 *		 UU - Unit number of the device; for example hard drive 0
 *		PPP - pseudo device partition number (0-4)
 *			0: whole disk, except boot track
 *			1: file system (if any)
 *			2: remainder of disk (potential swap area) (if any)
 *			3: boot track
 *			4: /dev/recover --  a place for fsck to put output
 */

#define	NPSEUDO	5
#define	BOOTPART 3
#define	RCVRPART 4	/* Right now, the recover partition *must* be BOOT+1 */
#define	FLSHIFT	2	/* log (base 2) number of floppies */
#define	NMAJ	3
#define	NDISK	4
#define DRSHIFT	3	/* size of shift to convert dev to/from driveno */
#define	MJSHIFT	2	/* size of shift to convert dev to/from disk maj nr */

/* Device Major numbers */
#define	FLOP	0
#define	HARD	1
#define	CART	2

/* Handy macros to get at useful info */
#define	driveno(dev)	(((dev) >> DRSHIFT) & 15)
#define	pseudo(dev)	((dev) & 7)
#define	dmaj(dn)	(((dn) >> MJSHIFT) & 3)
#define	dmin(dn)	((dn) & 3)
#define	_hard(dmj)	(dmj == HARD)
#define	_flop(dmj)	(dmj == FLOP)
#define	_cart(dmj)	(dmj == CART)
#define	hard(dn)	(_hard(dmaj(dn)))
#define	flop(dn)	(_flop(dmaj(dn)))
#define	cart(dn)	(_cart(dmaj(dn)))
#define	tabindex(dn)	(hard(dn)? 0 : flop(dn)? 1 : 2)



/* Device misc, non I/O commands [function codes for dkcommand()] */
#define	dkINIT	  0		/* Run an INIT/MOUNT command on this drive */
#define	dkUMOUNT  1		/* Run a UMOUNT command on this drive */
#define	dkFORMAT  2		/* Run a FORMAT command on this drive */

#define b_command  b_blkno	/* No `command' field, so use blkno, which
				 *   is unused for disk commands anyway */


/* Misc struct definitions for disk.c and their defines */
struct sizes {
	daddr_t	nsects;		/* number of sectors */
	unsigned short trkoff;	/* track offset for start of minor dev */
	unsigned short nspt;	/* number of sectors per track */
	unsigned short sshift;	/* log (base 2) number byte per sector */
};

struct diskconf {
	unsigned short
		dc_ncyl,	/* invariant description of disk */
		dc_secsiz,	/* doesn't necessarily apply to every track */
		dc_trksiz;	/* doesn't necessarily apply to every track */
	unsigned char
		dc_nhead,	/* invariant description of disk */
		dc_fmt,		/* zero or more of MIXED_FMT, DBL_DENS */
		dc_seek,
		dc_mode;
	struct sizes dc_sizes[NPSEUDO];
};


struct dkflags {
	unsigned char openfl[2];
	unsigned short status;
};
#define	DKOPEN		0x01		/* currently open */
#define	DKREADONLY	0x02		/* currently read-only */
#define	DKERROR		0x04		/* error detected, must be closed */
#define	DKINITTING	0x08		/* init is in progress */
#define	DKWANTED	0x10		/* process waiting for other	*/
					/*	process's init to complete */
#define	DKWARNED	0x20		/* printed out a warning message */
