/***********************************

z80clk.h - interface to the Z80 real time clock

Copyright (C) Microsoft Corporation, 1982

This file must be kept consistent with z80/ctl/68kdefs.h and z80/ctl/bss.s

************************************

In normal operation, the Z80 interrupts the 68k every 1/30 sec and sets
CLKINT in z_clock.  The 68K acknowledges the interrupt by resetting it.

If XENIX does not acknowledge an interrupt before the next 1/30 sec tick
by resetting CLKINT, the ticks are lost.  The Z80 does nothing to insure
that XENIX does not lose clock ticks.  However, it is worth noting that
the Z80 does keep track of unacknowledged time, and if sufficient time
elapses without an acknowledgement, it assumes the 68K has crashed.

The clock hardware is actually the vertical sync from the console screen;
thus the z80 tells the 68000 what the closest frequency (25hz or 30hz)
that it should use for Hz.   Thus HZ is now auto-configured on boot.

Another item worth noting is the fact that, because of parts tolerence
and other factors, the vertical retrace frequency cannot be programmed
for EXACTLY 25 or 30 hz.  Thus in mdep/clkreld() you will note some code
designed to "fudge" the clock.  clkreld now returns the number of ticks
to add to the internal counter in clock() to keep the time accurate.

*******/

struct z80clk {
	unsigned char
		z_clock,	/* clock csr */
		z_count;	/* The acutal clock frequency */
} z80clk;

/* bits in z_clock */
#define CLKINT	0x01	/* z80 sets on clock interrupt, 68k clears to ack */
#define CLKIENAB 0x02	/* 68k sets to enable clk interrupts */

/*
   68k clock intr handler:
	z_clock &= ~CLKINT;
*/
