/*********************************

z80graph.h -- interface to z80 graphics drivers

Copyright (C) Tandy Corporation, 1983.

This file must be kept consistent with z80/ctl/68kdefs.h and z80/ctl/bss.s

*********************************

The z80 graphics driver is really two drivers under one roof.  One
handles direct screen I/O  and the other handles graphics requests,
that is, it allows more-or-less direct I/O to the graphics board.

****/


struct zg_iopb {
  /* these are set by the 68k for each transfer */
	caddr_t	zg_addr;	/* 68k data address, must be first entry */
	unsigned int
		zg_count;	/* Length of data block to be "written" */
	union {
		unsigned int
			offset;	/* address to write data at (for screen) */
		struct {
		    unsigned char
			x,	/* `X' coordinate for graphics board */
			y;	/* `Y' coordinate for graphics board */
		} zg_xy;
	} zg_u;
  /* these are set by 68k and z80 for each cmd */
	unsigned char
		zg_csr,		/* command/status register */
		zg_xcsr,	/* one's complement of zd_csr for safety */
		zg_dev,		/* GRAPHIC or SCREEN I/O select */
		zg_err;		/* error code if zg_csr & ERROR */
	unsigned short
		zg_ctl;		/* Control word for graphics */
};
/* defines for easy access to everything */
#define zg_offset  zg_u.offset
#define zg_x	   zg_u.zg_xy.x
#define zg_y	   zg_u.zg_xy.y

/* defines for zg_csr --
	To start an operation the 68k first writes the one's complement
	of the csr bits to zd_xcsr and then writes the bits to zd_csr.
	The 68k can't touch the csr after it sets GO until the z80 clears GO.
	The z80 can't touch the csr after it clears GO until the 68k sets GO. 
 */
#define	GO	0x01	/* set by 68k to begin xfer, cleared by z80 when done */
#define	INTENAB	0x02	/* set by 68k to enable interrupt on -GO */

#define	INIT	0x00	/* requested by 68k to initialize device */
#define READ	0x04	/* 68k requested read (ONLY valid for graphics) */
#define	WRITE	0x08	/* set by 68k for graphics/screen write */
#define CLOSE	0x0C	/* device close/cleanup command */

#define	INTPEND	0x20	/* set by z80 when interrupting 68k, 68k can clear it
				to acknowledge interrupt */
#define	ERROR	0x40	/* set by z80 if an error occured, this and the error
				code in zg_err are valid from when the z80 
				clears GO to when 68k sets GO */

/* Error codes seen at zg_err -- set only by z80 */
#define ILLCMD	0x01	/* 68k command error */
#define BADPARM	0x02	/* Bad write offset or length given */

/* Device minor numbers seen at zg_dev */
#define	GRAPHIC	0x00	/* The graphics board */
#define	SCREEN	0x01	/* The virtual screen */

/* Defines for control word */
#define ENABLE	0x0001	/* Enables graphics */
#define	WAITS	0x0002	/* Selects video waits on reads and writes */
#define INC_X	0x0000	/* Selects `increment' mode on the `X' register */
#define DEC_X	0x0004	/* Selects `decrement' mode on the `X' register */
#define INC_Y	0x0000	/* Selects `increment' mode on the `Y' register */
#define DEC_Y	0x0008	/* Selects `decrement' mode on the `Y' register */
#define RDCLK_X	0x0010	/* Selects inc/dec mode on `X' AFTER reads */
#define RDCLK_Y	0x0020	/* Selects inc/dec mode on `Y' AFTER reads */
#define WRCLK_X	0x0040	/* Selects inc/dec mode on `X' AFTER writes */
#define WRCLK_Y	0x0080	/* Selects inc/dec mode on `Y' AFTER writes */
#define	AUTOCLK	0x0100	/* Automagically increment `Y' when `X' wraps */
#define	W_ASYNC	0x0200	/* Don't wait for screen to update on writes */
#define	RAW	0x0800	/* X/Y positioning is embedded in each write */
#define	NOT	0x8000	/* Bytes are NOT masks (may be ORed with below) */
#define	AND	0x1000	/* Bytes are AND masks */
#define	OR	0x2000	/* Bytes are OR masks */
#define	XOR	0x3000	/* Bytes are XOR masks */
  /* RD/WRCLK_X/Y are actually active on `0', so we invert them */
#define bmap(wrd)	((wrd) ^ (RDCLK_X|RDCLK_Y|WRCLK_X|WRCLK_Y))

/* Format of an individual request for RAW mode */
struct gr_raw {
	unsigned char
		x, y;			/* New screen position */
	unsigned char
		c;			/* What goes there */
};


/* the actual iopb */
struct zg_iopb z80graph;
