/*
 *  "graphics.c"	Z80 graphics device driver
 *
 *   Copyright 1986 Tandy Corporation
 *
 *	09/10/1983	jeIV 
 *
 * $Log:	graphics.c,v $
 * Revision 6.0  87/10/12  17:04:42  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.8  87/05/06  11:23:14  iv
 * 	- Got rid of unneeded initializtions to `0' (bss is cleared by mch.s)
 * 
 * Revision 5.7  86/11/19  10:34:55  iv
 * 	- Turned EXCLUSIVE back ON, found the bug in XENIX (sys/sys2@copen)
 * 
 * Revision 5.6  86/11/11  18:18:43  iv
 * 	- Improved debug output.
 * 	- Added new EXCLUSIVE conditional & turned it OFF.
 * 	  A kLUDGE around a bug somewhere else which causes grclose()
 * 	  not to be called when doing floppy I/O & hard disk I/O.
 * 	- Removed silly check for BUSY in grclose()
 * 	- Made DEBUG error check in close another conditional: CDEBUG
 * 
 * Revision 5.5  86/10/16  15:22:56  iv
 * 	- Moved splx(s) calls to AFTER we sleep waiting for the mailbox.
 * 	  Turns out splx()ing before we sleep has interrupt holes!
 * 	  Should've known better than that.
 * 	- Added a test of IOASYNC to end of grcomm() so we don't release
 * 	  the mailbox prematurely
 * 	- Removed GO test in g_unlock().  Don't need it now.
 * 
 * Revision 5.4  86/10/01  17:36:11  iv
 * 	- Added new lockout system for the mailbox; this way, processes
 * 	  which need to look at the iopb after the I/O has completed
 * 	  won't get it changed on them by another proc which grabs it.
 * 	- Fixed up some conditional code to be more readable
 * 	- Moved a bunch of #defines and declarations into ../h/graphics.h
 * 	- Added some #ifdef GRAPHICS that were missing
 * 
 * Revision 5.3  86/09/25  15:24:53  iv
 * 	- Fixed bug in ioctls GIOCCBIS and GIOCCBIC that used bmap incorrectly
 * 	- Added support for W_ASYNC; writes don't have to wait for completion
 * 
 * Revision 5.2  86/08/29  15:05:51  iv
 * 	- Rewrote to add support for /dev/graphics
 * 	- read/write now call grcomm & share code 
 * 	- Changed screen code so the screen can be read now
 * 	- Added ioctls on /dev/graphics for manipulation of control word
 * 	- Changed debugging statements to new dbug.h macro package
 * 
 * Revision 5.1  85/12/04  10:01:27  iv
 * 	- Put interrupt fix in (it was possible for the 68000 code to see
 * 	  INTPEND on, and set up the next I/O *before* the Z80 had reset
 * 	  the GO bit, and then all I/O would die, because it reset GO)
 * 
 * Revision 5.0  85/08/23  10:30:18  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.1  85/03/29  10:21:40  iv
 * 	- Moved the initialization of "zp" to BEFORE the first reference!
 * 	- Removed code for the Z80 interrupt ID array; now use iopb & INTPEND
 * 
 * Revision 4.0  85/03/13  14:43:36  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:48:10  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:57:42  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.3  84/12/19  10:27:51  iv
 * 	- Re-wrote debug stuff for sys III
 * 
 * Revision 1.2  84/04/16  16:59:07  iv
 * 	- Added include of "../h/ioctl.h" to pick up ioctl cmd defines
 * 
 * Revision 1.1  84/04/11  14:00:26  iv
 * 	- Initial checkin for 3.0
 */


#define GRAPHICS	/* Compiles in Graphics code */
#define EXCLUSIVE	/* Devices are exclusive (only 1 user at a time) */

/*#define DEBUG	4	/* Compiles in Debugging code */
/*#define CDEBUG  		/* Compiles in device close debugging code */
#ifdef DEBUG
#  define grOC	0x0001			/* Debug Open/Close calls */
#  define grRW	0x0002			/* Debug Read/Write calls */
#endif /*DEBUG*/
#define DEBUGS grOC
#define DbugVar	gr_debug


#include  "../h/param.h"
#include  "../h/conf.h"
#include  "../h/dir.h"
#include  "../h/user.h"
#include  "../h/tty.h"
#include  "../h/ioctl.h"
#include  "../h/z80graph.h"
#include  "../h/graphics.h"
#include  "../h/z80attn.h"
#include  "../h/dbug.h"
#ifdef DEBUG
#  include  "../h/proc.h"
#endif /*DEBUG*/


unsigned char	ginfo[2];	/* Open/Close management flags */
unsigned short	grflags;	/* iopb management flags */



/*ARGSUSED*/
gropen(dev,flag)
  register dev_t  dev;
  int	flag;
{
	register	 int	d, s;
	register struct zg_iopb	*zp;
	register unsigned char	*gip;

	dbug1(grOC,4,"%d: gropen: executing...\n", u.u_procp->p_pid);
/*
 *  Do we have a legal minor number?
 *  And is the control terminal /dev/console?
 */
	d = minor(dev);
	if (!devIsOK(d) || u.u_ttyp != &ser_tty[0]) {
		dbug(grOC,0,"gropen:  Bad minor number, exit(ENXIO)...\n");
		u.u_error = ENXIO;
		return;
	}
	gip = &ginfo[d];
#ifdef EXCLUSIVE
	if ((*gip & (GBUSY|GLOCK)) == (GBUSY|GLOCK)) {
		dbug1(grOC,0,"gropen(%d): Busy and locked, exit(EBUSY)\n", d);
		u.u_error = EBUSY;
		return;
	}
#endif /*EXCLUSIVE*/

	if (*gip & GBUSY)
		return;		/* Don't need to initialize dev again */
	
	/* Default to exclusive use (if compiled in) */
	*gip |= GLOCK|GBUSY;

	zp = &z80graph;
	s = SPL();
	g_lock();		/* Get iopb */

#ifdef GRAPHICS
	/* set up some crunchy default options */
	zp->zg_ctl = bmap(ENABLE|WAITS|AUTOCLK|INC_X|RDCLK_X|WRCLK_X);
#endif /*GRAPHICS*/

	zp->zg_dev = d;
	zp->zg_err = 0;
	zp->zg_xcsr = ~(GO | INIT | INTENAB);
	zp->zg_csr = (GO | INIT | INTENAB);
	z80attn(GRdev, d, RqStart);

	while (zp->zg_csr & GO)
	    sleep((caddr_t)zp, GRPRI);	/* Wait for request to complete */
	splx(s);

	if (zp->zg_csr & ERROR) {
		*gip = 0;		/* Free for other folk to try */
		dbug(grOC,0,"gropen: Z80 reported open error, exit(ENXIO)\n");
		u.u_error = ENXIO;
	} else
		dbug(grOC,5,"gropen: Z80 reported open OK, exit(OK)\n");

	g_unlock();
}


grclose(dev)
  dev_t	dev;
{
	register int	d, s;
	register struct zg_iopb	*zp;

/*
 *  Do we have a legal minor number?
 */
	dbug1(grOC,4,"%d: grclose: executing...\n", u.u_procp->p_pid);

	d = minor(dev);
	if (!devIsOK(d)) {
		dbug(grOC,0,"grclose: Bad minor #; return (ENXIO)...\n");
		u.u_error = ENXIO;
		return;
	}

	zp = &z80graph;

	s = SPL();
	g_lock();		/* Get iopb */
	ginfo[d] &= ~GBUSY;

	zp->zg_dev = d;
	zp->zg_err = 0;
	zp->zg_xcsr = ~(GO | CLOSE | INTENAB);
	zp->zg_csr = (GO | CLOSE | INTENAB);
#ifndef CDEBUG
	grflags |= IOASYNC;		/* Free iopb for us when done */
#endif /*CDEBUG*/
	z80attn(GRdev, d, RqStart);
#ifdef CDEBUG
	while (zp->zg_csr & GO)
		sleep((caddr_t)zp, GRPRI);  /* Wait for request to complete */

	if (zp->zg_csr & ERROR)
		dbug(grOC,0,"grclose: Device reported close I/O error!...\n");

	g_unlock();
#endif /*CDEBUG*/

	splx(s);
}


grread(dev)
  dev_t	dev;
{
	grcomm(minor(dev), 1);
}

grwrite(dev)
  dev_t	dev;
{
	grcomm(minor(dev), 0);
}

grcomm(d, rdflag)
  register int	d;
{
	register struct zg_iopb	*zp;
	register caddr_t	xaddr;
	register int	s;
	register unsigned int	size;
	register unsigned char	cmd;

	dbug1(grRW,4,"grcomm(%s):  executing...\n", rdflag? "READ":"WRITE");
	if (!devIsOK(d)) {
		dbug(grRW,0,"grcomm:  Bad minor number, exit(ENXIO)...\n");
		u.u_error = ENXIO;
		return;
	}

/*
 *  check parameter validity
 */
	xaddr = (u.u_segflg == 1)? u.u_base : umap(u.u_base, u.u_count);
	if (xaddr == (caddr_t)-1) {
		dbug(grRW,0,"grcomm:  Bad user buffer:\n");
		dbug1(grRW,0,"	  u.u_base   = %X,\n", u.u_base);
		dbug2(grRW,0,"	  u.u_offset = %X [%d],\n", 
					 u.u_offset, u.u_offset);
		dbug2(grRW,0,"	  u.u_count  = %X [%d]\n", 
					u.u_count, u.u_count);
		dbug(grRW,0,"exit(EFAULT)...\n");
		u.u_error = EFAULT;
		return;
	} 

	size = (d == SCREEN)? SCREENSIZE : GSCREENSIZE;
	if (u.u_count > size  ||  u.u_offset > size-1) {
		if (!rdflag || u.u_offset > size) {
			dbug(grRW,0,"grcomm:  Bad offset/count:\n");
			dbug1(grRW,0,"	  u.u_base   = %X,\n", u.u_base);
			dbug2(grRW,0,"	  u.u_offset = %X [%d],\n", 
						 u.u_offset, u.u_offset);
			dbug2(grRW,0,"	  u.u_count  = %X [%d]\n", 
						u.u_count, u.u_count);
			dbug(grRW,0,"exit(EINVAL)...\n");
			u.u_error = EINVAL;
			return;
		}

		if (u.u_offset == size)
			return;			/* return (EOF) for reads */
	}

	if (u.u_count == 0)
		return;			/* if count = 0, exit (we're done) */
	

	s = SPL();
	g_lock();		/* Get iopb */

	zp = &z80graph;

	/* Setup count and offset */
	zp->zg_count =
		(u.u_count + u.u_offset > size)? size - u.u_offset : u.u_count;

#ifdef GRAPHICS
	if (d != SCREEN) {
		zp->zg_x = u.u_offset % GSCREENWID;
		zp->zg_y = u.u_offset / GSCREENWID;
	}
	else
#endif /*GRAPHICS*/
		zp->zg_offset = u.u_offset;

	dbug(grRW,4,"grcomm:  Initiating request...\n");
	zp->zg_dev = d;
	zp->zg_addr = xaddr;

	cmd = (GO | INTENAB) | (rdflag? READ : WRITE);
	zp->zg_xcsr = ~cmd;
	zp->zg_csr = cmd;
	z80attn(GRdev, d, RqStart);

#ifdef GRAPHICS
	if (d == GRAPHIC && !rdflag && (zp->zg_ctl & W_ASYNC))
		grflags |= IOASYNC;
	else
#endif /*GRAPHICS*/
	{
		while (zp->zg_csr & GO)
		    sleep((caddr_t)zp, GRPRI);/* Wait for request to complete */
	
		if (zp->zg_csr & ERROR) {
		    dbug(grRW,0,"grcomm:  Device reported parameter error:\n");
		    dbug1(grRW,0,"	  u.u_base   = %X,\n", u.u_base);
		    dbug1(grRW,0,"	umap(u.u_base) = %X,\n",
						umap(u.u_base, u.u_count));
		    dbug2(grRW,0,"	  u.u_offset = %X [%d],\n",
						u.u_offset,
						u.u_offset);
		    dbug2(grRW,0,"	  u.u_count  = %X [%d]\n",
						u.u_count,
						u.u_count);
		    dbug(grRW,0,"exit(EINVAL)...\n");
		    u.u_error = EINVAL;
		}
	}

	if (u.u_error == 0) {
		u.u_count -= zp->zg_count;
		u.u_offset += zp->zg_count;
	}
	if (!(grflags & IOASYNC))
		g_unlock();

	splx(s);
}


/*ARGSUSED*/
grioctl(dev, cmd, addr, flag)
  dev_t	dev;
  caddr_t addr;
{
	register unsigned char	*gip;
	register struct zg_iopb	*zp;
	register int	c;
	int	d;

	d = minor(dev);

	if (!devIsOK(d)) {
		u.u_error = ENXIO;
		return;
	}

	gip = &ginfo[d];
	zp = &z80graph;

	switch (cmd) {
	    case TIOCNXCL:
		*gip &= ~GLOCK;
		break;

	    case TIOCEXCL:
		*gip |= GLOCK;
		break;

#ifdef GRAPHICS
	    case GIOCSETC:
		c = fuword(addr);
		if (c < 0)
			u.u_error = EFAULT;
		else {
			zp->zg_ctl = bmap((unsigned short)c);
			z80attn(GRdev, d, RqSTwrite);
		}
		break;

	    case GIOCGETC:
	    {				/* MUST be same type as zg_ctl */
		unsigned short	c;	/* and can't be a register */

		c = bmap(zp->zg_ctl);
		if (copyout((caddr_t)&c, addr, sizeof (c)))
			u.u_error = EFAULT;
		break;
	    }

	    case GIOCCBIS:
	    case GIOCCBIC:
		c = fuword(addr);
		if (c < 0)
			u.u_error = EFAULT;
		else 
		{
			register unsigned short	val;

			val = bmap(zp->zg_ctl);
			if (cmd == GIOCCBIS)
				val |= (unsigned short)c;
			else
				val &= ~((unsigned short)c);
			zp->zg_ctl = bmap(val);
			z80attn(GRdev, d, RqSTwrite);
		}
		break;
#endif /*GRAPHICS*/

	    default:
		u.u_error = EINVAL;
	}
}


grintr()
{
	register struct zg_iopb	*zp;

	zp = &z80graph;
	if ((zp->zg_csr & (GO|INTPEND)) == INTPEND) {
		zp->zg_csr &= ~INTPEND;
#ifdef GRAPHICS
		if (grflags & IOASYNC)
			g_unlock();
		else
#endif /*GRAPHICS*/
			wakeup(zp);
	}
}

g_lock()
{
	while (grflags & IOBUSY) {
		grflags |= IOWANTED;
		  /* Wait for iopb to become free */
		sleep((caddr_t)&grflags, GRPRI);
	}
	grflags &= ~IOWANTED;
	grflags |= IOBUSY;
}

g_unlock()
{
	if (grflags & IOWANTED)
		wakeup(&grflags);
	grflags &= ~(IOBUSY|IOWANTED|IOASYNC);
}
