/*
 *  "acu.c"	Dialout device pseudo-driver
 *
 *     Copyright 1984 Tandy Corporation
 *
 *	05/02/1983	jeIV 
 *
 * Modification log:
 *   $Log:	acu.c,v $
 * Revision 6.1  87/12/29  13:40:48  iv
 * 	- Fixed dialout code.  Was anding off speed bits.
 * 	  Now save speed bits in `isacu' to be compatible with patch.
 * 
 * Revision 6.0  87/10/12  17:04:28  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.3  87/05/06  11:19:19  iv
 * 	- Made `_aculbolt' visible externally (for qps & etc)
 * 	- Got rid of unused extern decl for `lbolt'
 * 
 * Revision 5.2  86/10/01  17:17:35  iv
 * 	- Re-arranged some code to remove duplicated effort in code paths
 * 	  (now do it earlier, before logic forks)
 * 	- Changed ACU logic in serial.c to not need ACUINIT bit, which
 * 	  was stolen for other immortal porpoises.
 * 
 * Revision 5.1  86/05/19  21:52:26  iv
 * 	- Fixed multiple defined errors from ../h/serial.h
 * 
 * Revision 5.0  85/08/23  10:29:31  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.2  85/08/22  16:26:26  iv
 * 	- Got rid of some stupid calls to modem(OFF) where I was blowing
 * 	  off the open (which meant that the legitimate user of the line
 * 	  would get blown off the line when we told the modem to hang up!)
 * 	- Got rid of serinit() in serial.c and added a flag to the status
 * 	  structure which tells open that all we want is the line set up
 * 	  for I/O. 
 * 	- Changed drain() call to grab the queue pointer and use that
 * 	- Added initialization of VMIN/VTIME to RAW mode code
 * 
 * Revision 4.1  85/03/24  13:50:22  iv
 * 	- Condensed queue-drain code into new routine "acudrain()"
 * 	- Added set-word-length code so we don't try to dial with
 * 	  5-bit characters (this doesn't work too well on some modems)
 * 
 * Revision 4.0  85/03/13  14:43:03  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.2  85/02/18  10:41:51  iv
 * 	- Replaced status array "acustatus[]" with new array "serstat[]"
 * 
 * Revision 3.1  85/01/22  15:25:11  iv
 * 	- Added a 1 second delay after turning DTR off in reset_modem()
 * 	  so that the modem has a chance to notice we turned it off
 * 	- Increased the delay after turning DTR back ON in reset_modem()
 * 	  to 2 seconds to allow the modem more time to unscramble
 * 	  its alleged brains
 * 	- Removed the output of the NULL in reset_modem(), as it was
 * 	  unnecessary and it caused PCI channels to hang
 * 	- Changed the delay in acu_puts to 2 seconds
 * 
 * Revision 3.0  85/01/17  15:47:51  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.1  85/01/14  18:15:43  iv
 * 	- **sigh**  DEBUG got left turned on in the 3.0(22) release.
 * 	  Turned it back OFF.  *SIGH*
 * 
 * Revision 2.0  85/01/10  13:57:24  iv
 * 	- Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.4  84/12/18  16:41:57  iv
 * 	- New ACU driver for system III
 */

/* #define DEBUG   		/* Compiles debugging code IN */

#ifdef DEBUG
   char  acuDbgLvl = 2;
#  define  dOK(n)	((acuDbgLvl > 0 && n < 2) || n == acuDbgLvl)
#  define  debug(n,s)		if dOK(n)   printf(s)
#  define  debug1(n,s,a)	if dOK(n)   printf(s,a)
#  define  debug2(n,s,a,b)	if dOK(n)   printf(s,a,b)
#  define  debug3(n,s,a,b,c)	if dOK(n)   printf(s,a,b,c)
#  define  debug4(n,s,a,b,c,d)	if dOK(n)   printf(s,a,b,c,d)
#  define  debug5(n,s,a,b,c,d,e) if dOK(n)  printf(s,a,b,c,d,e)
#  define  debug6(n,s,a,b,c,d,e,f) if dOK(n)  printf(s,a,b,c,d,e,f)
#else /*DEBUG*/
#  define  debug(n,x)
#  define  debug1(n,s,x)
#  define  debug2(n,s,a,b)
#  define  debug3(n,s,a,b,c)
#  define  debug4(n,s,a,b,c,d)
#  define  debug5(n,s,a,b,c,d,e)
#  define  debug6(n,s,a,b,c,d,e,f)
#endif /*DEBUG*/

/*
 *  Global variables & stuff goes here
 */
#include  "../h/param.h"
#include  "../h/conf.h"
#include  "../h/dir.h"
#include  "../h/user.h"
#include  "../h/tty.h"
#include  "../h/z80ser.h"
#include  "../h/ioctl.h"
#include  "../h/acu.h"
#include  "../h/serial.h"

#define	DLSPEED		(HZ / 4)
#define serstart(tp)	serproc(tp, T_OUTPUT)

	 char	aculbolt;
static   char	acuioerror;
static	 short	acu_speed;

extern int	Hz;
extern unsigned char serstat[];
int		wakeup();



/*ARGSUSED*/
acuopen(d, flag)
  register int	d;
  int	flag;
{
	register struct	tty	*tp;
	register unsigned char	*ap;
	register int		isacu;

	extern	 struct tty  ser_tty[];

#ifdef DEBUG
	static	short	firsttime = 1;

	if (firsttime) {
		firsttime = 0;
		debug1(0,"\nacuopen:  acuDbgLvl is at 0x%X\n", &acuDbgLvl);
	}
#endif DEBUG

	isacu = (d & ~ACUMINR);
	d &= ACUMINR;			/* Mask out device # */
	if (d >= NZ80SER) {
		debug1(0,"acuACUopen:  d = %d, Taking ENXIO exit...\n", d);
		u.u_error = ENXIO;
		return;
	}

	tp = &ser_tty[d];
	ap = &serstat[d];
	debug6(4,"acu%sopen(%d):  serial line is %s%s; ACU is %s and %s\n",
				isacu? "ACU":"LINE",
				(d & ACUMINR),
				(tp->t_state & ISOPEN)? "OPEN" : "CLOSED",
				(*ap & LINEOPEN)? " (by us)" : "",
				(*ap & ACUOPEN)? "OPEN" : "CLOSED",
				(*ap & ONLINE)? "ONLINE" : "OFFLINE");
	if ( ((tp->t_state & ISOPEN) && !(*ap & LINEOPEN))
	     || (*ap & ONLINE)
	     || isacu && (*ap & ACUOPEN) )
	{
		u.u_error = EBUSY;
		return;
	}

/*
 *  *****************************************************
 *  *	Do a LINE open					*
 *  *****************************************************
 */
	if (!isacu) {			/* are they opening the line? */
		debug(2,"acuLINEopen called\n");
		debug3(4,"acuLINEopen:  d=%d, ACU is %s and %s\n", d,
			(*ap & ACUOPEN)? "OPEN" : "CLOSED",
			(*ap & ONLINE)? "ONLINE" : "OFFLINE");

		debug(4,"acuLINEopen:  calling seropen()...\n");
		/*
		 *  What's going on here with 'flag' is that cu(1) [v7]
		 *  opens the line TWICE, once which sleeps and then alarms
		 *  and the second time for real.  The only way to tell the
		 *  two apart was they used `flag' differently...
		 */
		if (flag)
		    *ap |= (ACURQST|LINEOPEN);  /* Remember WE opened line */
		else
		    *ap = (*ap & ~LINEOPEN) | (ONLINE|ACURQST);
		seropen(d, flag);		/* sleep waiting for carrier */
		debug(4,"acuLINEopen:  returned from seropen()...\n");

		if (u.u_error) {
			*ap &= ~(LINEOPEN|ONLINE);
			debug1(0,"acuLINEopen:  seropen returned error %d!\n",
								u.u_error);
		}
		debug(4,"acuLINEopen:  returning...\n");
		return;
	}

/*
 *  *****************************************************
 *  *	Do an open on the ACU  (hey, that's us!)	*
 *  *****************************************************
 */
	debug(2,"acuACUopen called\n");
	acu_speed = ASPEED(isacu);	/* Grab the speed code */
	switch (acu_speed) {
	    case AS_AUTO:
		debug(1, "acuACUopen:  Automagic speed sense selected...\n");
		acu_speed = 0;		/* A speed of 0 means auto */
		break;
	    case AS_300:
		debug(1, "acuACUopen:  300 baud selected...\n");
		acu_speed = B300;	/* Dial out at 300 baud */
		break;
	    case AS_1200:
		debug(1, "acuACUopen:  1200 baud selected...\n");
		acu_speed = B1200;	/* Dial out at 1200 baud */
		break;
	    case AS_UNDEF:
	    default:
		debug(0, "acuACUopen:  Bad speed code; exit(ENXIO)...\n");
		u.u_error = ENXIO;
		return;
	}

/*
 *  Now we let seropen() know that the dialer is using the line 
 *  and even superusers may not interfere...
 */
	*ap |= (ACURQST|ACUOPEN);
	seropen(d, 0);			/* initialize serial information */
	acuioerror = FALSE;

	reset_modem(d);

	debug(0,"acuACUopen:  Taking normal exit...\n");
}

reset_modem(d)
  register int	d;
{
	register struct	tty	*tp = &ser_tty[d];

	modem(d, MODEMOFF);
	  /* Wait for disable to take effect */
	timeout(wakeup, (caddr_t)&aculbolt, HZ);
	sleep((caddr_t)&aculbolt, TTOPRI);

	modem(d, MODEMON);
	  /* Give modem a chance to unscramble its brains */
	timeout(wakeup, (caddr_t)&aculbolt, HZ * 2);
	sleep(&aculbolt, TTOPRI);
}


acuclose(d)
  register int	d;
{
	register u_char	*sp;

	sp = &serstat[d & ACUMINR];

	if (d & ISACU) {
		debug(4,"acuACUclose:  closing acu...\n");
		*sp &= ~ACUOPEN;
	}
	else {
		debug(4,"acuLINEclose:  closing line...\n");
		*sp &= ~(ONLINE | ACURQST | LINEOPEN);
		serioctl(d, TIOCNXCL, (caddr_t)0, 0);
		serclose(d, 0);
	}
}


acuread(d)
  register int	d;
{
	if (d & ISACU)
		u.u_error = ENXIO;
	else
		serread(d);
}


acuwrite(d)
  register int	d;
{
	if (!(d & ISACU)) {
		serwrite(d);
		return;
	}
    {
	static	 char	numberbuf[ACUMAXNR];
	register char	*ptr = numberbuf;
	register struct tty  *tp;
	register int	numberlen = u.u_count;
	register short	speed = acu_speed;
	register int	c;

	debug(2,"acuACUwrite called\n");
	d &= ACUMINR;
	tp = &ser_tty[d];

	debug1(4,"acuACUwrite:  u.u_count = %d\n", u.u_count);
	if (u.u_count < 1  ||  u.u_count > ACUMAXNR) {
	debug(0,"acuACUwrite:  Bad u.u_count, taking EINVAL exit...\n");
		u.u_error = EINVAL;
		modem(d, MODEMOFF);
		return;
	}

	ptr = numberbuf;
	while((c = cpass()) != -1)
		*ptr++ = c;
	*ptr = '\0';
	debug1(4,"acuACUwrite:  numberbuf = '%s'\n", numberbuf);

/*
 *  Examine the presented string
 */
	ptr = numberbuf;
	do {
		if (*ptr == 'e')
			*ptr = '<';
	} while(*ptr != '<'  &&  ++ptr - numberbuf <= numberlen);

/*
 *  If the presented string did not end properly, blow 'em off with an error
 */
	if (*ptr != '<') {
		debug1(4,"acuACUwrite: Number does NOT end with '<', ptr = '%s'\n", ptr);
		debug(2,"acuACUwrite:  Number didn't end with '<', taking EINVAL exit...\n");
		u.u_error = EINVAL;
		modem(d, MODEMOFF);
		return;
	} 

	/*
	 *  Now we are ready...
	 *  If we're in automagic mode or in 300 baud mode, the first
	 *  thing we need to do after resetting is to check to see if
	 *  the modem is ALREADY in programming mode by outputting a 'D'.
	 */
	if (speed != B1200) {
		debug(4,"\nacuACUwrite:  About to check for modem II...\n");

		debug(4,"acuACUwrite:  Outputting a 'D'\n");
		acu_puts(B300, "D", tp);

		debug(4,"acuACUwrite:  Checking queue for a 'D'...\n");
		debug1(4,"              [number of chars in rawq = %d]\n",
							tp->t_rawq.c_cc);

	/*
	 *  The ONLY modem which will echo back a 'D' after being reset is the
	 *  most totally evil modem in the galaxy, the Modem II.  The fact that
	 *  we get a 'D' echoed back at us means that the Modem II is ALREADY
	 *  in programming mode, and that we shouldn't try to init it with "*."
	 */
		if ((c = getc(&tp->t_rawq) & 0x7f) == 'D') {
			debug(1,"acuACUwrite:  Modem speed is 300 baud\n");
			debug(2,"\nDialing... ");
			acuputc('O', tp);  /* put Modem II in Originate mode */
			goto start2;
		}
	}

	/*
	 *  Now we know we are not talking to a Modem II.
	 *  We see if we can wake up the modem with a "*." to put it in
	 *  programming mode. [if it echos a '*' back at us, we're set]
	 */

	if (speed != B300) {
		debug(4,"\nacuACUwrite:  trying at 1200 baud\n");
		if (acu_wakeup(B1200, d)) {
			debug(1,"acuACUwrite:  Modem speed is 1200 baud\n");
			goto start;
		}
	}

	if (speed != B1200) {
		debug(4,"\nacuACUwrite:  trying at 300 baud\n");
		if (acu_wakeup(B300, d)) {
			debug(1,"acuACUwrite:  Modem speed is 300 baud\n");
			goto start;
		}
	}

	/*
	 *  At this point, we've failed.
	 *  We couldn't wake up the modem at the speed(s) requested.
	 */
	debug(0,"\nacuACUwrite:  Couldn't wake up modem, taking EIO exit...\n");
	u.u_error = EIO;
	modem(d, MODEMOFF);
	return;

/* ========================================================================= */

   start:
	debug(2,"\nDialing... ");
   start2:
	acudrain(tp);

	acuputc('D', tp);		/* tell it to <D>ial */
	acudrain(tp);

	acuputc('T', tp);		/*  in <T>one mode */

	ptr = numberbuf;
	do {
		switch (*ptr) {
		   case '-':
		   case 'w':
		   case '=':
			acuputc('P', tp);
			acuputc('P', tp);
			break;

		   case '<':
			acuputc('X', tp);
			break;
		   
		   default:
			acuputc((unsigned char)*ptr, tp);
			break;
		}

	} while (*ptr++ != '<');
	debug(4,"acuwrite:  End of sequence, exit\n");
	if (u.u_error)
		modem(d, MODEMOFF);
    }
}


acudrain(tp)
  register struct tty	*tp;
{
	register struct	clist	*qp;

	qp = &tp->t_rawq;
	debug(4,"acudrain:  Draining queue...\n");
	debug1(4,"          [number of chars in rawq = %d]\n", qp->c_cc);
	while (qp->c_cc > 0)
		getc(qp);
}


acu_wakeup(speed, d)
  register short  speed;
  register int	d;
{
	register struct tty	*tp;
	register unsigned char	c;

	tp = &ser_tty[d];

	debug(4,"acu_wakeup:  resetting modem\n");
	reset_modem(d);

	debug(4,"acu_wakeup:  outputting \"*.\"\n");
	acu_puts(speed, "*.", tp);

	debug(4,"acuACUwrite:  Checking queue for a '*'...\n");
	debug1(4,"              [number of chars in rawq = %d]\n",
							tp->t_rawq.c_cc);

	if ((c = getc(&tp->t_rawq) & 0x7f) != '*') {
		debug(4,"acu_wakeup:  '*' was not echoed back! ");
		debug2(4,"             [c = '%c'(0x%x)]\n",(c < ' ')?'.':c, c);
	}

	return (c == '*');
}


acu_puts(speed, s, tp)
   register short	speed;
   register char	*s;
   register struct tty	*tp;
{
	register  int	d = tp - ser_tty;

	debug(4,"acu_puts:  Called to output");
	debug2(4," string \"%s\" on channel %d\n", s, d);

	tp->t_cflag = (tp->t_cflag & ~(CBAUD | CSIZE)) | (speed | CS8);
	/*
	 *   Believe it or not, the following mess puts us into RAW mode...
	 */
	tp->t_oflag &= ~(OPOST|ONLCR);
	tp->t_iflag &= ~(ICRNL|ISTRIP|IXON);
	tp->t_lflag &= ~(ECHO|ICANON|ISIG);
	tp->t_cc[VTIME] = 0;
	tp->t_cc[VMIN] = 1;
	debug(4,"acu_puts:  Calling serparam()...\n");
	serparam(d);
	timeout(wakeup, (caddr_t)&aculbolt, DLSPEED);
	sleep(&aculbolt, TTOPRI);

	acudrain(tp);
	while (*s)
		putc(*s++, &tp->t_outq);

	debug(4,"acu_puts:  Calling serstart()...\n");
	serstart(tp);
	debug(4,"acu_puts:  Going to sleep...\n");
	timeout(wakeup, (caddr_t)&aculbolt, (HZ * 2));
	sleep(&aculbolt, TTIPRI);
}


acuputc(c, tp)
   register unsigned char	c;
   register struct tty		*tp;
{
	register unsigned char	d;

	if (acuioerror)
		return;

	debug1(4,"acuputc:  dialing a '%c'...\n", c);
	debug2(2,"%c%c", c, (c == 'X')? '\n' : '\0');
	putc((char)c, &tp->t_outq);
	serstart(tp);
	timeout(wakeup, (caddr_t)&aculbolt, DLSPEED);
	sleep(&aculbolt, TTIPRI);
	debug1(4,"          [number of chars in rawq = %d]\n", tp->t_rawq.c_cc);
	/*
	 *  ***  WARNING  YUCK! GROSS!  KLUDGE AROUND COMPLER BUG!!  WARNING ***
	 *  \vvvvvvvvvvvvvvvvvvv/
	 */
	d = (unsigned char)((int)getc(&tp->t_rawq) & 0x7f);
	debug2(4,"acuputc:  Char recieved was '%c' (0x%x)\n",
							(d < ' ')? '.' : d, d);
	if (d != c) {
		debug2(4,"acuputc:  Char NOT echoed back!  [d = '%c'(0x%x)]\n",
							(d < ' ')? '.' : d, d);
		debug(2,"acuputc:  Character verify error, taking EIO exit.\n");
		u.u_error = EIO;
		acuioerror = TRUE;
	}
}



acuioctl(d, cmd, addr, flag)
  register int	d;
  caddr_t addr;
{
	if (d & ISACU) {
		debug(4,"acuACUioctl:  executing...\n");
		u.u_error = ENOTTY;
	}
	else {
		debug(4,"acuLINEioctl:  executing...\n");
		serioctl(d, cmd, addr, flag);
	}
}
