/*
 *	Tandy 68000/XENIX virtual console driver
 *
 * $Header: /usr/sys/io/RCS/console.c,v 1.2 88/08/17 14:52:04 root Exp $
 */

/*#define DEBUG	1		/* Enable console debugging messages */
#define DbugVar conDbg
#ifdef DEBUG
#  define cnIN		0x0001		/* Debug Init code */
#  define cnOC		0x0002		/* Debug Open/Close code */
#  define cnPR		0x0004		/* Debug PRoc code */
#  define cnRW		0x0008		/* Debug Read/Write code */
#  define cnIO		0x0010		/* Debug IOctl code */
#  define cnBL		0x0020		/* Debug blit code */
#  define vrIN		0x0100		/* Debug virtual input code */
#  define vrOU		0x0200		/* Debug virtual output code */
#endif /*DEBUG*/
#define DEBUGS		(cnIN|cnBL|cnIO|vrIN|vrOU)

#ifdef DEBUGC
#  define copybuf(from, to, len)  copytrc(__LINE__, from, to, len)
#endif /*DEBUGC*/

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/conf.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/tty.h"
#include "../h/proc.h"
#include "../h/file.h"
#include "../h/mmu.h"
#include "../h/z80graph.h"
#include "../h/z80attn.h"
#include "../h/graphics.h"
#include "../h/console.h"
#include "../h/dbug.h"

short 		conDefs = C_SCRNSVR|C_CURBLOK|C_CURBLNK|C_CURVIS;
short		nscr = 4;
struct tty	*con_tty;		/* Array of all tty structs */
struct vscr	*con_scr,		/* Array of all virtual screens */
		*confocus;		/* Pointer to current input focus */

extern unsigned char	ginfo[];
extern struct tty	ser_tty[];
extern int	(*servinp)();
extern char	*umem;

int	conproc(), vinput(), vscan(), vputc();


coninit()
{
	register struct vscr	*vp;
	register struct tty	*tp;
	register int	size;

	if (nscr < 0)
		nscr = 0;			/* Gotta have at least 0! */
	size = nscr * CONSIZE;

	printf("[con driver v1.0(03).  %d screen%s ",
				nscr,
				(nscr == 1)? "" : "s");
	if (nscr > 0)
		printf("(using %dk) ", ((size+(MMPGSZ-1)) & ~(MMPGSZ-1))/1024);
	printf("available.]\n");
	if (nscr < 1) {
		dbug(cnIN,1,"coninit: 0 screens configured.  returning.\n");
		return;
	}

	dbug1(cnIN,4, "coninit: _umem = 0x%X\n", umem);
	servinp = vinput;	/* Tell the serial driver where to send input */

	con_tty = (struct tty *)umem;
	con_scr = (struct vscr *)&con_tty[nscr];
	dbug2(cnIN,5, "coninit: con_tty is at 0x%X, con_scr is at 0x%X\n",
					con_tty,
					con_scr);

	/* Round umem up to next page */
	umem = (char *)(((unsigned)&con_scr[nscr] + (MMPGSZ-1)) & ~(MMPGSZ-1));
	dbug1(cnIN,3,"coninit: new _umem is at 0x%X\n", umem);

	dbug(cnIN,3,"coninit: Initializing screen(s)...");
	clear((caddr_t)con_tty, size);
	for (vp = con_scr, tp = con_tty; vp < &con_scr[nscr]; ++vp, ++tp) {
		dbug1(cnIN,4," %d", vp - con_scr);
		tp->t_proc = conproc;
		tp->t_state |= CARR_ON;

#ifdef SCRNSAVE
		vp->savetime = (uchar)('@' + 10);	/* 10-minute default */
#endif /*SCRNSAVE*/
		vp->flags = conDefs;
		vp->disp = vputc;
		vp->input = vscan;

		vputc(vp, CTRL('L'));	/* Start out with a blank screen */
	}
	dbug(cnIN,3,"\n");

	/* Fake out /dev/screen driver */
	ginfo[SCREEN] = GBUSY|GLOCK;
#if 0
	tp = u.u_ttyp;
	u.u_ttyp = &ser_tty[0];
	dbug(cnIN,6, "coninit: gropen(SCREEN, 1)\n");
	gropen(SCREEN, 1);		/* Grab /dev/screen for ourselves */
	u.u_ttyp = tp;

	if (u.u_error != 0)
		dbug1(cnIN,0,"coninit: error ret %d from gropen\n", u.u_error);
#endif /*0*/
}

/*ARGSUSED*/
conopen(d, flag)
  int  d;
  int  flag;
{
	register struct vscr	*vp;
	register struct tty	*tp;
	int	x;

	d = minor(d);
	if (d >= nscr) {
		dbug1(cnOC,0,"conopen: bad chan %d; return (ENXIO)\n", d);
		u.u_error = ENXIO;
		return;
	}

	dbug1(cnOC,2,"*** conopen(%d)\n", d);

	vp = &con_scr[d];
	tp = &con_tty[d];

	if ((tp->t_lflag & XCLUDE) && ((tp->t_state & ISOPEN) || !suser())) {
		dbug2(cnOC,1,"conopen: XCLUDE set/ uid = %d, tp->t_state = %s\n", u.u_uid, (tp->t_state & ISOPEN)? "OPEN":"CLOSED");
		u.u_error = EBUSY;
		return;
	}
	
	x = SPL();

	if (!(tp->t_state & ISOPEN)) {
		/*
		 *  Special hack because console has no "del" key,
		 *  and the "break" key sends a control-C; and it is
		 *  nice to enable cannonical processing CRT-style
		 */
		dbug(cnOC,3,"conopen: Initializing tty defaults...\n");
		vp->disp = vputc;
		vp->input = vscan;
		ttinit(tp);
		tp->t_cc[VINTR] = '\03';
		tp->t_oflag = OPOST|ONLCR;
		tp->t_iflag = ICRNL|ISTRIP|IXON|IXANY;
		tp->t_lflag = ECHO|ECHOE|ICANON|ISIG;
	}

	(*linesw[tp->t_line].l_open)(tp);
	if (u.u_error != 0)
		dbug1(cnOC,0,"conopen: error ret %d from ttopen\n", u.u_error);

	/* If the input focus hasn't been set yet,
	 *   make this one the current focus */
	if (confocus == (struct vscr *)0)
		conswtch(confocus, vp);

	splx(x);
}

/*ARGSUSED*/
conclose(d, flag)
  int	 d;
  int  flag;
{
	register struct tty	*tp;

	dbug1(cnOC,2,"*** conclose(%d)\n", d);
	tp = &con_tty[d];
	tp->t_lflag &= ~XCLUDE;		/* ttclose should be doing this */
	(*linesw[tp->t_line].l_close)(tp);

	/* Look to see if there's any screens left that are open */
	for (tp = &con_tty[0]; tp < &con_tty[nscr]; ++tp) {
		if (tp->t_state & ISOPEN)
			break;
	}

	if (tp >= &con_tty[nscr])  {
		/* Oh, well; no screens are left open,
		 *  so send input to /dev/console */
		confocus = (struct vscr *)0;
	}
}

conread(d)
{
	struct tty *tp;
	
	tp = &con_tty[d];
	dbug1(cnRW,8,"*** conread(%d)\n", d);
	(*linesw[tp->t_line].l_read)(tp);
}

conwrite(d)
{
	struct tty *tp;

	tp = &con_tty[d];
	dbug1(cnRW,8,"*** conwrite(%d)\n", d);
	(*linesw[tp->t_line].l_write)(tp);
}

/*ARGSUSED*/
conioctl(d, cmd, addr, flag)
{
	register struct tty *tp;
	register status;

	tp = &con_tty[d];
	dbug2(cnIO,2,"*** conioctl(%d, 0x%x)\n", d, cmd);
	status = ttiocom(tp, cmd, addr, flag);
	if (status != 1 && status != 0) {
		dbug(cnIO,0,"conioctl: bad sts; return (ENOTTY)\n");
		u.u_error = ENOTTY;
	}
}

extern int	ttrstrt();
conproc(tp, cmd)
  register struct tty *tp;
{
	register int	nch, s, d;
	register struct vscr	*vp;
	register struct tty	*outq;

	d = tp - con_tty;
	vp = &con_scr[d];
	outq = &ser_tty[0];

	switch (cmd) {
	    case T_BREAK:
		dbug(cnPR,3,"conproc: T_BREAK requested\n");
		break;

	    case T_SUSPEND:
		dbug(cnPR,3,"\nconproc: T_SUSPEND requested\n");
		s = SPL();
		tp->t_state |= TTSTOP;
		goto run_cmd;

	    case T_TIME:
		dbug(cnPR,3,"conproc: T_TIME requested\n");
		tp->t_state &= ~TIMEOUT;
		goto start;

	    case T_WFLUSH:
		dbug(cnPR,3,"conproc: T_WFLUSH requested\n");
		if (confocus == vp)
			ttyflush(outq, FWRITE);
		break;

	    case T_RESUME:
		dbug(cnPR,3,"conproc: T_RESUME requested\n");
		tp->t_state &= ~TTSTOP;
		if (confocus == vp)
			(*ser_tty[0].t_proc)(&ser_tty[0], cmd);

	    case T_OUTPUT:
		/*
		 *  Start (or restart) transmission on a line
		 */
		dbug(cnPR,9,"\nconproc: T_OUTPUT requested\n");

	     start:
		if (confocus == vp)
			(*ser_tty[0].t_proc)(&ser_tty[0], cmd);
		s = SPL();
		if (tp->t_state&(TIMEOUT|BUSY|TTSTOP))
			goto out;

		tp->t_state &= ~(TTXON|TTXOFF);

		/*
		 * If the writer was sleeping on output overflow,
		 * wake him when low tide is reached.
		 */
		if (tp->t_state & OASLP
			  && tp->t_outq.c_cc <= ttlowat[tp->t_cflag & CBAUD]) {
			tp->t_state &= ~OASLP;
			wakeup((caddr_t)&tp->t_outq);
		}

		if (tp->t_outq.c_cc == 0 && (tp->t_state & TTIOW)) {
			tp->t_state &= ~TTIOW;
			wakeup((caddr_t)&tp->t_oflag);
		}

		while (tp->t_outq.c_cc > 0) {
			if (outq->t_outq.c_cc > tthiwat[outq->t_cflag & CBAUD])
			{
				tp->t_state |= TIMEOUT;
				timeout(ttrstrt, (caddr_t)tp, Hz / 4);
				break;
			}

			d = getc(&tp->t_outq);
			(*vp->disp)(vp, (d & 0x7f));
			if (confocus == vp && !(vp->flags & C_MONITOR)
			    || d == '\007')
			{
				_conputc(d);
			}
		}
		if (!(ser_tty[0].t_state & BUSY))
	     run_cmd:
		if (confocus == vp)
			(*ser_tty[0].t_proc)(&ser_tty[0], cmd);
	     out:
		splx(s);
		break;
	} /*end switch*/
}

conputs(s)
  register char	*s;
{
	while (*s)
		_conputc(*s++);
}

#if 0
/* This won't work 'cause we can't sleep at interrupt time, alas. */
_conputc(c)
  register char	c;
{
	register struct tty	*tp;
	int	s;

	tp = &ser_tty[0];
	s = spl5();
	while (tp->t_outq.c_cc > tthiwat[tp->t_cflag & CBAUD]) {
		(*tp->t_proc)(tp, T_OUTPUT);
		/* Don't sleep if it was drained all in one draught */
		if (tp->t_outq.c_cc != 0) {
		    tp->t_state |= OASLP;
		    sleep((caddr_t)&tp->t_outq, TTOPRI);
		}
	}
	putc(c, &tp->t_outq);
	splx(s);
}
#endif /* 0 */

conputc(c)
  char	c;
{
	_conputc(c);
	(*ser_tty[0].t_proc)(&ser_tty[0], T_OUTPUT);
}

conswtch(fp, tp)
  register struct vscr	*fp, *tp;
{
	int	s;

	/* Not allowed to try to switch again before we're done */
	s = SPL();

	/* First, attempt to blit the new screen into place */
	conblit(tp);

	/* Next, make tp foreground screen */
	confocus = tp;

	/* Finally, position cursor and switch to attributes of new screen */
	  /* Position cursor */
	conputs("Y"); _conputc(tp->y + ' '); _conputc(tp->x + ' ');

#ifdef SCRNSAVE
	  /* Set screen-saver */
	conputs((tp->flags & C_SCRNSVR)? "^" : "_");
	conputs("`"); _conputc(tp->savetime);
#endif /*SCRNSAVE*/

	  /* Set cursor blink */
	conputs("[?33"); _conputc((tp->flags & C_CURBLNK)? 'h' : 'l');

	  /* Set cursor type */
	conputs("[");
	_conputc((tp->flags & C_CURBLOK)? DEL : '_');
	conputs(" q");

	  /* Turn cursor on/off */
	conputs("R"); _conputc((tp->flags & C_CURVIS)? 'C' : 'c');

	  /* Set/reset graphics mode */
	conputs("R"); _conputc((tp->flags & C_GRAPHIC)? 'G' : 'g');

	  /* Set/reset inverse-video mode */
	conputs("R"); _conputc((tp->flags & C_REVERSE)? 'D' : '@');

	/* Have the real serial driver output the setup string */
	(*ser_tty[0].t_proc)(&ser_tty[0], T_OUTPUT);

	splx(s);
}

extern unsigned short	grflags;
conblit(vp)
  register struct vscr	*vp;
{
	register struct zg_iopb	*zp;
	struct tty	*tp;
	u_char	cmd;
	int	s;

	/* Do SPL() again here in case we were called from timeout */
	s = SPL();
	dbug1(cnBL,2,"conblit: start %d\n", vp - con_scr);

	/* Wait for output to drain, wait for graphics iopb */
	tp = &ser_tty[0];
	if (tp->t_outq.c_cc > 0 || (grflags & IOBUSY)) {
		dbug1(cnBL,1,"conblit: timeout(conblit, %d)\n", Hz / 4);
		timeout(conblit, (caddr_t)vp, Hz / 4); 
		splx(s);
		return;
	}

	grflags |= (IOBUSY|IOASYNC);
	zp = &z80graph;

	/* Setup count for blit of whole screen */
	zp->zg_count = CWIDTH * CDEPTH;
	zp->zg_offset = 0;
	zp->zg_dev = SCREEN;
	zp->zg_addr = (caddr_t)vp->scr;

	cmd = GO | INTENAB | WRITE;
	zp->zg_xcsr = ~cmd;
	zp->zg_csr = cmd;
	z80attn(GRdev, SCREEN, RqStart);

	dbug1(cnBL,2,"conblit: DONE %d\n", vp - con_scr);
	splx(s);
}

u_char	vold_c;
vswtch(tp, c)
  struct tty	*tp;
  u_char  c;
{
	int	n;
	int	vscan();

	n = tp - con_tty;
	dbug2(vrIN,5,"\rvswtch(%d, %x)\n", n, c);
	switch (c) {
	    /* HUP the terminal */
	    case 0x1B:		/* For the old keyboards */
	    case 0x8B:		/* ALT-<ESC> */
			signal(tp->t_pgrp, SIGHUP);
			ttyflush(tp, (FREAD|FWRITE));
			break;

	    /* (Re)set monitor mode */
	    case 0xFC:		/* For the old keyboards */
	    case 0x8C:		/* ALT-<BKSP> */
			confocus->flags ^= C_MONITOR;
			if (confocus->flags & C_MONITOR)
				confocus->flags |= C_WASMON;
			dbug(vrIN,2,(confocus->flags & C_MONITOR)? "M\b":"m\b");
			break;

	    /* Skip forward/back a screen */
	    case '+':
	    case '=': case 0x83:  ++n; goto cswtch;	/* ALT-<+> */

	    case '-': case 0x82:  --n; goto cswtch;	/* ALT-<-> */

	    /* Switch directly to a screen */
	     /* Orig   New        screen
	      * kybd   kybd       number  */
	    case '1': case 'x':   n = 0; goto cswtch;	/* ALT-1 */
	    case '2': case 'y':   n = 1; goto cswtch;	/* ALT-2 */
	    case '3': case 'z':   n = 2; goto cswtch;	/* ALT-3 */
	    case '4': case '{':   n = 3; goto cswtch;	/* ALT-4 */
	    case '5': case '|':   n = 4; goto cswtch;	/* ALT-5 */
	    case '6': case '}':   n = 5; goto cswtch;	/* ALT-6 */
	    case '7': case '~':   n = 6; goto cswtch;	/* ALT-7 */
	    case '8': case DEL:   n = 7; goto cswtch;	/* ALT-8 */
	    case '9': case 0x80:  n = 8; goto cswtch;	/* ALT-9 */
	    case '0': case 0x81:  n = 9;		/* ALT-0 */

	     cswtch:
		if (n < nscr && n >= 0)
			conswtch(confocus, &con_scr[n]);
		else
			conputc('\007');
		break;

	    default:
		if (tp->t_state & ISOPEN) {
			(*linesw[tp->t_line].l_input)(tp, vold_c, 0);
			(*linesw[tp->t_line].l_input)(tp, c, 0);
		}
		else {
			dbug2(vrIN,1,"\rvswtch: %x rec'd on closed chan %d\n",
					0xFF<<8|c,
					confocus - con_scr);
		}
		break;
	}

	/* Reset OLD screen's handler (OLD if we switched) */
	con_scr[tp - con_tty].input = vscan;
}

vscan(tp, c)
  struct tty	*tp;
  u_char  c;
{
	dbug2(vrIN,5,"\rvscan(%d, %x)\n", tp - con_tty, c);
	if (c == 0xFF || c == 0xFC) {	/* 0xFC == <CTRL><1> on old keybd */
		vold_c = c;
		con_scr[tp - con_tty].input = vswtch;
	}
	else if (tp->t_state & ISOPEN)
		(*linesw[tp->t_line].l_input)(tp, c, 0);
	else {
		dbug2(vrIN,1,"\rvscan: %x rec'd on closed chan %d\n",
				c,
				confocus - con_scr);
	}
}

vinput(tp, c)
  struct tty	*tp;
  u_char  c;
{
	if (confocus == (struct vscr *)0) {
		(*linesw[tp->t_line].l_input)(tp, c, 0);
		return;
	}

	dbug2(vrIN,5,"\rvinput(%d, %x)\n", confocus - con_scr, c);
	(*confocus->input)(&con_tty[confocus - con_scr], c);
}

int	vputc();
static char	curse;

vANSIblnk(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	switch (c) {
	    case 'h':	vp->flags |= C_CURBLNK;
			break;
	    case 'l':	vp->flags &= ~C_CURBLNK;
			break;
	}
	vp->disp = vputc;
}

vANSI3b(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	switch (c) {
	    case '3':	vp->disp = vANSIblnk;
			break;
	    default:	vp->disp = vputc;
			break;
	}
}

vANSI3a(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	switch (c) {
	    case '3':	vp->disp = vANSI3b;
			break;
	    default:	vp->disp = vputc;
			break;
	}
}

vANSIq(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	if (c == 'q') {
		if (curse == '_')
			vp->flags &= ~C_CURBLOK;
		else
			vp->flags |= C_CURBLOK;
	}
	vp->disp = vputc;
}

vANSIsp(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	switch (c) {
	    case ' ':	vp->disp = vANSIq;
			break;
	    default:	vp->disp = vputc;
			break;
	}
}

vescANSI(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	switch (c) {
	    case '?':	vp->disp = vANSI3a;
			break;
	    case '_':
	    case DEL:	vp->disp = vANSIsp;
			curse = c;
			break;
	    default:	vp->disp = vputc;
			break;
	}
}

vescR(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	switch (c) {
	    case '@':	vp->flags &= ~C_REVERSE;
			break;
	    case 'D':	vp->flags |= C_REVERSE;
			break;
	    case 'C':	vp->flags |= C_CURVIS;
			break;
	    case 'c':	vp->flags &= ~C_CURVIS;
			break;
	    case 'G':	vp->flags |= C_GRAPHIC;
			break;
	    case 'g':	vp->flags &= ~C_GRAPHIC;
			break;
	    case 'M':	vp->flags |= (C_MONITOR|C_WASMONR);
			dbug(vrOU,2,"M\b");
			break;
	    case 'm':	vp->flags &= ~C_MONITOR;
			dbug(vrOU,2,"m\b");
			break;
	}
	vp->disp = vputc;
}

vposcol(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	vp->x = (c - ' ') % CWIDTH;
	vp->disp = vputc;
}

vposrow(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	vp->y = (c - ' ') % CDEPTH;
	vp->disp = vposcol;
}

#ifdef SCRNSAVE
vsetsave(vp, c)
  register struct vscr	*vp;
  u_char	c;
{
	vp->savetime = c;
	vp->disp = vputc;
}
#endif /*SCRNSAVE*/

vesc(vp, c)
  register struct vscr	*vp;
  register char	c;
{
	switch (c) {
	    case '[':	vp->disp = vescANSI;
			return;
#ifdef SCRNSAVE
	    case '^':	vp->flags |= C_SCRNSVR;
			break;
	    case '_':	vp->flags &= ~C_SCRNSVR;
			break;
	    case '`':	vp->disp = vsetsave;
			return;
#endif /*SCRNSAVE*/
	    case 'A':	if (vp->y > 0)
				y_inc(vp, -1);
			break;
	    case 'B':	y_inc(vp, 1);
			break;
	    case 'C':	x_inc(vp, 1);
			break;
	    case 'D':	if (vp->x > 0)
				x_inc(vp, -1);
			break;
	    case 'E':	vcl2eof(vp, 0, 0);
	    case 'H':	vp->x = vp->y = 0;
			break;
	    case 'J':	vcl2eof(vp, vp->x, vp->y);
			break;
	    case 'K':	vcl2eol(vp);
			break;
	    case 'L':	vinsln(vp, 1);
			break;
	    case 'M':	vdelln(vp, vp->y, 1);
			break;
	    case 'P':	vinsc(vp);
			break;
	    case 'Q':	vdelc(vp);
			break;
	    case 'R':	vp->disp = vescR;
			return;
	    case 'Y':	vp->disp = vposrow;
			return;
	}

	vp->disp = vputc;
}

x_inc(vp, n)
  register struct vscr	*vp;
  int	n;
{
	vp->x += n;
	if (vp->x >= CWIDTH) {
		y_inc(vp, vp->x / CWIDTH);
		vp->x %= CWIDTH;
	}
}

y_inc(vp, n)
  register struct vscr	*vp;
  int	n;
{
	vp->y += n;
	if (vp->y >= CDEPTH) {
		vdelln(vp, 0, vp->y - CDEPTH + 1);
		vp->y = CDEPTH - 1;
	}
}

vcl2eof(vp, x, y)
  struct vscr	*vp;
  int		x, y;
{
	vclrbuf((char *)&vp->scr[y][x], vp->scr[CDEPTH] - &vp->scr[y][x], ' ');
}

vcl2eol(vp)
  register struct vscr	*vp;
{
	vclrbuf((char *)&vp->scr[vp->y][vp->x], CWIDTH - vp->x, ' ');
}

vinsc(vp)
  register struct vscr	*vp;
{
	u_char	*cp;

	cp = &vp->scr[vp->y][vp->x];
	copybuf(cp, &cp[1], CWIDTH - vp->x - 1);
	*cp = ' ';
}

vdelc(vp)
  register struct vscr	*vp;
{
	u_char	*cp;

	cp = &vp->scr[vp->y][vp->x];
	copybuf(&cp[1], cp, CWIDTH - vp->x - 1);
	vp->scr[vp->y][CWIDTH - 1] = ' ';
}

vinsln(vp, n)
  register struct vscr	*vp;
  int		n;
{
	int	size;

	size = vp->scr[CDEPTH] - vp->scr[vp->y + n];
	copybuf(vp->scr[vp->y], vp->scr[vp->y + n], size);
	vclrbuf(vp->scr[vp->y], n * CWIDTH, ' ');
}

vdelln(vp, y, n)
  register struct vscr	*vp;
  int		y, n;
{
	int	size;

	size = vp->scr[CDEPTH - y] - vp->scr[n];
	copybuf(vp->scr[y + n], vp->scr[y], size);
	vclrbuf(vp->scr[CDEPTH - n], n * CWIDTH, ' ');
}

vputc(vp, c)
  register struct vscr	*vp;
  register u_char  c;
{
	c &= 0x7F;
	if (vp->flags & C_MONITOR) {
		if (c <= 0x1f) {
			c += '@';
			if (!(vp->flags & C_REVERSE)) {
				vp->flags |= C_REVERSE;
				conputs("RD");
			}
		}
		else {
			if (vp->flags & C_REVERSE) {
				vp->flags &= ~C_REVERSE;
				conputs("R@");
			}
		}

		vp->scr[vp->y][vp->x] = c | (vp->flags & C_REVERSE);
		_conputc(c);
		x_inc(vp, 1);

		return;
	}

	/* Cancel any special goings on when coming out of monitor mode */
	if (vp->flags & C_WASMONR) {
		vp->flags &= ~(C_WASMONR|C_GRAPHIC|C_REVERSE);
		conputs("R@Rg");
	}
	if (isprint(c)) {
		/* Notice the REVERSE flag is cleverly 0x80 so
		 *  we can just OR it into the characters */
		if ((vp->flags & C_GRAPHIC) && c >= '@' && c <= '_')
			c &= 0x1f;
		vp->scr[vp->y][vp->x] = c | (vp->flags & C_REVERSE);
		x_inc(vp, 1);
	}
	else switch (c) {
	    case '\b':	if (vp->x > 0)
				x_inc(vp, -1);
			break;
	    case '\t':	x_inc(vp, 8 - (vp->x & 7));
			break;
	    case '\r':	vp->x = 0;
			break;
	    case '\n':	y_inc(vp, 1);
			break;
	    case '\f':	vcl2eof(vp, 0, 0);
	    		vp->x = vp->y = 0;
			break;
	    case ESC:	vp->disp = vesc;
			break;
	    case '\0':
	    default:	break;		/* Nothing to do... */
	}
}

vclrbuf(buff, size, c)
  register u_char  *buff;
  register int  size;
  register u_char  c;
{
#ifdef DEBUGC
	if (buff < (u_char *)con_scr && buff + size > (u_char *)&con_scr[nscr])
		printf("vclrbuf: bad parms (0x%X, %d)\n", buff, size);
#endif /*DEBUGC*/
	while (size--)
		*buff++ = c;
}

#ifdef SDEBUG
condump(d)
  int	d;
{
	printf("p_pid = %d\n", u.u_procp->p_pid);
	printf("p_pgrp = %d\n", u.u_procp->p_pgrp);
	printf("t_pgrp = %d\n", con_tty[d].t_pgrp);
	printf("&con_tty[%d] = 0x%X\n", d, &con_tty[d]);
	printf("u.u_ttyp = 0x%X\n", u.u_ttyp);
}

sdump(s)
  int	s;
{
	register struct vscr	*vp;
	register struct tty	*tp;
	register struct proc	*pp;

	if (s > nscr)
		return;

	tp = &con_tty[--s];
	vp = &con_scr[s];
	pp = u.u_procp;
}
#endif /*SDEBUG*/

#ifdef DEBUGC
copytrc(line, from, to, len)
  int	line;
  caddr_t  from, to;
  int	len;
{
	/* do a quick sanity check */
	if (from >= (caddr_t)con_scr && from + len <= (caddr_t)&con_scr[nscr]
	    && to >= (caddr_t)con_scr && to + len <= (caddr_t)&con_scr[nscr]
	    && len < sizeof (struct vscr))
	{
		return (bcopy(from, to, len));
	}

	printf("copytrc, line %d: bad copybuf(0x%X, 0x%X, %d)\n",
						line, from, to, len);
	return (-1);
}
#endif /*DEBUGC*/
