/*
 *	Tandy 68000/XENIX disk driver
 *	Hacked from PDP-11 driver.
 */

/*
 * Modifications
 *
 * $Log:	disk.c,v $
 * Revision 6.1  88/04/05  11:05:12  iv
 * 	- Fixed up some debug stuff
 * 	  (no code changes)
 * 
 * Revision 6.0  87/10/12  17:04:33  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.8  87/08/26  12:19:32  iv
 * 	- Updated error message format comments to reflect reality.
 * 	  No code changes.
 * 
 * Revision 5.7  87/05/06  11:21:17  iv
 * 	- Re-arranged some code for better optimization
 * 	- Changed open-anyway code to only fail for NOTREADY and CMDERR
 * 	- Changed format ioctl to allow floppy format too
 * 	  [z80ctl now supports online floppy format]
 * 
 * Revision 5.6  86/11/11  17:07:28  iv
 * 	- Improved debugging; separated dkcommand() debugging from the rest
 * 	  Also separated Open/Close debugging from the rest
 * 	- Changed open so only BADFORM errors are opened anyway
 * 	  (if the error is more serious, the user should run diskutil)
 * 	- Changed strategy to blow off I/O on unopened devices
 * 	  unless it is a read of the superblock [dkopen() does this]
 * 
 * Revision 5.5  86/10/16  15:21:11  iv
 * 	- Changed some debugging stuff.  No active code changes.
 * 
 * Revision 5.4  86/10/01  17:22:37  iv
 * 	- Changed reINIT logic to only do so if it is a hard disk
 * 	  (users can't change write-protect status on floppies or
 * 	  carts while they're still mounted; and besides, it confuses
 * 	  the poor Z80)
 * 	- Added more debug code
 * 	- Tightened up some of the error code.  Hopefully smaller now.
 * 
 * Revision 5.3  86/08/29  14:58:21  iv
 * 	- Added include of ioctl.h; DKIOCFMT was moved there
 * 	- Added include of z80halt.h; we now set the MSGSAFE bit if the
 * 	  driver is asked to do a write to the message save sector (block
 * 	  16 on /dev/hdbt0)
 * 	- Added support for /dev/recover; a new partition on hard disks
 * 	  (used to be the 2nd track on the boot partition [diag track])
 * 
 * Revision 5.2  86/08/13  16:49:24  iv
 * 	- Redid debug macro calls to use new partitioned debug
 * 	- Reworked whole driver for speed; mostly saved popular macro
 * 	  results in registers instead of recalculating over and over
 * 	- Changed dkcommand() to always make cylin FFFF, so disksort()
 * 	  will always add to end of queue to run cmds last
 * 	- Fixed bug in disk-write-protected message
 * 	- Fixed bug in dkIOerr() where we had a bp test backwards
 * 
 * Revision 5.1  86/07/16  15:23:45  iv
 *	- Re-wrote all special case code for INIT, UMOUNT, & FORMAT
 *	  to use new dkcommand() routine which uses a buf to handle
 *	  the command in a normal [ie using dkstrategy() & dkuintr()]
 *	  manner.
 *	- Fixed bug where if disk is mounted read-only, and a user tries
 *	  to mount it for write, the disk would be umounted (and spun down
 *	  in the case of cartridge).
 *	  Now test to make sure disk is not already mounted at end of dkopen()
 *	- Now have 1 cdkbuf for each major device, so if 1 device is doing
 *	  raw I/O, another will not have to wait for it.
 *	- Redid error messages to use common copies of strings
 *	- Redid error output format to be more like the 3000's
 *
 * Revision 5.0  85/08/23  10:29:51  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.2  85/08/22  16:49:12  iv
 * 	- Structure rewrite for new IOmega cartridge drive.
 * 	  Many hacks and changes.
 * 
 * Revision 4.1  85/03/27  15:34:57  iv
 * 	- Changed "drive closed; error cleared" message to "Device closed; ..."
 * 	- Only output the above message if the user was warned about the error
 * 
 * Revision 4.0  85/03/13  14:43:17  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.1  85/02/18  10:39:58  iv
 * 	- Corrected some error handling and fixed args to deverr()
 * 
 * Revision 3.0  85/01/17  15:47:59  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:57:31  iv
 * 	- Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.4  84/12/18  17:53:33  iv
 * 	- Worked over good for sys III.  Unnoted.
 * 
 * Revision 1.3  84/04/16  16:56:34  iv
 * 	- Debugged v7 -> sys3 conversion.  Many changes.
 * 
 * Revision 1.2  84/04/12  17:00:41  iv
 * 	- Converted driver from v7 to sys 3  per Microsoft 3.0 conversion guide
 * 
 * Revision 1.1  84/04/11  13:58:35  iv
 * Initial revision
 */


/*#define DEBUG	6		/* Enable driver debugging messages */
#define DbugVar	dk_debug
#ifdef DEBUG
#  define dkCN	0x0001		    /* Configuration debugging */
#  define dkPR	0x0002		    /* Partitioning debugging */
#  define dkSR	0x0004		    /* Strategy debugging */
#  define dkST	0x0008		    /* Start debugging */
#  define dkI	0x0010		    /* Interrupt debugging */
#  define dkIN	0x0020		    /* Init debugging */
#  define dkIO	0x0040		    /* Ioctl debugging */
#  define dkOC	0x0080		    /* Open/close debugging */
#  define dkCMD	0x0100		    /* dkcommand debugging */
#  define debug()	/* For ASSERT */
#endif /*DEBUG*/
#define DEBUGS	(dkST|dkSR|dkCMD)	/* Which sections I want to hear from */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/filsys.h"
#include "../h/buf.h"
#include "../h/iobuf.h"
#include "../h/dir.h"
#include "../h/conf.h"
#include "../h/user.h"
#include "../h/assert.h"
#include "../h/file.h"
#include "../h/proc.h"
#include "../h/ioctl.h"
#include "../h/z80disk.h"
#include "../h/z80boot.h"
#include "../h/z80halt.h"
#include "../h/z80attn.h"
#include "../h/disk.h"
#include "../h/dbug.h"


/*
 *  Misc strings
 */
char	Hard[] = "Hard";
char	Soft[] = "Soft";
char	Mbox[] = "Mailbox";
char	Unclass[] = "Unclassified";
char	dk_prot[] = "active drive write protected";
char	dk_chg[] = "active disk changed";
char	dk_nrdy[] = "active drive not ready";
char	*dkNames[] = { "Floppy", "Hard", "Cartridge" };
#ifdef DEBUG
char	*dkCmds[] = { "INIT", "UMOUNT", "FORMAT" };
#endif /*DEBUG*/


/*
 * If dkverfy[FOO] is set, then all output requests for
 * device FOO have the WRVERIFY bit set.  This makes it
 * simple for a user program with access to /dev/kmem
 * to turn on or off write verification.
 */
short dkverfy[NMAJ] = {
	0, 0, 0
};

struct diskconf confmap[NMAJ][NDISK] = {
	{ 0, 0, 0, 0, 0, 2, MOTORWAIT }, /* Floppies */
	{ 0, 0, 0, 0, 0, 3, MOTORWAIT },
	{ 0, 0, 0, 0, 0, 3, MOTORWAIT },
	{ 0, 0, 0, 0, 0, 3, MOTORWAIT },
	{ 0, 0, 0, 0, 0, 0, 0 },	 /* Hard Disks */
	{ 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0 },	 /* Cartridge drives */
	{ 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0 }
};


/*
 * dkflags:
 *
 *     DKERROR, DKWARNED, and DKREADONLY can get changed at interrupt time
 *     DKERROR can get set only at interrupt time.
 *     DKERROR is cleared only at close time.
 *
 *     The XENIX close sequence is purge, then close.  This means that it
 *     impossible for a queued request to find a cleared DKOPEN bit.
 * 
 *     dkclose clears DKERROR on last close.
 *
 *     !DKOPEN ==> no queued requests ==> the only possible disk interrupt
 *	    for given drive can come from init
 */
struct	dkflags	dkflags[NMAJ][NDISK];


struct  iobuf	dktab[NMAJ];
struct  buf	rdkbuf[NMAJ];
struct	buf	cdkbuf[NMAJ];	/* used for non I/O commands [eg UMOUNT] */

/*
 * The open routine needs to know the major device number of
 * the disk.
 *
 * If rootdev is a dk with minor==255, then rootdev, swapdev, pipedev,
 *	are determined at boot time.
 */
static int majorno = -1;
static dev_t	bootdev;

/*
 * The comments associated with the DK status bits are crucial to
 * the working of dkopen.
 */
dkopen(dev, flag, class)
{
	register struct iobuf *dp;
	register struct buf *bp;
	register struct dkflags *flagp;
	register int dn;
	register int partition;
	register int opened;
	register unsigned char	dmj, dmn;
	int	x;
	int	initerr;
	static unsigned char onceonly = 0;
	static unsigned char rootonce = 0;
	static unsigned char swaponce = 0;

	initerr = 0;
	/*
	 * On the first call, find the major device number.
	 */
	if (onceonly == 0) {
		register struct bdevsw *bdp;

		onceonly++;
		for (bdp = &bdevsw[bdevcnt]; --bdp >= &bdevsw[0]; ) 
			if (bdp->d_tab == dktab) {
				majorno = bdp - &bdevsw[0];
				break;
			}
		if (majorno == -1)
			panic("disk major number");

		dev = makedev(majorno, minor(dev));
		bootdev = makedev(majorno, 35);
		/*
		 * Autoconfigure rootdev, swapdev, pipedev.
		 */
		if (dev == rootdev) {
			if (minor(dev) == 0377) {
				dbug(dkCN,0, "[Auto-configing rootdev");
				rootdev = dev = makedev(majorno,
				    (z80boot.zb_root << DRSHIFT) + 1);
				dbug2(dkCN,1,": rootdev = %d/%d",
					major(rootdev), minor(rootdev));
				dbug(dkCN,0,"]\n");
			}
			if (major(swapdev)==majorno && minor(swapdev)==0377) {
				dbug(dkCN,0, "[Auto-configing swapdev");
				swapdev = makedev(majorno,
				    (z80boot.zb_root << DRSHIFT) + 2);
				dbug2(dkCN,1,": swapdev = %d/%d",
					major(swapdev), minor(swapdev));
				dbug(dkCN,0,"]\n");
			}
			if (major(pipedev)==majorno && minor(pipedev)==0377) {
				dbug(dkCN,0, "[Auto-configing pipedev");
				pipedev = makedev(majorno,
				    (z80boot.zb_root << DRSHIFT) + 1);
				dbug2(dkCN,1,": pipedev = %d/%d",
					major(pipedev), minor(pipedev));
				dbug(dkCN,0,"]\n");
			}
		}
	}

	dn = driveno(dev);
	partition = pseudo(dev);
	dmj = dmaj(dn);
	dmn = dmin(dn);
	flagp = &dkflags[dmj][dmn];

	dbug5(dkOC,1,"*** dkopen on dev %d/%d [%s:%d;%d]",
			majorno,
			minor(dev),
			dkNames[dmj],
			dmn,
			partition);
	dbug2(dkOC,1," flag/class = %d/%d\n", flag, class);

loop:
	if (flagp->status & DKERROR) {
		dbug(dkOC,0, "*** dkopen: DKERROR set; return (EIO)\n");
		u.u_error = EIO;
		goto out;
	}
	opened = (flagp->status & DKOPEN);
	if (opened == 0 
	    || (flag & FWRITE) && (flagp->status & DKREADONLY) && _hard(dmn))
	{
		/*
		 * Only one process per drive may enter this code.
		 */
		if (flagp->status & DKINITTING) {
			dbug(dkOC,4, "*** dkopen: DKINITTING sleep...\n");
			flagp->status |= DKWANTED;
			sleep((caddr_t)flagp, PRIBIO);
			dbug3(dkOC,5,"*** dkopen: %s:%d;%d DKINITTING awake\n",
					dkNames[dmj],
					dmn,
					partition);
			goto loop;
		}
		dbug(dkOC,2, "*** dkopen: going to INIT drive...\n");
		flagp->status |= DKINITTING;

		/*
		 * Ask Z80 for physical configuration of disk
		 */
		initerr = dkcommand(dev, dkINIT, 0);
		if (initerr == 0 || initerr == SOFTERR) {
			initerr = 0;
			dbug3(dkIN,4,"*** dkopen: %s:%d;%d INIT okay\n",
					dkNames[dmj],
					dmn,
					partition);
		}
		else {
			dbug3(dkIN,0,"*** dkopen: %s:%d;%d INIT failed\n",
					dkNames[dmj],
					dmn,
					partition);
		}

		/*
		 * If necessary, look for a file system to determine
		 * partition information.
		 */
		if (initerr == 0 && opened == 0) {
			ASSERT((flagp->status & (DKERROR|DKWARNED))==0);
			/*
			 * Set up partition info for whole disk partition.
			 * (Also sets up boot track partitition, if any).
			 */
			if (dkpart0(dn)) {
				/*
				 * Look for superblock.
				 */
				dbug3(dkOC,4,"*** dkopen: %s:%d;%d read SB\n",
						dkNames[dmj],
						dmn,
						partition);
				bp = bread(makedev(majorno, dn<<DRSHIFT), SUPERB);
				/*
				 * Set up partitition info for rest of disk.
				 */
				dkpartition(dn, bp);
				bp->b_flags |= (B_AGE|B_STALE); /* Throw away */
				brelse(bp);
				/*
				 * Because the bread may have produced an error,
				 * we must clear these status bits.
				 */
				flagp->status &= ~(DKERROR|DKWARNED);
			} else
				initerr = INVALID;
		}
		/*
		 * Allow other processes into this code.
		 */
		if (flagp->status & DKWANTED)
			wakeup((caddr_t)flagp);
		flagp->status &= ~(DKINITTING|DKWANTED);
	}

	if (partition == BOOTPART && initerr != NOTREADY && initerr != CMDERR)
		goto ckwrp;		/* Allow opens for format */
	if (initerr) {
		dbug(dkOC,0, "*** dkopen: `initerr' set; setting EIO...\n");
		u.u_error = EIO;
	}
	else if (partition >= NPSEUDO 
		 || confmap[dmj][dmn].dc_sizes[partition].nsects == 0) 
	{
		dbug(dkOC,0, "*** dkopen: bad or 0 partition; err ENXIO...\n");
		u.u_error = ENXIO;
	}
	else {
    ckwrp:	if ((flag&FWRITE) && (flagp->status & DKREADONLY)) {
			dkerror(dn, &dk_prot[7]);
			u.u_error = EIO;
		}
		else if (u.u_error == 0 || partition == BOOTPART) {
			flagp->status |= DKOPEN;
			flagp->openfl[class] |= (1 << partition);
			if (initerr || u.u_error) {
			    dbug(dkOC,0, "*** dkopen: BOOT error open OKd\n");
			    flagp->status |= DKERROR;
			    u.u_error = initerr = 0;
			}
		}
	}
#ifdef DEBUG
	if (rootonce == 0 && class && makedev(majorno,minor(dev)) == rootdev) {
		rootonce++;
		if (pseudo(rootdev) == 1) {
			register struct sizes *sizep =
				&confmap[dmj][dmn].dc_sizes[1];
			register blocks;

			blocks = sizep->nsects >> (BSHIFT - sizep->sshift);
			dbug1(dkPR,0, "Root area %d blocks\n", blocks);
		}
	}
#endif /*DEBUG*/
	if (swaponce == 0 && class && makedev(majorno,minor(dev)) == swapdev) {
		swaponce++;
		if (pseudo(swapdev) == 2) {
			register struct sizes *sizep =
				&confmap[dmj][dmn].dc_sizes[2];

			/*
			 * Automatically calculate swap parameters
			 * from disk information.
			 */
			swplo = 0;
			nswap = sizep->nsects >> (BSHIFT - sizep->sshift);
			dbug1(dkPR,0, "Swap area %d blocks\n", nswap);
		}
	}

    out:
	if (u.u_error) {
		/* Just like close */
		if (flagp->openfl[0] == 0 && flagp->openfl[1] == 0) {
			dbug3(dkOC,1,"*** dkopen: %s:%d;%d calling dkumount\n",
					dkNames[dmj],
					dmn,
					partition);
			flagp->status &= ~(DKOPEN|DKERROR|DKWARNED);
			dkumount(dev);
		}
	}
}

dkpart0(dn)
register int dn;
{
	register struct diskconf *confp;
	register struct sizes *sizep;
	register int ntracks;
	register int trkoff;
	register int sshift;

	confp = &confmap[dmaj(dn)][dmin(dn)];
	sizep = &confp->dc_sizes[BOOTPART];	/* boot track descriptor */
	ntracks = confp->dc_ncyl * confp->dc_nhead;

	confp->dc_sizes[RCVRPART].nsects = 0;	/* Recover partition */

	switch (confp->dc_secsiz) {
		case 512:
			sshift = 9;
			break;
		case 256:
			sshift = 8;
			break;
		case 128:
			sshift = 7;
			break;
		case 64:
			sshift = 6;
			break;
		default:
			dkerror(dn, "bad sector size");
			sizep->nspt = 0;
			return (0);
	}

	if (confp->dc_fmt & MIXED_FMT) {
		ASSERT(!cart(dn) && !hard(dn));
		ntracks--;
		trkoff = 1;
		sizep->trkoff = 0;
		sizep->nsects = 26;		/* MAGIC NUMBER */
		sizep->nspt = 26;		/* MAGIC NUMBER */
		sizep->sshift = 7;		/* MAGIC NUMBER */
	}
	else if (hard(dn)) {
		ntracks -= 2;			/* boot track & diag track */
		trkoff = 2;

		sizep->trkoff = 0;			/* Boot track */
		sizep->nsects = confp->dc_trksiz;
		sizep->nspt = confp->dc_trksiz;
		sizep->sshift = sshift;

		++sizep;
		sizep->nsects = confp->dc_trksiz;	/* Recover track */
		sizep->trkoff = 1;
		sizep->nspt = confp->dc_trksiz;
		sizep->sshift = sshift;
	} 
	else if (cart(dn)) {
		--ntracks;			/* boot track */
		trkoff = 1;
		sizep->trkoff = 0;
		sizep->nsects = confp->dc_trksiz;
		sizep->nspt = confp->dc_trksiz;
		sizep->sshift = sshift;
	}
	else {
		sizep->nsects = 0;
		trkoff = 0;
	}

	sizep = &confp->dc_sizes[0];		/* non-boot part */
	sizep->nsects = confp->dc_trksiz*ntracks;
	sizep->trkoff = trkoff;
	sizep->nspt = confp->dc_trksiz;
	sizep->sshift = sshift;
#ifdef DEBUG
	if (DbugOK(dkPR, 2)) {
		register int i;

		dbug(dkPR,2, "dkpart0:\n");
		dbug5(dkPR,2,"\tncyl=%d secsiz=%d trksiz=%d ntrk=%d fmt=0x%x\n",
			confp->dc_ncyl,
			confp->dc_secsiz,
			confp->dc_trksiz,
			confp->dc_nhead,
			confp->dc_fmt);
		for (i = 0; i < NPSEUDO; i += NPSEUDO - 1) {
			sizep = &confp->dc_sizes[i];
			dbug5(dkPR,2,"\tpartition %d nsects=%d trkoff=%d nspt=%d sshift=%d\n",
				i,
				sizep->nsects,
				sizep->trkoff,
				sizep->nspt,
				sizep->sshift);
		}
	}
#endif /*DEBUG*/
	return (1);
}

dkpartition(dn, bp)
  register int dn;
  register struct buf *bp;
{
	register struct diskconf *confp = &confmap[dmaj(dn)][dmin(dn)];
	register struct sizes *sizep = &confp->dc_sizes[0];
	register struct filsys *fsp = (struct filsys *)paddr(bp);
	register int nspt = sizep->nspt;
	register int fssize;
	register int temp;

	fssize = fsp->s_fsize;
	temp = fsp->s_isize;
	dbug4(dkPR,2,"dkpartition: bp->b_flags=0%o s_fsize=0x%x s_isize=0x%x s_clean=0%o\n",
		bp->b_flags,
		fssize,
		temp,
		fsp->s_clean);

	if ((bp->b_flags & B_ERROR) == 0
	    && fssize > temp
	    && temp > 0
	    && ((temp = fsp->s_clean) == S_CLEAN || temp == 0))
	{
		/*
		 * convert to sector count
		 */
		fssize <<= (BSHIFT - sizep->sshift);
		/*
		 * Round up to whole number of tracks.
		 */
		fssize += nspt - 1;
		fssize /= nspt;
		fssize *= nspt;
		if (fssize > sizep->nsects)
			fssize = 0;
	} else
		fssize = 0;

	copybuf(&sizep[0], &sizep[1], sizeof (sizep[0]));
	copybuf(&sizep[0], &sizep[2], sizeof (sizep[0]));

	sizep++;			/* points to file system now */
	sizep->nsects = fssize;

	sizep++;			/* points to rest of disk now */
	sizep->nsects -= fssize;
	sizep->trkoff += fssize/nspt;
#ifdef DEBUG
	if (DbugOK(dkPR, 2)) {
		register int i;

		dbug(dkPR,2, "dkpartition:\n");
		dbug5(dkPR,2,"\tncyl=%d secsiz=%d trksiz=%d ntrk=%d fmt=0x%x\n",
			confp->dc_ncyl,
			confp->dc_secsiz,
			confp->dc_trksiz,
			confp->dc_nhead,
			confp->dc_fmt);
		for (i = 0; i < NPSEUDO; i++) {
			sizep = &confp->dc_sizes[i];
			dbug5(dkPR,2,"\tpartition %d nsects=%d trkoff=%d nspt=%d sshift=%d\n",
				i,
				sizep->nsects,
				sizep->trkoff,
				sizep->nspt,
				sizep->sshift);
		}
	/*	dbug(dkPR,3,"Press any key to continue: ");
		if (DbugOK(dkPR, 3))
			getchar();
	*/
	}
#endif /*DEBUG*/
}

/*ARGSUSED*/
dkclose(dev, flag, class)
{
	register int	dn = driveno(dev);
	register struct dkflags *flagp = &dkflags[dmaj(dn)][dmin(dn)];

	dbug3(dkOC,2, "*** dkclose(%d/%d, %d)\n", majorno, minor(dev), class);
	flagp->openfl[class] &= ~(1 << pseudo(dev));
	if (flagp->openfl[0] == 0 && flagp->openfl[1] == 0) {
		if (flagp->status & DKWARNED)
			dkerror(dn, "Device closed; error cleared");
		flagp->status &= ~(DKOPEN|DKERROR|DKWARNED);
		dkumount(dev);
	}
}

dkumount(dev)
  int	dev;
{
	if (cart(driveno(dev)))
		dkcommand(dev, dkUMOUNT, 0);
}


dkcommand(dev, cmd, arg)
  int	dev, cmd, arg;
{
	register struct buf	*bp;
	int	s;

	dbug4(dkCMD,2, "*** dkcommand (%s:%d, %s, %d)\n",
				dkNames[dmaj(driveno(dev))],
				dmin(driveno(dev)),
				dkCmds[cmd],
				arg);

	bp = &cdkbuf[dmaj(driveno(dev))];

	s = SPL();
	while (bp->b_flags & B_BUSY) {
		bp->b_flags |= B_WANTED;
		sleep((caddr_t)bp, PRIBIO);
	}
	bp->b_flags = B_BUSY | B_READ;
	splx(s);

	/* Make sure any commands we run will be done LAST, because
	 *   if we're doing a UMOUNT, it HAS to be done last. */
	bp->b_cylin = (ushort)-1;	/* the biggest # that'll fit... */
	bp->b_dev = dev;
	bp->b_error = 0;
	bp->b_command = cmd;
	bp->b_bcount = arg;
	dkstrategy(bp);

	iowait(bp);
	if (bp->b_flags & B_WANTED)
		wakeup((caddr_t)bp);
	bp->b_flags &= B_ERROR;

	return (bp->b_resid);		/* Return error code (if any) */
}


/*
 *    Strategy Routine:
 *    Arguments:
 *      Pointer to buffer structure
 *    Function:
 *      Check validity of request
 *      Queue the request
 *      Start-up the device if idle
 *
 */

dkstrategy(bp)
  register struct buf *bp;
{
	register int	dev;
	register int	dn;
	register unsigned char	dmj, dmn;
	struct buf	*cbp;

	dev = bp->b_dev;
	dn = driveno(dev);
	dmj = dmaj(dn);
	dmn = dmin(dn);
	cbp = &cdkbuf[dmj];
	dbug2(dkSR,5,"*** dkstrategy %s:%d ", dkNames[dmj], dmn);

	if (bp != cbp)		/* Only applicible for I/O requests */
	{
		register struct dkflags *flagp;

		dbug(dkSR,5,":1");
		flagp = &dkflags[dmj][dmn];
		if ((flagp->status & DKERROR) 
		    || !(flagp->status & DKOPEN)
			 && (!(bp->b_flags & B_READ) || bp->b_blkno != SUPERB))
		{
			dbug(dkSR,5," EIO");
			if (!(flagp->status & DKOPEN)) {
				dbug5(dbANY,0, "%s[pid %d] I/O on closed %s:%d;%d !!\n",
					DbugOK(dkSR, 5)? "\n" : "dkstrategy: ",
					u.u_procp->p_pid,
					dkNames[dmj],
					dmn,
					pseudo(dev));
				ICEbrk();
			}
			else if ((flagp->status & DKWARNED) == 0) {
				flagp->status |= DKWARNED;
				dkerror(dn, "Further I/O aborted until device closed");
			}

			bp->b_flags |= B_ERROR;
			bp->b_resid = bp->b_bcount;
			iodone(bp);
			return;
		}
	}

	if (bp != cbp)		/* Only applicible for I/O requests */
	{
		register int shift;
		register struct diskconf *confp;
		register struct sizes *sizep;
		register int secsize;
		register int secno;
		register int sz;

		dbug(dkSR,5,":2");
		confp = &confmap[dmj][dmn];
		sizep = &confp->dc_sizes[pseudo(dev)];
		shift = sizep->sshift;
		secsize = 1 << shift;

		sz = (bp->b_bcount + secsize - 1) >> shift;
		secno = bp->b_blkno << (BSHIFT - shift);

		dbug2(dkSR,6," sec %d, count %d", secno, sz);
		bp->b_resid = 0;
		if (secno + sz > sizep->nsects) {
			dbug(dkSR,5, " E2BIG");
			if (secno >= sizep->nsects) {
				if ((bp->b_flags & B_READ) == 0) {
					dbug(dkSR,5, " ENXIO");
					bp->b_error = ENXIO;
					bp->b_flags |= B_ERROR;
				} else
					dbug(dkSR,5, " EOF");
				bp->b_resid = bp->b_bcount;
				iodone(bp);
				return;
			}
			else {
				bp->b_bcount -=
				    bp->b_resid =
					(secno + sz - sizep->nsects) << shift;
				if (bp->b_resid > 0) {
					dbug2(dkSR,6, " bcount=%d resid=%d", 
							bp->b_bcount,
							bp->b_resid);
				}
			}
		}

		bp->av_forw = NULL;
		bp->b_cylin = sizep->trkoff + (secno / sizep->nspt);
	}

	{
		register struct iobuf	*dp;
		register int	s;

		dbug(dkSR,5,":3\n");

		dp = &dktab[tabindex(dn)];
		s = SPL();
		disksort(dp, bp);
		if (!dp->b_active)
			dkustart(dp);
		splx(s);
	}
}

/*
 * device start routine
 * always called at hi pri
 */

dkustart(dp)
  struct iobuf	*dp;
{
	register struct buf *bp;
	register struct zd_iopb *iop;
	register struct diskconf *confp;
	register struct sizes	*sizep;
	register unsigned shift;
	register unsigned short nspt;
	register unsigned daddr_t sn;
	register unsigned short cn;
	register unsigned char	dmj, dmn;
	unsigned short	dn;
	unsigned char	cmd;
	struct buf	*cbp;
	int		dev;


dkloop:
	if ((bp = dp->b_actf) == NULL)
		return;
	dev = bp->b_dev;
	dn = driveno(dev);
	dmj = dmaj(dn);
	dmn = dmin(dn);
	cbp = &cdkbuf[dmj];

	/* It's OK to try a non I/O command when DKERROR is set,
	 *   we might be doing a FORMAT or UMOUNT...  */
	if (bp != cbp) {
		if (dkflags[dmj][dmn].status & DKERROR) {
			dp->b_errcnt = 0;
			dp->b_actf = bp->av_forw;
			bp->b_resid = bp->b_bcount;
			bp->b_flags |= B_ERROR;
			iodone(bp);
			goto dkloop;
		}
	}

	confp = &confmap[dmj][dmn];
	iop = &z80disk[dmj];

	if (iop->zd_csr & GO) {
		dbug2(dbANY,0,"dkustart: GO set on %s:%d\n", dkNames[dmj],dmn);
		panic("GO set at dkustart");
	}

	dp->b_active++;                    /* ACTIVATE DEVICE */
	if (bp == cbp) {
		dbug3(dkST|dkCMD,2, "*** dkustart: RUN (%s:%d, %s)\n",
				dkNames[dmj],
				dmn,
				dkCmds[bp->b_command]);

		switch (bp->b_command) {
		    case dkINIT:
		      /* Because we don't |= UMOUNT, this is a MOUNT request */
			cmd = INTENAB|INIT|GO;
			break;

		    case dkUMOUNT:
			cmd = INTENAB|UMOUNT|INIT|GO;
			break;

		    case dkFORMAT:
			cmd = FORMAT|INTENAB|GO;
			iop->zd_count = bp->b_bcount;	/* Set interleave */
			break;
		}
	} 
	else {
		sizep = &confp->dc_sizes[pseudo(dev)];
		shift = sizep->sshift;
		nspt = sizep->nspt;

		cn = bp->b_cylin / confp->dc_nhead;
		sn = bp->b_blkno << (BSHIFT - shift);

		cmd = (INTENAB | GO) | ((bp->b_flags & B_READ)? READ : WRITE);
if (!_hard(dmj))
		dbug(dkST,3,"*** dkustart: ");

		iop->zd_addr = (caddr_t)paddr(bp);
		iop->zd_count = (bp->b_bcount + (1 << shift) - 1) >> shift;
if (!_hard(dmj))
		dbug5(dkST,3, "%s:%d %s addr=0x%X count=%d",
				dkNames[dmj],
				dmn,
				(cmd & READ)? "RD" : "WR",
				iop->zd_addr,
				iop->zd_count);

		if (_cart(dmj)) {
		    iop->zd_blk = (sizep->trkoff * nspt) + sn;
		    dbug1(dkST,3, " blk=0x%X", iop->zd_blk);
		}
		else {
		    iop->zd_cyl = cn;
		    iop->zd_hd = 
			(sn / nspt) + sizep->trkoff - (confp->dc_nhead * cn);
		    iop->zd_sec = sn % nspt;
if (!_hard(dmj))
		    dbug3(dkST,3, " cyl=0x%x hd=%d sec=%d",
					iop->zd_cyl,
					iop->zd_hd,
					iop->zd_sec);
		}

if (!_hard(dmj))
		dbug3(dkST,4," ncyl=0x%x nhead=%d trksiz=%d",
			iop->zd_ncyl,
			iop->zd_nhead,
			iop->zd_trksiz);

if (!_hard(dmj))
		dbug(dkST,3,"\n");
	}

	iop->zd_mode = confp->dc_mode;
	iop->zd_seek = confp->dc_seek;
	iop->zd_dev = dmn;
	if (dkverfy[dmj])
		iop->zd_mode |= WRVERIFY;

if (!_hard(dmj))
	dbug4(dkST,5,"*** dkustart: dev=%d mode=0x%x seek=%d csr=0x%x\n",
					iop->zd_dev,
					iop->zd_mode,
					iop->zd_seek,
					cmd);
	iop->zd_xcsr = ~cmd;
	iop->zd_csr = cmd;

	dmj = _hard(dmj)? HDdev :
	      _cart(dmj)? SCSIdev :
			FDdev;
	z80attn(dmj, dmn, RqStart);

	/*
	 *  Wierd hook for message save code in the z80
	 */
	if (bp->b_dev == bootdev && bp->b_blkno == 16 && (cmd & WRITE))
		z80halt |= MSGSAFE;
}

struct intptrs {
	struct zd_iopb	*iop;
	struct iobuf	*dp;
} iptrs[NMAJ] = {
	{ &z80disk[HARD], &dktab[0] },		/* Hard disk first, */
	{ &z80disk[CART], &dktab[2] },		/* then cartridge, */
	{ &z80disk[FLOP], &dktab[1] }		/* then floppy */
};
dkintr()
{
	register struct zd_iopb	*iop;
	register struct intptrs *ip;

	for (ip = &iptrs[0]; ip < &iptrs[NMAJ]; ++ip) {
		iop = ip->iop;
		if ((iop->zd_csr & (GO|INTPEND)) == INTPEND) {
			iop->zd_csr &= ~INTPEND;
			dkuintr(ip->dp, iop);
		}
	}
}


dkuintr(dp, iop)
  struct iobuf	*dp;
  register struct zd_iopb *iop;
{
	register struct buf	 *bp;
	register struct diskconf *confp;
	register struct dkflags	 *flagp;
	register unsigned char	dmj, dmn;
	int	dn;
	char	*msg;

	dbug(dkI,6,"*** dkuintr()\n");
	if (dp->b_active == 0)
		return;

	bp = dp->b_actf;
	dn = driveno(bp->b_dev);
	dmj = dmaj(dn);
	dmn = dmin(dn);

#ifdef DEBUG
	if (iop->zd_csr & GO) {
		dbug2(dkI,0,"dkuintr: GO set on %s:%d!\n", dkNames[dmj], dmn);
		return;
	}
#endif /*DEBUG*/

	confp = &confmap[dmj][dmn];
	flagp = &dkflags[dmj][dmn];

	dp->b_active = 0;
	if (bp == &cdkbuf[dmj]) {
		if (iop->zd_csr & ERROR)
			bp->b_resid = iop->zd_err;
		else
			bp->b_resid = 0;

		switch (bp->b_command) {
		    case dkINIT:
			if (iop->zd_csr & ERROR) switch (iop->zd_err) {
			    case 0:
				/* This can happen if the disk is already 
				 *  mounted.  The z80 is preventing us from
				 *  below trashing our copy of the disk's
				 *  physical geometry in the conf, which
				 *  it didn't load in this case. */
				break;

			    case SOFTERR:
				iop->zd_csr &= ~ERROR;
				msg = Soft;
				goto IOprint;

			    default:
				msg = Unclass;		/* "Hunh?" */
				goto IOprint;

			    case CMDERR:
				msg = Mbox;
				goto IOprint;

			    case HARDERR:
				msg = Hard;
			IOprint:
				dkIOerr(bp, iop, dn, msg);
				break;

			    case INVALID:
				msg = &dk_chg[7];
				goto INprint;

			    case NOTREADY:
				msg = &dk_nrdy[7];
				goto INprint;

			    case BADFORM:
				msg = "bad format";
			  INprint:
				dkerror(dn, msg);
				break;
			}

			dbug(dkIN,2,"dkINIT:");
			dbug3(dkI|dkIN,2,
				"  zd_ncyl=%d zd_secsiz=%d zd_trksiz=%d ",
					iop->zd_ncyl,
					iop->zd_secsiz,
					iop->zd_trksiz);
			dbug2(dkI|dkIN,2,
				"  zd_nhead=%d zd_fmt=0x%x\n",
					iop->zd_nhead,
					iop->zd_fmt);

			if (!(iop->zd_csr & ERROR)) {
				confp->dc_ncyl = iop->zd_ncyl;
				confp->dc_secsiz = iop->zd_secsiz;
				confp->dc_trksiz = iop->zd_trksiz;
				confp->dc_nhead = iop->zd_nhead;
				confp->dc_fmt = 
					(iop->zd_fmt & (MIXED_FMT|DBL_DENS));
			}

			if (iop->zd_fmt & WRPROT)
				flagp->status |= DKREADONLY;
			else
				flagp->status &= ~DKREADONLY;
			break;

		    case dkUMOUNT:
			/* Don't care about the rest except that it is done */
			break;

		    case dkFORMAT:
			if (iop->zd_csr & ERROR) switch (iop->zd_err) {
			    case WRPROT:
				flagp->status |= DKREADONLY;
				msg = dk_prot;
				goto FMprint;

			    case INVALID:
				msg = dk_chg;
				goto FMprint;

			    case HARDERR:
				msg = Hard;
				goto Fprint;

			    case SOFTERR:
				msg = Soft;
				goto Fprint;

			    case NOTREADY:
				msg = dk_nrdy;
			FMprint:
				dkerror(dn, msg);
				break;

			    case BADFORM:
			    case 0:
				break;		/* Oops, no error! */

			    case CMDERR:
				msg = Mbox;
				goto Fprint;

			    default:
				msg = Unclass;
			  Fprint:
				dkIOerr((struct buf *)NULL, iop, dn, msg);
				break;
			}
			break;
		}
	}
	else if (iop->zd_csr & ERROR)
	    switch (iop->zd_err) {
		case WRPROT:
			/*
			 * We don't set DKREADONLY here on a hard disk
			 * because we want the user to be able
			 * to change the write-protect button
			 * without having to close all references.
			 * We are not nearly as cautious with hard
			 * disks as we are with floppies and carts.
			 */
			if (!_hard(dmj))
				flagp->status |= DKREADONLY;
			msg = dk_prot;
			goto IMprint;

		case INVALID:
			ASSERT(!hard(dn));
			flagp->status |= DKERROR;
			msg = dk_chg;
		  IMprint:
			dkerror(dn, msg);
			goto errset;

		case SOFTERR:
			dkIOerr(bp, iop, dn, Soft);
			break;

		case NOTREADY:
		/*  M004:  Do retry here so we don't display NRDY msg  */
			if (_hard(dmj) && dp->b_errcnt++ < RETRIES) {
				dkustart(dp);	/*M004  Do retry here */
				return;
			}
			else {
				if (!_hard(dmj))
					flagp->status |= DKERROR;
				dkerror(dn, dk_nrdy);
				goto errset;
			}

		case 0:
			break;

		case HARDERR:
			msg = Hard;
			goto Iprint;

		case CMDERR:
			msg = Mbox;
			goto Iprint;

		default:
			msg = Unclass;
		  Iprint:
			dkIOerr(bp, iop, dn, msg);

		retry:
			/* retry if haven't exceeded limit else return error */
			if (dp->b_errcnt++ < RETRIES) {
				dkustart(dp);
				return;
			}
		errset:
			bp->b_flags |= B_ERROR;
	}
	dp->b_errcnt = 0;
	dp->b_actf = bp->av_forw;
	/*
	 * resid already set up by strategy routine,
	 * or with error code when initting disk.
	 */
	dbug(dkI,6,"*** dkintr iodone\n");
	iodone(bp);
	dkustart(dp);
}

dkread(dev)
{
	dbug2(dbANY,4,"\n*** dkread %s:%d\n",
			dkNames[dmaj(driveno(dev))],
			dmin(driveno(dev)));
	physio(dkstrategy, &rdkbuf[dmaj(driveno(dev))], dev, B_READ);
}

dkwrite(dev)
{
	dbug2(dbANY,4,"\n*** dkwrite %s:%d\n", 
			dkNames[dmaj(driveno(dev))],
			dmin(driveno(dev)));
	physio(dkstrategy, &rdkbuf[dmaj(driveno(dev))], dev, B_WRITE);
}

dkprint(dn)
  register int	dn;
{
	printf("%s Drive %d: ", dkNames[dmaj(dn)], dmin(dn));
}

dkerror(dn, msg)
  int	dn;
{
	dkprint(dn);
	printf("%s\n", msg);
}

/*
 * Cartridge Drive N: TYPE error XXXX while Writing Block NNNNN
 * Cartridge Drive N: TYPE error XXXX while Reading Block NNNNN
 * Cartridge Drive N: TYPE error XXXX while Formatting
 *
 * Hard Drive N: TYPE error XXXX while Reading Cylinder NNN Head N Sector N
 * Hard Drive N: TYPE error XXXX while Writing Cylinder NNN Head N Sector N
 *
 * Floppy Drive N: TYPE error XXXX while Reading Cylinder NNN Side N Sector N
 * Floppy Drive N: TYPE error XXXX while Writing Cylinder NNN Side N Sector N
 * Floppy Drive N: TYPE error XXXX while Formatting
 */
dkIOerr(bp, iop, dn, type)
  register struct zd_iopb *iop;
  struct buf	*bp;
  register int	dn;
  char		*type;
{
	/* bp is NULL if we are formatting */
	dkprint(dn);
	printf("%s error %x while %sing%c",
			type,
			iop->zd_diag,
			!bp? "Formatt"
			     : (bp->b_flags & B_READ)? "Read" 
			     : "Writ",
			!bp? '\n' : ' ');

	if (bp) {
		if (cart(dn))
			printf("Block %d\n", iop->zd_blk);
		else {
			printf("Cylinder %d %s %d Sector %d\n",
				iop->zd_cyl,
				hard(dn) ? "Head" : "Side",
				iop->zd_hd,
				iop->zd_sec);
		}
	}

	/* printf(" [Status: %x]\n", iop->zd_diag); */
}


/*ARGSUSED*/
dkioctl(dev, cmd, addr, flag)
  caddr_t addr;
{
	register struct dkflags	 *flagp;
	register int	dn;
	register short	err;
	int		s;

	dn = driveno(dev);
	flagp = &dkflags[dmaj(dn)][dmin(dn)];

	dbug2(dkIO,0, "*** %s Drive %d: ", dkNames[dmaj(dn)], dmin(dn));

	if (cmd != DKIOCFMT) {
		dbug1(dkIO,1,"dkioctl bad cmd = 0x%x\n", cmd);
		u.u_error = ENOTTY;
		return;
	}

	/* Can only format floppies and cartridges */
	if (hard(dn) || pseudo(dev) != BOOTPART) {
		dbug4(dkIO,1,"dkioctl can't format [hard(%d)=%s, pseudo(0x%x)=%d]\n",
					dn, hard(dn)? "TRUE" : "FALSE",
					dev, pseudo(dev) );
		u.u_error = ENODEV;
		return;
	}

	if (flagp->openfl[0] != (1 << pseudo(dev)) || flagp->openfl[1] != 0) {
		dbug(dkIO,1,"device busy; format denied");
		dbug2(dkIO,1,"  (openfl[0]=0x%x, openfl[1]=0x%x)\n",
						flagp->openfl[0],
						flagp->openfl[1]);
		u.u_error = EBUSY;
		return;
	}

	dbug(dkIO,2,"format drive\n");

	err = dkcommand(dev, dkFORMAT, (int)addr);
	flagp->status |= DKERROR;	/* Can't do anything after a FORMAT */

	switch (err) {
	    case WRPROT:
		u.u_error = EROFS;
		break;

	    case INVALID:
		u.u_error = EINVAL;
		break;

	    case HARDERR:
		u.u_error = EIO;
		break;

	    case 0:
	    case SOFTERR:
		u.u_error = 0;
		break;

	    case NOTREADY:
		u.u_error = ENXIO;
		break;

	    case BADFORM:
		u.u_error = ENOSPC;
		break;

	    case CMDERR:
	    default:
		u.u_error = EIO;
		break;
	}
}
