/*
 *	Memory special file
 *	minor device 0 is physical memory
 *	minor device 1 is kernel memory
 *	minor device 2 is EOF/RATHOLE
 *	minor device 3 is new proc reader  (returns struct w/proc data)
 *
 */

/*
 * Modifications
 *	JJD	01 Mar 83	M000
 *		- fixed bug in validity checking range of mem accesses
 */

/*
 *	This code is Radio-Shack specific -- in particular,
 *	it depends on the working Upage being part of the kernel's
 *	normal data space.  To modify for use on another system,
 *	the rangecheck code would have to be slightly modified.
 */

/*#define DEBUG	4		/* Compile in debug statments */
#define DbugVar	pr_debug
#define DEBUGS	dbALL

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/conf.h"
#include "../h/mmu.h"
#include "../h/buf.h"
#include "../h/var.h"
#include "../h/proc.h"
#include "../h/uproc.h"
#include "../h/dbug.h"

#define PHYSMEM 0             /* minor dev 0 is physical memory  /dev/mem    */
#define KERNMEM 1             /* minor dev 1 is kernel memory    /dev/kmem   */
#define EOFRATH 2             /* minor dev 2 is eof/rathole      /dev/null   */
#define GETPROC	3             /* minor dev 3 is process info     /dev/procs  */


/*
 *      open and close are not needed, use nulldev in devsw (see c.c)
 */


/*
 *      read/write eof or memory
 */

mmread(dev)
{
	if (minor(dev) == GETPROC)
		prread();
	else
		mmcomm(dev, 0);
}

mmwrite(dev)
{
	mmcomm(dev, 1);
}

mmcomm(dev, wrflag)
  register wrflag;
{
	register c = minor(dev);
	register count = u.u_count;
	register char *base = (char *)u.u_base;
	register char *offset = (char *) u.u_offset;
	register char *limit = offset + count;

	if (c == EOFRATH) {		/* rathole read returns 0 */
		if (wrflag)
			u.u_count = 0;
		return;
	}
	if (c >= GETPROC || offset > limit || limit > (c == KERNMEM ? umem : maxaddr))
		u.u_error = ENXIO;
	else {
		if (wrflag ? copyin(base, offset, count)
		   	   : copyout(offset, base, count))
			u.u_error = EFAULT;
		else {
			u.u_offset += count;
			u.u_base += count;
			u.u_count = 0;
		}
	}
}

/*
 *  /dev/procs
 */
#define UPSIZE (sizeof (struct uproc))
prread()
{
	register struct	proc	*pp;
	register struct uproc	*upp;
	register struct user	*up;
	register struct buf	*bp;
	int		slot;
	struct uproc	upr;

	slot = u.u_offset / UPSIZE;
	if (slot > v.v_proc || u.u_offset % UPSIZE != 0 || u.u_count % UPSIZE != 0) {
		u.u_error = ENXIO;
		return;
	}
	
	pp = &proc[slot];
	upp = &upr;

	for ( ; u.u_count; ++pp) {
		if (pp == v.ve_proc)
			return;			/* EOF */

		if (pp == u.u_procp)
			up = &u;		/* Looking at ourselves */
		else {
			bp = (struct buf *)NULL;
			if (pp->p_stat == 0 || pp->p_stat == SZOMB || pp->p_stat == SIDL) {
			    /* Empty slot, Zombie proc, or Uninitialized proc */
			    /* Empty slot */
				up = (struct user *)NULL;
			}
			else if (pp->p_flag & SLOAD) {
			    /* Process is in memory */
				up = (struct user *)mltoa(pp->p_addr.p_caddr);
			} 
			else {
			    /* Oh, bother; it's swapped out... */
				dbug4(dbANY,1, "*** prread: bread(%d/%d, %X + %X) ",
						major(swapdev),
						minor(swapdev),
						swplo,
						pp->p_addr.p_daddr);
				bp = bread(swapdev, swplo + pp->p_addr.p_daddr);
				if (bp->b_flags & B_ERROR) {
					dbug(dbANY,1, "ERROR\n");
					brelse(bp);
					bp = (struct buf *)NULL;
				} 
				else {
					dbug(dbANY,1, "OK\n");
					up = (struct user *)paddr(bp);
				}
			}
		}
		upp->up_flag = pp->p_flag;
		upp->up_stat = pp->p_stat;
		upp->up_uid = pp->p_uid;
		upp->up_pid = pp->p_pid;
		upp->up_ppid = pp->p_ppid;
		upp->up_wchan = pp->p_wchan;

		upp->up_name[0] = '\0';
		if (up != (struct user *)NULL) {
		    upp->up_hastty = (up->u_ttyp != (struct tty *)NULL);
		    upp->up_ttyd = up->u_ttyd;

		    /* sizeof (struct user) > BSIZE on this system.  *sigh*
		     *  So, to get at stuff in "2nd half" of the U when
		     *  the process is swapped out, we have to go get it,
		     *  and then kludge the U pointer.  *SIGH* */
		    if (bp != (struct buf *)NULL) {
			/* It's swapped out */
			dbug4(dbANY,1, "*** prread: bread(%d/%d, %X + %X) ",
					major(swapdev),
					minor(swapdev),
					swplo,
					pp->p_addr.p_daddr + 1);
			brelse(bp);
			bp = bread(swapdev, swplo + pp->p_addr.p_daddr + 1);
			if (bp->b_flags & B_ERROR) {
				dbug(dbANY,1, "ERROR\n");
				goto null;
			}
			dbug(dbANY,1, "OK\n");

			/* ========== *SUPER* Dirty =========> vvvvvvv */
			up = (struct user *)((char *)paddr(bp) - BSIZE);
		    }

		    copybuf(up->u_comm, upp->up_name, sizeof (upp->up_name));
		}
		else {
	null:
			upp->up_hastty = 0;
		}

		if (bp != (struct buf *)NULL)
			brelse(bp);

		if (copyout((caddr_t)upp, u.u_base, UPSIZE))
			u.u_error = EFAULT;
		else {
			u.u_offset += UPSIZE;
			u.u_base += UPSIZE;
			u.u_count -= UPSIZE;
		}
	}
}
