/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: /usr/sys/mdep/RCS/machdep.c,v 6.2 88/07/18 11:22:35 iv Exp $
 *
 * $Log:	machdep.c,v $
 * Revision 6.2  88/07/18  11:22:35  iv
 * 	- Fixed a couple of comments
 * 	- Fixed bug in code which puts all of memory (except the last two
 * 	  pages) into the free map.  It frees backwards, highest to lowest,
 * 	  and by using `>' instead of `>=', forgot to free the 1st page.  Oops.
 * 
 * Revision 6.1  88/07/15  09:57:38  iv
 * 	- Finally fixed clkfudg to be the same as the release.
 * 
 * Revision 6.0  87/10/12  17:07:31  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.5  87/05/06  15:09:29  iv
 * 	- Fixed bug where we calculated # texts ONLY if they were specified
 * 	  (# texts came up 0 if you left it at AutoConfig.  Oops.)
 * 	- Changed memory free code to save TWO pages for icode
 * 	  (icode needs a couple of pages saved for its use for safety)
 * 
 * Revision 5.4  86/11/12  14:56:30  iv
 * 	- Commoned up all the size calculations to one section which saves
 * 	  them in stack variables
 * 	- Added a text table size calculation
 * 	- Removed stupid printf() from cfginit() & set a flag instead
 * 	  startup() can complain just before it displays the memory sizes.
 * 
 * Revision 5.3  86/10/23  15:35:56  iv
 * 	- Added inodes/files/locks/procs/clists to what is AutoConfigurable
 * 	  Pretty crunchy.
 * 
 * Revision 5.2  86/10/01  17:45:04  iv
 * 	- Changed startup code to free last page of memory last;
 * 	  this means (hopefully) that icode[] will be placed in the
 * 	  last page of memory, and if the MMU can't go there, XENIX
 * 	  won't boot.
 * 	- Changed "System" memory display to %dk+%dk to show how many
 * 	  buffers were configured
 * 	- Added autoconfig code for buffers
 * 
 * Revision 5.1  86/08/28  17:58:17  iv
 * 	- Redid icode[] to put in new reboot icode
 * 	- Added ricodech() to change from normal boot to a reboot
 * 	- Cleaned up some code (changed from commented out to #ifdef NEVER)
 * 	- Spedup (hopefully) addupc()
 * 
 * Revision 5.0  85/12/03  17:09:42  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:45:44  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:46:37  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:17  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.6  84/12/21  13:41:37  iv
 * 	- Added automagic initialization of 'Hz' from Z80's reported clk speed
 * 	- Added clock fudge code to clkreld() and return of number of ticks
 * 	- Added routine rclock(), which accepts correct arguments from mch.S
 * 	  and then calls clock() properly
 * 
 * Revision 1.5  84/10/04  11:28:51  iv
 * 	- Fixed clock correction routine 
 * 
 * Revision 1.4  84/10/03  18:51:38  iv
 * 	- Re-wrote clkreld() to return the number of ticks that have passed
 * 	  since the last interrupt (from the Z80)
 * 
 * Revision 1.3  84/04/18  10:45:53  iv
 * 	- Fixed bad ref to "pagoff", which is a Lisa mmu variable.
 * 
 * Revision 1.2  84/04/11  09:57:44  iv
 * 	- Changed 'z80clock.h' to 'z80clk.h', which is its name
 * 
 * Revision 1.1  84/04/10  17:36:30  iv
 * -	Fixed clock routines "clkstart" and "clkreld" so they talk to the
 * 	proper z80 mailbox  (iopb)  to turn ON clock interrupts
 * 
 */


/*
 * Machine dependent code.
 *
 * 03/22/81 JGL
 *	Added icodech routine for auto file recovery
 * 12/23/81 JJD
 *	Mods for Grasshopper prototype -- greater machine independence.
 * 01/10/82 JJD
 *	Added DEBUG code.  Changed backup to get stack pointer from
 *	ar0[USP] rather than ar0[A7].
 *
 *	JJD	13 May 82	M004
 *		- added stkprobe() routine, changed backup().  See
 *		  comments below.
 *		- fixed bug (misplaced parenthesis) in addupc
 *		- replace [fs]uword in addupc() with [fs]ushort
 * 	JJD	27 Aug 82	M005
 *		- added dummy mapfree routine
 *	JJD	20 Sep 82	M006
 *		- modified for greater portability
 *	arw	31 Oct 82	M007
 *		- AL clock support
 *	JJD	24 Dec 82	M008
 *		- moved swap initialization from startup into iinit, in main.c
 *	JJD	26 Dec 82	M009
 *		- fixed handling of maxmem. c.c initializes Maxmem
 *		  in units of Kbytes; mmu.c treats maxmem as a page
 *		  count, for simplicity.
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/acct.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/inode.h"
#include "../h/file.h"
#include "../h/locking.h"
#include "../h/proc.h"
#include "../h/text.h"
#include "../h/map.h"
#include "../h/tty.h"
#include "../h/reg.h"
#include "../h/buf.h"
#include "../h/mmu.h"
#include "../h/var.h"
#include "../h/z80clk.h"
#include "../h/z80ser.h"	/* to get NZ80SER */


/*#define DEBUG	7	/* Compile in debugging statments */
#define DbugVar	md_debug
#ifdef DEBUG
#  define dbBI	0x0001		/* Debug cfginit() */
#endif /*DEBUG*/
#define DEBUGS	dbBI
#include "../h/dbug.h"


#define	ICODEHI		0	/* hi part of address where icode is copied */
#define	ICODELO		0	/* lo part of address where icode is copied */

/*
 * Icode is the hex bootstrap
 * program executed in user mode
 * to bring up the system.
 * M006
 */
short icode[] = {
	0x2E7C,	ICODEHI,ICODELO+0xFE,	/* 0000	start:	movl	#start+/FE,sp*/
	0x2F3C,	ICODEHI,ICODELO+0x2A,	/* 0006		movl	#envp,-(sp)  */
	0x2F3C,	ICODEHI,ICODELO+0x22,	/* 000C		movl	#argp,-(sp)  */
	0x2F3C,	ICODEHI,ICODELO+0x2E,	/* 0012 	movl	#name,-(sp)  */
	0x42A7,				/* 0018 	clrl	-(sp)	     */
	0x303C,	0x003B,			/* 001A 	movw	#59.,d0	     */
	0x4E40,				/* 001E		trap	#0	     */
	0x60FE,				/* 0020		jra	.	     */
	ICODEHI,ICODELO+0x2E,		/* 0022	argp:	.long	name	     */
	ICODEHI,ICODELO+0x38,		/* 0026		.long	reboot	     */
	0,	0,			/* 002A	envp:	.long	0	     */
	0x2F65,	0x7463,	0x2F69,		/* 002E	name:	.asciz	"/etc/init"  */
	0x6E69,	0x7400,			/* 0034				     */
	0x2D6E,	0			/* 0038	reboot:	.asciz	"-n"         */
					/* 003C				     */
};
unsigned short	szicode = sizeof(icode);

/*      icodech - change the icode program.
 *
 *      icodech changes the icode program to invoke /etc/inir
 *      instead of /etc/init.  `inir' is run when the system
 *      discovers that the root file structure is unsafe.
 *      `inir' causes the root structure to be checked, and
 *	then calls /etc/init.
 *
 *      ricodech changes the icode program to invoke /etc/init
 *	or /etc/inir with the argument -r to indicate an auto-
 *	boot instead of a manual boot.   ricodech()  is called
 *	from `z80init.c'.
 */

icodech()
{
	icode[27] = 0x7200;		/* make it /etc/inir */
}

ricodech()
{
	icode[28] = 0x2D72;		/* change to "-r" for auto-boot */
}



/*
 * Machine-dependent startup code
 */

/*#define TEST0   0x4df4          /* test -512.(sp) word zero */
/*#define TEST1   0xFE00          /* test -512.(sp) word one  */

short	system = 0, bufs = 0, bad_cfg = 0;
mloc_t	icodepg;

startup()
{
	register mloc_t	first, pg;
	short	rest;
	extern char *umem, *maxaddr;

	/*
	 * NOTE:
	 *	The assembler support (conf/mch.s) has placed the physical
	 * address of the start of the page following the last available
	 * page in the global maxaddr. (first un-addressable page)
	 * It also placed the address of the first user page in umem.
	 */

	mmuinit();	/* initalize maps (hardware and software) */

	/* initialize any device that need initialization */
	dinit();

	/*
	 * By freeing pages in reverse order, we build core free list in
	 * ascending order (on those systems which care).
	 */
	first = atoml(umem);
	icodepg = atoml(maxaddr) - 2;
	pg = icodepg - 1;
	/* Free all but two last pages (we're going to run icode there) */
	do
		mmufree(pg, 1);
	while ((unsigned) --pg >= (unsigned) first);

	bufbase = (paddr_t)&sabuf[0][0];

	cputype = 68000;

	if (bad_cfg)
		printf("WARNING: Bad configuration, resetting to defaults.\n");
	rest = ((unsigned)umem/1024) - (system  + bufs);
	printf("System %dk+%dk+%dk=%dk, User %dk\n",
		    system,
		    bufs,
		    rest,
		    system + bufs + rest,
		    maxmem = ((maxaddr-umem)/1024));
	if (Maxmem < maxmem)
		maxmem = Maxmem;
	maxmem = (int)btoms(maxmem*1024);	/* convert to page count */

	/* talk to memory boards */
	meminit();
}

struct mems {
	short	memsiz;
	short	nbufs;
	short	ninode;
	short	nfile;
	short	nlocks;
	short	nproc;
	short	nclist;
} cfgtbl[] = {	       /* Mem  bufs inode file lock proc clst */
    /* 3.0 meg */	{ 3072, 300, 200, 200, 400, 100, 200 },
    /* 2.5 meg */	{ 2560, 300, 160, 160, 320,  90, 200 },
    /* 2.0 meg */	{ 2048, 300, 140, 140, 280,  80, 200 },
    /* 1.5 meg */	{ 1536, 200, 120, 120, 240,  70, 200 },
    /* 1.0 meg */	{ 1024, 100, 100, 100, 200,  60, 200 },
    /* 0.5 meg */	{  512,  50, 100, 100, 200,  60, 200 },
    /* Spare */		{    0,  32,  75,  75, 150,  50, 150 },
    /* Spare */		{    0,  32,  75,  75, 150,  50, 150 },
/* Don't change this line: */
    /* Minimums */	{    0,  32,  60,  60,  20,  20,  50 }
};

extern caddr_t	maxaddr;
caddr_t
cfginit()
{
	register struct var	*vp;
	register struct mems	*mp, *lp;
	register caddr_t	ep, tp;
	register int		i, tmp;
	int	sbufs, sinodes, sfiles, slocks, sprocs, stexts, smaps, sclists;
	int	reset = 0;

	tmp = (int)maxaddr/1024;	/* System memory size in K */
	for (mp = cfgtbl; mp->memsiz; ++mp) {
		if (tmp >= mp->memsiz)
			break;
	}

	dbug1(dbBI,1, "Profile for %dk:", tmp);
	dbug3(dbBI,1, " %dk bufs  %d inode  %d file",
				mp->nbufs >> 1,
				mp->ninode,
				mp->nfile);
	dbug3(dbBI,1, "  %d lock  %d proc  %d clst\n",
				mp->nlocks,
				mp->nproc,
				mp->nclist);

	/* Check for AutoConfig on each tunable */
	vp = &v;
	lp = &cfgtbl[7];		/* Minimum limits */
	if (vp->v_sabuf < lp->nbufs) {
		dbug(dbBI,2, "AutoConfiging sabufs\n");
		vp->v_sabuf = mp->nbufs;
	}
	if (vp->v_inode < lp->ninode) {
		dbug(dbBI,2, "AutoConfiging inodes\n");
		vp->v_inode = mp->ninode;
	}
	if (vp->v_file < lp->nfile) {
		dbug(dbBI,2, "AutoConfiging files\n");
		vp->v_file = mp->nfile;
	}
	if (vp->v_lock < lp->nlocks) {
		dbug(dbBI,2, "AutoConfiging locks\n");
		vp->v_lock = mp->nlocks;
	}
	if (vp->v_proc < lp->nproc) {
		dbug(dbBI,2, "AutoConfiging procs\n");
		vp->v_proc = mp->nproc;
	}
	if (vp->v_clist < lp->nclist) {
		dbug(dbBI,2, "AutoConfiging clists\n");
		vp->v_clist = mp->nclist;
	}
	vp->v_text = (vp->v_proc * 2) / 3;

	system = ((unsigned)end)/1024;		/* Save for startup() */
	/* Available memory */
	tmp = (maxaddr - end) / 1024;
	dbug1(dbBI,2, "cfginit: Avail mem %dk,", tmp);

	/* Calculate usage (in bytes) for each table */
      calc:
	sbufs = (vp->v_sabuf * (sizeof (struct buf) + BSIZE));
	sinodes = (vp->v_inode * sizeof (struct inode));
	sfiles = (vp->v_file * sizeof (struct file));
	slocks = (vp->v_lock * sizeof (struct locklist));
	sprocs = (vp->v_proc * sizeof (struct proc));
	stexts = (vp->v_text * sizeof (struct text));
	smaps = ((vp->v_proc/2) * sizeof (struct map));
	sclists = (vp->v_clist * sizeof (struct cblock));

	/* Config'd memory (in K) */
	i = sbufs
	    + sinodes
	    + sfiles
	    + slocks
	    + sprocs
	    + stexts
	    + smaps + smaps		/* Coremap & Swapmap */
	    + sclists;
	i /= 1024;
	dbug1(dbBI,2, " Used mem %dk\n", i);

	if (!reset && tmp - i < 256) {
		/* Should have at least ~256k left to run programs */
		bad_cfg = 1;
		vp->v_sabuf = mp->nbufs;
		vp->v_inode = mp->ninode;
		vp->v_file = mp->nfile;
		vp->v_lock = mp->nlocks;
		vp->v_proc = mp->nproc;
		vp->v_text = (vp->v_proc * 2) / 3;
		vp->v_clist = mp->nclist;
		reset = 1;
		goto calc;
	}

	/* Start actual configuring */
	ep = (caddr_t)end;

	/* Point to space for hbufs */
	hbuf = (struct hbuf *)ep;
	i = 0; tmp = vp->v_sabuf;
	do {
		++i;
		tmp >>= 1;
	} while (tmp > 0);	/* Take log2(#-buffers) */
	tmp = 1 << i;
	if (vp->v_sabuf > tmp)
		tmp <<= 1;
	vp->v_hbuf = tmp;
	vp->v_hmask = tmp - 1;
	dbug2(dbBI,2, "         %d hbufs @ 0x%X\n", vp->v_hbuf, ep);
	ep += sizeof (struct hbuf) * tmp;

	/* Point to space for inodes */
	inode = (struct inode *)ep;
	dbug2(dbBI,2, "         %d inodes @ 0x%X\n", vp->v_inode, ep);
	ep += sinodes;
	vp->ve_inode = (struct inode *)ep;

	/* Point to space for files */
	file = (struct file *)ep;
	dbug2(dbBI,2, "         %d files @ 0x%X\n", vp->v_file, ep);
	ep += sfiles;
	vp->ve_file = (struct file *)ep;

	/* Point to space for locks */
	locklist = (struct locklist *)ep;
	dbug2(dbBI,2, "         %d locks @ 0x%X\n", vp->v_lock, ep);
	ep += slocks;

	/* Point to space for procs */
	proc = (struct proc *)ep;
	dbug2(dbBI,2, "         %d procs @ 0x%X\n", vp->v_proc, ep);
	ep += sprocs;
	vp->ve_proc = (struct proc *)ep;

	/* Point to space for texts */
	text = (struct text *)ep;
	dbug2(dbBI,2, "         %d texts @ 0x%X\n", vp->v_text, ep);
	ep += stexts;
	vp->ve_text = (struct text *)ep;

	/* Point to space for coremap */
	coremap = (struct map *)ep;
	dbug2(dbBI,2, "         %d coremap @ 0x%X\n", vp->v_proc/2, ep);
	ep += smaps;

	/* Point to space for swapmap */
	swapmap = (struct map *)ep;
	dbug2(dbBI,2, "         %d swapmap @ 0x%X\n", vp->v_proc/2, ep);
	ep += smaps;

	/* Point to space for cblocks */
	cfree = (struct cblock *)ep;
	dbug2(dbBI,2, "         %d cblocks @ 0x%X\n", vp->v_clist, ep);
	ttyhog = ((vp->v_clist - (NZ80SER * 4)) * CLSIZE) / 4;
	ep += sclists;

	/* Do buffers last so we can round # of buffers to nearest MMPGSZ */
	tp = ep + sbufs;
	dbug1(dbBI,2, "           Projected _end @ 0x%X\n", tp);
	tp = (caddr_t)(((unsigned)tp + (MMPGSZ-1)) & ~(MMPGSZ-1));
	dbug1(dbBI,2, "           Rounded _end @ 0x%X\n", tp);
	tmp = (tp - ep) / (sizeof (struct buf) + BSIZE);
	dbug1(dbBI,2, "           Room for %d buffers\n", tmp);
	vp->v_sabuf = tmp;
	if (vp->v_sabuf > vp->v_hbuf)
		vp->v_sabuf = vp->v_hbuf;
	bufs = vp->v_sabuf >> 1;		/* Save for startup() */

	/* Point to space for bufs */
	buf = (struct buf *)ep;
	dbug2(dbBI,2, "         %d bufs @ 0x%X\n", vp->v_sabuf, ep);
	ep += sizeof (struct buf) * vp->v_sabuf;

	/* Point to space for sabuf[][] */
	_sabufs = (struct sysbufs *)ep;
	dbug1(dbBI,2, "         sbufs @ 0x%X\n", ep);
	ep += BSIZE * vp->v_sabuf;

	/* Tell mch.S where new `_end' is */
	dbug1(dbBI,2, "         return (0x%X);\n", ep);
	return (ep);
}

/*
 * dinit
 *	Initialize devices.
 *
 * Called early in intialization process to allow devices to initialize
 * BEFORE interrupts come on.
 */

dinit()
{
	extern int (*dinitsw[])();
	register i;

	for(i = 0; dinitsw[i]; i++)
		(*dinitsw[i])();
}


/*
 * set up a physical address
 * into users virtual address space.
 */

 /*
  * CURRENTLY UNIMPLEMENTED ON PAGED 68000 VERSION
  */
sysphys()
{
#ifdef mmuNEVER
	register i, s;
	register struct a {
		int	segno;
		int	size;
		int	phys;
	} *uap;

	if(!suser())
		return;
	uap = (struct a *)u.u_ap;
	i = uap->segno;
	if(i < 0 || i >= 8)
		goto bad;
	s = uap->size;
	if(s < 0 || s > 128)
		goto bad;
    /******************************* PERHAPS NOT A GOOD FEATURE ******/
	return;
  bad:
	u.u_error = EINVAL;
#endif /*mmuNEVER*/
}

/*      Start the clock   */
extern	int	Hz;
clkstart()
{
	register int	zHz;

	zHz = z80clk.z_count;
	if (zHz == 25 || zHz == 30)
		Hz = z80clk.z_count;

	z80clk.z_clock |= CLKIENAB;
}

/*      Restart the clock and tell clock() number of ticks...  */
unsigned short	clkfudge = 1950;
clkreld()
{
	register int	s, lboltinc;
	static unsigned short	realticks = 0;

	lboltinc = 1;
	z80clk.z_clock &= ~CLKINT;

	/*
	 *  It's fudge time, boys and girls!  *sigh*
	 *
	 *  It seems that the Z80 CRTC cannot be programmed for exactly 30 hz.
	 *  The closest value we can get is slow by an amount which varies
	 *  depending on the variance in the parts in each machine!
	 */

	realticks = ++realticks % clkfudge;
	if (realticks == 0)
		++lboltinc;

	return (lboltinc);
}

/*
 *  sys/clock.c interface as our mch.s interface puts extra stuff on stack
 */
unsigned char	ointmask = 0;
/*ARGSUSED*/
rclock(intno, d0, d1, a0, a1, intmask, sr, pc)
  caddr_t	pc;
{
	ointmask = intmask;
	clock(pc, sr);
}

/*
 * Let a process handle a signal by simulating an interrupt
 */

/*ARGSUSED*/
sendsig(p, signo)
caddr_t p;
{
	register unsigned long n;

	n = u.u_ar0[USP] - 6;			/* decrement stack by 6 */
	sushort((caddr_t)n, u.u_ar0[SR]);	/* save status */
	suword((caddr_t)n+2, u.u_ar0[UPC]);	/* save user pc */
	u.u_ar0[USP] = n;			/* set user stack */
	u.u_ar0[SR] &= ~SR_TRACE;		/* set user status */
	u.u_ar0[UPC] = (int)p;			/* set user pc */
}

/* 
 * ***         RESTART INSTRUCTIONS 
 */
/* Begin M004 .... */

/*
 * Stkprobe checks if a memory fault was caused by a special stack
 * probe instruction.  If so, it returns the address to which the
 * stack should be grown so that the instruction would have executed
 * correctly.  It also does the backup work required to restart
 * the instruction.  Backup() is only called when stkprobe returns
 * 0 to indicate that it was not dealing with a stack probe.
 * On the current 68000 chip, backup() never succeeds.
 */
/*ARGSUSED*/
stkprobe(ar0)
register int *ar0;
{
#ifdef FIXEDSTACK
	return(0);			/* fixed size stacks */
#else
	register short *upc;
	register usp, probsiz;

	usp = ar0[USP];
	upc = (short *)ar0[UPC];
	if (fuishort(upc - 1) == 0x4a2f) {		/* tstb sp@(XX) */
		if ((probsiz = fuishort(upc)) > 0
	/* Compiler bug work around */
	/*	    && (short) probsiz < 0) {
	/* */
		    && (probsiz & 0x8000)) {
			usp += (short) probsiz;
			ar0[UPC] = (unsigned)(ar0[UPC]) - 2;
			return(usp);
		}
	}
	return(0);
#endif
}

/*ARGSUSED*/
backup(ar0)				/* return 1 if can't backup, else 0 */
int *ar0;				/* address of reg 0 on stack */
{
	return(1);
}
/* ... End M004 */

/* *** IMPLEMENT PROFILE */
/*
 *   Called in trap with:
 *      if(u.u_prof.pr_scale)
 *              addupc((caddr_t)pc, (int)&u.u_prof, (int)(u.u_stime-syst));
 *
 */

/*ARGSUSED*/
addupc(pc, profp, tim)
  caddr_t pc;             /* USER PC */
  int     profp;          /* Artifact from PDP11 version */
  int     tim;            /* incremental time */
{
	register ulong_t x;
	register unsigned y;

	/* increment the approiate word in user memory by tim */

	pc -= u.u_prof.pr_off;          /* subtract offset from pc */

	/* "fixed point fraction with binary point at left" */
	x = (ulong_t)((unsigned)pc/2) * (ulong_t)(u.u_prof.pr_scale/2);
	y = (unsigned)(x >> 14);

	y = (y+1)& ~1;	/* make sure even */

	if (y > u.u_prof.pr_size) return;        /* too big */
	sushort((caddr_t)(y + (unsigned)u.u_prof.pr_base),
		fushort((caddr_t)(y + (unsigned)u.u_prof.pr_base))+tim);
								/*M004*/
}

/*ARGSUSED*/
mapfree(bp)
struct buf *bp;
{
}

/* SAVE AND RESTOR FLOATING POINT */
/*ARGSUSED*/
savfp(fpregs)
int *fpregs;
{
}
#ifdef FPHARD
/*ARGSUSED*/
restfp(fpregs)
int *fpregs;
{
}
#endif /*FPHARD*/


/* machine dependent part of newproc */
/* create a duplicate process -- if not enough memory, swap current process */
procdup(rip,rpp)
    struct proc *rip, *rpp;
{
    mloc_t caddr;
	/*
	 * If there is not enough memory for the
	 * new process, swap out the current process to generate the
	 * copy.
	 */

	rpp->p_ssize = rip->p_ssize;
	if((caddr=mmupget(rpp))==MMUERR) {
		rip->p_stat = SIDL;
		rpp->p_addr.p_caddr = rip->p_addr.p_caddr; /* use moms u page */
		givemap(rip, rpp);      /* turn mapping over to child */
		xswap(rpp, 0, rpp->p_size);
		rip->p_stat = SRUN;
		u.u_procp = rip;                /* back to mom process */
		givemap(rpp, rip);      /* return mapping to parent */
		/*
		 * Make sure we didn't lose our map.
		 */
		mmumap();
	} else {
		/*
		 * There is core, so just copy.
		 */
		rpp->p_addr.p_caddr = caddr;
		mmupcopy(rip,rpp);
		u.u_procp = rip;
	}
}

/*
 *	MEMINIT - machine dependent memory init
 *
 *	This is called once after all of memory has been written into 
 *	it should do whatever it can to memory hardware
 */

#ifndef meminit
meminit()
{
}
#endif /*meminit*/

/*
 *	PARITY	- what to do when you get memory parity
 */

parity()
{
	panic("memory parity");
}
