/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: /usr/sys/mdep/RCS/main.c,v 6.2 88/08/17 15:14:21 root Exp $
 *
 * $Log:	main.c,v $
 * Revision 6.2  88/08/17  15:14:21  root
 * 	- Changed conditional names to differentiate main() DEBUG code
 * 	  and Kernel DEBUGger support
 * 	  (no code changes)
 * 
 * Revision 6.1  88/07/18  12:50:44  iv
 * 	- Changed define DEBUG to KDEBUG for kernel-debugger
 * 
 * Revision 6.0  87/10/12  17:07:37  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.2  87/05/06  15:16:38  iv
 * 	- Changed startup to have icode run in the mem pages saved for it
 * 	by startup() for safety.
 * 
 * Revision 5.1  86/08/28  18:05:41  iv
 * 	- Changed main() to put 2nd part of Legal[] out with puts(),
 * 	  which doesn't save stuff in msgbuf[].
 * 
 * Revision 5.0  85/12/03  17:09:52  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.1  85/08/22  17:26:40  iv
 * 	- Changed calls to d_open() to give correct READ/WRITE flags
 * 
 * Revision 4.0  85/03/13  14:45:51  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:46:43  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:23  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.3  84/12/21  14:26:54  iv
 * 	- Changed debugging stuff for TRS-16 kernel
 * 	  Misc pretty-ups on debug printf's
 * 	- Added initialization of u.u_addr
 * 	- Changed "tasktime(0)" to "tasktime()" so it'll call spl0()
 * 	- Added missing 3rd argument to the 3 ...d_open() calls 
 * 	- Added code to binit() under "#ifndef BUFMAPOUT" to check to make
 * 	  sure that no mapped out buffers have been configured in
 * 
 * Revision 1.2  84/04/20  17:28:41  iv
 * 	- Added debug check in case Legal[] doesn't checksum
 * 
 * Revision 1.1  84/04/10  17:37:56  iv
 * -	Moved the legal notice "Legal[]" into "version.c"
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/map.h"
#include "../h/mmu.h"
#include "../h/proc.h"
#include "../h/inode.h"
#include "../h/conf.h"
#include "../h/buf.h"
#include "../h/tty.h"
#include "../h/var.h"
#include "../h/file.h"
#include "version.c"

#ifdef	KDEBUG
int	debugmain = 0;
int     debugsw,debugsm,debugmm,debugwini;
int     debugsc;
#endif /*KDEBUG*/


/*
 * Initialization code.
 * Called from cold start routine as
 * soon as a stack and segmentation
 * have been established.
 *
 * Functions:
 *	clear and free user core
 *	turn on clock
 *	hand craft 0th process
 *	call all initialization routines
 *	fork - process 0 to schedule
 *	     - process 1 execute bootstrap
 *
 * loop at low address in user mode -- /etc/init
 *	cannot be executed.
 *
 *      NOTE that interrupts are turned off (spl7) upon entry,
 *      and stay turned off until the system controllers
 *      have been setup.  Some of the periph chips must be programmed
 *      with their vectors, if they somehow generate an interrupt
 *      before that, it will go to a garbage vector.
 *
 * main - main 'c' entry point for XENIX Kernel
 *
 *      At this point, quite a bit of work has alreay been done:
 *              1) the kernel has been entirely loaded into RAM
 *              2) the kernel map has been setup, mapping is turned on
 *
 */

main()
{
	char *cp;	/* character ptr for checksuming legal notice M013 */
	int sum;	/* checksum for legal string M013		   */
	extern char	*exenix;
	extern mloc_t	icodepg;	/* Saved startup location */

	printf(Legal);
	puts(&Legal[LOFFS]);
#ifdef KDEBUG
	printf("MAIN: debug is ON\n");
	debug();                      /* hook for setting breakpoints */
#endif /*KDEBUG*/
	startup();

	/*
	 * set up system process
	 */

	/*
	 * The assembler support has set exenix as the physical address
	 * of the system U.
	 */
	proc[0].p_addr.p_caddr = atoml(exenix);
	proc[0].p_size =  USIZE;
	proc[0].p_ssize = 0;
	proc[0].p_stat = SRUN;
	proc[0].p_flag |= SLOAD|SSYS;
	proc[0].p_nice = NZERO;
	u.u_procp = &proc[0];
	u.u_cmask = Cmask;
	u.u_addr = (caddr_t)&u;
	u.u_fpused = 0;
	u.u_limit = CDLIMIT;

	/*
	 * Checksum the copyright notice to make sure no one
	 * has zapped it.	M013
	 */
	for (sum = 0, cp = Legal; cp < &Legal[sizeof (Legal)]; cp++) {
		sum += *cp + 3;
	}
#ifndef DEBUG
	while (sum != LEGALSUM)
		;
#else /*DEBUG*/
	if (sum != LEGALSUM) {
		printf("MAIN: Copyright notice did not checksum;\n");
		printf("sum = %d/0x%X and LEGALSUM = %d/0x%X\n",
				sum, sum, LEGALSUM, LEGALSUM);
	}
#endif /*DEBUG*/

	/*
	 * Initialize devices and
	 * set up 'known' i-nodes
	 */

	cinit();
	binit();
	clkstart();		/* turn on the clock */
#ifdef DEBUG
	printf("turning on interrupts\n");
#endif /*DEBUG*/
/* WAS:	tasktime(0);            /* on interrupts */
	tasktime();            /* on interrupts */
#ifdef DEBUG
	printf("interrupts on!\n");
#endif /*DEBUG*/
	iinit();
	mfree(swapmap, nswap, 1);
	swplo--;

	rootdir = iget(rootdev, (ino_t)ROOTINO);
	rootdir->i_flag &= ~ILOCK;
	u.u_cdir = iget(rootdev, (ino_t)ROOTINO);
	u.u_cdir->i_flag &= ~ILOCK;
	u.u_rdir = NULL;

	/*
	 * make init process
	 * enter scheduling loop
	 * with system process
	 */
	if(newproc()) {
#ifdef DEBUG
		printf ("Main returned from newproc\n");
#endif /*DEBUG*/
		/*expand(USIZE + (int)btoms(szicode+255));*/
		if (!save(u.u_ssav)) 
		{
			register struct proc	*p;

			p = u.u_procp;
			/* Copy U to our reserved page */
			mmucopy(_paddr(p), icodepg, USIZE);
			/* Free the memory newproc() gave us, 'cause we're 
			 * going to use the two pages that startup() saved
			 * for us... */
			mmupfree(p);
			_paddr(p) = icodepg;
			p->p_size = USIZE + 1;
			/* Gotta do a resume() here, 'cause we want
			 * to run on the new U page... */
			resume(_paddr(p), u.u_ssav);
		}
		else {
			/* resume here: */
			mmumap();
			/* Don't clear the U, just the next page */
			mmuclear(_paddr(u.u_procp) + 1, (msize_t) 1); 
		}

		u.u_tsize = 0;
		u.u_dsize = btoms(szicode);
		u.u_ssize = 1;
		mmuset(RO);	/* force set of page reg */
		copyout((caddr_t)icode, DATASTART(0, 0), (int) szicode);
		/*
		 * Return goes to loc. 0 of user init
		 * code just copied out.
		 */
		return;
	}
#ifdef DEBUG
	printf ("Main going to sched\n");
#endif /*DEBUG*/
	sched();
}

/*
 * iinit is called once (from main)
 * very early in initialization.
 * It reads the root's super block
 * and initializes the current date
 * from the last modified date.
 *
 * icodech is called if the s_clean flag in the superblock shows
 * that the system was not properly brought down the last time.
 *
 *      Note that normally, we write the superblock unclean flag to the
 * disk immediately after a mount.  However, in the case of the root
 * filesystem, since we may be playing around, having troubles, or
 * otherwise be up and down a lot, we'll set unclean in the root
 * core superblock, but not flush through to the disk.  If he does any
 * meaningful things to the root, the unclean flag will quickly find its
 * way to the disk.
 *
 * panic: iinit -- cannot read the super
 * block. Usually because of an IO error.
 */
long    bootime;
iinit()
{
	register struct buf *bp;
	register struct filsys *fp;

#ifdef DEBUG
	printf("iinit: opening up root, pipe & swap devs\n");
#endif /*DEBUG*/
	(*bdevsw[major(rootdev)].d_open)(minor(rootdev), FREAD|FWRITE, 1);
	(*bdevsw[major(pipedev)].d_open)(minor(pipedev), FREAD|FWRITE, 1);
	(*bdevsw[major(swapdev)].d_open)(minor(swapdev), FREAD|FWRITE, 1);

#ifdef DEBUG
	printf("iinit: about to read in superblock\n");
#endif /*DEBUG*/
	bp = bread(rootdev, SUPERB);
	if(u.u_error)
		panic("iinit");
#ifdef DEBUG
	printf("iinit: successfully read superblock\n");
#endif /*DEBUG*/
	mount[0].m_bufp = getablk(1);
	fp = (struct filsys *)paddr(mount[0].m_bufp);
	copyio(paddr(bp), fp, sizeof(struct filsys), U_RKD);
	if(fp->s_magic!=S_S3MAGIC)
	    cvtv7superb(fp);
	brelse(bp);
	mount[0].m_flags = MINUSE;
	mount[0].m_dev = rootdev;
	fp->s_flock = 0;
	fp->s_ilock = 0;
	fp->s_ronly = 0;
	fp->s_ninode = 0;
	fp->s_inode[0] = 0;
	if (fp->s_clean != S_CLEAN)
		icodech();
	fp->s_clean = 0;                /* flag unclean */
	fp->s_fmod++;                   /* flag modified */
	time = fp->s_time;              /* init will sync */
	bootime = time;                 /* from UCB */
	printf("Root %Dk Swap %Dk\n",
		((unsigned) (fp->s_fsize * BSIZE))/1024,
		((unsigned) (nswap * BSIZE))/1024);
}

struct chead cfreelist;

/*
 * Initialize clist by freeing all character blocks.
 */
cinit()
{
	register n;
	register struct cblock *cp;

	for(n = 0, cp = &cfree[0]; n < v.v_clist; n++, cp++) {
		cp->c_next = cfreelist.c_next;
		cfreelist.c_next = cp;
	}
	cfreelist.c_size = CLSIZE;
}

/*
 * This is the set of buffers proper, whose heads
 * were declared in buf.h.  There can exist buffer
 * headers not pointing here that are used purely
 * as arguments to the I/O routines to describe
 * I/O to be done-- e.g. swbuf for
 * swapping.
 */

/*
 * Initialize the buffer I/O system by freeing
 * all buffers and setting all device buffer lists to empty.
 */
binit()
{
	register struct buf *bp;
	register struct buf *dp;
	register unsigned i;
	struct bdevsw *bdp;
	paddr_t nbase;

#ifndef BUFMAPOUT
	if (v.v_buf != 0)
		panic("binit");
#endif /*BUFMAPOUT*/

	dp = &bfreelist;
	dp->b_forw = dp->b_back =
	    dp->av_forw = dp->av_back = dp;
	nbase = bufbase;
	for (i=0, bp=buf; i<v.v_buf+v.v_sabuf; i++, bp++) {
		bp->b_dev = NODEV;
		if (i<v.v_sabuf)
			bp->b_paddr = (paddr_t)sabuf[i];
		else {
			bp->b_paddr = nbase;
			nbase += BSIZE+BSLOP;
		}
		bp->b_back = dp;
		bp->b_forw = dp->b_forw;
		dp->b_forw->b_back = bp;
		dp->b_forw = bp;
		bp->b_flags = B_BUSY;
		bp->b_bcount = BSIZE;
		brelse(bp);
	}

	for (i=0, bdp = bdevsw; i<bdevcnt; i++, bdp++) {
		dp = (struct buf *)bdp->d_tab;
		if (dp) {
			dp->b_forw = dp;
			dp->b_back = dp;
		}
	}
	for (i=0; i < v.v_hbuf; i++ )
		hbuf[i].b_forw = hbuf[i].b_back = (struct buf *)&hbuf[i];
}
