/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: preadi.c,v 6.0 87/10/12 17:07:49 iv Exp $
 *
 * $Log:	preadi.c,v $
Revision 6.0  87/10/12  17:07:49  iv
	- Checkin to lock down 3.2

Revision 5.1  86/08/28  14:27:10  iv
Initial revision

 */

#ifdef PREADI

#include "../h/param.h"
#include "../h/inode.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/proc.h"
#include "../h/systm.h"


/*#define DEBUG	7		/* Include debugging statements */
#ifdef DEBUG
#  define DbugVar  ex_debug
#  define dbPR	0x0001			/* Debug preadi() code */
#endif /*DEBUG*/
#define DEBUGS	dbALL
#include "../h/dbug.h"


/*
 * Read the file corresponding to the inode
 * pointed at by the argument into real memory
 * instead of pulling it in through buffers.
 * The actual read arguments are found
 * in the variables:
 *	u_base		core address for destination
 *	u_offset	byte offset in file
 *	u_count		number of bytes to read
 *	u_segflg	read to kernel/user/user I
 */
struct buf	*pbufp;
preadi(ip)
  register struct inode *ip;
{
	register unsigned  on, n;
	register	   type;
	register daddr_t   lbn, bn;
	dev_t	dev;
	off_t	diff;

	if (u.u_count == 0)
		return;
	if (u.u_offset < 0) {
		u.u_error = EINVAL;
		return;
	}

	dev = (dev_t)ip->i_rdev;
	type = ip->i_mode&IFMT;
	if (type != IFREG) {
		readi(ip);
		return;
	}

	pbufp = getablk(1);
	/* Save real paddr of buf */
	pbufp->av_back = (struct buf *)pbufp->b_paddr;

	do {
		lbn = bn = u.u_offset >> BSHIFT;
		on = u.u_offset & BMASK;
		n = min((unsigned)(BSIZE-on), u.u_count);
		diff = ip->i_size - u.u_offset;
		if (diff <= 0)
			break;
		if (diff < n)
			n = diff;
		bn = bmap(ip, bn, B_READ);
		if (u.u_error)
			break;
		dev = ip->i_dev;
		if ((long)bn < 0)
			pclear(n);
#ifdef PREADA
		else if (ip->i_lastr+1 == lbn && (on+n) == BSIZE)
			preada(dev, bn, on, n, rablock);
#endif /*PREADA*/
		else
			pread(dev, bn, on, n);
		ip->i_flag |= IACC;
	} while (u.u_error==0 && u.u_count!=0 && n!=0);

	/* Put real paddr of buf back */
	pbufp->b_paddr = (paddr_t)pbufp->av_back;
	brelse(pbufp);
}

pread(dev, blkno, offs, bytes)
  dev_t		dev;
  daddr_t	blkno;
  unsigned	offs, bytes;
{
	register struct buf	*bp;

	dbug5(dbPR,2, "*** pread(%d/%d, blk 0x%X, offs %d, %d bytes)\n",
					major(dev),
					minor(dev),
					blkno,
					offs,
					bytes);

	if (bytes < BSIZE && offs == 0) {
		/* SIGH.  Do it the slow way. */
		dbug3(dbPR,3, "    bread(%d/%d, blk 0x%X)\n",
					major(dev),
					minor(dev),
					blkno);
		bp = bread(dev, blkno);
		if (bp->b_resid == 0) {
			dbug2(dbPR,3, "    pimove(0x%X, %d, B_READ)\n", 
							paddr(bp),
							bytes);
			pimove(paddr(bp), bytes, B_READ);
		}
		brelse(bp);
		return;
	}

	bp = pbufp;
	bp->b_paddr = 
		(u.u_segflg == 1)? (paddr_t)u.u_base : umap(u.u_base, BSIZE);
	bp->b_bcount = 1;
	bp->b_blkno = blkno;
	bp->b_flags = B_BUSY | B_PHYS | B_READ;
	bp->b_dev = dev;
	dbug4(dbPR,3, "    physread(%d/%d, addr 0x%X, blk 0x%X)\n",
					major(dev),
					minor(dev),
					bp->b_paddr,
					bp->b_blkno);

	(*bdevsw[major(dev)].d_strategy)(bp);
	iowait(bp);

	if (bp->b_resid) {
		dbug1(dbPR,0, "*** pread: bad read; error = %d\n", u.u_error);
		return;		/* Oh, oh... */
	}

	/*
	 *  If there's a read offset, we go ahead and read the whole block
	 *  and then move it back from the offset point, because that's still
	 *  faster than allocating a block, reading, and copying the whole
	 *  block; because we're just copying part of one this way.
	 */
	if (offs > 0) {
		dbug3(dbPR,3, "    copybuf(from 0x%X, to 0x%X, %d)\n",
						bp->b_paddr + offs,
						bp->b_paddr,
						bytes);
		/* Use bp->b_paddr because it's already been translated
		 * if it was a user address and not a kernel address */
		if (copybuf(bp->b_paddr + offs, bp->b_paddr, bytes)) {
			dbug(dbPR,0,"*** pread: bad copy; return (EFAULT)\n");
			u.u_error = EFAULT;
			return;
		}
	}
	
	u.u_base += bytes;
	u.u_offset += bytes;
	u.u_count -= bytes;
}

pclear(bytes)
  register int	bytes;
{
	dbug2(dbPR,3, "    clearbuf(from 0x%X, %d)\n", u.u_base, bytes);
	clearbuf(u.u_base, bytes);
	
	u.u_base += bytes;
	u.u_offset += bytes;
	u.u_count -= bytes;
}

#endif /*PREADI*/
