/*
 *	%Z% %M% %I% %D% %Q%
 */


/*
 * THIS FILE CONTAINS CODE WHICH IS SPECIFIC TO THE
 * M68000 CPU. IT MAY REQUIRE CHANGING WHEN MOVING
 * XENIX TO A MACHINE CONTAINING A DIFFERENT CPU.
 *
 * $Header: ptrace.c,v 6.0 87/10/12 17:07:53 iv Exp $
 *
 * $Log:	ptrace.c,v $
 * Revision 6.0  87/10/12  17:07:53  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/12/03  17:10:23  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:46:11  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:46:59  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:36  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.2  84/04/11  10:57:47  iv
 * 	- Got rid of extra comment trailer (syntax error)
 * 
 * Revision 1.1  84/04/10  17:40:00  iv
 * 	- Initial checkin for TRS-XENIX 2.0
 * 
 */



#include "../h/param.h"
#include "../h/proc.h"
#include "../h/text.h"
#include "../h/mmu.h"
#include "../h/reg.h"
#include "../h/dir.h"   /* for user.h */
#include "../h/user.h"
#include "../h/inode.h"
#include "../h/systm.h"
#include "../h/var.h"

#define IPCPRI  PZERO   /* priority for tracing */
#define INTBOUND 2      /* boundary at which an integer must start */

/*
 * Tracing variables.
 * Used to pass trace command from
 * parent to child being traced.
 * This data base cannot be
 * shared and is locked
 * per user.
 */
struct
{
	int	ip_lock;
	int	ip_req;
	int     *ip_addr;
	int	ip_data;
} ipc;

/*
 * sys-trace system call.
 */
ptrace()
{
	register struct proc *p;
	register struct a {
		int	data;
		int	pid;
		int	*addr;
		int	req;
	} *uap;

	uap = (struct a *)u.u_ap;
#ifdef PTBUG
	if (pt_debug)			/*M001*/
printf("ptrace(data=%X pid=%X addr=%X req=%X)\n",uap->data,uap->pid,uap->addr,uap->req);
#endif
	if (uap->req <= 0) {
		u.u_procp->p_flag |= STRC;
		return;
	}
	for (p=proc; p < v.ve_proc; p++)
		if (p->p_stat==SSTOP
		 && p->p_pid==uap->pid
		 && p->p_ppid==u.u_procp->p_pid)
			goto found;
	u.u_error = ESRCH;
	return;

    found:
	while (ipc.ip_lock)
		sleep((caddr_t)&ipc, IPCPRI);
	ipc.ip_lock = p->p_pid;
	ipc.ip_data = uap->data;
	ipc.ip_addr = uap->addr;
	ipc.ip_req = uap->req;
	p->p_flag &= ~SWTED;
	setrun(p);
	while (ipc.ip_req > 0)
		sleep((caddr_t)&ipc, IPCPRI);
	u.u_rval1 = ipc.ip_data;
	if (ipc.ip_req < 0)
		u.u_error = EIO;
	ipc.ip_lock = 0;
	wakeup((caddr_t)&ipc);
}

/*
 * Code that the child process
 * executes to implement the command
 * of the parent process in tracing.
 */
procxmt()
{
	register int i;
	register *p;
	register struct text *xp;

	if (ipc.ip_lock != u.u_procp->p_pid)
		return(0);
	i = ipc.ip_req;
	ipc.ip_req = 0;
	wakeup((caddr_t)&ipc);
#ifdef PTBUG
if (pt_debug)					/*M000*/
printf("procxmt(i=%d)\n",i);
#endif
	switch (i) {

	/* read user I */
	case 1:
		if (fuibyte((caddr_t)ipc.ip_addr) == -1)
			goto error;
		ipc.ip_data = fuiword((caddr_t)ipc.ip_addr);
		break;

	/* read user D */
	case 2:
		if (fubyte((caddr_t)ipc.ip_addr) == -1)
			goto error;
		ipc.ip_data = fuword((caddr_t)ipc.ip_addr);
		break;

	/* read u */
	case 3:
		i = (int)ipc.ip_addr;
		if (i<0 || i > mstob(USIZE) - NBPW)
			goto error;
		ipc.ip_data =  *(int *)((((unsigned)&u) + i) & ~(INTBOUND-1));

		break;

	/* write user I */
	/* Must set up to allow writing */
	case 4:
		/*
		 * If text, must assure exclusive use
		 */
		if (xp = u.u_procp->p_textp) {
			if (xp->x_count!=1 || xp->x_iptr->i_mode&ISVTX)
				goto error;
			xp->x_iptr->i_flag &= ~ITEXT;
		}
		mmuset(RW);
		i = suiword((caddr_t)ipc.ip_addr, 0);
		suiword((caddr_t)ipc.ip_addr, ipc.ip_data);
		mmuset(RO);
		if (i<0)
			goto error;
		if (xp)
			xp->x_flag |= XWRIT;
		break;

	/* write user D */
	case 5:
		if (suword((caddr_t)ipc.ip_addr, 0) < 0)
			goto error;
		suword((caddr_t)ipc.ip_addr, ipc.ip_data);
		break;

	/* write u */
	case 6:
		i = (int)ipc.ip_addr;
		p = (int *) ((((unsigned)&u) + i) & ~(INTBOUND - 1));
		if (p >= (int *)&u.u_fps && p < ((int *)&u.u_fps +
				sizeof u.u_fps) )
			goto ok;
		if (p == &u.u_ar0[SR]) {
			ipc.ip_data &= ~SR_SUPER;       /* assure user space*/
			ipc.ip_data &= ~SR_IMASK;	/* priority 0 */
			goto ok;
		}
		for (i=0; i<reglocc; i++)
			if (p == &u.u_ar0[regloc[i]])
				goto ok;
		goto error;

	ok:
		*p = ipc.ip_data;
		break;

	/* set signal and continue */
	/*  one version causes a trace-trap */
	case 9:
		u.u_ar0[SR] |= SR_TRACE;

	case 7:
		if ((int)ipc.ip_addr != 1)
			u.u_ar0[UPC] = (int)ipc.ip_addr;
		u.u_procp->p_sig = 0;
		if (ipc.ip_data)
			psignal(u.u_procp, ipc.ip_data);
		return(1);

	/* force exit */
	case 8:
		exit(fsig(u.u_procp));

	default:
	error:
		ipc.ip_req = -1;
	}
	return(0);
}
