/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: text.c,v 6.0 87/10/12 17:08:06 iv Exp $
 *
 * Modifications
 *	JJD	12 Sep 81	M000
 *		Various changes to maintain consistency and transparency
 *		in routines manipulating text structures.
 *	JJD	26 Oct 81	M001
 *		If the scheduler skipped over a process with locked
 *		text which was otherwise ready to be swapped in,
 *		wake up the scheduler when unlocking text.
 *		Additionally, minor change to xalloc, with explanatory comment.
 *	JJD	20 Dec 81	M002
 *		Modifications to interface with new definitions of machine
 *		dependent portions of kernel.
 *	JJD	12 May 82	M003
 *		Adjust allocated text size to account for instruction prefetch.
 *	JJD	20 Sep 82	M004
 *		Adapted for contiguous allocation kernel.
 *	JJD	 1 Nov 82	M005
 *		Changed references to &text[NTEXT] into v.ve_text, for
 *		configurability.
 *		Changed occurrences of MSIZEZERO (M002) back to 0, because
 *		mmu system for this kernel depends upon their equivalence.
 *	JJD	14 Nov 82	M006
 *		Modified to support new x.out format.
 *
 * $Log:	text.c,v $
 * Revision 6.0  87/10/12  17:08:06  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.2  87/08/26  12:16:38  iv
 * 	- Changed some #defines for sections in text debug.
 * 	- Added some swap debug
 * 
 * Revision 5.1  86/08/28  14:07:16  iv
 * 	- Changed debugging messages over to new dbug.h package
 * 	- General readability overhaul.  No actual code changes.
 * 
 * Revision 5.0  85/12/03  17:10:47  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:46:27  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:47:14  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:51  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.1  84/04/10  17:41:50  iv
 * -	Initial checkin for TRS-XENIX 2.0
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/map.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/text.h"
#include "../h/inode.h"
#include "../h/buf.h"
#include "../h/mmu.h"
#include "../h/var.h"


/*#define DEBUG	7		/* Include debugging statements */
#ifdef DEBUG
#  define DbugVar  tx_debug
#  define txAF	0x0001			/* Trace attach/free code */
#  define txLK	0x0002			/* Trace lock/unlock code */
#  define txEX	0x0004			/* Trace xexpand() code */
#  define DEBUGS  (txEX)
#endif /*DEBUG*/
#include "../h/dbug.h"

/*
 * relinquish use of the shared text segment
 * of a process.        MUST MESS WITH SEGMENT ALLOCATION HERE TOO
 */
xfree()
{
	register struct text *xp;
	register struct inode *ip;

	if ((xp=u.u_procp->p_textp) == NULL)
		return;
	xlock(xp);
	xp->x_flag &= ~XLOCK;
	dbug2(2,txAF, "F-%d:%x ", u.u_procp->p_pid, xp);
	u.u_procp->p_textp = NULL;
	/*
	 * M000
	 * Tsize must be made to reflect lack of text structure, so that
	 * mmuset will never try to set up non-existent text pages.
	 */
	u.u_tsize = 0;
	ip = xp->x_iptr;
	if (--xp->x_count==0 && (ip->i_mode&ISVTX)==0) {
		xp->x_iptr = NULL;
		mfree(swapmap,(int)mstod(xp->x_size),(int)xp->x_daddr);
		mmufree(xp->x_caddr, xp->x_size);
		ip->i_flag &= ~ITEXT;
		if (ip->i_flag&ILOCK)
			ip->i_count--;
		else
			iput(ip);
	} else
		xccdec(xp);
}

/*
 * Attach to a shared text segment.
 * If there is no shared text, just return.
 * If there is, hook up to it:
 * if it is not currently being used, it has to be read
 * in from the inode (ip); the written bit is set to force it
 * to be written out as appropriate.
 * If it is being used, but is not currently in core,
 * a swap has to be done to get it back.
 */
xalloc(ip)
register struct inode *ip;
{
	register struct text *xp;
	register unsigned ts;
	register struct text *xp1;
	register msize_t ds, ss; /* dsize, ssize transparency */
	register int sep;	 /* sep transparency */

	if (u.u_exdata.x_text == 0)
		return;
	xp1 = NULL;
	for (xp = &text[0]; xp < v.ve_text; xp++) {
		if (xp->x_iptr == NULL) {
			if (xp1 == NULL)
				xp1 = xp;
			continue;
		}
		if (xp->x_iptr == ip) {
			xlock(xp);
			xp->x_count++;
			u.u_procp->p_textp = xp;
			if (xp->x_ccount == 0)
				/*
				 * M001
				 * Because xexpand unlocks the given text
				 * structure, there is no need to call xunlock
				 * upon return from the xexpand.  Therefore,
				 * the call on xunlock below has been moved into
				 * the second branch of the conditional;
				 * it used to follow the conditional.
				 */
				xexpand(xp);
			else {
				xp->x_ccount++;
				xunlock(xp);
			}
			/*
			 * Restore consistency.			M000
			 */
			u.u_tsize = xp->x_size;
			return;
		}
	}
	if ((xp=xp1) == NULL) {
		printf("out of text");
		psignal(u.u_procp, SIGKILL);
		return;
	}
	xp->x_flag = XLOAD|XLOCK;
	xp->x_count = 1;
	xp->x_ccount = 0;
	xp->x_iptr = ip;
	ip->i_flag |= ITEXT;
	ip->i_count++;
	ts = btoms(u.u_exdata.x_text + TEXTPAD);
	xp->x_size = ts;
	if ((xp->x_daddr = malloc(swapmap, (int)mstod(ts))) == NULL)
		panic("out of swap space");
	u.u_procp->p_textp = xp;
	xexpand(xp);
	u.u_tsize = ts;

	ds = u.u_dsize;
	ss = u.u_ssize;
	sep = u.u_sep;
	u.u_dsize = 0;
	u.u_ssize = 0;
	/* u.u_sep = 1; no i/d on 68000 */
	u.u_sep = 0;
	mmuset(RW);             /* set up memory management so I/O can write */
	u.u_count = u.u_exdata.x_text;
	u.u_offset = (sizeof u.u_exdata) + u.u_exdata.x_ext;
	u.u_base = TEXTSTART;
	u.u_segflg = 2;
	u.u_procp->p_flag |= SLOCK;
	readi(ip);
	u.u_procp->p_flag &= ~SLOCK;
	u.u_segflg = 0;
	u.u_error  = 0;
	u.u_dsize = ds;
	u.u_ssize = ss;
	u.u_sep = sep;
	if (xp->x_flag & (XLOCK|XWANT))
	    dbug1(0,txAF, "xalloc: clobbering x_flag with XWRIT, xp=%x\n", xp);
	xp->x_flag = XWRIT;
}

/*
 * Assure core for text segment
 * Text must be locked to keep someone else from
 * freeing it in the meantime.
 * x_ccount must be 0.
 */
xexpand(xp)
register struct text *xp;
{
	register mloc_t caddr;
	struct proc *p;

	dbug1(1,txEX, "XEXP %d - ", u.u_procp->p_pid);
	if ((caddr = mmuget(xp->x_size)) != MMUERR) {
		xp->x_caddr = caddr;
		if ((xp->x_flag&XLOAD)==0) {
			dbug(1,txEX, "swap()\n");
			swap(xp->x_daddr, xp->x_caddr, xp->x_size, B_READ);
		}
		xp->x_ccount++;
		xunlock(xp);
		return;
	}
	dbug(1,txEX, "no mem\n");
	if (save(u.u_ssav)) {
		mmumap();
		return;
	}
	p = u.u_procp;
	xswap(p, 1, p->p_size);
	xunlock(xp);
	p->p_flag |= SSWAP;
	qswtch();
	/* no return */
}

/*
 * Lock and unlock a text segment from swapping
 */
xlock(xp)
register struct text *xp;
{
	while (xp->x_flag&XLOCK) {
		xp->x_flag |= XWANT;
		dbug2(2,txLK, "LSLP-%d:%x ", u.u_procp->p_pid, xp);
		sleep((caddr_t)xp, PSWP);
	}
	dbug2(2,txLK, "L-%d:%x ", u.u_procp->p_pid, xp);
	xp->x_flag |= XLOCK;
}

xunlock(xp)
register struct text *xp;
{
	dbug2(2,txLK, "U-%d:%x ", u.u_procp->p_pid, xp);
	if (xp->x_flag&XWANT) {
		dbug2(2,txLK, "LWAK-%d:%x ", u.u_procp->p_pid, xp);
		wakeup((caddr_t)xp);
	}
	xp->x_flag &= ~(XLOCK|XWANT);
	/*
	 * M001
	 * If the scheduler is waiting because it couldn't find
	 * someone to swap in, and it detected a process
	 * which would have been swapped in but for the fact that it
	 * had locked text, we now jab the scheduler.
	 */
	if (runout && runtxt) {
		runout = 0;
		wakeup((caddr_t) &runout);
	}
}

/*
 * Decrement the in-core usage count of a shared text segment.
 * When it drops to zero, free the core space.
 */
xccdec(xp)
register struct text *xp;
{
	if (xp==NULL || xp->x_ccount==0)
		return;
	xlock(xp);
	if (--xp->x_ccount==0) {
		if (xp->x_flag&XWRIT) {
			xp->x_flag &= ~XWRIT;
			swap(xp->x_daddr, xp->x_caddr, xp->x_size, B_WRITE);
		}
		mmufree(xp->x_caddr, xp->x_size);
	}
	xunlock(xp);
}

/*
 * free the swap image of all unused saved-text text segments
 * which are from device dev (used by umount system call).
 */
xumount(dev)
register dev;
{
	register struct text *xp;

	for (xp = &text[0]; xp < v.ve_text; xp++)
		if (xp->x_iptr!=NULL && dev==xp->x_iptr->i_dev)
			xuntext(xp);
}

/*
 * remove a shared text segment from the text table, if possible.
 */
xrele(ip)
register struct inode *ip;
{
	register struct text *xp;

	/* this statement had been if (ip->i_flag&ITEXT==0) */
	if ((ip->i_flag&ITEXT)==0)
		return;
	for (xp = &text[0]; xp < v.ve_text; xp++)
		if (ip==xp->x_iptr)
			xuntext(xp);
}

/*
 * remove text image from the text table.
 * the use count must be zero.
 */
xuntext(xp)
register struct text *xp;
{
	register struct inode *ip;

	xlock(xp);
	if (xp->x_count) {
		xunlock(xp);
		return;
	}
	ip = xp->x_iptr;
	xp->x_flag &= ~XLOCK;

	dbug2(2,txAF, "N-%d:%x ", u.u_procp->p_pid, xp);
	if (xp->x_flag&XWANT)
		dbug1(0,txAF, "xuntext: ignoring XWANT xp=%x\n", xp);

	xp->x_iptr = NULL;
	mfree(swapmap, (int) mstod(xp->x_size), (int) xp->x_daddr);

	ip->i_flag &= ~ITEXT;
	if (ip->i_flag&ILOCK)
		ip->i_count--;
	else
		iput(ip);
}
