/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: trap.c,v 6.1 88/07/18 12:58:39 iv Exp $
 *
 * Modifications
 *	JJD	20 Feb 82	M000
 *		- Added USP, UPC, SR to regloc.
 *	JJD	12 May 82	M001
 *		- Corrected trace, breakpoint handling.
 *		- Fixed handling of bus error, addressing error
 *		- Explicit check for stack probe instruction
 *	JJD	18 May 82	M002
 *		- reassigned interrupt vectors
 *	JJD	20 May 82	M003
 *		- special handling for trace-trap on system call
 *	JJD	28 May 82	M004
 *		- mods to support floating point package
 *		- includes change in args to trap
 *	JJD	02 Jun 82	M005
 *		- corrected special case to handle trace trap
 *		  in combination with other exception
 *		- fixed so that u.u_ar0 is always set on traps
 *		  from user mode
 *	JJD	25 Aug 82	M006
 *		- added vecintsw use
 *	JJD	01 Sep 82	M007
 *		- Further debug, trace fixes.
 *	JJD	22 Sep 82	M008
 *		- hook for interrupt handler to simulate scheduling trap
 *		- cast argument to debugsub
 *	JJD	10 Oct 82	M009
 *		- modified handling of unexpected trap when debugging
 *	JJD	11 Oct 82	M010
 *		- added user-mode trace facility for debugging
 *	JJD	19 Oct 82	M011
 *		- convert to use setjmp/longjmp with u.u_qsav
 *	JJD	30 Oct 82	M012
 *		- made floating point emulation conditional on SOFTFP
 *	JJD	 8 Jan 83	M013
 *		- print out access address on bus error or memory fault
 *		- on kernel trap, print out real kernel sp in a7
 *	vwh	 6 Dec 83	M014
 *		- changed idleflg to Idlef, as per sys/clock.c
 *
 * $Log:	trap.c,v $
 * Revision 6.1  88/07/18  12:58:39  iv
 * 	- Changed define DEBUG to KDEBUG for kernel-debugger code
 * 
 * Revision 6.0  87/10/12  17:08:10  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.3  86/11/25  11:11:32  iv
 * 	- Fixed display of the illegal word (Didn't need to back up PC)
 * 
 * Revision 5.2  86/11/21  13:40:27  iv
 * 	- Improved some debugging handling with the t)race command
 * 	- Added an "inst word=%x" printf when trap == 4 (Illegal instruction)
 * 
 * Revision 5.1  86/08/29  10:50:06  iv
 * 	- Took out all the silly trace stuff, which was commented out
 * 	  and didn't work anyway
 * 	- Minor readability changes here and there
 * 	- Reworked for system call speed when FASTSC is not defined
 * 	  (ie, mch.s is not helping us make fast system calls)
 * 	- Added FASTSC support code (mch.s sends system calls directly
 * 	  to systrap(), who's only job is to quickly process system calls)
 * 
 * Revision 5.0  85/12/03  17:10:56  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:46:31  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:47:20  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:55  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.3  84/12/21  14:33:23  iv
 * 	- Improved debugging code in misc places
 * 	- Added new interface code for improved debug.c
 * 
 * Revision 1.2  84/04/18  10:47:39  iv
 * 	- Got rid of an extern declaration of "int  (*vecintsw[]) ()", not used
 * 
 * Revision 1.1  84/04/10  17:42:05  iv
 * -	Initial checkin for TRS-XENIX 2.0
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/reg.h"
#include "../h/mmu.h"

extern  struct {
	short   fill0;
	int     access;
	short   iwd;
} fltinfo[2];

#define NMI_NUM	31		/* todays hardware is doing this */

/*
 * Offsets of the user's registers relative to
 * the saved r0. See reg.h
 */
char    regloc[] = {
	D0,D1,D2,D3,D4,D5,D6,D7,A0,A1,A2,A3,A4,A5,A6,USP,UPC,SR
};
short reglocc = 18;

/*
 *	SYSTEMS CALLS MADE WITH THE FOLLOWING CODE:
 *
 *		movw	#system_call_number,d0
 *		trap	#0
 *
 *	where system-call arguments are on the user stack as follows:
 *
 *		sp ->	return pc
 *			arg1
 *			arg2
 *			...
 */

#ifdef KDEBUG
extern	int	*ss;
extern	int	supertrace, strflag;
int	debugsw = 1;
#define BRKPT	0x002f		/* trap number for break point */
#define	DEBTRAP	0x002e		/* trap number to enter debugger */
#endif /*KDEBUG*/

#ifdef DEBUGSC
int     debugsc = 0;            /* print system call info */
char *sysnam[] = {		/* names for debug printout */
	"=ERROR=",              /*  0 = indir */
	"REXIT",		/*  1 = exit */
	"FORK",			/*  2 = fork */
	"READ",			/*  3 = read */
	"WRITE",		/*  4 = write */
	"OPEN",			/*  5 = open */
	"CLOSE",		/*  6 = close */
	"WAIT",			/*  7 = wait */
	"CREAT",		/*  8 = creat */
	"LINK",			/*  9 = link */
	"UNLINK",		/* 10 = unlink */
	"EXEC",			/* 11 = exec */
	"CHDIR",		/* 12 = chdir */
	"GTIME",		/* 13 = time */
	"MKNOD",		/* 14 = mknod */
	"CHMOD",		/* 15 = chmod */
	"CHOWN",		/* 16 = chown; now 3 args */
	"SBREAK",		/* 17 = break */
	"STAT",			/* 18 = stat */
	"SEEK",			/* 19 = seek; now 3 args */
	"GETPID",		/* 20 = getpid */
	"SMOUNT",		/* 21 = mount */
	"SUMOUNT",		/* 22 = umount */
	"SETUID",		/* 23 = setuid */
	"GETUID",		/* 24 = getuid */
	"STIME",		/* 25 = stime */
	"PTRACE",		/* 26 = ptrace */
	"ALARM",		/* 27 = alarm */
	"FSTAT",		/* 28 = fstat */
	"PAUSE",		/* 29 = pause */
	"UTIME",		/* 30 = utime */
	"STTY",			/* 31 = stty */
	"GTTY",			/* 32 = gtty */
	"SACCESS",		/* 33 = access */
	"NICE",			/* 34 = nice */
	"FTIME/NULLSYS",        /* 35 = ftime; formerly sleep */
	"SYNC",			/* 36 = sync */
	"KILL",			/* 37 = kill */
	"NULLSYS/CLOCAL",       /* 38 = s3 clocal */
	"NULLSYS/SETPGRP",      /* 39 = setpgrp (s3 only) */
	"CXENIX/=ERROR=",       /* 40 = s3 cxenix */
	"DUP",			/* 41 = dup */
	"PIPE",			/* 42 = pipe */
	"TIMES",		/* 43 = times */
	"PROFIL",		/* 44 = prof */
	"LOCK",                 /* 45 = lock process in core */
	"SETGID",		/* 46 = setgid */
	"GETGID",		/* 47 = getgid */
	"SSIG",			/* 48 = sig */
	"=ERROR=",              /* 49 = reserved for USG */
	"=ERROR=",              /* 50 = reserved for USG */
	"SYSACCT",		/* 51 = turn acct off/on */
	"=ERROR=",              /* 52 = set user physical addresses */
	"=ERROR=",              /* 53 = lock user in core */
	"IOCTL",		/* 54 = ioctl */
	"=ERROR=",              /* 55 = readwrite (in abeyance) */
	"=ERROR=",              /* 56 = creat mpx comm channel */
	"=ERROR=/UTSSYS",       /* 57 = utssys (uname & ustat) */
	"=ERROR=",              /* 58 = reserved for USG */
	"EXECE",		/* 59 = exece */
	"UMASK",		/* 60 = umask */
	"CHROOT",		/* 61 = chroot */
	"CLOCAL/FCNTL",         /* 62 = custom local calls */
	"CXENIX/ULIMIT"};       /* 63 = custom xenix calls */

char *cxenam[] = {              /* names for CXENIX debug printout */
	"SHUTUP",               /* 0 - shutup OS */
	"LOCKING",              /* 1 - record locking */
	"CREATSEM",             /* 2 - dummy for place-holding */
	"OPENSEM",              /* 3 - dummy for place-holding */
	"SIGSEM",               /* 4 - dummy for place-holding */
	"WAITSEM",              /* 5 - dummy for place-holding */
	"NBWAITSEM",            /* 6 - dummy for place-holding */
	"RDCHK",                /* 7 - blocking read check */
	"STKGROW",              /* 8 - set stack limit */
	"XPTRACE",              /* 9 - extended ptrace */
	"FTRUNC",               /* 10- truncate file */
	"FTIME",                /* 11- ftime for sys3 progs */
	"NAP",                  /* 12 - sleep for short interval */
	"SDGET",                /* 13 - attach to shared data */
	"SDFREE",               /* 14 - release shared data */
	"SDENTER",              /* 15 - enter critical region */
	"SDLEAVE",              /* 16 - leave critical region */
	"SDGETV",               /* 17 - get SD version # */
	"SDWAITV"};             /* 18 - wait for new SD version */
#define CXEMAX         18               /* max subfunction # */
#endif /*DEBUGSC*/

/*
 * Called from mch.s when a processor trap occurs.
 * The arguments are the words saved on the system stack
 * by the hardware and software during the trap processing.
 * Their order is dictated by the hardware and the details
 * of C's calling sequence. They are peculiar in that
 * this call is not 'by value' and changed user registers
 * get copied back on return.
 *	Your basic stack is:
 *	pchi,pclo	long
 *	sr		word
 *	phi,plo		long from jsr
 *	mmu		mmu control reg
 *	regs		16 longs
 *	trap no		long
 *
 * trap(no,d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,usp,mmu,sr,tpc)
*/


extern	short	Idlef;
/*VARARGS0*/
/*ARGSUSED*/
trap(no,d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,usp,mmu,sr,tpc)
  int	no,d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,usp,mmu,sr,tpc;
{
	register struct user	*up = &u;
	short	i;
	time_t	syst;

	syst = up->u_stime;
	
#ifdef KDEBUG
	switch(no) {
		case 31:
			printf("Trap %d = NMI\n",no);
			dostate(&d0);	/* don't want to mess up u_ar0 */
			ss = &d0;
			debugsub((ushort *) 0);
			printf("NMI return\n");
			Idlef = 0;		/* don't loop in idle (mch.s) */
			return;
		case BRKPT:
			printf("Trap %d = BREAK\n",no);
			dostate(&d0);	/* don't want to mess up u_ar0 */
			ss = &d0;
			if (debugsub((ushort *)tpc)) tpc = tpc - 2;

			Idlef = 0;		/* don't loop in idle (mch.s) */
			return;
		case DEBTRAP:
			if (sr & SR_SUPER) {
				printf("Trap %d = DEBUG ENTRY\n", no);
				dostate(&d0);
				ss = &d0;
				debugsub((ushort *)0);
				Idlef = 0;	/* don't loop in idle (mch.s) */
				return;
			}
	}
#endif /*KDEBUG*/


	if(sr & SR_SUPER) {	/* trap in supervisor state */
		register int *ar0 = &d0;
		if (no == 9) { /* trace trap */
			/* firstvec and endvec delimit the system trap  */
			/* vectors.					*/
			/* If the user attempts to single step a system */
			/* call, we won't see the trace trap until after*/
			/* entering system mode to service the trap     */
			/* instruction.  In this case, we just ignore   */
			/* the trap.  (If we sent a signal to the	*/
			/* process now, it would likely produce an EINTR*/
			/* error on the system call.)			*/
			/* The same logic applies to illegal		*/
			/* instructions, etc.				*/
			extern int firstvec, endvec;

			if (tpc >= (int)&firstvec && tpc < (int)&endvec)
				return;
#ifdef KDEBUG
			if (supertrace) {
				if (supertrace < 2) {
					putchar('\n');
					dostate(ar0);
				}
				strflag = 1;
				ss = ar0;
				debugsub((ushort *)ss[UPC]);
				/* don't reschedule when debugging */
				return;
			}
#endif /*KDEBUG*/
		}

		printf("Supervisor Trap %d\n",no);
		if (no == 2 || no == 3){
			printf("Access address=%X",
				fltinfo[no-2].access);
			printf("  Instr wd=%x\n", fltinfo[no-2].iwd);
		} 
		else if (no == 4)
			printf("Instr wd=%x\n", *((unsigned short *)ar0[UPC]));

		printf("&d0=%X mmu=%X sr=%X pc=%X usp=%X\n",
			ar0,
			ar0[MMU],
			ar0[SR],
			ar0[UPC],
			ar0[USP]);
			
		printf("d0-d7 %X %X %X %X  %X %X %X %X\n",
			ar0[D0],
			ar0[D1],
			ar0[D2],
			ar0[D3],
			ar0[D4],
			ar0[D5],
			ar0[D6],
			ar0[D7]);
		printf("a0-a7 %X %X %X %X  %X %X %X %X\n",
			ar0[A0],
			ar0[A1],
			ar0[A2],
			ar0[A3],
			ar0[A4],
			ar0[A5],
			ar0[A6],
			(int) ((int *)&tpc + 1));
#ifdef KDEBUG
		ss = ar0;
		debugsub((ushort *)0);
#endif /*KDEBUG*/
		panic("trap in sys");
	}

	/* trap from user state */
	up->u_ar0 = &d0;
		
#ifndef FASTSC
	if (no == 32) {		/* Trap Instruction "trap #0" System Call */
		sr = do_scall(d0,usp,sr,tpc);
		goto out;
	}
#endif /*FASTSC*/
	switch (no) {
		default:	/* all other traps from user state */
			i = SIGBUS;		/* bus error for others now */
			break;
		case 1:				/* rescheduling trap */
			goto out;
		case 4:
			i = SIGILL;             /* illegal instruction */
			break;
		case 9:
			sr &= ~SR_TRACE;	/* turn off trace bit */
			/* FALL THRU */
		case 34:	/* "trap #2" is breakpoint */
			i = SIGTRAP;            /* trace or breakpoint */
			break;
		case 10:
			i = SIGIOT;		/* iot  = line 1010 */
			break;
		case 11:
#ifdef SOFTFP
			softfp(up->u_ar0, &up->u_fps, up->u_procp);
			goto out;
#else /*SOFTFP*/
			i = SIGEMT;
			break;
#endif /*SOFTFP*/

		/*
		 * M001
		 * What the 68000 documentation calls an "Address Error"
		 * (fetch word, long, or instruction from odd address)
		 * corresponds to the UNIX concept of a "Bus Error."
		 * Similarly, when the 68000 docs call a "Bus Error"
		 * corresponds to the UNIX concept of a "Memory Fault"
		 * or "Segmentation violation".  Confusing, but true.
		 */
		case 3:	
			i = SIGBUS;		/* UNIX bus error */
			break;

		case 2:				/* Addr Error */
			i = SIGSEGV;            /* segmentation violation */
			{
			register unsigned osp;
		
				up->u_ar0 = &d0;

				/* Begin M001 ... */
				/*
				 * Special check for stack probe
				 * instruction.  See machdep.c.
				 */
				if ((osp = stkprobe(up->u_ar0)) != 0
				    || (osp = usp, (backup(up->u_ar0) == 0)))
					if (grow((unsigned)osp)) {
						goto out;
					}
				/* ... End M001 */
#ifdef KDEBUG
				if (debugsw>1) {
					printf("segmentation\n");
					debugsub((ushort *)tpc);
				}
#endif /*KDEBUG*/
				break;
			}

	}

#ifdef KDEBUG
	if (debugsw>0){
	    printf("TRAP %d in USER mode\n",no);
	    if (no == 2 || no == 3){
		printf("Access address=%X",
			fltinfo[no-2].access);
		printf("  Instr wd=%x\n", fltinfo[no-2].iwd);
	    }
	    dostate(&d0);
	    ss = &d0;
	    debugsub((ushort *)0);
	}
#endif /*KDEBUG*/
	psignal(up->u_procp, i);

out:
	if (issig())
		psig();
	if (runrun)
		qswtch();
	if (up->u_prof.pr_scale)
		addupc((caddr_t)tpc, (int)&up->u_prof, (int)(up->u_stime-syst));
#ifdef FPHARD
	if (up->u_fpsaved)
		restfp((int *)&up->u_fps);
#endif /*FPHARD*/
	if ((up->u_sdcnt > 0) && (up->u_sdsaved != 0)){
		up->u_sdsaved = 0;
		sdswtch(1);
	}

	Idlef = 0;			/* don't loop in idle (mch.s) */
}

#ifdef FASTSC
systrap(d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,usp,mmu,sr,tpc)
  int	d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,usp,mmu,sr,tpc;
{
	register struct user	*up = &u;
	time_t	syst;

	syst = up->u_stime;

	up->u_ar0 = &d0;
	sr = do_scall(d0,usp,sr,tpc);

	if ((up->u_procp->p_sig)? issig() : 0)
		psig();
	if (runrun)
		qswtch();
	if (up->u_prof.pr_scale)
		addupc((caddr_t)tpc, (int)&up->u_prof, (int)(up->u_stime-syst));
#  ifdef FPHARD
	if (up->u_fpsaved)
		restfp((int *)&up->u_fps);
#  endif /*FPHARD*/
	if ((up->u_sdcnt > 0) && (up->u_sdsaved != 0)){
		up->u_sdsaved = 0;
		sdswtch(1);
	}

	Idlef = 0;			/* don't loop in idle (mch.s) */
}
#endif /*FASTSC*/


/*
 * system call trap
 */
int
do_scall(d0,usp,sr,tpc)
  int	d0, usp, sr, tpc;
{
	extern struct sysent sysent[], v7sysent[], s3sysent[];
	register struct sysent *callp;
	register struct proc *pp;
	register short i;
	short	scallno;

	u.u_error = 0;
	scallno =  d0 & 0x3f;
	callp = &sysent[scallno];
	if (callp->sy_call == ((int (*)()) 0) )	/* conflict between versions */
		if ((u.u_exdata.x_renv & XE_VERS) == XE_V3)
			callp = &s3sysent[callp->sy_narg];
		else
			callp = &v7sysent[callp->sy_narg];

	for (i=0; i<callp->sy_narg; i++)
	    u.u_arg[i] = fuword((caddr_t) usp + ((i+1) *4));
	/* hack: on v7 the cxenix subcode is arg[0] */
	if ((u.u_exdata.x_renv&XE_V3)==0 && scallno==63) {
	    d0 = scallno + (u.u_arg[0]<<8);
	    for (i=0; i<4; i++) u.u_arg[i] = u.u_arg[i+1];
	}
	u.u_arg[4] = d0;  /*last arg is scall no+subcode*/
	u.u_dirp = (caddr_t)u.u_arg[0];
#ifdef DEBUGSC
	if (debugsc) {
		char *strptr;
		extern cxenix();
		strptr = "=== ILLEGAL ===";
		if (scallno < 0x40)
			strptr = sysnam[scallno];
		if ((callp->sy_call == &cxenix) && ((d0>>8) <= CXEMAX))
			strptr = cxenam[d0>>8];
		printf("%d %s (%x/%d) SI=%x/%x ", scallno, strptr,
			u.u_procp,u.u_procp-proc, usp, tpc);
		if (debugsc>1)
			debugsub((ushort *)0);
	}
#endif /*DEBUGSC*/
	u.u_rval1 = 0;
	u.u_rval2 = u.u_ar0[D0];
	u.u_ap = u.u_arg;
	if (setjmp(u.u_qsav)) {
		if (u.u_error==0)
			u.u_error = EINTR;
	} else {
		(*callp->sy_call)();
	}
	if (u.u_error) {
		sr |= SR_CARRY;
		u.u_ar0[D0] = u.u_error;
	} else {
		sr &= ~SR_CARRY;
		u.u_ar0[D0] = u.u_rval1;
		u.u_ar0[D1] = u.u_rval2;
	}
#ifdef DEBUGSC
	if (debugsc) {
		printf("eSC (%x/%d) %s %x %x\n", u.u_procp, u.u_procp-proc,
			(sr&SR_CARRY)?"** ERROR **":"OK",
			(sr&SR_CARRY)?(u.u_error&0377):u.u_rval1, u.u_rval2);
		if (debugsc>1)
			debugsub((ushort *)0);
	}
#endif /*DEBUGSC*/
	pp = u.u_procp;
	pp->p_pri=(pp->p_cpu>>1) + PUSER + pp->p_nice - NZERO;
	curpri = pp->p_pri;
	return (sr);
}


/*
 * nonexistent system call -- set fatal error code.
 */
nosys()
{
	u.u_error = EINVAL;
}

/*
 * Ignored system call
 */
nullsys()
{
}
