/*
 *	@(#) sys/bio.c 1.2 83/10/23 Copyright Microsoft Corporation, 1983 [Microsoft SID]
 *	@(#) sys/bio.c 2.1 86/08/23 Copyright (c) The Santa Cruz Operation, Inc., 1985, 1986; All rights reserved.
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 *
 *  MODIFICATION HISTORY
 *	27 March 1984	blf	M000	post alpha (Cromemco)
 *		- Changed test for a buffer being "directly addressable"
 *		  to use the newly-created issabuf() macro in param.h.
 *	23 April 1984	cliff	M001
 *		- in iodone(), add conditional call to mapfree()
 *		For some reason, Microsoft or Bell chose to delete
 *		the call to mapfree() in the 3.0 or Sys III change.
 *	7 August 1986	blf	S002
 *		- Replace bflush() with a new version by sco!buckm,
 *		  to solve a problem found by easi!kd several years
 *		  ago:  Bflush took a very long time if any pipes
 *		  were active (and did not allow any sorting of the
 *		  write requests).
 *		- Replace disksort(), which did "escelator" sorting,
 *		  by a new version (also by sco!buckm) which does
 *		  "elevator" sorting.
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/sysinfo.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/buf.h"
#include "../h/iobuf.h"
#include "../h/conf.h"
#include "../h/proc.h"
#include "../h/var.h"

#define	b_acount	b_resid

struct buf bfreelist;
short bfcount	= 0;	/* S002 */

/*
 * The following several routines allocate and free
 * buffers with various side effects.  In general the
 * arguments to an allocate routine are a device and
 * a block number, and the value is a pointer to
 * to the buffer header; the buffer is marked "busy"
 * so that no one else can touch it.  If the block was
 * already in core, no I/O need be done; if it is
 * already busy, the process waits until it becomes free.
 * The following routines allocate a buffer:
 *	getblk
 *	getablk
 *	bread
 *	breada
 * Eventually the buffer must be released, possibly with the
 * side effect of writing it out, by using one of
 *	bwrite
 *	bdwrite
 *	bawrite
 *	brelse
 */


/*
 * Read in (if necessary) the block and return a buffer pointer.
 */
struct buf *
bread(dev, blkno)
dev_t dev;
daddr_t blkno;
{
	register struct buf *bp;

	sysinfo.lread++;
	bp = getblk(dev, blkno);
	if (bp->b_flags&B_DONE)
		return(bp);
	bp->b_flags |= B_READ;
	(*bdevsw[major(dev)].d_strategy)(bp);
	u.u_ior++;
	sysinfo.bread++;
	iowait(bp);
	return(bp);
}

/*
 * Read in the block, like bread, but also start I/O on the
 * read-ahead block (which is not allocated to the caller)
 */
struct buf *
breada(dev, blkno, rablkno)
dev_t dev;
daddr_t blkno, rablkno;
{
	register struct buf *bp, *rabp;

	bp = NULL;
	if (!incore(dev, blkno)) {
		sysinfo.lread++;
		bp = getblk(dev, blkno);
		if ((bp->b_flags&B_DONE) == 0) {
			bp->b_flags |= B_READ;
			(*bdevsw[major(dev)].d_strategy)(bp);
			u.u_ior++;
			sysinfo.bread++;
		}
	}
	if (rablkno && bfreelist.b_bcount>1 && !incore(dev, rablkno)) {
		rabp = getblk(dev, rablkno);
		if (rabp->b_flags & B_DONE)
			brelse(rabp);
		else {
			rabp->b_flags |= B_READ|B_ASYNC;
			(*bdevsw[major(dev)].d_strategy)(rabp);
			u.u_ior++;
			sysinfo.bread++;
		}
	}
	if(bp == NULL)
		return(bread(dev, blkno));
	iowait(bp);
	return(bp);
}

/*
 * Write the buffer, waiting for completion.
 * Then release the buffer.
 */
bwrite(bp)
register struct buf *bp;
{
	register flag;

	sysinfo.lwrite++;
	flag = bp->b_flags;
	bp->b_flags &= ~(B_READ | B_DONE | B_ERROR | B_DELWRI | B_AGE);
	(*bdevsw[major(bp->b_dev)].d_strategy)(bp);
	u.u_iow++;
	sysinfo.bwrite++;
	if ((flag&B_ASYNC) == 0) {
		iowait(bp);
		brelse(bp);
	} else if (flag & B_DELWRI)
		bp->b_flags |= B_AGE;
	else
		geterror(bp);
}

/*
 * Release the buffer, marking it so that if it is grabbed
 * for another purpose it will be written out before being
 * given up (e.g. when writing a partial block where it is
 * assumed that another write for the same block will soon follow).
 * This can't be done for magtape, since writes must be done
 * in the same order as requested.
 */
bdwrite(bp)
register struct buf *bp;
{
	register struct iobuf *dp;

	dp = bdevsw[major(bp->b_dev)].d_tab;
	if(dp->b_flags & B_TAPE)
		bawrite(bp);
	else {
		sysinfo.lwrite++;
		bp->b_flags |= B_DELWRI | B_DONE;
		bp->b_resid = 0;
		brelse(bp);
	}
}

/*
 * Release the buffer, start I/O on it, but don't wait for completion.
 */
bawrite(bp)
register struct buf *bp;
{

	if(bfreelist.b_bcount>4)
		bp->b_flags |= B_ASYNC;
	bwrite(bp);
}

/*
 * Release the buffer, waiting for I/O to complete.
 */
bswrite(bp)
register struct buf *bp;
{
	bp->b_flags &= ~B_ASYNC;
	bwrite(bp);
}

/*
 * release the buffer, with no I/O implied.
 */
brelse(bp)
register struct buf *bp;
{
	register struct buf **backp;
	register s;

	if (bp->b_flags&B_WANTED)
		wakeup((caddr_t)bp);
	if (bfreelist.b_flags&B_WANTED) {
		bfreelist.b_flags &= ~B_WANTED;
		wakeup((caddr_t)&bfreelist);
	}
	if (bp->b_flags&B_ERROR) {
		bp->b_flags |= B_STALE|B_AGE;
		bp->b_flags &= ~(B_ERROR|B_DELWRI);
		bp->b_error = 0;
	}
	s = spl6();
	if(bp->b_flags & B_AGE) {
		backp = &bfreelist.av_forw;
		(*backp)->av_back = bp;
		bp->av_forw = *backp;
		*backp = bp;
		bp->av_back = &bfreelist;
	} else {
		backp = &bfreelist.av_back;
		(*backp)->av_forw = bp;
		bp->av_back = *backp;
		*backp = bp;
		bp->av_forw = &bfreelist;
	}
	bp->b_flags &= ~(B_WANTED|B_BUSY|B_ASYNC|B_AGE);
	bfreelist.b_bcount++;
/* M000	if (bp->b_paddr <= (paddr_t)(unsigned)sabuf[v.v_sabuf])	***/
	if (issabuf( bp ))					/* M000 */
		bfreelist.b_acount++;
	splx(s);
}

/*
 * See if the block is associated with some buffer
 * (mainly to avoid getting hung up on a wait in breada)
 */
incore(dev, blkno)
dev_t dev;
daddr_t blkno;
{
	register struct buf *bp;
	register struct iobuf *dp;

	dp = (struct iobuf *) bhash(dev, blkno);
	for (bp=dp->b_forw; bp != (struct buf *)dp; bp = bp->b_forw)
		if (bp->b_blkno==blkno && bp->b_dev==dev && (bp->b_flags&B_STALE)==0)
			return(1);
	return(0);
}

/*
 * Assign a buffer for the given block.  If the appropriate
 * block is already associated, return it; otherwise search
 * for the oldest non-busy buffer and reassign it.
 */
struct buf *
getblk(dev, blkno)
dev_t dev;
daddr_t blkno;
{
	register struct buf *bp;
	register struct iobuf *dp;

	if(major(dev) >= bdevcnt)
		panic("blkdev");

    loop:
	spl0();
	dp = (struct iobuf *) bhash(dev, blkno);
	if(dp == NULL)
		panic("devtab");
	for (bp=dp->b_forw; bp != (struct buf *)dp; bp = bp->b_forw) {
		if (bp->b_blkno!=blkno || bp->b_dev!=dev || bp->b_flags&B_STALE)
			continue;
		spl6();
		if (bp->b_flags&B_BUSY) {
			bp->b_flags |= B_WANTED;
			syswait.iowait++;
			sleep((caddr_t)bp, PRIBIO+1);
			syswait.iowait--;
			goto loop;
		}
		spl0();
		notavail(bp);
		return(bp);
	}
	spl6();
	if (bfreelist.av_forw == &bfreelist) {
		bfreelist.b_flags |= B_WANTED;
		sleep((caddr_t)&bfreelist, PRIBIO+1);
		goto loop;
	}
	spl0();
	notavail(bp = bfreelist.av_forw);
	if (bp->b_flags & B_DELWRI) {
		bp->b_flags |= B_ASYNC;
		bwrite(bp);
		goto loop;
	}
	bp->b_flags = B_BUSY;
	bp->b_back->b_forw = bp->b_forw;
	bp->b_forw->b_back = bp->b_back;
	bp->b_forw = dp->b_forw;
	bp->b_back = (struct buf *)dp;
	dp->b_forw->b_back = bp;
	dp->b_forw = bp;
	bp->b_dev = dev;
	bp->b_blkno = blkno;
	bp->b_bcount = BSIZE;		/* Make sure we read a whole block */
	return(bp);
}

/*
 * get an empty block,
 * not assigned to any particular device
 * and if flag non-zero, directly addressable
 */
struct buf *
getablk(flag)
{
	register struct buf *bp;
	register struct buf *dp;

loop:
	dp = &bfreelist;
	spl6();
	while (dp->av_forw == dp || (flag && dp->b_acount == 0)) {
		dp->b_flags |= B_WANTED;
		sleep((caddr_t)dp, PRIBIO+1);
	}
	for (bp = dp->av_forw; flag; bp = bp->av_forw)
/* M000		if (bp->b_paddr <= (paddr_t)(unsigned)sabuf[v.v_sabuf])	***/
		if (issabuf( bp ))				/* M000 */
			break;
	spl0();
	notavail(bp);
	if (bp->b_flags & B_DELWRI) {
		bp->b_flags |= B_ASYNC;
		bwrite(bp);
		goto loop;
	}
	bp->b_flags = B_BUSY|B_AGE;
	bp->b_back->b_forw = bp->b_forw;
	bp->b_forw->b_back = bp->b_back;
	bp->b_forw = dp->b_forw;
	bp->b_back = dp;
	dp->b_forw->b_back = bp;
	dp->b_forw = bp;
	bp->b_dev = (dev_t)NODEV;
	return(bp);
}

/*
 * Wait for I/O completion on the buffer; return errors
 * to the user.
 */
iowait(bp)
register struct buf *bp;
{

	syswait.iowait++;
	spl6();
	while ((bp->b_flags&B_DONE)==0)
		sleep((caddr_t)bp, PRIBIO);
	spl0();
	syswait.iowait--;
	geterror(bp);
}

/*
 * Unlink a buffer from the available list and mark it busy.
 * (internal interface)
 */
notavail(bp)
register struct buf *bp;
{
	register s;

	s = spl6();
	bp->av_back->av_forw = bp->av_forw;
	bp->av_forw->av_back = bp->av_back;
	bp->b_flags |= B_BUSY;
	bfreelist.b_bcount--;
/* M000	if (bp->b_paddr <= (paddr_t)(unsigned)sabuf[v.v_sabuf])	***/
	if (issabuf( bp ))					/* M000 */
		bfreelist.b_acount--;
	splx(s);
}

/*
 * Mark I/O complete on a buffer, release it if I/O is asynchronous,
 * and wake up anyone waiting for it.
 */
iodone(bp)
register struct buf *bp;
{

	if (bp->b_flags & B_MAP)			/*M001*/
		mapfree(bp);				/*M001*/
	bp->b_flags |= B_DONE;
	if (bp->b_flags&B_ASYNC)
		brelse(bp);
	else {
		bp->b_flags &= ~B_WANTED;
		wakeup((caddr_t)bp);
	}
	if (bp->b_flags & B_FLUSH) {			/* S002 begin... */
		bp->b_flags &= ~B_FLUSH;
		if (--bfcount == 0)
			wakeup((caddr_t)&bfcount);
	}						/* ...end S002 */
}


/*
 * make sure all write-behind blocks
 * on dev (or NODEV for all)
 * are flushed out.
 * (from umount and update)
 */
bflush(dev)						/* S002 begin... */
dev_t dev;
{
	register struct buf *bp, *bf;

	spl6();
	while (bfcount)
		sleep((caddr_t)&bfcount, PRIBIO);
	bf = NULL;
	for (bp = bfreelist.av_forw; bp != &bfreelist; bp = bp->av_forw) {
		if (bp->b_flags&B_DELWRI && (dev == NODEV||dev==bp->b_dev)) {
			notavail(bp);
			bp->av_back = bf;
			bf = bp;
			bfcount++;
		}
	}
	spl0();
	while (bf) {
		bp = bf->av_back;
		bf->b_flags |= B_FLUSH|B_ASYNC;
		bwrite(bf);
		bf = bp;
	}
	spl6();
	if (bfcount)
		sleep((caddr_t)&bfcount, PRIBIO);
	spl0();
}							/* ...end S002 */


/*
 * Pick up the device's error number and pass it to the user;
 * if there is an error but the number is 0 set a generalized
 * code.  Actually the latter is always true because devices
 * don't yet return specific errors.
 */
geterror(bp)
register struct buf *bp;
{

	if (bp->b_flags&B_ERROR)
		if ((u.u_error = bp->b_error)==0)
			u.u_error = EIO;
}

/*
 * Invalidate blocks for a dev after last close.
 */
binval(dev)
{
	register struct buf *dp;
	register struct buf *bp;
	register i;

	for (i=0; i<v.v_hbuf; i++) {
		dp = (struct buf *)&hbuf[i];
		for (bp = dp->b_forw; bp != dp; bp = bp->b_forw)
			if (bp->b_dev == dev)
				bp->b_flags |= B_STALE|B_AGE;
	}
}

/* S002 begin...
 *  Generalized seek sort for disk ("elevator" algorithm).
 */

int Dsort = 0;                  /* == 1 for special read preference */

disksort(dp, bp)
struct iobuf *dp;
register struct buf *bp;
{
	register struct buf *ap;
	struct buf *tp;

	tp = dp->b_actf;
	if (tp == NULL) {
		dp->b_actf = bp;
		dp->b_actl = bp;
		bp->av_forw = NULL;
		return;
	}
	/* Special case for cylin == 0xFFFF;
	 *   0xFFFF means bp is a driver command & must be run LAST */
	if (bp->b_cylin + 1 == 0) {
		dp->b_actl->av_forw = bp;
		dp->b_actl = bp;
		bp->av_forw = NULL;
		return;
	}
	for (ap = tp->av_forw; ap != NULL; tp = ap, ap = ap->av_forw) {
		if (Dsort) {
			if ((bp->b_flags&B_READ) && (ap->b_flags&B_READ) == 0)
			    break; /* incoming reads preceed queued writes */
			if ((bp->b_flags&B_READ) == 0 && (ap->b_flags&B_READ))
			    continue; /* incoming writes follow queued reads */
		}
		/* classic elevator up-down sorting */
		if (ap->b_cylin < bp->b_cylin  &&  bp->b_cylin <= tp->b_cylin)
			break;
		if (ap->b_cylin > bp->b_cylin  &&  bp->b_cylin >= tp->b_cylin)
			break;
	}
	bp->av_forw = ap;
	tp->av_forw = bp;
	if (tp == dp->b_actl)
		dp->b_actl = bp;
}
