/*
 *	@(#) sys/sig.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: sig.c,v 6.0 87/10/12 17:09:52 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */

/*
 * MODIFICATION HISTORY:
 *
 *	21 Feb 1984	cliff	M000
 *		- add missing synchronous-write flag to writei() calls.
 *
 * $Log:	sig.c,v $
 * Revision 6.0  87/10/12  17:09:52  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.1  86/08/29  13:10:34  iv
 * 	-  Prettied up code some
 * 	- Changed U references to register pointer for speed
 * 	  (part of system call performance speed up)
 * 
 * Revision 5.0  85/08/23  10:36:19  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:49:09  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:44:24  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:23:14  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 0.1  84/05/17  15:19:15  iv
 * 	- Missing synchronous-write argument added to writei() calls.
 * 
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/inode.h"
#include "../h/reg.h"
#include "../h/text.h"
#include "../h/mmu.h"
#include "../h/var.h"


/*
 * Send the specified signal to
 * all processes with 'pgrp' as
 * process group.
 * Called by tty.c for quits and
 * interrupts.
 */
signal(pgrp, sig)
register pgrp;
{
	register struct proc *p;

	if(pgrp == 0)
		return;
	for(p = &proc[1]; p < (struct proc *)v.ve_proc; p++)
		if(p->p_pgrp == pgrp)
			psignal(p, sig);
}

/*
 * Send the specified signal to
 * the specified process.
 */
psignal(p, sig)
register struct proc *p;
register sig;
{

	sig--;
	if (sig < 0 || sig >= NSIG)
		return;
	p->p_sig |= 1L<<sig;
	if(p->p_stat == SSLEEP && p->p_pri > PZERO) {
		if(p->p_pri > PUSER)
			p->p_pri = PUSER;
		setrun(p);
	}
}

/*
 * Returns true if the current
 * process has a signal to process.
 * This is asked at least once
 * each time a process enters the
 * system.
 * A signal does not do anything
 * directly to a process; it sets
 * a flag that asks the process to
 * do something to itself.
 */
issig()
{
	register struct proc *p, *q;
	register n;

	p = u.u_procp;
	while (p->p_sig) {
		n = fsig(p);
		if (n == SIGCLD) {
			if (u.u_signal[SIGCLD-1]&01) {
				for (q = &proc[1];
				     q < (struct proc *)v.ve_proc;
				     q++)
				{
			   		if (p->p_pid == q->p_ppid 
					    && q->p_stat == SZOMB)
						freeproc(q, 0);
				}
			} else
				if (u.u_signal[SIGCLD-1])
					return (n);
		} else
			if ((u.u_signal[n-1]&1) == 0 || (p->p_flag&STRC))
				return (n);
		p->p_sig &= ~(1L<<(n-1));
	}
	return (0);
}

/*
 * Enter the tracing STOP state.
 * In this state, the parent is
 * informed and the process is able to
 * receive commands from the parent.
 */
stop()
{
	register struct proc *pp, *cp;

loop:
	cp = u.u_procp;
	if(cp->p_ppid != 1)
	for (pp = &proc[0]; pp < (struct proc *)v.ve_proc; pp++)
		if (pp->p_pid == cp->p_ppid) {
			wakeup((caddr_t)pp);
			cp->p_stat = SSTOP;
			swtch();
			if ((cp->p_flag&STRC)==0 || procxmt())
				return;
			goto loop;
		}
	exit(fsig(u.u_procp));
}

/*
 * Perform the action specified by
 * the current signal.
 * The usual sequence is:
 *	if (issig())
 *		psig();
 */
psig()
{
	register n, p;
	register struct proc *rp;
	register struct user *up;

	up = &u;
	rp = up->u_procp;
	if (up->u_fpsaved==0) {
		savfp(&up->u_fps);
		up->u_fpsaved = 1;
	}
	if (rp->p_flag&STRC)
		stop();
	n = fsig(rp);
	if (n==0)
		return;
	rp->p_sig &= ~(1L<<(n-1));
	if ((p=up->u_signal[n-1]) != 0) {
		up->u_error = 0;
		if (n != SIGILL && n != SIGTRAP)
			up->u_signal[n-1] = 0;
		sendsig(p, n);
		return;
	}
	switch (n) {
	    case SIGQUIT:
	    case SIGILL:
	    case SIGTRAP:
	    case SIGIOT:
	    case SIGEMT:
	    case SIGFPE:
	    case SIGBUS:
	    case SIGSEGV:
	    case SIGSYS:
		up->u_rval1 = n;  /* put sig # in core image */
		if (core())
			n += 0200;
	}
	exit(n);
}

/*
 * find the signal in bit-position
 * representation in p_sig.
 */
fsig(p)
struct proc *p;
{
	register i;
	register long n;

	n = p->p_sig;
	for (i=1; i<=NSIG; i++) {
		if (n & 1L)
			return (i);
		n >>= 1;
	}
	return (0);
}

/*
 * Create a core image on the file "core"
 *
 * It writes USIZE block of the
 * user.h area followed by the entire
 * data+stack segments.
 */
core()
{
	register struct inode *ip;
	register unsigned s;
	extern schar();

	if (u.u_uid != u.u_ruid)
		return(0);
	u.u_error = 0;
	u.u_dirp = "core";
	ip = namei(schar, 1);
	if(ip == NULL) {
		if(u.u_error)
			return(0);
		ip = maknode(0666);
		if (ip==NULL)
			return(0);
	}
	if(!access(ip, IWRITE) && (ip->i_mode&IFMT) == IFREG) {
		itrunc(ip,0L);
		u.u_offset = 0;
		u.u_base = (caddr_t)&u;
		u.u_count = mstob(USIZE);
		u.u_segflg = 1;
		u.u_limit = (daddr_t)mstod(maxmem);
		writei(ip, 0);				/* M000 non-sync */
		u.u_base = DATASTART(u.u_tsize,u.u_sep);
		u.u_count = mstob(u.u_dsize);
		u.u_segflg = 0;

#ifdef FIXEDSTACK
#ifdef SHORTINT
		/* data & stack are in one segment. */
		/* If data segment is full, u.u_count will overflow on */
		/* machines with short integers. Must break in two pieces */
		if (u.u_count == 0 && u.u_dsize!=0) {
			register s;
			s = u.u_dsize/2;
			u.u_count = mstob(s);
			writei(ip, 0);			/* M000 non-sync */
			u.u_base = mstob(s);
			u.u_count = mstob(u.u_dsize-s);
		}
#endif
#endif
		writei(ip, 0);				/* M000 non-sync */
		/* if fixedstack, u.u_ssize is 0 so this does nothing */
		u.u_base = STACKSTART(u.u_ssize);
		u.u_count = mstob(u.u_ssize);
		writei(ip, 0);				/* M000 non-sync */
	} else
		u.u_error = EACCES;
	iput(ip);
	return(u.u_error==0);
}
