/*
 *	@(#) sys/subr.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: subr.c,v 6.0 87/10/12 17:09:59 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */

/* MODIFICATION HISTORY
 *
 * M000	lion	2-13-84
 *	- Actually, there is a routine in here called novec, which
 *		panics on an unexpected inturrupt. In one port,
 *		the vectors seem to vary depending on some hardware
 *		control register values. Because of this, novec was
 *		moved to mdep/trap.c, the arguments expanded, and the
 *		automatic panic was removed.
 *		The old novec should probably be retained for other ports.
 *
 * $Log:	subr.c,v $
 * Revision 6.0  87/10/12  17:09:59  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/08/23  10:36:38  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:49:28  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:44:33  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:23:24  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.1  84/05/17  16:43:20  iv
 * 	- Put back novec() routine, like I said I would.
 * 
 * Revision 0.1  84/05/17  15:20:14  iv
 * 	- novec() code was taken out.  Will put back in next revision.
 * 
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/inode.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/buf.h"

/*
 * Bmap defines the structure of file system storage
 * by returning the physical block number on a device given the
 * inode and the logical block number in a file.
 * When convenient, it also leaves the physical
 * block number of the next block of the file in rablock
 * for use in read-ahead.
 */
daddr_t
bmap(ip, bn, rwflg)
register struct inode *ip;
daddr_t bn;
{
	register i;
	struct buf *bp, *nbp;
	int j, sh;
	daddr_t nb;
	dev_t dev;
	int sflg;

	sflg = (rwflg & B_ASYNC)==0;
	rwflg &= ~B_ASYNC;

	if ((bn < 0) || (rwflg==B_WRITE && bn*(BSIZE/512) >= u.u_limit)){
		u.u_error = EFBIG;
		return((daddr_t)-1);
	}
	dev = ip->i_dev;
	rablock = 0;

	/*
	 * blocks 0..NADDR-4 are direct blocks
	 */
	if(bn < NADDR-3) {
		i = bn;
		nb = ip->i_addr[i];
		if(nb == 0) {
			if(rwflg==B_READ || (bp = alloc(dev,sflg))==NULL)
				return((daddr_t)-1);
			nb = bp->b_blkno;
			bdwrite(bp);
			ip->i_addr[i] = nb;
			ip->i_flag |= IUPD|ICHG;
		}
		if(i < NADDR-4)
			rablock = ip->i_addr[i+1];
		return(nb);
	}

	/*
	 * addresses NADDR-3, NADDR-2, and NADDR-1
	 * have single, double, triple indirect blocks.
	 * the first step is to determine
	 * how many levels of indirection.
	 */
	sh = 0;
	nb = 1;
	bn -= NADDR-3;
	for(j=3; j>0; j--) {
		sh += NSHIFT;
		nb <<= NSHIFT;
		if(bn < nb)
			break;
		bn -= nb;
	}
	if(j == 0) {
		u.u_error = EFBIG;
		return((daddr_t)-1);
	}

	/*
	 * fetch the address from the inode
	 */
	nb = ip->i_addr[NADDR-j];
	if(nb == 0) {
		if(rwflg==B_READ || (bp = alloc(dev,sflg))==NULL)
			return((daddr_t)-1);
		nb = bp->b_blkno;
		bdwrite(bp);
		ip->i_addr[NADDR-j] = nb;
		ip->i_flag |= IUPD|ICHG;
	}

	/*
	 * fetch through the indirect blocks
	 */
	for(; j<=3; j++) {
		bp = bread(dev, nb);
		if(u.u_error) {
			brelse(bp);
			return((daddr_t)-1);
		}
		sh -= NSHIFT;
		i = (bn>>sh) & NMASK;
		nb = bigetl(bp, i*sizeof(daddr_t));
		if(nb == 0) {
			if(rwflg==B_READ || (nbp = alloc(dev,sflg))==NULL){
				brelse(bp);
				return((daddr_t)-1);
			}
			nb = nbp->b_blkno;
			bdwrite(nbp);
			biputl(bp, i*sizeof(daddr_t), nb);
			if(sflg==0) bdwrite(bp); else bswrite(bp);
		} else
			brelse(bp);
	}

	/*
	 * calculate read-ahead.
	 */
	if(i < NINDIR-1)
		rablock = bigetl(bp, (i+1)*sizeof(daddr_t));
	return(nb);
}

/*
 * Pass back  c  to the user at his location u_base;
 * update u_base, u_count, and u_offset.  Return -1
 * on the last character of the user's read.
 * u_base is in the user data space.
 */
passc(c)
register c;
{
	if(subyte(u.u_base, c) < 0) {
		u.u_error = EFAULT;
		return(-1);
	}
	u.u_count--;
	u.u_offset++;
	u.u_base++;
	return(u.u_count == 0? -1: 0);
}

/*
 * Pick up and return the next character from the user's
 * write call at location u_base;
 * update u_base, u_count, and u_offset.  Return -1
 * when u_count is exhausted.
 * u_base is in the user data space.
 */
cpass()
{
	register c;

	if(u.u_count == 0)
		return(-1);
	if ((c = fubyte(u.u_base)) < 0) {
		u.u_error = EFAULT;
		return(-1);
	}
	u.u_count--;
	u.u_offset++;
	u.u_base++;
	return(c);
}

/*
 * Routine which sets a user error; placed in
 * illegal entries in the bdevsw and cdevsw tables.
 */
nodev()
{

	u.u_error = ENODEV;
}

/*
 * Null routine; placed in insignificant entries
 * in the bdevsw and cdevsw tables.
 */
nulldev()
{
}

/*
 * Routine which logs a system error; placed in
 * illegal entries in the vecintsw table.
 */

novec(vec)
    int vec;
{
	printf("interrupt from unknown device, vec=%d\n",vec);
#ifdef DEBUG
	debug(0);
#else
	panic("unknown interrupt\n");
#endif
}
