/*
 *	@(#) sys/tty.c 1.2 84/01/02 Copyright Microsoft Corporation, 1983
 *
 *  $Header: tty.c,v 6.0 87/10/12 17:10:36 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 *
 * 2-24-84	M001	bobt
 *		only set ECHOK in v7 conversion if ECHO mode (not raw)
 * 3-13-84	M002	pr
 *		added ECHOE
 * 3-29-84      M003    amys
 *              only set INPCK ( input parity check ) in v7 conversion
 *              routine if parity is enabled
 *   M004	14 Apr 1984	blf
 *		- Removed M002, above, as ill-advised.
 *   M005	26 Apr 1984	blf
 *		- Another "fix" from the IBM'ers:
 *		  If not checking parity, use an 8-char data path.
 *   M006	17 Dec 1984	jeIV  (Tandy)
 *		- Changed references to HZ to Hz for configurability
 *		- Put M002 back IN, as we need ECHOE for compatibility
 *		- When canonical processing was turned OFF, if there were any
 *		  characters in tp->t_rawq awaiting a delimiter (such as ^D
 *		  or ^J), the process which had switched it off would not wake
 *		  up and read the characters until another character arrived.
 *		  Solution:  When switching canonical processing OFF,
 *		  set tp->delct to 1 and wakeup(&tp->t_rawq).
 *		- In fixing above bug, found and fixed another bug.
 *		  In the v7 emulation code for TIOCSETN, if we were in 'cooked'
 *		  mode [ie, (l_flag&ICANON) == ICANON], and RAW or CBREAK is
 *		  being requested, then we tried to save the EOF and EOL chars.
 *		  However, this code was never being executed because 'l_flag' 
 *		  had been set to 0 earlier in the code.  What we really wanted
 *		  was to test what l_flag USED to be, and so code was added to
 *		  save the old value, which we needed for the above fix anyway.
 */


/*
 * general TTY subroutines
 */
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/tty.h"
#include "../h/ttold.h"
#include "../h/proc.h"
#include "../h/file.h"
#include "../h/conf.h"
#include "../h/ioctl.h"
#include "../h/sysinfo.h"

/*
 * tty low and high water marks
 * high < TTYHOG
 */
int	tthiwat[16] = {
	0, 60, 60, 60,
	60, 60, 60, 120,
	120, 180, 180, 240,
	240, 240, 100, 100,
};
int	ttlowat[16] = {
	0, 20, 20, 20,
	20, 20, 20, 40,
	40, 60, 60, 80,
	80, 80, 50, 50,
};

char    ttcchar[NCC+2] = {
	CINTR,
	CQUIT,
	CERASE,
	CKILL,
	CEOF,
	0,
	0,
	0,
	CEOF,
	0
};

/* null clist header */
struct clist ttnulq;

/* canon buffer */
char	canonb[CANBSIZ];
/*
 * Input mapping table-- if an entry is non-zero, when the
 * corresponding character is typed preceded by "\" the escape
 * sequence is replaced by the table value.  Mostly used for
 * upper-case only terminals.
 */
char    maptab[] = {
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,'|',000,000,000,000,000,'`',
	'{','}',000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,000,000,
	000,000,000,000,000,000,'~',000,
	000,'A','B','C','D','E','F','G',
	'H','I','J','K','L','M','N','O',
	'P','Q','R','S','T','U','V','W',
	'X','Y','Z',000,000,000,000,000,
};

#ifndef MAXTTYS
#define MAXTTYS 16
#endif

short maxttys=0;

/*
 * routine called on first teletype open.
 * establishes a process group for distribution
 * of quits and interrupts from the tty.
 */
ttopen(tp)
register struct tty *tp;
{
	register struct proc *pp;

	pp = u.u_procp;
	if ((pp->p_pid == pp->p_pgrp)
	 && (u.u_ttyp == NULL)
	 && (tp->t_pgrp == 0)) {
		if(maxttys >= MAXTTYS){
			u.u_error = ENOTTY;
			return;
		}
		maxttys++;
		u.u_ttyp = tp;
		tp->t_pgrp = pp->p_pgrp;
	}
	tp->t_state &= ~WOPEN;
	tp->t_state |= ISOPEN;
}

ttclose(tp)
register struct tty *tp;
{
	spl5();
	(*tp->t_proc)(tp, T_RESUME);
	spl0();
	ttyflush(tp, FREAD);
	ttywait(tp);
	tp->t_state &= ~(ISOPEN|WOPEN);
}

/*
 * common ioctl tty code
 */
ttiocom(tp, cmd, arg, mode)
register struct tty *tp;
{
	register flag;
	struct termio cb;
	struct sgttyb tb;
	struct tc tun;          /* V7 */
	unsigned t;             /* V7 */

	switch(cmd) {
	case IOCTYPE:
		u.u_rval1 = TIOC;
		break;

	case TCSETAW:
	case TCSETAF:
		ttywait(tp);
		if (cmd == TCSETAF)
			ttyflush(tp, (FREAD|FWRITE));
	case TCSETA:
		if (copyin(arg, &cb, sizeof cb)) {
			u.u_error = EFAULT;
			break;
		}
		if (tp->t_line != cb.c_line) {
			if (cb.c_line < 0 || cb.c_line >= linecnt) {
				u.u_error = EINVAL;
				break;
			}
			(*linesw[tp->t_line].l_ioctl)(tp, LDCLOSE, 0, mode);
		}
		flag = tp->t_lflag;
		tp->t_iflag = cb.c_iflag;
		tp->t_oflag = cb.c_oflag;
		tp->t_cflag = cb.c_cflag;
		tp->t_lflag = cb.c_lflag;
		bcopy(cb.c_ccs, tp->t_cc, NCC);
		if(tp->t_lflag&ICANON){
		    tp->t_cc[VCEOF]=tp->t_cc[VEOF];
		    tp->t_cc[VCEOL]=tp->t_cc[VEOL];
		}
		else if (flag&ICANON) {				/*M006*/
			if (tp->t_rawq.c_cc > 0) {		/*M006*/
				tp->t_delct = 1;		/*M006*/
				wakeup(&tp->t_rawq);		/*M006*/
			}					/*M006*/
		}						/*M006*/
		if (tp->t_line != cb.c_line) {
			tp->t_line = cb.c_line;
			(*linesw[tp->t_line].l_ioctl)(tp, LDOPEN, 0, mode);
		} else if (tp->t_lflag != flag) {
			(*linesw[tp->t_line].l_ioctl)(tp, LDCHG, flag, mode);
		}
		return(1);

	case TCGETA:
		cb.c_iflag = tp->t_iflag;
		cb.c_oflag = tp->t_oflag;
		cb.c_cflag = tp->t_cflag;
		cb.c_lflag = tp->t_lflag;
		cb.c_line = tp->t_line;
		bcopy(tp->t_cc, cb.c_ccs, NCC);
		if (copyout(&cb, arg, sizeof cb))
			u.u_error = EFAULT;
		break;

	case TCSBRK:
		ttywait(tp);
		if (arg == 0)
			(*tp->t_proc)(tp, T_BREAK);
		break;

	case TCXONC:
		switch (arg) {
		case 0:
			(*tp->t_proc)(tp, T_SUSPEND);
			break;
		case 1:
			(*tp->t_proc)(tp, T_RESUME);
			break;
		case 2:
			(*tp->t_proc)(tp, T_BLOCK);
			break;
		case 3:
			(*tp->t_proc)(tp, T_UNBLOCK);
			break;
		default:
			u.u_error = EINVAL;
		}
		break;

	case TCFLSH:
		switch (arg) {
		case 0:
		case 1:
		case 2:
			ttyflush(tp, (arg - FOPEN)&(FREAD|FWRITE));
			break;

		default:
			u.u_error = EINVAL;
		}
		break;

/* conversion aide only */
	case TIOCSETP:
		ttywait(tp);
		ttyflush(tp, (FREAD|FWRITE));
	case TIOCSETN:
		if (copyin(arg, &tb, sizeof(tb))) {
			u.u_error = EFAULT;
			break;
		}
		t = tp->t_lflag;				/*M006*/
		tp->t_iflag = (tp->t_iflag & (IGNBRK|IXANY));
		tp->t_oflag = (tp->t_oflag & (OFILL|OFDEL));
		tp->t_lflag = (tp->t_lflag & (NOFLSH|XCLUDE));
		tp->t_cflag = ((tb.sg_ospeed&CBAUD)|CREAD)|(tp->t_cflag&(HUPCL|CLOCAL));
		if ((tb.sg_ospeed&CBAUD)==B110)
			tp->t_cflag |= CSTOPB;
		tp->t_cc[VERASE] = tb.sg_erase;
		tp->t_cc[VKILL] = tb.sg_kill;
		flag = tb.sg_flags;
/*****          if (flag&O_HUPCL)               *****/
/*****                  tp->t_cflag |= HUPCL;   *****/
		if (flag & O_TANDEM)
			tp->t_iflag |= IXOFF;
		if (flag&O_XTABS)
			tp->t_oflag |= TAB3;
		else if (flag&O_TBDELAY)
			if(flag&O_TAB1)
			    tp->t_oflag |= TAB1;
			else tp->t_oflag |= TAB2;
		if (flag&O_LCASE) {
			tp->t_iflag |= IUCLC;
			tp->t_oflag |= OLCUC;
			tp->t_lflag |= XCASE;
		}
		if (flag&O_ECHO)
			tp->t_lflag |= ( ECHO | ECHOK | ECHOE );  /* M002 */
		if (flag&O_CRMOD) {
			tp->t_iflag |= ICRNL;
			tp->t_oflag |= ONLCR;
			if (flag&O_CR1)
				tp->t_oflag |= CR1;
			if (flag&O_CR2)
				tp->t_oflag |= ONOCR|CR2;
		} else {
			tp->t_oflag |= ONLRET;
			if (flag&O_NL1)
				tp->t_oflag |= CR1;
			if (flag&O_NL2)
				tp->t_oflag |= CR2;
		}
		if (flag&O_RAW || flag&O_CBREAK) {
			if(t&ICANON){				/*M006*/
				tp->t_cc[VCEOF] = tp->t_cc[VEOF];
				tp->t_cc[VCEOL] = tp->t_cc[VEOL];
				if (tp->t_rawq.c_cc > 0) {	/*M006*/
					tp->t_delct = 1;	/*M006*/
					wakeup(&tp->t_rawq);	/*M006*/
				}				/*M006*/
			}					/*M006*/
			tp->t_cc[VTIME] = 0;
			tp->t_cc[VMIN] = 1;
			if (flag&O_RAW) {
				tp->t_iflag &= ~(ICRNL|IUCLC);
				tp->t_cflag |= CS8;
			} else {
				tp->t_iflag |= BRKINT|IGNPAR|ISTRIP|IXON;
				tp->t_oflag |= OPOST;
				tp->t_cflag |= CS7|PARENB;
				tp->t_lflag |= ISIG;
			}
		} else {        /* cooked mode */
			tp->t_cc[VEOF] = tp->t_cc[VCEOF];
			tp->t_cc[VEOL] = tp->t_cc[VCEOL];
			tp->t_iflag |= BRKINT|IGNPAR|ISTRIP|IXON;
			tp->t_oflag |= OPOST;
			tp->t_cflag |= CS7|PARENB;
			tp->t_lflag |= ICANON|ISIG;
		}
		tp->t_iflag |= INPCK;
		if (flag&O_ODDP)
			if (flag&O_EVENP)
			/* M003 only enable parity checking of 
			 * input if either odd or even parity is 
			 * requested
			 * M005 also, if no parity checking,
			 * accept the full 8 bits
			 */
		        /* tp->t_iflag &= ~INPCK; */
		        {
				tp->t_iflag &= ~INPCK; 
				tp->t_cflag &= ~(PARENB|CSIZE);  /*M003 M005*/
				tp->t_cflag |= CS8;
			}
			else
				tp->t_cflag |= PARODD;
		if (flag&O_VTDELAY)
			tp->t_oflag |= FFDLY;
		if (flag&O_BSDELAY)
			tp->t_oflag |= BSDLY;
		return(1);

	case TIOCGETP:
		tb.sg_ospeed = tp->t_cflag&CBAUD;
		tb.sg_ispeed = tb.sg_ospeed;
		tb.sg_erase = tp->t_cc[VERASE];
		tb.sg_kill = tp->t_cc[VKILL];
		flag = 0;
/*****          if (tp->t_cflag&HUPCL)       ******/
/*****                  flag |= O_HUPCL;     ******/
		if (tp->t_iflag & IXOFF)
			flag |= O_TANDEM;
		if (!(tp->t_lflag&ICANON))
		   if((tp->t_oflag&OPOST)==0)
			flag |= O_RAW;
		   else flag |= O_CBREAK;
		if (tp->t_lflag&XCASE)
			flag |= O_LCASE;
		if (tp->t_lflag&ECHO)
			flag |= O_ECHO;
/*****          if (!(tp->t_lflag&ECHOK))    ******/
/*****                  flag |= O_NOAL;      ******/
		if (tp->t_cflag&PARODD)
			flag |= O_ODDP;
		else if (tp->t_iflag&INPCK)
			flag |= O_EVENP;
		else
			flag |= O_ODDP|O_EVENP;
		if (tp->t_oflag&ONLCR) {
			flag |= O_CRMOD;
			if (tp->t_oflag&CR1)
				flag |= O_CR1;
			if (tp->t_oflag&CR2)
				flag |= O_CR2;
		} else {
			if (tp->t_oflag&CR1)
				flag |= O_NL1;
			if (tp->t_oflag&CR2)
				flag |= O_NL2;
		}
		if ((tp->t_oflag&TABDLY)==TAB3)
			flag |= O_XTABS;
		else if (tp->t_oflag&TAB1)
			flag |= O_TAB1;
		else if (tp->t_oflag&TAB2)
			flag |= O_TAB2;
		if (tp->t_oflag&FFDLY)
			flag |= O_VTDELAY;
		if (tp->t_oflag&BSDLY)
			flag |= O_BSDELAY;
		tb.sg_flags = flag;
		if (copyout(&tb, arg, sizeof(tb)))
			u.u_error = EFAULT;
		break;
	case TIOCSETC:
		if (copyin(arg, (caddr_t)&tun, sizeof(struct tc)))
			u.u_error = EFAULT;
		tp->t_cc[VINTR]=tun.t_intrc;
		tp->t_cc[VQUIT]=tun.t_quitc;
		tp->t_cc[VCEOF]=tun.t_eofc;
		tp->t_cc[VCEOL]=tun.t_brkc;
		if(tp->t_lflag&ICANON){
		    tp->t_cc[VEOF]=tun.t_eofc;
		    tp->t_cc[VEOL]=tun.t_brkc;
		}
		/* changing start & stop not supported */
		break;
	case TIOCGETC:
		tun.t_intrc=tp->t_cc[VINTR];
		tun.t_quitc=tp->t_cc[VQUIT];
		tun.t_eofc=tp->t_cc[VCEOF];
		tun.t_brkc=tp->t_cc[VCEOL];
		tun.t_startc=CSTART;
		tun.t_stopc=CSTOP;
		if (copyout((caddr_t)&tun, arg, sizeof(struct tc)))
			u.u_error = EFAULT;
		break;
	case TIOCEXCL:
		tp->t_lflag |= XCLUDE;
		break;
	case TIOCNXCL:
		tp->t_lflag &= ~XCLUDE;
		break;
	case TIOCHPCL:
		tp->t_cflag |= HUPCL;
		break;
	case TIOCFLUSH:
		ttyflush(tp, (FREAD|FWRITE));
		break;
	case TIOCGETD:
		t = tp->t_line;
		if (copyout((caddr_t)&t, arg, sizeof(t)))
			u.u_error = EFAULT;
		break;
	case TIOCSETD:
		if (copyin(arg, (caddr_t)&t, sizeof(t))) {
			u.u_error = EFAULT;
			break;
		}
		if (t >= linecnt) {
			u.u_error = ENXIO;
			break;
		}
		if (tp->t_line)
			(*linesw[tp->t_line].l_ioctl)(tp, LDCLOSE, 0, mode);
		if (t)
			(*linesw[tp->t_line].l_ioctl)(tp, LDOPEN, 0, mode);
		if (u.u_error==0)
			tp->t_line = t;
		break;
	case FIORDCHK:
		if ((tp->t_state&CARR_ON)==0)
			break;
		if(tp->t_canq.c_cc != 0)
			u.u_rval1 = 1;
		else if((tp->t_lflag&ICANON)==0 && tp->t_rawq.c_cc > 0)
			u.u_rval1 = 1;
		else if((tp->t_lflag&ICANON)!=0 && tp->t_delct >0)
			u.u_rval1 = 1;
		break;
	case DIOCSETP:
	case DIOCGETP:
		break;  /* what to do here? */

/* New, local functions */
	/*
	 * Return number of characters in
	 * the output.
	 */
	case TIOCOUTQ:
		if (copyout((caddr_t)&tp->t_outq.c_cc, arg, sizeof(tp->t_outq.c_cc)))
			u.u_error = EFAULT;
		break;

	/*
	 * Simulate typing of a character at the terminal.
	 */
	case TIOCSTI: 
	{
		int	c;

		c = fubyte(arg);
		if (u.u_uid && u.u_ttyp != tp || c < 0)
			u.u_error = EFAULT;
		else
			(*linesw[tp->t_line].l_input)(tp, c, 0);
		break;
	}
/* end of locals */

	default:
		if ((cmd&IOCTYPE) == LDIOC)
			(*linesw[tp->t_line].l_ioctl)(tp, cmd, arg, mode);
		else
			u.u_error = EINVAL;
		break;
	}
	return(0);
}

ttinit(tp)
register struct tty *tp;
{
	tp->t_line = 0;
	tp->t_iflag = 0;
	tp->t_oflag = OPOST;
	tp->t_cflag = SSPEED|CS8|CREAD|HUPCL;
	tp->t_lflag = 0;
	bcopy(ttcchar, tp->t_cc, NCC+2);
}

ttywait(tp)
register struct tty *tp;
{
	spl5();
	while (tp->t_outq.c_cc || (tp->t_state&(BUSY|TIMEOUT))) {
		tp->t_state |= TTIOW;
		sleep((caddr_t)&tp->t_oflag, TTOPRI);
	}
	spl0();
	delay(Hz/15);
}

/*
 * flush TTY queues
 */
ttyflush(tp, cmd)
register struct tty *tp;
{
	register struct cblock *cp;
	register s;

	if (cmd&FWRITE) {
		while ((cp = getcb(&tp->t_outq)) != NULL)
			putcf(cp);
		(*tp->t_proc)(tp, T_WFLUSH);
		if (tp->t_state&OASLP) {
			tp->t_state &= ~OASLP;
			wakeup((caddr_t)&tp->t_outq);
		}
		if (tp->t_state&TTIOW) {
			tp->t_state &= ~TTIOW;
			wakeup((caddr_t)&tp->t_oflag);
		}
	}
	if (cmd&FREAD) {
		while ((cp = getcb(&tp->t_canq)) != NULL)
			putcf(cp);
		s = spl5();
		while ((cp = getcb(&tp->t_rawq)) != NULL)
			putcf(cp);
		tp->t_delct = 0;
		splx(s);
		(*tp->t_proc)(tp, T_RFLUSH);
		if (tp->t_state&IASLP) {
			tp->t_state &= ~IASLP;
			wakeup((caddr_t)&tp->t_rawq);
		}
	}
}

/*
 * Transfer raw input list to canonical list,
 * doing erase-kill processing and handling escapes.
 */
canon(tp)
register struct tty *tp;
{
	register char *bp;
	register c, esc;

	spl5();
	if (tp->t_rawq.c_cc == 0)
		tp->t_delct = 0;
	while (tp->t_delct == 0) {
		if ((tp->t_state&CARR_ON)==0) {
			spl0();
			return;
		}
		if (u.u_fmode & FNDELAY)
			return;
		tp->t_state |= IASLP;
		sleep((caddr_t)&tp->t_rawq, TTIPRI);
	}
	if (!(tp->t_lflag&ICANON)) {
		tp->t_canq = tp->t_rawq;
		tp->t_rawq = ttnulq;
		tp->t_delct = 0;
		spl0();
		return;
	}
	spl0();
	bp = canonb;
	esc = 0;
	while ((c=getc(&tp->t_rawq)) >= 0) {
		if (!esc) {
			if (c == '\\') {
				esc++;
			} else if (c == tp->t_cc[VERASE]) {
				if (bp > canonb)
					bp--;
				continue;
			} else if (c == tp->t_cc[VKILL]) {
				bp = canonb;
				continue;
			} else if (c == tp->t_cc[VEOF]) {
				break;
			}
		} else {
			esc = 0;
			if (c == tp->t_cc[VERASE] ||
			    c == tp->t_cc[VKILL] ||
			    c == tp->t_cc[VEOF])
				bp--;
			else if (tp->t_lflag&XCASE) {
				if ((c < 0200) && maptab[c]) {
					bp--;
					c = maptab[c];
				} else if (c == '\\')
					continue;
			} else if (c == '\\')
				esc++;
		}
		*bp++ = c;
		if (c == '\n' || c == tp->t_cc[VEOL])
			break;
		if (bp >= &canonb[CANBSIZ])
			bp--;
	}
	tp->t_delct--;
	c = bp - canonb;
	sysinfo.canch += c;
	bp = canonb;
/* faster copy ? */
	while (c--)
		putc(*bp++, &tp->t_canq);
	return;
}

/*
 * Restart typewriter output following a delay timeout.
 * The name of the routine is passed to the timeout
 * subroutine and it is called during a clock interrupt.
 */
ttrstrt(tp)
register struct tty *tp;
{

	(*tp->t_proc)(tp, T_TIME);
}

/*
 * Called from device's read routine after it has
 * calculated the tty-structure given as argument.
 */
ttread(tp)
register struct tty *tp;
{
	register struct clist *tq;

	tq = &tp->t_canq;
	if (tq->c_cc == 0)
		canon(tp);
	if (tq->c_cc == 0 && (u.u_fmode & FNDELAY))
		return;

	while (u.u_count!=0 && u.u_error==0) {
		if (u.u_count >= CLSIZE) {
			register n;
			register struct cblock *cp;

			cp = getcb(tq);
			if (cp == NULL)
				break;
			n = min(u.u_count, cp->c_last - cp->c_first);
			if (copyout(&cp->c_data[cp->c_first], u.u_base, n))
				u.u_error = EFAULT;
			putcf(cp);
			u.u_base += n;
			u.u_count -= n;
		} else {
			register c;

			if ((c = getc(tq)) < 0)
				break;
			if (subyte(u.u_base++, c))
				u.u_error = EFAULT;
			u.u_count--;
		}
	}
	if (tp->t_state&TBLOCK) {
		if (tp->t_rawq.c_cc<TTXOLO) {
			(*tp->t_proc)(tp, T_UNBLOCK);
		}
	}
}

/*
 * Called from device's write routine after it has
 * calculated the tty-structure given as argument.
 */
ttwrite(tp)
register struct tty *tp;
{

	if ((tp->t_state&CARR_ON)==0)
		return;
	while (u.u_count) {
		spl5();
		while (tp->t_outq.c_cc > tthiwat[tp->t_cflag&CBAUD]) {
			(*tp->t_proc)(tp, T_OUTPUT);
			/* don't sleep if drained all at once */
			if(tp->t_outq.c_cc != 0){
			    tp->t_state |= OASLP;
			    sleep((caddr_t)&tp->t_outq, TTOPRI);
			}
		}
		spl0();
		if (u.u_count >= (CLSIZE/2)) {
			register n;
			register struct cblock *cp;

			if ((cp = getcf()) == NULL)
				break;
			n = min(u.u_count, cp->c_last);
			if (copyin(u.u_base, cp->c_data, n)) {
				u.u_error = EFAULT;
				putcf(cp);
				break;
			}
			u.u_base += n;
			u.u_count -= n;
			cp->c_last = n;
			ttxput(tp, cp, n);
		} else {
			register c;

			c = fubyte(u.u_base++);
			if (c<0) {
				u.u_error = EFAULT;
				break;
			}
			u.u_count--;
			ttxput(tp, c, 0);
		}
	}
	spl5();
	(*tp->t_proc)(tp, T_OUTPUT);
	spl0();
}
