/*
**  MMU defines for Tandy 68000/XENIX on the Model 16/6000
**
**	Totally and inexorably machine dependant.
*/

/*#define MMDEBUG 3			/* Enable debugging statements */


/************************

   #defines for mmu routines which can be macros for efficiency

************************/

/* Misc "constants" to make defines below easier to read */
#define _paddr(pp)	(pp)->p_addr.p_caddr
#define _mmu_ulimit	((long)maxmem - USIZE - u.u_ssize - u.u_tsize)

#define meminit()		/* No memory initialization necessary */
#ifndef MMDEBUG
#  define mmuinit()		/* No MMU initialization necessary */
#  define mmuget(msize)		((mloc_t)malloc(coremap, (int)msize))
#  define mmupget(p)		mmuget((p)->p_size)
#  define mmupfree(p)		mmufree(_paddr(p), (p)->p_size)
#  define mmupcopy(fr, to)	copybuf(mltoa(_paddr(fr)), mltoa(_paddr(to)), mstob((fr)->p_size))
#  define mmuclear(mloc, msize)	clearbuf(mltoa(mloc), mstob(msize))
#  define mmucopy(fr, to, sz)	copybuf(mltoa(fr), mltoa(to), mstob(sz))
#  define givemap(rp1, rp2)	/* No MMU action necessary (or possible) */
#  define mmumap()		mmusub(u.u_procp, (int)u.u_lxrw)
#  define mmuset(rwfl)		mmusub(u.u_procp, (int)(u.u_lxrw = (rwfl)))
#  define mmu_ulimit()		((long)(mstob(_mmu_ulimit)))
#endif /*MMDEBUG*/



/************************

   #defines for where the world is and how it wants to be programmed

	The 16/6000's MMU is programmed with 3 words: 
	    o The lower 8 bits of the Lo segment limit ORd
	      with the lower 8 bits of the Lo segment offset
	    o The lower 8 bits of the Hi segment limit ORd
	      with the lower 8 bits of the Hi segment offset
	    o The upper 3 bits of each of the above, 
	      each in a nybble (see below)

	The reason it is done this way is that the CPU didn't used to 
	have the upper 3 bits of the address.  When it was designed, the
	engineer used an 8-bit adder for the MMU.   So, the MMU could only
	handle memory up to 1 megabyte, which was thought to be plenty.

	Later, an MMU upgrade was designed to add the remaining 3 bits 
	needed to fully utilize the 16/6000's 22 address bits, which 
	brought the memory usable by the MMU to 7 megabytes.

************************/

/* Defines for programming the lower 8 bits of the MMU */
#define	MMU_MkLow(offset, limit)  (((limit) << 8) | (offset))
#define	MMU_HI		(* (unsigned short *) 0x7800d2)
#define	MMU_LO		(* (unsigned short *) 0x7800d4)

/* I know, I know; this is hard to read.  I find it difficult, and I wrote it!
 * It used to be MUCH worse; _MkByte's code used to be inline in MMU_MkRest()!
 *
 *	The upper 3 bits of each MMU register are programmed in one last
 *	word write.  This word is layed out like this:
 *
 *		      Hi-seg | Lo-seg
 *		    =========|=========
 *		     Lim|Off | Lim|Off
 *		    ----|----|----|----
 *	    7800D8: x111|x111|x111|x111
 *
 *	So, we grab the top 3 bits from the high byte of each offset and limit
 *	and shift/OR them into place.
 *
 *****************************************************************************/

/* Assemble byte for last word */
#define _MkByte(Off, Lim)  ((((Lim)&7) << 4) | ((Off)&7))
/* Put the whole word together */
#define MMU_MkRest(HiOff, HiLim, LoOff, LoLim)  ( (_MkByte(HiOff,HiLim) << 8) |  _MkByte(LoOff, LoLim) )
/* Address where final word goes (for the MMU upgrade) */
#define	MMU_REST	(* (unsigned short *) 0x7800d8)


/* Read-Only and Read-Write function code defines */
#define	RO		0
#define	RW		1

/* This is assumed to be a zero, because mmu routines use
 *   malloc's error return of 0 */
#define	MMUERR		((mloc_t) 0)	/* mmuget() & mmupget() error return */



/*******************************

    Misc extern declarations

*******************************/

int	xrwflag;		/* flag holding text protection status */
				/*	(unused)		       */

extern msize_t pagoff;

extern struct map *coremap;

#ifdef MMDEBUG
extern mloc_t mmuget();
extern mloc_t mmupget();
#endif /*MMDEBUG*/
