/*
 *	%Z% %M% %I% %D% %Q%
 */



/*
 * THIS FILE CONTAINS CODE WHICH IS SPECIFIC TO THE
 * Tandy Model 16 COMPUTER. IT MAY REQUIRE CHANGING WHEN
 * MOVING XENIX TO DIFFERENT MACHINE ARCHITECTURES
 * AND CONFIGURATIONS.
 */



/* Conditional compilation */
/*#define PAGEMMU*/             /* mmu is divided in pages */
/*#define SHORTINT*/            /* sizeof(int)==sizeof(short) */
#define BYTESWAP                /* byte ordering is opposite that of pdp11 */
#define OLDFORK                 /* return to parent by bumping pc by 2 */
#define FIXEDSTACK		/* stack cannot grow */
/*#define BUFMAPOUT*/           /* buffers are mapped out */
/*#define LARGECODE*/           /* kernel supports multisegment code */


/*
 * Adjustable parameters
 */
#ifdef FIXEDSTACK
#  define SSIZE  0x2000		/* Large initial stack */
#else
#  define SSIZE  2048		/* initial stack size (bytes) */
#  define SINCR  1024		/* increment of stack (bytes) */
#endif
#define NOFILE  20              /* max open files per process */
#define CANBSIZ 256		/* max size of typewriter line */
#define NCARGS	5120		/* # characters in exec arglist */
#define MAXTTYS 16              /* Max # open ttys */
#define NIOSTAT 50              /* max number of bufs to keep stats for */
#define	MAXPID	30000		/* max process id */
#define	MAXUID	60000		/* max user id */
#define	MAXLINK	1000		/* max links */
#define	CDLIMIT	(1L<<21)	/* default max write address */
#define NSDSEGS 5               /* number of shared data segs/proc */
/* WARNING:  Do not change this number without changing the Z80 code which
 *	     saves msgbuf[] to disk on shutdown/crash */
#define MSGBUFS 504		/* Characters saved from error messages */


/*
 * priorities
 * probably should not be
 * altered too much
 */

#define PSWP	0
#define PINOD	10
#define PRIBIO	20
#define PZERO	25
#define NZERO	20
#define PPIPE	26
#define PWAIT	30
#define PSLEP	40
#define PUSER	50
#define PIDLE   127

/*
 * signals
 * dont change
 */

#define	NSIG	20
/*
 * No more than 32 signals (1-32) because they are
 * stored in bits in a long.
 */
#define	SIGHUP	1	/* hangup */
#define	SIGINT	2	/* interrupt (rubout) */
#define	SIGQUIT	3	/* quit (ASCII FS) */
#define	SIGILL	4	/* illegal instruction (not reset when caught) */
#define	SIGTRAP	5	/* trace trap (not reset when caught) */
#define	SIGIOT	6	/* IOT instruction */
#define	SIGEMT	7	/* EMT instruction */
#define	SIGFPE	8	/* floating point exception */
#define	SIGKILL	9	/* kill (cannot be caught or ignored) */
#define	SIGBUS	10	/* bus error */
#define	SIGSEGV	11	/* segmentation violation */
#define	SIGSYS	12	/* bad argument to system call */
#define	SIGPIPE	13	/* write on a pipe with no one to read it */
#define	SIGALRM	14	/* alarm clock */
#define	SIGTERM	15	/* software termination signal from kill */
#define	SIGUSR1	16	/* user defined signal 1 */
#define	SIGUSR2	17	/* user defined signal 2 */
#define	SIGCLD	18	/* death of a child */
#define	SIGPWR	19	/* power-fail restart */

/*
 * fundamental constants of the implementation--
 * cannot be changed easily
 */

#define NBPW	sizeof(int)	/* number of bytes in an integer */
#define BSIZE   512             /* size of secondary block (bytes) */
/* BSLOP can be 0 unless you have a TIU/Spider */
#define BSLOP   0             /* must be 0 for current copyio implementation */
#define NINDIR	(BSIZE/sizeof(daddr_t))
#define BMASK   00777           /* BSIZE-1 */
#define BSHIFT  9               /* LOG2(BSIZE) */
#define NMASK   0177            /* NINDIR-1 */
#define NSHIFT  7               /* LOG2(NINDIR) */
#define	INOPB	(BSIZE/sizeof(struct dinode))	/* # inodes per block */
#define LINOPB  3               /* LOG2(INOPB) */
#define NULL	0
#define NODEV	(dev_t)(-1)
#define ROOTINO ((ino_t)2)	/* i number of all roots */
#define SUPERB	((daddr_t)1)	/* block number of the super block */
#define DIRSIZ	14		/* max characters per directory */
#define NICINOD 100		/* number of superblock inodes */
#define NICFREE 50              /* number of superblock free blocks */
#define NSBFILL 51              /* aligns s_magic & s_type at end of SB blk */

/*
 * MMU parameters.
 */

#define MMPGSZ	4096		/* bytes/page in the MMU */
#define LMMPGSZ	12		/* log2(MMPGSZ) */

/*
 * Some macros for units conversion
 */

/* mem size to disk blocks */
#define mstod(x) ((x)*(MMPGSZ/BSIZE))

/* inumber to disk address */
#define itod(x) (daddr_t)(((unsigned)(x)+(INOPB+INOPB-1))>>LINOPB)

/* inumber to disk offset */
#define itoo(x) (int)(((x)+(INOPB+INOPB-1))&(INOPB-1))

/* mem size to bytes */
#define mstob(x) ((x)<<LMMPGSZ)

/* bytes to mem size */
#define btoms(x) (((unsigned)(x)+(MMPGSZ-1))>>LMMPGSZ)

/* bytes to mem size (truncated)*/
#define btomst(x) (((unsigned)(x))>>LMMPGSZ)

/* Memory location to logical address */
#define mltoa(x) ((char *) ((unsigned) (x) << LMMPGSZ))

/* physical address to memory location -- used only at startup */
#define atoml(x) ((mloc_t) ((unsigned) (x) >> LMMPGSZ))

/* bytes to page number */
#define btopn(x) (((unsigned)(x))>>LMMPGSZ)

/* address (long (32 bit)) to page number (int)*/
#define atopn(x) ((int)(((long)(x))>>LMMPGSZ))

/* address (long (32 bit)) to page count (int)*/
#define atop(x)  ((int)(((long)(x)+(MMPGSZ-1))>>LMMPGSZ))

/* address (long (32 bit)) to offset (int) get bits LMMPGSZ-1 - 0 */
#define atoo(x) ((int)((x)&(MMPGSZ-1)))

/* long address to short address (get low 16 bits of long address */
#define atos(x) ((int)( (x) & 0x0000FFFF))

/* long address to short address (get low 16 bits of long address */
#define atoh(x) ((int)( (x) >> 16))

/* mem size to long address */
#define mstol(x) ((long)((int)(x)&(MMPGSZ-1))<<LMMPGSZ)

#include "types.h"

/*
 * Machine-dependent bits and macros
 */

#ifndef KERNEL
#  define  HZ	30
#endif KERNEL
#define SMODE   0x2000         /* system mode bit */
#define USERMODE(sr)    (((sr) & SMODE)!=SMODE)

#ifdef KERNEL
#  define  UBASE 	/* BUG  I sure hope this isn't used by anybody! */
#else KERNEL
#  define  UBASE  0	/* BUG  I sure hope this isn't used by anybody! */
#endif KERNEL
#define USIZE   1	/* size of user block is 4096 bytes */
#define	UBYTES	4096	/* number of actual bytes from &u to end of page */

/* rclock() saves the old mask passed it by assembly support at _ointmask */
extern	unsigned char	ointmask;
#define INTPRI  0xE7			/* mask bits for used interrupts */
#define CLKONLY(ps)	(((ointmask) & INTPRI) == 0)
#define	tasktime(n)	/* BUG  spl0() 	/* Enable interrupts */

#define	TEXTSTART	((caddr_t)0x800000)
#define DATASTART(ts,sep) ((caddr_t)0)	/* start of user data */
#define	STACKSTART(ss)	((caddr_t)0)
#define USRSTACK        ((caddr_t)0)	/* base of user stack */
#define	TEXTPAD		2
#define URSBC           0               /* User stk Reg Save area Byte Cnt */
#define SHWDSZ          (NBPW+6)        /* shell word */
