.\" ===========================================================================
.\"  68000/XENIX 3.2 Kernel Message list and Description
.\"
.\"	John Elliott IV		Revised 12 May 1987
.\"
.\"		Format me with `-ms', please
.\"
.\" =================================================
.\"  Local macro definitions
.de P
.sp \\n(PDu
..
.de SS
.sp
.ne 1.1
.KS
.if t .B
..
.de SE
.if t .R
.KE
.RS
..
.de MT
.I
.if !"\\$1"" \&\\$1 \\$2 \\$3 \\$4 \\$5 \\$6
.R
.RE
..
.\" =================================================
.\"   A little set up...
.ds & \*([.\\n*\*(.]\"		Show current footnote number, no increment
.\"
.if n .ds X XENIX\"		Vanilla XENIX
.if n .ds FT "*\"		Footnote symbol	(using `*' if nroff)
.\"
.if t .ds X \s-2XENIX\s+2\"	Nice XENIX
.if t .ds FT \(dg\"		Footnote symbol	(using `dagger' if troff)
.\"
.cs L 21	\" Set character size to mono for mono "display" font
.\" ===========================================================================
.RP
.TL
.UL
68000/\*X\*(FT Kernel Message List and Description
.R
.FS
\*(FT \*X is a trademark of Microsoft Corporation.
.FE
.AU 
John Elliott IV
.AI 
Tandy Systems Software
.AB
This is a description of the warning and error messages put out by 
various parts of the 68000/\*X 3.2 kernel, and what they mean.
.PP
It does 
.UL not
include a list or description of the messages from 
the Z80 Control Program, these are described in another document.
.PP
Included with the description of each message and its format are
possible causes and their locations.
.PP
.B
This information is valid only for 68000/\*X versions 3.2(42) and later.
.R
For releases of 68000/\*X later than 3.2 the information presented here
may be incomplete, but it will generally be correct.
.AE
.\" ====================================
.PP
This document describes the warning and error messages put out by 
various parts of the 68000/\*X 3.2 kernel, and what they mean.
The messages are categorized as follows:
.sp
.I fatal
.RS
Recovery is impossible, and \*X ceases execution.
.RE
.sp
.I
System Inconsistency
.R
.RS
An unexpected or contradictory situation 
(a \*Qthis can't happen\*U situation) exists in
the kernel.
.RE
.sp
.I Abnormal
.RS
A probably legitimate but extreme situation exists.
.RE
.sp
.I Hardware
.RS
Indicates a hardware problem exists.
.RE
.sp
.I
User error
.R
.RS
The user has caused the problem.
.RE
.sp 2
.PP
Fatal system messages begin with \*Qpanic:\*U and indicate 
hardware problems or kernel inconsistencies that are too 
severe for continued operation.  After displaying a fatal 
message, the system will stop.  Rebooting is required. 
.PP
System inconsistency messages indicate problems usually
traceable to hardware malfunction, such as memory failure.
These messages rarely occur since associated
hardware problems are generally detected before such
an inconsistency can occur.  
.PP
Abnormal messages represent kernel operation problems, 
such as the overflow of critical tables.  It takes extreme 
situations to bring these problems about, so they 
should never occur in normal system use.
.PP
User errors are situations caused by the user doing something 
contradictory, eg, write-protecting a hard drive and then trying
to mount it for write.
.sp
.KS
.SH
System Start-up Messages
.R
.PP
The following text is displayed every time the system is booted:
.DS
.ft L
Tandy 68000/XENIX version 3.\fIn\fP(\fIee\fP)

Microsoft  XENIX  v3.0

Copyright  Microsoft  Corporation,  1984.  All  rights reserved.
Licensed to Tandy Corporation.
Portions copyright 1985,1987 Tandy Corporation.  All rights reserved.
Restricted rights:  Use, duplication,  and  disclosure  are  subject
to the terms stated in the customer Non-Disclosure Agreement.

System \fIkk\fPk+\fIbb\fPk+\fIaa\fPk=\fIttt\fPk User \fIuuu\fPk
Root \fIrrrrr\fPk Swap \fIssss\fPk
.ft P
.DE
.KE
.PP
The first line tells you what version of the kernel you are running.
The 
.B
3.\fIn\fP(\fIee\fP)
.R
tells you you are running release \fIn\fP of version 3 \*X.
The number in parenthesis is the \fIedit\fP number of the kernel. 
Any time System Software changes the kernel for any reason,
this number is incremented.
In this way, you may tell the difference between
various versions of the same release of \*X.
.PP
The next block of text is the copyright and non-disclosure warning.
After that, the system displays how much memory it is using.  This 
information is the sum of three parts:
.DS
.IP \fIkk\fP
the amount of fixed memory the system takes
.IP \fIbb\fP
the amount of memory allocated for buffers
.IP \fIaa\fP
the amount of memory allocated for the other configurable tables \**
.IP \fIttt\fP
is the total amount of memory the system is using
.DE
.FS
For more information on configurable tables, see \fIcfg\fP(C)
.FE
Next, \*X displays how much memory is available for users,
how big the disk partition it is booted on is,
and how much swap area is available.
.PP
From release to release, the size of \*X (\fIkk\fP) does not usually
change, as if it did change, it would probably grow and this would
leave less memory available to the users.  This number is
.UL not
guaranteed to remain the same, however.  Some releases of \*X may add
features which require the system to use more memory.
.sp 2
.KS
.SH
Device Driver Messages
.R
.PP
Device drivers display error messages in four different formats 
(one of which is special and related to cartridge disk operation only):
.\"	Start Display section
.DS L
.RS
.\" Foo Drive n: specific error message
\fIdevice\fP Drive \fIn\fP: \fIspecific error message\fP
.br
.\" Foo Drive n: TYPE error hhhh while {read,writ}ing
.\"					Cylinder CCC Head HH Sector SSSS
\fIdevice\fP Drive \fIn\fP: \fItype\fP error \fIhhhh\fP while Read/Writing Cylinder \fIcyl\fP Head \fIhead\fP Sector \fIsector\fP
.br
.\" Foo Drive n: TYPE error hhhh while formatting
\fIdevice\fP Drive \fIn\fP: \fItype\fP error \fIhhhh\fP while formatting
.br
.\" Cartridge Drive n: TYPE error hhhh while Read/Writing Block BBBBBB
Cartridge Drive \fIn\fP: \fItype\fP error \fIhhhh\fP while Read/Writing Block \fIblock\fP
.RE
.DE				\" End display section
.KE
.sp
.if t .B
\fIdevice\fP Drive \fIn\fP: \fIspecific error message\fP
.if t .R
.br
.RS
This format is used for general errors from the driver which
affect all operations.  For example, the 
message \*QHard Drive 0: active drive not ready\*U is displayed when the
driver tried some hard disk I/O operation on an already active (open)
drive, and the controller reported the drive busy or not ready.
The following messages are possible from each type of drive:
.DS
Further I/O aborted until device closed
Device closed; error cleared
active disk changed
active drive not ready
active drive write protected
bad format
bad sector size
drive not ready
write protected
.DE
.SS
Further I/O aborted until device closed
.br
Device closed; error cleared
.SE
These messages are seen when an unrecoverable error has occurred on
a drive and the driver is refusing to do any more I/O until the device
is closed (or until \*X shuts down).
The first usually appears after the error messages describing the
unrecoverable error, and the second after the device driver is closed.
.MT Hardware \fRor\fP User error.
.SS
active disk changed
.SE
This error occurs when a disk is changed while \*X was doing I/O on it
(ie, it was mounted).  This message is normally seen on floppy drives.
It cannot occur on hard drives, and it is very difficult to cause on
cartridge drives.
.MT User error.
.SS
active drive not ready
.SE
This error occurs when \*X tried some I/O operation on an already active
(open) drive, and the controller reported the drive busy or not ready.
.MT Hardware.
.SS
active drive write protected
.SE
This error occurs when, after the drive was opened for I/O, the user
write-protects the media.  This error most often occurs on a hard disk,
where the error is ignored until the problem is corrected (ie, the user
removes the write-protection).  It is very difficult to create this
error on a floppy or cartridge disk as the write-protection devices on
these are inaccessible while they are in their drives.
.MT User error.
.SS
bad format
.SE
This error indicates that when \*X first examined a disk, the media
format was incorrect.   This usually indicates that the user did not
format the media with \fIdiskutil\fP or \fIformat\fP(C)
before attempting to use it.
.MT User error.
.SS
bad sector size
.SE
This is a very unusual error.  It indicates that when \*X examined a
disk, it had an unusual (ie, non-standard) sector size.
This error is unusual because for raw I/O, \*X will handle most sector
sizes supported by the controller.
This error usually indicates something is wrong with the Z80 hardware.
.MT Hardware.
.SS
drive not ready
.SE
This error is displayed when \*X first attempts I/O on a drive and the
controller says the drive is not ready.  This is a common error when, 
for example, the user tries to read a floppy or cartridge drive with
no disk in it.
.P
This error is more serious on a hard drive, and usually indicates
hardware problems.
.MT User error \fRor\fP Hardware.
.SS
write protected
.SE
This error is displayed when \*X looks at a drive and the user 
wants to write to the drive and the media is write-protected.
.MT User error.
.RE
.sp 2
.DS L
.if t .B
.\" Foo Drive n: TYPE error hhhh while {read,writ}ing
.\"					Cylinder CCC Head HH Sector SSSS
\fIdevice\fP Drive \fIn\fP: \fItype\fP error \fIhhhh\fP while Read/Writing Cylinder \fIcyl\fP Head \fIhead\fP Sector \fIsector\fP
.br
.\" Foo Drive n: TYPE error hhhh while Formatting
\fIdevice\fP Drive \fIn\fP: \fItype\fP err \fIhhhh\fP on format
.\" Cartridge Drive n: TYPE err hhhh on rd/wrt of blk BBBBBB
Cartridge Drive \fIn\fP: \fItype\fP error \fIhhhh\fP while Read/Writing Block \fIblock\fP
.br
.if t .R
.DE
.RS
This format is used for errors which occur during I/O of some sort.
You are shown the type of error, some hardware-related status, what type
of I/O it occurred on (read, write, or format), and where it occurred.  
The \fItype\fP of error will be one of the following:
\fBhard\fP, \fBsoft\fP, \fBmailbox\fP, or \fBunclassified\fP.
.RS
.P
A \fIhard\fP error indicates that the driver retried the operation 
and was unable to complete the requested I/O.
.P
A \fIsoft\fP error indicates that an error occurred, the driver 
retried the operation and it succeeded.  The data is
.UL valid .
These are usually indications that the media is aging.
.P
A \fImailbox\fP error is a rare error indicating that \*X handed the
Z80 Control Program an invalid request of some sort.  This usually
indicates a bug in the 68000 software, or, in extreme cases, a 68000
hardware problem.
.P
An \fIunclassified\fP error is an extremely rare error type indicating
that the Z80 Control Program returned an undefined error type and \*X
was unable to determine what the problem was.  This usually indicates a
Z80 hardware problem, or, in some cases, a bug in the Z80 Control Program.
.RE
.sp
The status displayed (\fIhhhh\fP)
is data that might be useful from the controller or drive.
.sp
.RS
For the hard drive, the data displayed is the error register and the
control/status register from the controller.  For example, you might see a
.B 405B .
If you look these numbers up in the Western Digital WD-1010 book, you will
see that the
.B 40
means \*QID not found\*U and the 
.B 5B
means \*QReady, Seek complete, Data Request, Command-in-Progress, Error.\*U
.P
For the floppy drive, the data displayed is the drive select port 
and the FDC status port.\**
.P
For the disk cartridge drive, the data displayed is the sense key, class,
code, and address valid bit.\*&
.FS
For more information on floppy or disk cartridge error status numbers,
see the document \fIZ80CTL Error Message Descriptions for Tandy 6000\fP.
.FE
.RE
.sp
Lastly, where the error occurred is displayed.
For the floppy and hard disk, the cylinder, head, and sector are
displayed.  For the IOmega, the block number is displayed.
.\" ======================================================================
.sp 2
.SH
Specific or Specialized Device Driver Error Messages
.RS
.SS
dkustart:  GO set on \fIdevice\fP drive \fIn\fP
.br
panic: GO set at dkustart
.SE
These two messages are produced by the disk driver when \*X
asks for some I/O and it discovers that the Z80 already claims to
be busy doing something which \*X doesn't think it asked for.
This usually indicates a bug in the 68000 software.
.MT System inconsistency, Fatal.
.SS
panic: disk major number
.SE
This message will only be seen when \*X is put together improperly,
or there is an unusual 68000 hardware problem.  It indicates that 
when \*X started up, it was unable to find the disk driver listed 
in the driver tables.
.MT System inconsistency, Fatal.
.SS
panic: ndflush
.SE
This message indicates that, during some type of serial I/O, the system
tried to flush some bytes from an output queue, and the queue was
corrupted.
.MT System inconsistency, Fatal.
.RE
.sp
.\" ======================================================================
.SH
Miscellaneous Hardware-Related Messages
.R
.RS
.KS	\" Would use `.SS', but don't need a `.sp' here...  (`.SE' ends keep)
.DS L
.if t .B
Supervisor Trap \fId\fP
Access address=\fIdddddddd\fP  Instr wd=\fIiiii\fP
&d0=\fIdddddddd\fP mmu=\fIdddddddd\fP sr=\fIdddddddd\fP pc=\fIdddddddd\fP usp=\fIdddddddd\fP
a0\-a7 \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP  \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP
d0\-d7 \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP  \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP \fIdddddddd\fP
panic: trap in sys
.nr xx \n(DD
.nr DD 0
.if t .R
.DE
.SE
.nr DD \n(xx
This message indicates that 68000 TRAP number
.I d
occurred in
.I supervisor
mode (ie, while the processor was running \*X code and not user code).
For specifics on the trap, consult a 68000 manual.
.P
This error can be caused by an infinite myriad of problems, ranging from
bugs in \*X to line-power problems.  The data presented is a nearly 
complete dump of the processor's registers and may be used to find where 
the 68000 was executing when the trap occurred.
.MT Hardware \fRor\fP System inconsistency, Fatal.
.SS
panic: memory parity
.SE
This message occurs when a 68000 board set equipped with parity chips
gives a parity trap.  There is nothing that can be done to correct the
error, so \*X ceases processing.
.MT Hardware \fRor\fP System inconsistency, Fatal.
.SS
panic: mmusub: chk
.SE
This message indicates that, while trying to allocate space for a user
program, \*X discovered that a previously checked and \*QOK\*U program 
has an illegal or impossible size or illegal segments (eg, separate data and 
stack segments).
.MT System inconsistency, Fatal.
.SS
interrupt from unknown device, vec=\fId\fP
.br
panic: unknown interrupt
.SE
This message indicates that an unknown, unused interrupt (number \fId\fP)
occurred on the 68000.  This usually indicates a 68000, or, more rarely, 
a Z80 hardware problem.
.MT Hardware, Fatal.
.RE
.sp
.\" ===========================================================================
.SH
Other System Messages
.R
.RS
.SS
WARNING: bad configuration, resetting to defaults.
.SE
The parameters specified with the \fIcfg\fP(C) command require an 
unreasonable amount of memory be used for \*X, and would leave the 
available user memory area too small to be usable.   The system ignores 
the table size parameters and uses the defaults.  This message can only
occur at boot time.  This message may be caused by running a system
with less memory than it ordinarily has, or with an abnormally small 
(read: unsupported) amount of memory (eg, 256k). \**
.MT Abnormal.
.FS
Believe it or not, \*X \fIwill\fP boot on a 256k machine.
See also \fIcfg\fP(C).
.FE
.SS
** ABNORMAL System Shutdown **
.SE
This message appears when errors occur during 
system shutdown.  It is usually accompanied by other
system messages.
.MT System inconsistency, Fatal
.SS
bad block on dev \fInn\fP/\fImm\fP
.SE
A nonexistent disk block was found on,
or is being inserted in, the structure's free list.
.MT System inconsistency.
.SS
bad count on dev \fInn\fP/\fImm\fP
.SE
A structural inconsistency in the superblock of a file system.  
The system attempts a repair, but this message will probably be 
followed by more complaints about this file system.
.MT System inconsistency.
.SS
Bad free count on dev \fInn\fP/\fImm\fP
.SE
A structural inconsistency in the superblock of a file system.  
The system attempts a repair, but this message will probably 
be followed by more complaints about this file system.
.MT System inconsistency.
.SS
error on dev \fIname\fP (\fInn\fP/\fImm\fP)
.SE
This is a way that device driver diagnostic messages can be displayed.
In reality, none of Tandy's drivers use this format, and this style
message is almost never seen.\**  The message will indicate the specific
driver and complaint.  \fIname\fP is a word identifying the device.
.MT
.FS
See the previous section on device drivers.
.FE
.SS
iaddress > 2^24
.SE
This indicates an attempted reference to an illegal block number,
one so large that it could only occur on a file system larger 
than 8 billion bytes.
.MT Abnormal.
.SS
Inode table overflow
.SE
Each open file requires an inode entry to be kept in memory.  When
this table overflows the specific request (usually
.I "open (S)"
or
.I "creat (S)" )
is refused.  Although not fatal to the system, this event may 
damage the operation of various spoolers, daemons, the mailer, 
and other important utilities.  Anomalous results and missing 
data files are a common result.  
.MT Abnormal.
.SS
no file
.SE
There are too many open files, the system has run out of 
entries in its \*Qopen file\*U table.  The warnings given for 
the message \*Qinode\ table\ overflow\*U apply here.
.MT Abnormal.
.SS
no space on dev \fInn\fP/\fImm\fP
.SE
This message means that the specified file system
has run out of free blocks.  Although not normally as 
serious, the warnings discussed for \*Qinode table overflow\*U 
apply: often user programs are written casually and ignore the 
error code returned when they tried to write to the disk; 
this results in missing data and \*Qholes\*U in data files.  
The system administrator should keep close watch on the 
amount of free disk space and take steps to avoid this situation.
.MT Abnormal.
.SS
**  Normal System Shutdown  **
.br
[Z80 Control System Halted] \**
.SE
This message appears when the system has been shutdown properly.
It indicates that the machine may now be rebooted or powered down.
.FS
See also the 
document \fIZ80CTL Error Message Descriptions for Tandy 6000\fP.
.FE
.MT
.SS
Out of inodes on dev \fInn\fP/\fImm\fP
.SE
The indicated file system has run out of free inodes.  
The number of inodes available on a file system is 
determined when the file system is created (using
.I "mkfs (C)" ).
The default number is quite generous, this message 
should be very rare.  The only recourse is to remove some 
worthless files from that file system, or dump the entire 
system to a backup device, run
.I "mkfs (C)"
with more inodes specified, and restore the files from backup.
.MT Abnormal.
.SS
out of text
.SE
When programs linked with the 
.I ld
.B \-i 
or 
.B \-n 
switch are run, a table entry is made so that only one copy 
of the pure text will be in memory even if there are multiple 
copies of the program running.  This message appears when this 
table is full.  The system refuses to run the program
which caused the overflow.  Note that there is only one entry 
in this table for each different pure text program.  Multiple 
copies of one program will not require multiple table entries.
Each \*Qsticky\*U program [see \fIchmod\fP(C)]
requires a permanent entry in this table; nonsticky
pure text programs require an entry only when there is at 
least one copy being executed.
.MT Abnormal.
.SS
panic: blkdev
.SE
An internal disk I/O request, already verified as valid, is 
discovered to be referring to a nonexistent disk.
.MT System inconsistency, Fatal.
.SS
panic: devtab
.SE
An internal disk I/O request, already verified as valid, is discovered
to be referring to a nonexistent disk.
.MT System inconsistency, Fatal.
.SS
panic: iinit
.SE
The super-block of the root file system could not be read.  This
message occurs only at boot time.
.MT Hardware, Fatal.
.SS
panic: IO err in swap
.SE
A \fIhard\fP\|\** I/O error occurred while reading or writing the swap area.
.FS
See the above section on device driver error messages.
.FE
.MT Hardware, Fatal.
.SS
panic: no fs
.SE
A file system descriptor has disappeared from its table.
.MT System inconsistency, Fatal.
.SS
panic: no imt
.SE
A mounted file system has disappeared from the mount table.
.MT System inconsistency, Fatal.
.SS
panic: no procs
.SE
Each user is limited in the amount of simultaneous processes 
he can have; an attempt to create a new process when none is 
available or when the user's limit is exceeded is refused.  
That is an occasional event and produces no console messages; 
this panic occurs when the kernel has certified
that a free process table entry is available and 
yet can't find one when it goes to get it.
.MT System inconsistency, Fatal.
.SS
panic: Out of swap 
.SE
There is insufficient space on the swap disk to hold a task that is
scheduled to be swapped.
The system refuses to create tasks when it feels there is 
insufficient disk space, but it is possible to create situations 
to fool this mechanism.
.MT Abnormal, Fatal.
.SS
panic: Timeout table overflow
.SE
The timeout table is full.  Timeout requests are generated by device
drivers, there should usually be room for one entry per system serial
line plus ten more for other usages.  
.P
This is usually caused by a bug in a device driver.
.MT Abnormal, Fatal.
.SS
proc on q
.SE
The system attempts to queue a process already on
the process ready-to-run queue. 
.MT System inconsistency, Fatal.
.RE
.\"
.\" ====== End Document =======================================================
