/*
 *  cfg.c --  Configure /xenix
 *
 *	cfg [-n xenix] [Tables] [Misc]
 *	    Where: "Misc" may be one or more of the following:
 *		[maxprocs=<m>] [maxmem=<k>] [nodename=<name>] [screens=<s>]
 *		[swapdev=a | swapdev=<mm,nn> swplo=<dd> nswap=<nn>];
 *	    where: "Tables" may be "-d" or one or more of the following:
 *	    	buffers=n inodes=n files=n locks=n procs=n clists=n
 *	      and: `n' may be "a" or an integer greater than 0.
 *
 *   Written	3 September 1986	John Elliott IV
 */

#include <stdio.h>
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/var.h>
#include <sys/relsym.h>
#include <sys/utsname.h>


#ifdef DEBUG
#  define debug(s)		fprintf(stderr, s)
#  define debug1(s, a)		fprintf(stderr, s, a)
#  define debug2(s, a,b)	fprintf(stderr, s, a,b)
#  define debug4(s, a,b,c,d)	fprintf(stderr, s, a,b,c,d)
#  define debug7(s, a,b,c,d,e,f,g)  fprintf(stderr, s, a,b,c,d,e,f,g)
#else /*DEBUG*/
#  define debug(s)
#  define debug1(s, a)
#  define debug2(s, a,b)
#  define debug4(s, a,b,c,d)
#  define debug7(s, a,b,c,d,e,f,g)
#endif /*DEBUG*/

/* Misc defines & tunables */
#define isdigit(c)	((c) >= '0' && (c) <= '9')
#define islower(c)	((c) >= 'a' && (c) <= 'z')
#define isupper(c)	((c) >= 'A' && (c) <= 'Z')
#define tolower(c)	(isupper(c)? (c) + ('a' - 'A') : (c))
#define	REALBIG	8192

#ifdef SYS3
#  define  index	strchr
#  define  rindex	strrchr
#  include <fcntl.h>
#else /*SYS3*/
#  define O_RDONLY	0
#  define O_WRONLY	1
#  define O_RDWR	2
#endif /*SYS3*/

extern char	*index(), *rindex();


char	id[] = "Copyright 1988 Tandy Corporation.  All Rights Reserved.\n";
char	version[] = "cfg v1.2(00)\n";
#include "date.h"

char	AutoConfig[] = "AutoConfig\n";


struct nlist nl[] = {
  /*0*/	{ "_v" },
  /*1*/	{ "_Maxmem" },
  /*2*/	{ "_cfgtbl" },
  /*3*/	{ "_utsname" },
  /*4*/	{ "_swapdev" },
  /*5*/	{ "_swplo" },
  /*6*/	{ "_nswap" },
  /*7*/	{ "_nscr" },
	{ "" },
};
#define NLSIZE	(sizeof (nl) / (sizeof nlist))


char		*kernel = "/xenix";
char		*myname;
int		changed = 0;
struct xexec	xh;			/* x.out header */

int	maxmem;			/* Must be the same size as _Maxmem in kernel */
int	newmem = -1;		/* Must be the same size as _Maxmem in kernel */

short	nscr;			/* Must be the same size as _nscr in kernel */
short	newscr = -1;		/* Must be the same size as _nscr in kernel */

struct utsname utsname;
char	nodename[sizeof (utsname.nodename)] = "\"";

struct var	v;
struct var	new = {
	0,
	-1,	/* NSABUF */
	0,
	-1,	/* NINODE */
	0, 
	-1,	/* NFILE */
	0,
	0, 0,
	-1,	/* NPROC */
	0, 
	-1,	/* NTEXT */
	0,
	-1,	/* NCLIST */
	-1,	/* MAXUPRC */
	-1,	/* NFLOCKS */
	0, 0
};

dev_t	swapdev;	/* MUST be the same size as _swapdev in kernel */
daddr_t	swplo, nswap;	/* MUST be the same size as _swplo & _nswap in kernel */
daddr_t	nswplo = -1;	/* MUST be the same size as _swplo in kernel */
daddr_t	nnswap = -1;	/* MUST be the same size as _nswap in kernel */
int	nswapdev = -1;


main(argc, argv)
  char **argv;
{
	caddr_t base;
	char	*scan;

	scan = rindex(*argv, '/');
	if (scan == (char *)NULL)
		myname = *argv;
	else
		myname = ++scan;
	++argv, --argc;

	if (getuid() != 0)
		errexit("You must be the superuser to run this program");

	/* This isn't in crackargs() 'cause we need the kernel name to call
	 *   nlist(), and we need to call nlist() before we call crackargs() */
	if (strcmp(*argv, "-n") == 0) {
		++argv, --argc;
		if (argc > 0)
			kernel = *argv;
		else
			bad("Missing argument for `-n'");
		++argv, --argc;
	}

	nlist(kernel, nl);

	crackargs(argc, argv);

	  /* 0	{ "_v" }
	   * 1	{ "_Maxmem" }
	   * 2	{ "_cfgtbl" }
	   * 3	{ "_utsname" }
	   * 4	{ "_swapdev" }
	   * 5	{ "_swplo" }
	   * 6	{ "_nswap" }
	   * 7	{ "_nscr" }
	   */
	if (nl[0].n_type == 0 
	    || nl[1].n_type == 0 
	    || nl[3].n_type == 0 
	    || nl[4].n_type == 0 
	    || nl[5].n_type == 0 
	    || nl[6].n_type == 0)
	{
		errexit("No namelist");
	}

	if (nl[2].n_type == 0)
		errexit("cfg not supported in that xenix");

	kpatch();

	exit(0);
}

extern long lseek();
#define XESIZE	sizeof (struct xexec)
kpatch()
{
	register struct nlist	*np;
	register int	fd;
	char		ndbuff[12];

	/*
	 * now patch /xenix, if there's a data seg there.
	 */
	if ((fd = open(kernel, O_RDWR)) < 0)
		errexit("Can't open xenix");

	if (read(fd, (char *)&xh, XESIZE) != XESIZE) {
		close(fd);
		errexit("Can't read xenix");
	}

	if ((xh.x_data == (long)0) || (xh.x_text+xh.x_data < nl[0].n_value)) {
		close(fd);
		errexit("Not enough data in xenix");	
		/* not enough data in /xenix (no error) */
	}

	  /* 0	{ "_v" }
	   * 1	{ "_Maxmem" }
	   * 2	{ "_cfgtbl" }
	   * 3	{ "_utsname" }
	   * 4	{ "_swapdev" }
	   * 5	{ "_swplo" }
	   * 6	{ "_nswap" }
	   * 7	{ "_nscr" }
	   */
	read_kvar(fd, &nl[0], (char *)&v, sizeof (v));
	read_kvar(fd, &nl[1], (char *)&maxmem, sizeof (maxmem));
	read_kvar(fd, &nl[3], (char *)&utsname, sizeof (utsname));
	read_kvar(fd, &nl[4], (char *)&swapdev, sizeof (swapdev));
	read_kvar(fd, &nl[5], (char *)&swplo, sizeof (swplo));
	read_kvar(fd, &nl[6], (char *)&nswap, sizeof (nswap));
	if (nl[7].n_type != 0)
	    read_kvar(fd, &nl[7], (char *)&nscr, sizeof (nscr));

	debug7("utsname = {\n\tsysname=\"%s\"\n\tnodename=\"%s\"\n\trelease=\"%s\"\n\tversion=\"%s\"\n\tsysorigin=%d\n\tsysoem=%d\n\tsysserial=%ld\n};\n",
		utsname.sysname,
		utsname.nodename,
		utsname.release,
		utsname.version,
		utsname.sysorigin,
		utsname.sysoem,
		utsname.sysserial);
	debug4("swapdev=%d,%d / swplo = %ld / nswap = %ld\n",
		major(swapdev),
		minor(swapdev),
		swplo,
		nswap);

	if (changed) {
		if (new.v_sabuf >= 0)  v.v_sabuf = new.v_sabuf;
		if (new.v_inode >= 0)  v.v_inode = new.v_inode;
		if (new.v_file >= 0)  v.v_file = new.v_file;
		if (new.v_lock >= 0)  v.v_lock = new.v_lock;
		if (new.v_proc >= 0)  v.v_proc = new.v_proc;
		if (new.v_clist >= 0)  v.v_clist = new.v_clist;
		if (new.v_maxup >= 0)  v.v_maxup = new.v_maxup;
		write_kvar(fd, &nl[0], (char *)&v, sizeof (v));

		if (newmem != -1) {
		    write_kvar(fd, &nl[1], (char *)&newmem, sizeof (newmem));
		    maxmem = newmem;		/* For display routine */
		}

		if (newscr != -1) {
		    write_kvar(fd, &nl[7], (char *)&newscr, sizeof (newscr));
		    nscr = newscr;		/* For display routine */
		}

		if (nodename[0] != '"') {
		     strcpy(utsname.nodename, nodename);
		     write_kvar(fd, &nl[3], (char *)&utsname, sizeof (utsname));
		}

		if (nswplo >= 0) {
		    swapdev = nswapdev;
		    swplo = nswplo;
		    nswap = nnswap;
		    write_kvar(fd, &nl[4], (char *)&swapdev, sizeof (swapdev));
		    write_kvar(fd, &nl[5], (char *)&swplo, sizeof (swplo));
		    write_kvar(fd, &nl[6], (char *)&nswap, sizeof (nswap));
		}
	}

	close(fd);

	printf("%s is%s configured as follows:\n",
			kernel,
			changed? " now" : "");
	printf("    Buffers\t");
	if (v.v_sabuf == 0)  	printf(AutoConfig);
	  else  printf("%10d\n", v.v_sabuf);

	printf("    Inodes\t");
	if (v.v_inode == 0)  	printf(AutoConfig);
	  else  printf("%10d\n", v.v_inode);

	printf("    Files\t");
	if (v.v_file == 0)  	printf(AutoConfig);
	  else  printf("%10d\n", v.v_file);

	printf("    Locks\t");
	if (v.v_lock == 0)  	printf(AutoConfig);
	  else  printf("%10d\n", v.v_lock);

	printf("    Procs\t");
	if (v.v_proc == 0)  	printf(AutoConfig);
	  else  printf("%10d\n", v.v_proc);

	printf("    Clists\t");
	if (v.v_clist == 0)  	printf(AutoConfig);
	  else  printf("%10d\n", v.v_clist);

	printf("    Maxprocs\t");
	  printf("%10d\n", v.v_maxup);

	printf("    Maxmem\t");
	  printf("%9dk\n", maxmem);

	if (nl[7].n_type != 0) {
	  printf("    Screens\t");
	   printf("%10d\n", nscr);
	}

	printf("    Nodename\t");
	  sprintf(ndbuff, "\"%s\"", utsname.nodename);
	  printf("%10s\n", ndbuff);

	printf("    Swapdev\t");
	if (minor(swapdev) == 255)
		printf(AutoConfig);
	  else {
		printf("%6d,%3d\n", major(swapdev), minor(swapdev));

		printf("      swplo\t");
		  printf("%10ld\n", swplo);

		printf("      nswap\t");
		  printf("%10ld\n", nswap);
	  }
}

read_kvar(fd, np, bp, size)
  int		fd;
  struct nlist *np;
  char		*bp;
  int		size;
{
	register long	seekpos;

	seekpos = XTEXTPOS(&xh) + (long)np->n_value;
	if (lseek(fd, seekpos, 0) != seekpos) {
		close(fd);
		errexit("Can't seek for read");
	}
	if (read(fd, bp, size) != size) {
		close(fd);
		errexit("Can't read data");
	}
}

write_kvar(fd, np, bp, size)
  int		fd;
  struct nlist *np;
  char		*bp;
  register int	size;
{
	register long	seekpos;

	seekpos = XTEXTPOS(&xh) + (long)np->n_value;
	if (lseek(fd, seekpos, 0) != seekpos) {
		close(fd);
		errexit("Can't seek for write");
	}
	if (write(fd, bp, size) != size) {
		close(fd);
		errexit("Can't write data");
	}
}

crackargs(argc, argv)
  register int argc;
  register char **argv;
{
	register char	*scan;

	while (argc > 0) {
		**argv = tolower(**argv);
		if (strcmp(*argv, "-d") == 0) {
			if (changed)
				bad("You must specify \"-d\" alone");
			
			new.v_sabuf = 0;
			new.v_inode = 0;
			new.v_file = 0;
			new.v_lock = 0;
			new.v_proc = 0;
			new.v_clist = 0;
			new.v_maxup = 15;
			newmem = 256;

			changed = 1;
		}
		else if (strncmp(*argv, "buffers=", 8) == 0) {
			if (new.v_sabuf >= 0)
				too_many("buffer");
			if (argv[0][8] == 'a' && argv[0][9] == '\0')
				new.v_sabuf = 0;
			else {
				new.v_sabuf = atoi(&argv[0][8]);
				if (new.v_sabuf < 20 || new.v_sabuf > REALBIG)
					out_range(" of buffers", 20, REALBIG);
			}
			changed = 1;
		}
		else if (strncmp(*argv, "inodes=", 7) == 0) {
			if (new.v_inode >= 0)
				too_many("inode");
			if (argv[0][7] == 'a' && argv[0][8] == '\0')
				new.v_inode = 0;
			else {
				new.v_inode = atoi(&argv[0][7]);
				if (new.v_inode < 60 || new.v_inode > REALBIG)
					out_range(" of inodes", 60, REALBIG);
			}
			changed = 1;
		}
		else if (strncmp(*argv, "files=", 6) == 0) {
			if (new.v_file >= 0)
				too_many("file");
			if (argv[0][6] == 'a' && argv[0][7] == '\0')
				new.v_file = 0;
			else {
				new.v_file = atoi(&argv[0][6]);
				if (new.v_file < 60 || new.v_file > REALBIG)
					out_range(" of files", 60, REALBIG);
			}
			changed = 1;
		}
		else if (strncmp(*argv, "locks=", 6) == 0) {
			if (new.v_lock >= 0)
				too_many("lock");
			if (argv[0][6] == 'a' && argv[0][7] == '\0')
				new.v_lock = 0;
			else {
				new.v_lock = atoi(&argv[0][6]);
				if (new.v_lock < 20 || new.v_lock > REALBIG)
					out_range(" of locks", 20, REALBIG);
			}
			changed = 1;
		}
		else if (strncmp(*argv, "procs=", 6) == 0) {
			if (new.v_proc >= 0)
				too_many("proc");
			if (argv[0][6] == 'a' && argv[0][7] == '\0')
				new.v_proc = 0;
			else {
				new.v_proc = atoi(&argv[0][6]);
				if (new.v_proc < 20 || new.v_proc > REALBIG)
					out_range(" of procs", 20, REALBIG);
			}
			changed = 1;
		}
		else if (strncmp(*argv, "clists=", 7) == 0) {
			if (new.v_clist >= 0)
				too_many("clist");
			if (argv[0][7] == 'a' && argv[0][8] == '\0')
				new.v_clist = 0;
			else {
				new.v_clist = atoi(&argv[0][7]);
				/* The reason we're using `5500' here instead
				 * of `REALBIG' is that if we allow values
				 * larger than 5500, `ttyhog' goes negative
				 * and all serial queues get flushed
				 * constantly. */
				if (new.v_clist < 50 || new.v_clist > 5500)
					out_range(" of clists", 50, 5500);
			}
			changed = 1;
		}
		else if (strncmp(*argv, "maxprocs=", 9) == 0) {
			if (new.v_maxup >= 0)
				too_many("maxprocs");
			new.v_maxup = atoi(&argv[0][9]);
			if (new.v_maxup < 5 || new.v_maxup > 100)
				out_range(" for maxprocs", 5, 100);
			changed = 1;
		}
		else if (strncmp(*argv, "maxmem=", 7) == 0) {
			if (newmem > 0)
				too_many("maxmem");
			newmem = atoi(&argv[0][7]);
			if (newmem < 128 || newmem > 8192)
				out_range(" for maxmem", 128, 8192);
			changed = 1;
		}
		else if (strncmp(*argv, "swplo=", 6) == 0) {
			if (nswplo >= 0)
				too_many("swplo");
			nswplo = atol(&argv[0][6]);
			if (nswplo < 0)
				out_range(" for swplo", 0, 0x7FFFFFFF);
			changed = 1;
		}
		else if (strncmp(*argv, "nswap=", 6) == 0) {
			if (nnswap > 0)
				too_many("nswap");
			nnswap = atol(&argv[0][6]);
			if (nnswap < 512)
				out_range(" for nswap", 512, 0x7FFFFFFF);
			changed = 1;
		}
		else if (strncmp(*argv, "swapdev=", 8) == 0) 
		{
			int	maj, min;

			if (nswapdev != -1)
				too_many("swapdev");
			scan = &argv[0][8];
			if (scan[0] == 'a' && scan[1] == '\0') {
			    nswapdev = ((int)makedev(1, 255)) & 0xFFFF;
			    nswplo = 0L;
			    nnswap = 1024L;
			}
			else {
			    if (!isdigit(*scan))
				bad("Missing major number for `swapdev'");
			    maj = atoi(scan);
			    scan = index(scan, ',');
			    if (scan == (char *)NULL)
				bad("Missing minor number for `swapdev'");
			    min = atoi(++scan);
			    if (maj < 0 || maj > 255)
				out_range(" for swapdev major #", 0, 255);
			    if (min < 0 || min > 255)
				out_range(" for swapdev minor #", 0, 255);
			    nswapdev = ((int)makedev(maj, min)) & 0xFFFF;
			}
			changed = 1;
		}
		else if (strncmp(*argv, "nodename=", 9) == 0) {
			if (nodename[0] != '"')
				too_many("nodename");
			scan = &argv[0][9];
			if (strlen(scan) > 8)
				errexit("Too many characters in nodename");
			strncpy(nodename, scan, sizeof (nodename) - 1);
			nodename[sizeof (nodename) - 1] = '\0';
			if (!check_chars(nodename))
				errexit("Illegal character in nodename");
			changed = 1;
		}
		else if (strncmp(*argv, "screens=", 8) == 0) {
			if (nl[7].n_type == 0)
			    errexit("\"screens\" not supported in that xenix");
			if (newscr > 0)
				too_many("screens");
			newscr = atoi(&argv[0][8]);
			if (newscr < 0 || newscr > 10)
				out_range(" for screens", 0, 10);
			changed = 1;
		}
		else {
			fprintf(stderr, "%s: Unknown option \"%s\"\n",
							myname,
							*argv);
			usage();
		}

		--argc, ++argv;
	}

	if (changed) {
		if (nswapdev != -1 && (nswplo < 0 || nnswap < 0)
		    || nswplo >= 0 && (nswapdev == -1 || nnswap < 0)
		    || nnswap >= 0 && (nswapdev == -1 || nswplo < 0))
		{
		    bad("You must specify swapdev, swplo, and nswap together");
		}
	}
}

/* Check for illegal characters in a `nodename' */
check_chars(np)
  register char	*np;
{
	while (*np) {
		if (!islower(*np) && !isdigit(*np) && *np != '-' && *np != '_')
			return (0);
		++np;
	}

	return (1);
}

out_range(s, l, h)
  char	*s;
{
	char	buff[80];

	sprintf(buff, "Please type a number%s between %d and %d", s, l, h);
	bad(buff);
}

too_many(s)
  char	*s;
{
	char	buff[80];

	sprintf(buff, "Too many %s change specifications", s);
	bad(buff);
}

bad(s)
  char *s;
{
	fprintf(stderr, "%s: %s\n", myname, s);
	usage();
}

errexit(s)
  char *s;
{
	fprintf(stderr, "%s: %s\n", myname, s);
	exit(1);
}

char	UsgMsg[] = "Usage: %s [-n xenix] [Tables] [Misc]\nWhere: \"Misc\" may be one or more of the following:\n\t[maxprocs=<m>] [maxmem=<k>] [nodename=<name>] [screens=<s>]\n\t[swapdev=a | swapdev=<mm,nn> swplo=<dd> nswap=<nn>];\nwhere: \"Tables\" may be \"-d\" or one or more of the following:\n\tbuffers=n inodes=n files=n locks=n procs=n clists=n\n  and: `n' may be \"a\" or an integer greater than 0.\n";

usage()
{
	fprintf(stderr, UsgMsg, myname);
	exit(1);
}
