/*
 *  "copytest.c" -	A structure-copy test program for the compiler
 *
 *	Written  20 January 1987		John Elliott IV
 *
 *
 *	Compile this with:	-DREGISTER=register	to use register pointers
 *				-DSIZE=xxx		to set the struct size
 *
 *	Example:  "cc -DSIZE=1024 -O copytest.c -n -o copytest"
 */

/* Dictionary */
char	DestClobber[] = "ERROR: Destination structure pointer clobbered\n";
char	SrcClobber[] = "ERROR: Source structure pointer clobbered\n";
char	DestBad[] = "ERROR: Destination structure corrupted\n";
char	SourceBad[] = "ERROR: Source structure corrupted\n";
char	BackFlow[] = "ERROR: Copy overflowed backwards\n";
char	OverFlow[] = "ERROR: Copy overflowed forwards\n";
char	Okay[] = "All checks okay.\n";


/* Actual data area */
struct foo {
	unsigned char	f_a[SIZE];
};

int	a_slop[] = { -1, -2, -3, -4, -5, -6 };

struct foo	a, b;

int	b_slop[] = { -1, -2, -3, -4, -5, -6 };

#ifndef REGISTER
#  define REGISTER
#endif /*REGISTER*/


main()
{
	REGISTER struct foo	*ap, *bp;
	int	bad;

	bad = 0;
	ap = &a;
	bp = &b;

	fill(bp);

	*ap = *bp;
	/* Hopefully at this point we haven't clobbered our stack or text */

	if (ap != &a)
		++bad, write(2, DestClobber, sizeof (DestClobber));
	if (bp != &b)
		++bad, write(2, SrcClobber, sizeof (SrcClobber));

	if (!check(bp))
		++bad, write(2, SourceBad, sizeof (SourceBad));
	if (!check(ap))
		++bad, write(2, DestBad, sizeof (DestBad));

	if (!slop_check(a_slop))
		++bad, write(2, BackFlow, sizeof (BackFlow));
	if (!slop_check(b_slop))
		++bad, write(2, OverFlow, sizeof (OverFlow));

	if (bad)
		exit(1);

	write(2, Okay, sizeof (Okay));
	exit(0);
}

fill(fp)
  struct foo	*fp;
{
	register unsigned char	*sp, c;

	c = 0;
	for (sp = &fp->f_a[sizeof fp->f_a - 1]; sp >= fp->f_a; --sp)
		*sp = c++;
}

check(fp)
  struct foo	*fp;
{
	register unsigned char	*sp, c;

	c = 0;
	for (sp = &fp->f_a[sizeof fp->f_a - 1]; sp >= fp->f_a; --sp) {
		if (*sp != c++)
			return (0);
	}

	return (1);
}

slop_check(sp)
  register int	*sp;
{
	if (sp[0] == -1 
	    && sp[1] == -2 
	    && sp[2] == -3 
	    && sp[3] == -4
	    && sp[4] == -5
	    && sp[5] == -6)
	{
		return (1);
	}

	return (0);
}
