|***	Startup for XENIX
|***	
|***	Create the environment for a call to main and upon return start
|***	the initial user process.
|***
|***	Originally written by Jerry J. Dunietz, Microsoft Corp.; January 1983
|***	Maintained locally by John A. Elliott IV
|***	Copyright (c) Tandy Corporation, 1986, 1987
|***
|***	Assembly Language Assist Routines
|***   
|***	===========================================================
|***   
|***	Tandy Radio Shack Model 16
|***   
|***	The hardware provides no relocation in supervisor mode (!!).
|***	For initial development, we reserve an area in which the currently
|***	active U and stack manipulated.  Whenever a save is done, we copy
|***	the currently active U and stack back to where they belong; whenever
|***	a resume is done, the newly selected U and stack are copied into 
|***	the active area.
|***
|***	The memory management in user mode supports two segments;
|***	one starts at location 0, the other at hex 800000.
|***	Segments can be any multiple of 4Kbytes in size.  They are
|***	defined by set of offset registers and a set of limit registers.
|***
|***	All I/O is done thru the auxiliary Z80 processor.  The 68000
|***	informs the Z80 of the start of their communications buffers
|***	by loading this address into location 4.  The Z80 initializes 
|***	them and resets the contents of location 4 back to its original 
|***	value.  This informs the 68000 to proceed.
|***
|*** ========================================================================
|***  Modification record:
|***
|***   $Log:	mch.S,v $
|***   Revision 6.2  88/08/17  15:25:18  root
|***   	- Changed all DEBUG putchar() calls to _putcha()
|***   	  [so it won't get put in the output buffer]
|***   
|***   Revision 6.1  88/04/05  11:16:38  iv
|***   	- Added MC68010 conditional support
|***   	- Added DEBUG code for startup under conditional DEBUGS
|***   	- Changed kernel-debugger conditional to KDEBUG
|***   
|***   Revision 6.0  87/10/12  17:00:10  iv
|***   	- Checkin to lock down 3.2
|***   
|***   Revision 5.6  86/10/16  15:17:28  iv
|***   	- Replaced `movb's with `moveq's in _spl? calls
|***   	- Renamed bufinit() to cfginit()  [now handles more than buffers]
|***   
|***   Revision 5.5  86/10/01  16:59:37  iv
|***   	- Re-arranged startup code to support auto-configurable buffers
|***   	  (and maybe other things later!)
|***   
|***   Revision 5.4  86/08/28  17:22:14  iv
|***   	- Changed mmu-related code to support extra 3 bits of resolution
|***   	  (see mmu.h and mmu.c comments and change log)
|***   
|***   Revision 5.3  86/07/10  15:24:07  iv
|***   	- Cleaned up some of the RCS revsion comments
|***   	- Prettied up some of the code comments
|***   	- Changed copyright message to Tandy, 1986 [we've changed AT LEAST 10%]
|***   	- Added new 1 conditional, compiles in fast-system-call handling
|***   	  code.  Trimmed ~2 seconds off of our system call benchmark!
|***   	- Moved z80 mailboxes into mch.S, especially for Mr. Bob Snapp
|***   	  who likes patching our code to death
|***   	- Rewrote spl code & saved 4 clocks there & shortened local calls to it
|***   
|***   Revision 5.2  86/06/12  11:39:30  iv
|***   	- Added support code for auto-boot
|***   
|***   Revision 5.1  86/05/21  21:12:55  iv
|***   	- Converted the awful, awful, ea syntax that mch.s was written in
|***   	  to the one everyone else in the world uses.  Much more readable.
|***   
|***   Revision 5.0  85/08/23  10:00:48  iv
|***   	- Checkin to lock down XENIX 3.1 release
|***   
|***   Revision 5.0  85/08/23  09:56:21  iv
|***   	- Checkin to lock down XENIX 3.1 release
|***   
|***   Revision 4.2  85/08/22  16:06:14  iv
|***   	- Changed memory size code to lie and say 1 Meg if the actual
|***   	  main memory is really > 1 Meg, because the MMU can't handle
|***   	  addresses that are > 1 Meg.  *sigh*
|***   
|***   Revision 4.1  85/03/29  09:57:47  iv
|***    - Removed the Z80 interrupt ID array; it is not used
|***   
|***   Revision 4.0  85/03/13  14:38:03  iv
|***   	- Checkin to lock down XENIX 3.0(28) release
|***   
|***   Revision 3.0  85/01/17  15:37:59  iv
|***   	- Checkin to save XENIX v3.0(23)
|***   
|***   Revision 2.0  85/01/10  14:05:14  iv
|***   	- Checkin to save XENIX 3.0(22)
|***   
|***   Revision 1.5  84/12/21  15:58:10  iv
|***   	- Changed call to "clock()" to call to "rclock()", which gives
|***   	  clock() the arguments it expects
|***   	- Corrected misc typos (in comments and such)
|***   	- Added "_tasktime" as an alias for "_spl0"
|***   	- Added "_bcopy" entry for "copybuf"
|***   
|***   Revision 1.4  84/04/18  10:20:11  iv
|***   	- Fixed up some RCS comment leader errors, I forgot to set
|***   	  the comment header with 'rcs -c' before I checked in.
|***   
|***   Revision 1.3  84/04/18  10:12:52  iv
|***	- Changed "idleflg" to "Idlef", per sys/clock.c
|***
|*** ------------------------------------------------------------------------
|***	M000	07 Jul 1983	jeIV  (Tandy)
|***		- Added z80 interrupt routine and request fifo initialization
|***
|***	M001	19 Sep 1983	jeIV  (Tandy)
|***		- Added C interface to "umap" routine so drivers can hand
|***		  PHYSICAL user addresses to z80 for "DMA" z80 style
|***		- Added initialization of z80 interrupt ID mask
|***
|***	M002	01 Feb 1984	jeIV  (Tandy)
|***		- Added support for z80 notification of system shutdown
|***		  [done at haltcpu].  Added code to initalize shutdown enable
|***		  byte "z80halt" to a safe value.
|***
|*** ========================================================================

	.data
	.asciz	"Copyright 1987 Tandy Corporation\n"
	.even
	.text

USRSTART= /000000	| User Virtual Addr of First Execable Code

USIZE		=	/1000	| size of U structure
UMASK		=	/fff
PAGESIZE	=	/1000	| size of 1 page
PAGESHIFT	=	12	| shift required
SHIFT1		=	8	| fragments of PAGESHIFT
SHIFT2		=	4	|	for use in shift instructions

INTDATA		= 0x7800d0	| interrupt controller data register
INTCOMMAND	= 0x7800d1	| interrupt controller command register

CLRINT		= /60		| clear interrupt command
MASKWRITE	= /b0		| next data write is new mask register
MASKREAD	= /a4		| condition for reading mask register
INTPROGRAM	= /ffffffe0	| command to program first int with 1 byte
INTARM		= /a5		| arm the chip (and set for reading of mask)
INTRESET	= /00		| reset the interrupt controller
ISRREAD		= /a0		| condition reading interrupt service register

MMU_BASE	= 0x7800d2	| base of MMU registers
MMU_HILIM	= 0x7800d2	| 	limit register, hi segment
MMU_HIOFF	= 0x7800d3	|	offset register, hi segment
MMU_LOLIM	= 0x7800d4	|	limit register, lo segment
MMU_LOOFF	= 0x7800d5	|	offset register, lo segment

Z80IADDR	= /7800d6	| M000  z80 interrupt latch


|***
|***	We must set up software copies of the MMU registers...
|***
|***	     WARNING:	mmu.c *depends* on and uses the fact that
|***			_mmuHiOff follows _mmuHiLim and that
|***			_mmuLoOff follows _mmuLoLim.
|***
|***		[See the code and comments for mmusub() in mmu.c]
|***
	.data
	.globl	_mmuHiLim, _mmuHiOff, _mmuLoLim, _mmuLoOff
mmuhi:
_mmuHiLim:
	.word	0
_mmuHiOff:
	.word	0

mmulo:
_mmuLoLim:
	.word	0
_mmuLoOff:
	.word	0

	.text

|***
|***	Because a TIMERRI (bus timeout while 68000 is bus master) or ADERR
|***	ADERR (user MMU fault) interrupt would also generate a bus error,
|***	these are always masked off.
|***	Under no circumstances should the kernel ever get a TIMERRE (bus
|***	timeout while 68000 is *not* bus master) interrupt; therefore this
|***	mask bit is used to differentiate between MASK1 and MASK0.
|***
|***	MASK7-4 use negative numbers so that we can use a `moveq' in _spl
|***
MASK7	= -1	|/ff		| no ints enabled
MASK6	= -1	|/ff		| level six must disable clock; same as MASK7
MASK5	= -2	|/fe		| allow highest pri int only (clock)
MASK4   = -4	|/fc		| allow clock and disk
MASK1	= 	/38		| lowest pri which does not enable all ints
MASK0	= 	/18		| all ints enabled

|SPL7	= /2700		| status reg value disable all interupts, super mode
|SPL6	= /2600		| level 6
|SPL5	= /2500		| level 5
|SPL4	= /2400		| level 4
|SPL3	= /2300		| level 3
|SPL2	= /2200		| level 2
|SPL1	= /2100		| level 1
|SPL0	= /2000		| level 0

SSTATE	= /2000		| supervisor state bit
SBIT	= 13		| supervisor state bit of status register
TBIT	= 15		| trace bit of status register
NOINTR	= /2700		| no interrupts
TRACE	= /8000		| trace mode bit

|***
|***	Location 0	Interrupt and Trap Vectors
|***
zero:
	.long	_end+0x100	| Initial SSP (FIX ME BUG)
	.long	start		| Initial PC

	.long	_aberr		| vector   2	BUS ERROR
	.long	_aberr		| vector   3	ADDRESS ERROR
	.long	jvector		| vector   4
	.long	jvector		| vector   5
	.long	jvector		| vector   6
	.long	jvector		| vector   7
	.long	jvector		| vector   8
	.long	jvector		| vector   9
	.long	jvector		| vector  10
	.long	jvector		| vector  11
	.long	jvector		| vector  12
	.long	jvector		| vector  13
	.long	jvector		| vector  14
	.long	jvector		| vector  15
	.long	jvector		| vector  16
	.long	jvector		| vector  17
	.long	jvector		| vector  18
	.long	jvector		| vector  19
	.long	jvector		| vector  20
	.long	jvector		| vector  21
	.long	jvector		| vector  22
	.long	jvector		| vector  23
	.long	jvector		| vector  24
	.long	jvector		| vector  25
	.long	jvector		| vector  26
	.long	jvector		| vector  27
	.long	jvector		| vector  28
	.long	jvector		| vector  29
	.long	jvector		| vector  30

	.long	nmi		| vector  31

	.long	do_scall	| vector  32	trap #0 -- System call


	.long	jvector		| vector  33
	.long	jvector		| vector  34
	.long	jvector		| vector  35
	.long	jvector		| vector  36
	.long	jvector		| vector  37
	.long	jvector		| vector  38
	.long	jvector		| vector  39
	.long	jvector		| vector  40
	.long	jvector		| vector  41
	.long	jvector		| vector  42
	.long	jvector		| vector  43
	.long	jvector		| vector  44
	.long	jvector		| vector  45
	.long	jvector		| vector  46
	.long	jvector		| vector  47
	.long	jvector		| vector  48
	.long	jvector		| vector  49
	.long	jvector		| vector  50
	.long	jvector		| vector  51
	.long	jvector		| vector  52
	.long	jvector		| vector  53
	.long	jvector		| vector  54
	.long	jvector		| vector  55
	.long	jvector		| vector  56
	.long	jvector		| vector  57
	.long	jvector		| vector  58
	.long	jvector		| vector  59
	.long	jvector		| vector  60
	.long	jvector		| vector  61
	.long	jvector		| vector  62
	.long	jvector		| vector  63


|***	We use the first user vectors for the interrupt
|***	contoller chip.  It must be programmed to generate
|**	vectors 64-71.

INTVEC	=	64

cont4:	.long	cont4encode	| initiated by z80
cont5:	.long	cont5encode	| initiated by z80
cont6:	.long	cont6encode	| initiated by z80
|***
|***	The following three interrupts cannot be generated without also
|***	generating a bus error.  We keep the first two masked off permanently.
|***	The third should absolutely never be generated while running XENIX;
|***	the mask bit for this interrupt is used for special purposes by
|***	XENIX.
|***
aderr:	.long	spu1encode	| MMU address error
timerri:
	.long	spu2encode	| timeout error, 68000 is bus master
timerre:
	.long	spu3encode	| timeout error, 68000 is not bus master
iper:	.long	ipencode	| parity error, 68000 is bus master
eper:	.long	epencode	| parity error, 68000 is not bus master


|***
|***	Just for fun, and to be totally evil to Mr. Snapp, we decided
|***	to put the mailboxes here instead of letting `ld' place them.
|***	[You hear maniacal laughter in the distance...]
|***

|	c68  - enhanced compiler  - Version 3(24)  29-Mar-88
	.text
	.globl	z80base
	.even
z80base:

	.globl	_z80clk
_z80clk:	.blkb	2
	.even

	.globl	_z80halt
_z80halt:	.blkb	1
	.even

	.globl	_z80disk
_z80disk:	.blkb	84
	.even

	.globl	_z80msgb
_z80msgb:	.blkb	636
	.even

	.globl	_z80grap
_z80grap:	.blkb	18
	.even

	.globl	_z80ser
_z80ser:	.blkb	130
	.even

	.globl	_z80rcv
_z80rcv:	.blkb	4
	.even

	.globl	_z80fifo
_z80fifo:	.blkb	512
	.even

	.globl	_z80req_
_z80req_:	.blkb	66
	.even

	.globl	_z80boot
_z80boot:	.blkb	2
	.even



|***  >SIGH<  trap() uses these symbols to determine if the user attempted
|***  to single-step a system call.  Since we go into supervisor mode at that
|***  point, we won't get the trace trap until after we're in supervisor mode.
|***  So, we need _firstvec and _endvec to be the range of addresses the PC
|***  will be in when we take the trace trap.  trap() will then discard it.

	.globl	_firstvec,_firstve,_endvec
jvector:
_firstvec:
_firstve:
	bra	_vector


cont4encode:
	.globl	_rclock
	jbsr	intr		| cont4
	.long	0
	.long	_rclock
cont5encode:
	.globl	_cont5
	jbsr	intr		| cont5
	.long	1
	.long	_cont5
cont6encode:
	.globl	_cont6
	jbsr	intr		| cont6
	.long	2
	.long	_cont6
spu1encode:
	.globl	_novec
	jbsr	intr		| unexpected interrupt
	.long	3
	.long	_novec
spu2encode:
	.globl	_novec
	jbsr	intr		| unexpected interrupt
	.long	4
	.long	_novec
spu3encode:
	.globl	_novec
	jbsr	intr		| unexpected interrupt
	.long	5
	.long	_novec
ipencode:
	.globl	_parity
	jbsr	intr
	.long	6
	.long	_parity
epencode:
	.globl	_parity
	jbsr	intr
	.long	7
	.long	_parity

	.globl	_endvec		| end of vectors
_endvec:

|***   
|***	START	- The very first XENIX code
|***
|***	Create a C environment and call main, when main returns start the
|***	first user task.
|***

|***
|***	Clear BSS, and initialize global variables as to size of system,
|***	base of working upage, address of first user memory, etc.
|***
	.text
	.globl	start,_end,_edata,_main,_cfginit
|***
|***	Size memory by trying to write a value and testing by reading it.
|***	Also check for memory wrap by looking at loc 0
|***
start:

	movl	#_end,d0	| point to start of user memory (more-or-less)
	addl	#USIZE-1,d0	| Round up one page
	andl	#~UMASK,d0	| to page boundry
	movl	d0,a0
	subl	a1,a1		| point to loc zero

	movl	#3$,nofault	| make memory faults go here
2$:
	movw	#/1234,(a0)	| try to store 1234
	movw	#0,2(a0)	| put another value on the bus
	cmpw	#/1234,(a0)	| did it store ?
	jne	3$		| no, so this is top of memory
	cmpw	#/1234,(a1)	| did it set loc 0 ?
	jeq	3$		| yes, we have finished with all of memory
	addw	#PAGESIZE,a0	| next page boundary
	jra	2$		| loop for more memory

3$:	movl	a0,_maxaddr	| This is top of memory for XENIX

	clrl	nofault



	jsr	_cfginit	| Allocate configurable stuff as apropriate
				|  [cfginit() returns new _end in d0]


	addl	#USIZE-1,d0	| Round up one page
	andl	#~UMASK,d0	|   to page boundry
	movl	d0,_exenix	| addr of end of xenix
	movl	d0,uaddr	| initialize address of current process U
	addl	#USIZE,d0	| next page
	movl	d0,_umem	| first free page
	movl	#_edata,a0	| start clearing from here

|*** Clear BSS, first upage.  Because we may have parity equipped
|*** memory, we don't use a clr intruction on memory, which generates
|*** a read access before write access.
	moveq	#0,d1
1$:	movl	d1,(a0)+	| clear a long word and increment a0
	cmpl	a0,d0		| are we done
	jgt	1$		| loop while less than or equal













|***	Initialize all of physical memory, in case it is parity equipped.
	movl	_maxaddr,a0
	movl	a0,d0		| top of physical memory
	subl	_umem,d0	| number of bytes of user memory
	movl	#/4afc4afc,d1	| initialize memory with illegal instructions

4$:	movl	d1,-(a0)	| Initialize a word
	subql	#4,d0		| We did 4 bytes
	bnes	4$		| Loop for all words (we can't use dbra
				|   here because it might be > 16 bits)


|***	The initial system stack will grow from the end of the system U
|***	page down

	movl	#_u,a0		| physical address of system U
	movl	_exenix,uaddr	| where we keep the proc 0 upage when running
				|   other processes
	addw	#USIZE-2,a0
	movl	a0,sp

|***
|***	Initialize the Z80 interface and interrupt controller.
|***
	movw	#NOINTR,sr	| hi processor priority

	.globl	_z80init
	jsr	_z80init	| Initialize z80 structures

	movl	#z80base,4:w	| inform z80 of base address
8$:	cmpl	#z80base,4:w	| wait for z80 to intialize buffer space
	jeq	8$


	| We need to wait until this point to check the reboot-byte
	|   because the Z80 didn't know where the iopb's were until
	|   just now (we just told it).
	.globl	_z80rebo	| [In real life: z80reboot()]
	jsr	_z80rebo	| Check for reboot or normal boot




	|*** Check to make sure we're running on a 68010.  68010/XENIX will
	|***   appear to run on a 68000, until you get a bus or address error.
	|*** If we ARE running on a 68000, we should get an illegal
	|***   instruction trap here.
ill_loc:
	.word	/4e7a		| movec	vbr,a0
	.word	/8801


	lea	INTCOMMAND,a0	| pointer to interrupt command reg
	movl	a0,a1		| pointer to
	subql	#1,a1		|	interrupt data reg
	movb	#INTRESET,(a0)	| reset controller

	moveq	#INTPROGRAM,d0	| command to program first int level
	moveq	#INTVEC,d1	| interrupt vector to be programmed
	moveq	#7,d2		| number of vectors minus one

9$:	movb	d0,(a0)		| specify which register to program
	movb	d1,(a1)		| provide vector number
	addql	#1,d0		| next register
	addql	#1,d1		| next vector number
	dbra	d2,9$		| loop



	movb	#INTARM,(a0)	| arm the chip



	| make a stack so first proc can have a u.u_ar0
	moveml	#/ffff,-(sp)	| save all regs
	moveml	#/f800,-(sp)	| save 5 more
	movw	#SSTATE,sr	| lo processor priority



	jsr	_main		| call main

|***	return from main is "in" init process


	clrw	-(sp)		| Push format word

	movl	#USRSTART,-(sp)	| starting addr for user process
	clrw	-(sp)		| return to user mode
	rte

	.data
	.globl	_exenix,_umem,_maxaddr

_exenix: .long	0	| The first page after xenix (also the first U page)
_umem:	 .long	0	| The first free page
_maxaddr:.long	0	| The last page of memory

nmiflag:.byte	0

|***	Because we seem to be unable to read the mask register
|***	of the interrupt controller, we maintain copy in software.

	.globl	intmask		| so it will come out in the symbol table
intmask:.byte	/ff		| software copy of interrupt mask register
	.even


|***   
|***	Trap and Interrupt Handlers
|***
|***	Your basic stack is:
|***				pchi,pclo	long
|***				sr		word
|***				pchi,pclo	long from bsr
|***
|***	For traps, the "long from bsr" points to the trap number.
|***	For interrupts, the "long from bsr" points to the interrupt
|***		number, followed by a jmp instruction to the handler.
|***
|***	Your stack handed to _trap becomes:
|***				pchi,pclo	long
|***				sr		long
|***				intmask		long	(formerly unused "mmu")
|***				regs		16 longs
|***				trap no		long
|***
|***	Your stack handed to interrupt handlers is:
|***				pchi,pclo	long
|***				sr		long
|***				intmask		long
|***				saved a1	long
|***				saved a0	long
|***				saved d1	long
|***				saved d0	long
|***				intno		long
|***
|*** trap(no,d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,usp,mmu,sr,tpc)
|*** long no,d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,usp,mmu,sr,tpc;
|***
|*** devintr(no,d0,d1,a0,a1,intmask,sr,tpc)
|*** long no,d0,d1,a0,a1,sr,tpc;
|***

|***
|***	INTR -- handle vectored interrupts from interrupt controller
|***
	.text
intr:				| vectored interrupt handler
	clrw	-(sp)		| get to long alignment
	moveml	<a0,a1,d0,d1>,-(sp)

	movl	18(sp),a0	| where bsr put address of itself plus 4
	clrl	18(sp)		| make args nice

	movl	(a0)+,d0	| fetch vector number
	movl	d0,-(sp)	| push vector number as arg

	| Compute the appropriate bits to set in the mask register
	moveq	#-1,d1
	asll	d0,d1

	| d1 now holds bits to set in mask register
	movb	intmask,d0	| get old mask value
	movb	d0,23(sp)	| save as stack arg
	orb	d1,d0		| compute new mask

	| d0 now holds the new mask
	lea	INTCOMMAND,a1	| points after data register
	movb	#MASKWRITE,(a1)	| prepare to write new mask value
	movb	d0,-(a1)	| new mask into hardware
	movb	d0,intmask	| new mask into software copy
	addql	#1,a1		| points to command register again
	movb	#CLRINT,(a1)	| clear the interrupt -- trust the mask

	| we use d1 to figure what the new SR will be
	movw	26(sp),d1	| old sr
	bset	#SBIT,d1	| set supervisor bit
	bclr	#TBIT,d1	| clear trace bit
	movw	d1,sr		| new processor pri -- enable higher pri ints

	movl	(a0),a0		| Grab handler's address
	jsr	(a0)		| call handler(no,d0,d1,a0,a1,intmask,sr,tpc)

	addql	#4,sp		| pop int number
	btst	#SBIT,22(sp)	| came from supervisor mode?
	jne	1$		| yes: do not reschedule
	tstb	_runrun		| need to reschedule?
	jne	reschedule	| yes: branch
1$:
	clrw	_Idlef		| don't loop in idle
	movb	19(sp),d0	| saved mask register
	lea	INTCOMMAND,a0	| pointer to hardware register
	movw	#NOINTR,sr	| hi priority for mask fiddling
	movb	d0,intmask	| software copy
	movb	#MASKWRITE,(a0)	|
	movb	d0,-(a0)	| hardware copy
	moveml	(sp)+,<a0,a1,d0,d1>	| restore registers
	addql	#6,sp		| pop old intmask, hi sr
	rte

	.data
	.comm	_runrun,1
	.text


|***
|***	RESCHEDULE -- branched to from INTR to fake a rescheduling trap
|***	RESCHED1 -- branched to from DO_SCALL to fake a rescheduling trap
|***
	.text
reschedule:
	moveml	(sp)+,<a0,a1,d0,d1>	| restore registers, leaving intmask
resched1:
	moveml	#/ffff,-(sp)	| save ALL registers
	moveq	#1,d0		| fake trap 1
	jra	calltrap



|***
|***	DO_SCALL -- handle system call traps quickly
|***

do_scall:
	clrw	-(sp)		| Make sr a long
	clrl	-(sp)		| put old `mmu' arg on stack
	moveml	#/ffff,-(sp)	| save all regs
	movb	intmask,67(sp)	| save interrupt info on stack in old "mmu" slot

	btst	#SBIT,70(sp)	| Did we come from supervisor mode?
	jeq	1$		|   No:  proceed with system call

	moveq	#32,d0		| Yes:  Oh, oh; we executed a "trap #0" from
	bra	calltrap	|   SUPERVISOR mode...  (prepare to panic:)

1$:	movl	usp,a0
	movl	a0,60(sp)	| Replace saved a7 with user stack pointer

	jbsr	_systrap	| Run the system call

	movl	60(sp),a0	| Possibly altered user stack pointer
	movl	a0,usp		| Restore user stack pointer

|***  Program interrupt controller chip
	movb	67(sp),d0	| saved mask register
	lea	INTCOMMAND,a0	| pointer to hardware register
	movw	#NOINTR,sr	| hi priority for mask fiddling
	movb	d0,intmask	| Save software a copy
	movb	#MASKWRITE,(a0)	|
	movb	d0,-(a0)	| Give hardware its copy

	moveml	(sp)+,#/7fff	| restore all registers except sp
	addw	#10,sp		| Pop saved usp, intmask, and high word of sr
	clrb	nmiflag		| allow processing of NMI
	rte			| Back to the User



|***
|***	VECTOR -- handle processor traps
|***
	.globl	_vector,_trap

_vector:


	clrw	-(sp)		| Make sr a long
	clrl	-(sp)		| Make room for int-#
	moveml	#/ffff,-(sp)	| Save ALL regs
	moveq	#0,d0		| Clear upper-half of d0 for later
	movw	76(sp),d0	| Get format/vec-offset word (ffff00vvVvvvVvvv)
	clrw	76(sp)		| Make format 0000 for later legal rte
	rolw	#4,d0		| Put format nybble in lower bits
	lsrw	#6,d0		| Eat format code and div vec-offset by 4
				|   to get the vector-number (int-#)

	btst	#SBIT,70(sp)	| Did we come from supervisor mode?
	jeq	illtrap		|   no: proceed with trap
	cmpl	#ill_loc,72(sp)	| Did we generate the trap to check for a '010?
	jne	illtrap		|   nope: proceed with trap

	.data
pmsg:	.asciz	"not a 68010"
	.even
	.text

	.globl	_panic
	movl	#pmsg,-(sp)	| "panic: not a 68010"
	jsr	_panic
	jmp	_haltcpu	| Just in case...


illtrap:
	movb	intmask,67(sp)	| save interrupt info on stack in "mmu" arg

calltrap:
	movl	d0,-(sp)	| push trap number on stack
	movl	usp,a0		| get user stack pointer
	movl	a0,64(sp)	| replace saved a7 with user stack pointer

	jbsr	_trap		| call trap(number, regs)

	addql	#4,sp		| pop off vector number
	movl	60(sp),a0	| saved user stack pointer
	movl	a0,usp		| restore user stack pointer

	movb	67(sp),d0	| saved mask register
	lea	INTCOMMAND,a0	| pointer to hardware register
	movw	#NOINTR,sr	| hi priority for mask fiddling
	movb	d0,intmask	| software copy
	movb	#MASKWRITE,(a0)	|
	movb	d0,-(a0)	| hardware copy

	moveml	(sp)+,#/7fff	| restore all registers (not sp)
	addw	#10,sp		| saved usp, intmask, and hi of sr
	clrb	nmiflag		| allow processing of NMI
	rte			| back to somewhere

|*** Debounce NMI
nmi:
	bset	#0,nmiflag    | Figure out if we're already handling NMI
			      | uses kludge because some MMU's can't handle TAS

	jeq	_vector		| no -- call vector

	rte			| yes -- ignore this one

|***
|***	Bus error and Address error traps push 4 extra words on the stack.
|***	On a 68010, they push 22 extra words on the stack.
|***
|***	These routines take the words off the stack so that all traps look
|***	the same.
|***
	.data
	.globl	berrinfo,aerrinfo,_fltinfo,nofault
_fltinfo:
berrinfo:.word	0,0,0,0		| save error info in global
aerrinfo:.word	0,0,0,0		| save error info in global
nofault:.long	0		| address to go to on memory faults


	.text

	.globl	_aberr
_aberr:
	movl	a4,-(sp)	| Save scratch register

	movw	12(sp),a4	| Grab special status word to save it
	movw	a4,berrinfo	| Silliness because we don't know if we're
	movw	a4,aerrinfo	|   an address error or a bus error

	movl	14(sp),a4	| Grab fault address to save it
	movl	a4,berrinfo+2	| Silliness because we don't know if we're
	movl	a4,aerrinfo+2	|   an address error or a bus error

	movw	28(sp),a4	| Grab fault instruction to save it
	movw	a4,berrinfo+6	| Silliness because we don't know if we're
	movw	a4,aerrinfo+6	|   an address error or a bus error





	movl	(sp)+,a4	| Restore scratch reg

	movl	(sp),50(sp)	| Copy sr and pc-hi to bottom of frame
	movl	4(sp),54(sp)	| Copy pc-lo and format to bottom
	addw	#50,sp		| Pop 68010 Group 0 execption frame,
				|   except for last 4 words we copied
    |*** Voila!  Stack now looks like a "normal" exception frame

    |*** WARNING:  The format word still says there's a "long" exception
    |***	   frame on the stack.  Do NOT rte to this frame without 
    |***	   clearing the format word, or you will have one very
    |***	   confused and slightly angry 68010...
    |***	   We leave the old format/vector word on the stack so
    |***	   that _vector can use it to get the vector number.

	tstl	nofault		| nofault set ?
	jeq	1$

	movl	nofault,2(sp)	| Yes, so return there
	clrw	6(sp)		| Make format word legal
	rte			| Back to where nofault said to go...

    |*** Do an rte to _vector (which handles the fault) to avoid processor halt
1$:	clrw	-(sp)		| Push format word
	movl	#_vector,-(sp)	| push pc
	movw	sr,-(sp)	| push status word
	rte



|***
|***	SAVE(regs)
|***	struct label_t *regs;
|***
|***		Save register contents in U area pointed to by
|***		addr, in prepartion for context switch.
|***
|***		On TRS-16, must copy struct U and stack contents
|***		from working Upage to process's Upage.
|***
|***		  Note:	The M68010SR code does NOT work.  
|***			Do not define M68010SR.  23 Oct 87  jeIV
|***
	.data

	.globl	_dustack	| number of bytes from &u to &u.stack
				| initialized in C code
	.globl	_u
	.comm	_u,USIZE

	.globl	uaddr
uaddr:	.long	0		| address of U area for current process
				| set by last resume, initialized with
				| same value as _exenix
	.text
	.globl	_save

_save:				| save(regs)
	movl	(sp)+,a1	| return address off of stack
	movl	(sp),a0		| ptr to a label_t (place to save regs)
	moveml	#/fcfc,(a0)	| save d2-d7 and a2-a7
	movl	a1,48(a0)	| save return address
	movl	a1,-(sp)	| and back onto stack for safe keeping
|***
|***	Copy the stack and U-struct from the working area to the
|***	process's allocated U and stack
|***


    |*** First, copy the stack [from sp to (_u + USIZE - 1)]
	movl	sp,d0		| copy current stack pointer
	movl	#_u+USIZE,a0	| end of U
	subl	sp,a0		| number of bytes of stack
	movl	a0,-(sp)	| nbytes arg to copybuf

	movl	d0,d1		| stack pointer copy
	subl	#_u,d1		| calculate offset within U+stack page
	addl	uaddr,d1	| address within process's real U
	movl	d1,-(sp)	| toaddr arg to copybuf
	movl	d0,-(sp)	| copy of sp -- fromaddr arg to copybuf
	jbsr	_copybuf	| copy bytes
	addw	#12,sp		| pop args to copybuf

    |*** Then, copy the U struct proper  [from &u to &u.u_stack (end of the U)]
	movl	_dustack,-(sp)	| nbytes = (char *) u.u_stack - (char *) &u
	movl	uaddr,-(sp)	| toaddr
	movl	#_u,-(sp)	| fromaddr
	jbsr	_copybuf	| copy bytes
	addw	#12,sp		| pop args to copybuf


	moveq	#0,d0		| return value is 0 in saver
	rts			| return

|***
|***	RESUME(upage,regs)
|***	struct label_t *regs;
|***
|***		Switch contexts to kernel process whose U and stack
|***		are in physical page upage, and whose registers are
|***		saved in its U at the location corresponding to
|***		address regs.
|***
|***		On TRS-16, must copy new process's U-struct and stack
|***		into the working area, and must set variable (uaddr)
|***		to hold address of current process's allocated U-page.
|***
|***		  Note:	The M68010SR code does NOT work.  
|***			Do not define M68010SR.  23 Oct 87  jeIV
|***
	.globl	_resume
_resume:			| resume(upage,regs)
	movl	4(sp),d2	| address of user structure
	movl	8(sp),a2	| ptr to label_t (place where regs were saved)

	asll	#SHIFT1,d2	| convert upage number
	asll	#SHIFT2,d2	|	to physical address
	movl	d2,uaddr	| stow address of new U in global


	movl	_dustack,-(sp)	| nbytes
	movl	#_u,-(sp)	| toaddr
	movl	d2,-(sp)	| from addr
	jbsr	_copybuf	| copy bytes
	addw	#12,sp		| pop args


	movl	44(a2),sp	| stack pointer to be restored
				| note that current stack contents are now
				| 	invalid

	movl	sp,d3		| copy of new stack pointer
	movl	#_u+USIZE,a3	| end of working Upage
	subl	sp,a3		| calculate size of saved stack
	movl	a3,-(sp)	| nbytes

	movl	d3,-(sp)	| toaddress
	subl	#_u,d3		| convert sp copy to offset in page
	addl	d2,d3		| compute address within process's copy
	movl	d3,-(sp)	| fromaddress
	jbsr	_copybuf	| copy bytes
	addw	#12,sp		| pop args


	jbsr	_spl0		| we return in an interruptible condition
	movl	a2,a0		| reg pointer into a volatile register
	moveml	(a0)+,#/fcfc	| restore d2-d7 & a2-a7
	movl	(a0),a0		| return address
	moveq	#1,d0		| return 1
	jmp	(a0)		| return in new context

|***
|***	SETJMP(regs)
|***	LONGJMP(regs)
|***
|***		Similar to save and resume, but do not involve
|***		context switches, and therefore, don't require
|***		copying of any U information.
|***
	.globl	_setjmp,_longjmp
_setjmp:
	movl	(sp)+,a1	| return address
	movl	(sp),a0		| ptr to a label_t (place to save regs)
	moveml	#/fcfc,(a0)	| save d2-d7 and a2-a7
	movl	a1,48(a0)	| save return address
	moveq	#0,d0		| return value is 0 in saver
	jmp	(a1)		| return

_longjmp:
	movl	4(sp),a0	| ptr to label_t (place where regs were saved)
	moveml	(a0)+,#/fcfc	| restore d2-d7 & a2-a7
	movl	(a0),a1		| return address
	moveq	#1,d0		| return 1
	jmp	(a1)		| return

|***
|***	SPL	- spl7,6,5,4,1,0 set the interupt mask in the status reg
|***		also return the previous level so that a call to splx
|***		can be used to set a previous level
|***

	.globl	_spl7,_spl6,_spl5,_spl4,_spl1,_spl0,_splx,_tasktime

_spl7:
_spl6:				| spl6 and spl7 use the same interrupt mask
	moveq	#MASK6,d1
	jra	splcom

_spl5:
	moveq	#MASK5,d1
	jra	splcom

_spl4:
	moveq	#MASK4,d1
	jra	splcom

_spl1:
	moveq	#MASK1,d1
	jra	splcom

_tasktime:
_spl0:
	moveq	#MASK0,d1
	jra	splcom

_splx:
	movb	7(sp),d1


splcom:				| Common spl code: d1 has new mask
	lea	INTCOMMAND,a1	| points after data register
	movw	sr,d0		| save processor pri
	movw	d0,a0
	movw	#NOINTR,sr	| no interrupts while playing with mask
	movb	#MASKWRITE,(a1)	| prepare to write new mask value
	movb	d1,-(a1)	| new mask into hardware
	movb	intmask,d0	| old mask register
	movb	d1,intmask	| new mask into software copy
	movw	a0,d1		
	movw	d1,sr		| restore processor pri
	rts			| return


|***
|***	IDLE	- spl0 and wait for interrupt
|***
	.globl	_idle,_waitloc,_Idlef
	.data
_Idlef:
	.word	0
	.text

_idle:
	movw	#1,_Idlef
	jbsr	_spl0
_waitloc:
	tstw	_Idlef
	jne	_waitloc

	movb	d0,d1		| Fast call to splx()
	jbsr	splcom		| Don't have to mess with the stack this way

	rts

|***
|***	COPYBUF	- copy in kernel memory
|***
|***		Because this routine is used by copyin/out on TRS-16,
|***		copybuf must return 0 in d0.
|***
|***		Copyin/out make use of the fact that their arguments
|***		are logically equivalent to those of copybuf.
|***
|***		As presently coded, this routine is slightly inefficient
|***		for small transfers; however, it is highly efficient
|***		for large transfers, for which it is called often and
|***		for which efficiency is crucial.
|***
|***		NOTE:
|***		The bytecount argument must be in the range
|***		0 <= bytecount < 2^31.  (Because the 68000
|***		cannot address this much memory, this should
|***		not cause any problems.)
|***
|***		With the single exception noted above,
|***		copybuf will handled any combination of even, odd,
|***		zero, or nonzero arguments, as well as any overlap
|***		in the regions to be copied.
|***
|***		This routine is probably the most heavily used
|***		assembly language routine.  It is used by
|***			copyin
|***			copyout
|***			copyiin
|***			copyiout
|***			save
|***			resume
|***		A single call to copybuf is used in places where there
|***		would be multiple calls to
|***			copyseg
|***
	.globl	_copybuf
	.globl	_bcopy

_bcopy:
_copybuf:			| copybuf(from,to,count)
	movl	12(sp),d0	| d0 = count
	jeq	copyend		| if zero, no work to do
	movl	4(sp),a0	| a0 = from
	movl	8(sp),a1	| a1 = to
	cmpl	a0,a1		| compare from addr and to addr
	jeq	copyend		| if equal, don't do anything
	moveml	#/3f3e,-(sp)	| save reg d2-d7 and a2-a6
				| moveml does not affect condition codes!
	jgt	copyup		| if fromaddr < toaddr, must ripple correctly

| From address > To address: copy from low byte first

| If both from and to are odd, move one byte and proceed to even code
| But if one is even and the other odd proceed to the bytewise move

	movl	a0,d2		| get addr in data reg so btst can work
	movl	a1,d3		|
	moveq	#0,d4		| clear d4 for use in btst instructions
	btst	d4,d2		| test from bit 0
	jeq	1$		| from even
	btst	d4,d3		| test to bit 0
	jeq	70$		| to even but from odd

				| both odd
	movb	(a0)+,(a1)+	| move one byte
	subql	#1,d0		| cnt--
	jra	2$

1$:				| from even
	btst	d4,d3		| test to bit 0
	jne	70$		| from even but to odd

2$:				| both even
	moveq	#44.,d1		| use d1 to hold the constant
3$:
	subl	d1,d0		| subtract number of bytes to sub
	jle	4$		| if too few bytes, go to word-at-a-time
|*** Note that the moveml instruction generates an extra memory
|*** access; to avoid unexpected memory faults, if a move is exactly
|*** 44 bytes, we do it a word at a time.

	moveml	(a0)+,#/7cfc	| load 44 bytes
	moveml	#/7cfc,(a1)	| store 44 bytes
	addl	d1,a1		| can't autoinc the destination on moveml
	jra	3$		| loop

4$:				| 44 or few bytes to go
	addl	d1,d0
	asrl	#1,d0		| convert to word count
	jcs	61$		| if byte-count was odd, C-bit set
	jra	51$		| for proper iteration, enter end of loop

50$:	movw	(a0)+,(a1)+	| move word
51$:	dbra	d0,50$		| while words left
	jra	copypop		| pop regs and return

60$:	movw	(a0)+,(a1)+	| move words, then 1 byte at end
61$:	dbra	d0,60$
	movb	(a0)+,(a1)+
	jra	copypop		| pop regs and return
			
				| one even the other odd

| For the case where one number is even, the other odd,
| we cannot just use one big dbra loop.  This is because
| the dbra instruction operates only on the low order
| word of a register, not on the long register.  This
| makes certain gyrations necessary:
70$:
	swap	d0		| get high word of register into low word
	moveq	#-1,d1		| initialize d1 with FFFF
	bras	72$		| transfer 64K bytes some no. of times


71$:				| d1 holds FFFF, will transfer FFFF+1 bytes
	movb	(a0)+,(a1)+	| transfer byte
	dbra	d1,71$		| loop each of the 64K bytes
				| fall out of inner loop with FFFF in d1

72$:	dbra	d0,71$		| outer loop, 64K per iteration

	swap	d0		| get back low order half of d0
	bras	74$		| enter loop at bottom for proper iteration

73$:	movb	(a0)+,(a1)+	| transfer byte
74$:	dbra	d0,73$		| loop for each byte
	jra	copypop		| pop regs and return
	

| From address < To address: copy from hi byte first

copyup:
	addl	d0,a0		| pointer past end of from region
	addl	d0,a1		| pointer past end of to region

| If both from and to are odd, move one byte and proceed to even code
| But if one is even and the other odd proceed to the bytewise move

	movl	a0,d2		| get addr in data reg so btst can work
	movl	a1,d3		|
	moveq	#0,d4		| clear d4 for use in btst instructions
	btst	d4,d2		| test from bit 0
	jeq	1$		| from even
	btst	d4,d3		| test to bit 0
	jeq	80$		| to even but from odd

				| both odd
	movb	-(a0),-(a1)	| move one byte
	subql	#1,d0		| cnt--
	jra	2$

1$:				| from even
	btst	d4,d3		| test to bit 0
	jne	80$		| from even but to odd

2$:				| both even
	moveq	#44.,d1		| use d1 to hold the constant
	cmpl	d1,d0		| might we be doing a moveml?
	jle	5$		| no: jump around moveml loop (see note below)
	movw	-(a0),-(a1)	| yes: move 1 word because of moveml funniness
	subql	#2,d0		| indicate word moved in bytecount register
3$:
	subl	d1,d0		| subtract number of bytes to sub
	jlt	4$		| if too few bytes, go to word-at-a-time
|*** Note that the moveml instruction generates an extra memory
|*** access; to avoid unexpected memory faults, we must have assured
|*** that the first transfer instruction is not a moveml.

	subl	d1,a0		| move back from pointer
	moveml	(a0),#/7cfc	| load 44 bytes
	moveml	#/3f3e,-(a1)	| store 44 bytes
	jra	3$		| loop

4$:				| -44 <= d0 < 0
	addl	d1,d0		| get back real byte count

5$:				| fewer than 44 bytes to go
	asrl	#1,d0		| convert to word count
	jcs	71$		| if byte-count was odd, C-bit set
	jra	61$		| for proper iteration, enter end of loop

60$:	movw	-(a0),-(a1)	| move word
61$:	dbra	d0,60$		| while words left
	jra	copypop		| pop regs and return

70$:	movw	-(a0),-(a1)	| move words, then 1 byte at end
71$:	dbra	d0,70$
	movb	-(a0),-(a1)
	jra	copypop

				| one even the other odd

| For the case where one number is even, the other odd,
| we cannot just use one big dbra loop.  This is because
| the dbra instruction operates only on the low order
| word of a register, not on the long register.  This
| makes certain gyrations necessary:
80$:
	swap	d0		| get high word of register into low word
	moveq	#-1,d1		| initialize d1 with FFFF
	bras	82$		| transfer 64K bytes some no. of times


81$:				| d1 holds FFFF, will transfer FFFF+1 bytes
	movb	-(a0),-(a1)	| transfer byte
	dbra	d1,81$		| loop each of the 64K bytes
				| fall out of inner loop with FFFF in d1

82$:	dbra	d0,81$		| outer loop, 64K per iteration

	swap	d0		| get back low order half of d0
	bras	84$		| enter loop at bottom for proper iteration

83$:	movb	-(a0),-(a1)	| transfer byte
84$:	dbra	d0,83$		| loop for each byte

copypop:
	moveml	(sp)+,#/7cfc	| restor reg d2-d7 and a2-a6
copyend:
	moveq	#0,d0
	rts

|***
|***	CLEARBUF	- zero a region of physical memory
|***
|***		This code will work for any combination of (even, odd)
|***		address and bytecount.  The code is rather inefficient
|***		for small buffers, though.
|***
|***		NOTE: because of the way the dbra instruction is used,
|***		this code will only work for bytecount < 256 * 64K;
|***		bytecount must be less than 16 Megabytes.
|***
	.globl	_clear
	.globl	_clearbu
	.globl	_clearbuf
_clearbuf:			| clearbuf(addr, nbytes)
_clearbu:
_clear:
	movl	8(sp),d0	| nbytes
	jeq	9$		| no bytes to clear -- done
	movl	4(sp),a0	| addr
	moveml	#/3f3e,-(sp)	| save regs d2-d7,a2-a6

	moveq	#0,d2		| code *depends* on d2 being 0

	addl	d0,a0		| point after clear region
	movl	a0,d1		| get into data register for btst
	btst	d2,d1		| is address odd?
	beq	1$		| no: branch to code for even address
				| yes: odd address
	movb	d2,-(a0)	| clear one byte
	subql	#1,d0		| decrement byte count

1$:				| even address
	cmpl	#48,d0		| quick check for small transfer
	jlt	35$		| don't bother clearing registers!

	movl	d2,d1		| clear registers to be used for mem clearing
	movl	d1,d3
	movl	d1,d4
	movl	d1,d5
	movl	d1,d6
	movl	d1,d7
	movl	d1,a1
	movl	d1,a2
	movl	d1,a3
	movl	d1,a4
	movl	d1,a5
	movl	d1,a6

	movl	d0,-(sp)	| save byte count
	lsrl	#8,d0		| divide by 256
	jra	16$

15$:	moveml	#/7f7e,-(a0)	| clear 52 bytes
	moveml	#/7f7e,-(a0)	| 	+ 52 = 104
	moveml	#/7f7e,-(a0)	| 	+ 52 = 156
	moveml	#/7f7e,-(a0)	| 	+ 52 = 208
	moveml	#/3f7e,-(a0)	|	+ 48 = 256
16$:	dbra	d0,15$		| loop to clear 256 bytes

	movl	(sp)+,d0	| pop back byte count
	andl	#/FF,d0		| adjust for large transfer just completed
	
	moveq	#48,d1		| number of bytes transferred in next loop

2$:	subl	d1,d0		| subtract number of bytes cleared/iteration
	jlt	3$		| if result < 0, then proceed to slow loop
	moveml	#/3f7e,-(a0)	| clear memory
	jra	2$		| loop

3$:	addl	d1,d0		| get back number of bytes left to clear

35$:	asrl	#1,d0		| convert to word count
	jcs	7$		| carry set indicates odd byte
	jra	5$		| enter loop at bottom for proper indexing

4$:	movw	d2,-(a0)	| clear word
5$:	dbra	d0,4$		| loop for all words
	jra	8$		| done -- there is no odd byte

6$:	movw	d2,-(a0)	| clear word
7$:	dbra	d0,6$		| loop for all words
	movb	d2,-(a0)	| clear odd byte at beginning

8$:	moveml	(sp)+,#/7cfc	| restore regs d2-d7,a2-a6
9$:	rts


|***
|***	HALTCPU	- stop cpu
|***				M002  And stop the z80 & hard drives
|***
	.globl	_haltcpu
_haltcpu:
	movw	#NOINTR,sr	| lock out all interrupts
	jbsr	_zhalt		| M002  Ask z80 to shutdown
1$:	movw	#NOINTR,sr	| M002  Added extra DI just-in-case...
	bras	1$		| M002  Stop. Now.


|***	BEGIN M000...
|***
|***	Z80INTR	- interrupt the z80
|***
	.globl	_z80intr
_z80intr:
	movw	d0,Z80IADDR	| I/O latch interrupts z80 on access
	rts

|***	...END M000


|***
|***	FUBYTE,FUIBYTE,FUSHORT,FUISHORT,FUWORD,FUIWORD	- From User
|***	WORK IN PROGRESS
	.globl	_fubyte,_fuibyte,_fushort,_fuishor,_fuishort,_fuword,_fuiword

_fubyte:			| fu(i)byte(user_addr)
_fuibyte:
	movl	4(sp),a0
	moveq	#1,d0
	jbsr	umap
	jcs	1$

	moveq	#0,d0
	movb	(a0),d0
1$:	rts
	
_fushort:
_fuishor:			| fu(i)short(user_addr)
_fuishort:
	movl	4(sp),a0
	moveq	#2,d0
	jbsr	umap
	jcs	2$
	movl	a0,d1

	moveq	#0,d0			| cleared d0 used twice:
	btst	d0,d1			| test low order bit
	jeq	1$
	moveq	#-1,d0
	jra	2$
1$:	movw	(a0),d0			| result zero-extended by earlier clr
2$:	rts
	
_fuword:			| fu(i)word(user_addr)
_fuiword:
	movl	4(sp),a0
	moveq	#4,d0
	jbsr	umap
	jcs	2$
	movl	a0,d1
	btst	#0,d1
	jeq	1$
	moveq	#-1,d0
	jra	2$
1$:	movl	(a0),d0
2$:	rts

|***
|***	SUBYTE,SUIBYTE,SUSHORT,SUISHORT,SUWORD,SUIWORD	- Set User
|***
	.globl	_subyte,_suibyte,_sushort,_suishor,_suishort,_suword,_suiword

_subyte:			| su(i)byte(user_addr,byte)
	movl	4(sp),a0
	moveq	#1,d0
	jbsr	udmap
	jra	subytecom
_suibyte:
	movl	4(sp),a0
	moveq	#1,d0
	jbsr	umap
subytecom:
	jcs	suret
	movb	11(sp),(a0)
	jra	suok

_sushort:			| su(i)short(user_addr,word)
	movl	4(sp),a0	| user_address
	moveq	#2,d0
	jbsr	udmap
	jra	sushortcom

_suishor:
_suishort:
	movl	4(sp),a0	| user_address
	moveq	#2,d0
	jbsr	umap
sushortcom:
	jcs	suret
	movl	a0,d1
	btst	#0,d1
	jne	suerr
	movw	10(sp),(a0)
	jra	suok

_suword:			| su(i)word(user_addr,long)
	movl	4(sp),a0	| user_address
	moveq	#4,d0
	jbsr	udmap
	jra	suwordcom
_suiword:
	movl	4(sp),a0	| user_address
	moveq	#4,d0
	jbsr	umap
suwordcom:
	jcs	suret
	movl	a0,d1
	btst	#0,d1
	jne	suerr

	movl	8(sp),(a0)
suok:
	moveq	#0,d0
suret:
	rts
suerr:
	moveq	#-1,d0
	rts


|***
|***	UMAP -- check if address and nbytes given actually resides
|***		in current user's address space.  If so, return with user
|***		logical address replaced by physical address, and carry
|***		bit clear.
|***		If not, return with d0 = -1, and carry bit set.
|***	
|***	UDMAP -- like UMAP, but considers any access to text segment to
|***		an error.  Used by subyte, sushort, suword, copyout.
|***
|***	Arguments:	d0:	nbytes
|***			a0:	logical address
|***	Results:	Success:
|***				a0: physical address
|***				d0: nbytes (unchanged)
|***				carry bit clear
|***			Error:
|***				a0: undefined
|***				d0: -1
|***				carry bit set
|***
|***	Uses:		d1,a1
|***
|***	Note:		This routine returns with  -1 in d0 so that its
|***			callers need take no special action to indicate
|***			an error to their callers.
|***


|***	BEGIN M001...

|***	caddr_t  umap(addr, count)
|***	  caddr_t	addr;
|***	  int		count;
|***
	.globl	_umap
_umap:
	movl	4(sp),a0
	movl	8(sp),d0
	jbsr	umap
	bcs	1$			| Skip the move if carry flag
	movl	a0,d0			|   is set (error), -1 already in d0
1$:	rts

|***	...END M001


udmap:	movl	a0,d1			| address into d register for btst
	btst	#23,d1			| is this a text reference?
	jeq	umaplow			| no: use low segment
					| yes: fall thru to error return

umaperr:
	moveq	#-1,d0			| error return, set up d0
	movw	d0,cc			| indicate error
	rts

umap:
	movl	a0,d1			| address into d register for btst
	btst	#23,d1			| is this a text reference?
	jeq	umaplow			| no: use low segment

	movl	#mmuhi,a1		| yes: ptr to hi seg registers in a1
	bclr	#23,d1			|	clear hi seg bit
	movl	d1,a0			|	replace given address
	jra	umapcom			|	and continue with work
umaplow:
	movl	#mmulo,a1		| ptr to lo seg registers in a1

umapcom:
	tstl	d0			| test bytecount
	jeq	umapok			| bytecount of 0, indicate OK
	jlt	umaperr			| negative bytecount--error
	asrl	#SHIFT1,d1		| convert to page
	asrl	#SHIFT2,d1		|	number
	addw	2(a1),d1		| add offset register
	btst	#11,d1			| Did add overflow 11 bit MMU range?
	jne	umaperr			| if carry produced, then error

	cmpw	(a1),d1			| compare with limit register
	jhi	umaperr			| if (pageno > limit reg) then error

	cmpl	#PAGESIZE,d0		| if bytecount is <= 1 page
	jle	umapok			|	then done with verification

	movl	a0,d1			| starting address
	addl	d0,d1			| + bytecount
	subql	#1,d1			| - 1
	asrl	#SHIFT1,d1		| convert to page
	asrl	#SHIFT2,d1		|	number
	addw	2(a1),d1		| add offset register
	btst	#11,d1			| Did add overflow 11 bit MMU range?
	jne	umaperr			| if carry produced, then error

	cmpw	(a1),d1			| compare with limit register
	jhi	umaperr			| if (pageno > limit reg) then error

umapok:
	movw	2(a1),d1		| no error: take page offset
	asll	#SHIFT1,d1		| convert to
	asll	#SHIFT2,d1		|	bytecount
	addl	d1,a0			| and add to get physical addr
	movw	#0,cc			| clear condition codes for return
	rts


|***
|***	COPYIN/OUT	in and out of system space (from/to user space)
|***
|***		The current implementations is somewhat of a hack
|***		and depends on the arguments to copybuf corresponding
|***		to those to these routines.  We verify that the requested
|***		transfer involves entirely valid user address, replace
|***		the user address argument with a physical address, and
|***		use copybuf to do the actual copying.
|***
|***		The current implementation works for any combination
|***		of odd and even args.
|***
|***		The code here depends upon copybuf returning 0.
|***
	.globl	_copyin,_copyiin,_copyout,_copyiout,_copyiou
|***	copyin(user_address, kernel_address, nbytes)
|***	copyiin(user_address, kernel_address, nbytes)
|***	copyout(kernel_address, user_address, nbytes)
|***	copyiout(kernel_address, user_address, nbytes)
_copyin:
_copyiin:
	movl	4(sp),a0		| user-address
	movl	12(sp),d0		| byte-count
	jbsr	umap			| map and verify user address
	jcs	cperr			| error: return
	movl	a0,4(sp)		| replace logical addr w/physical
	jra	_copybuf		| use copybuf to do the copy!

cperr:	rts

_copyout:
	movl	8(sp),a0		| user-address
	movl	12(sp),d0		| byte-count
	jbsr	udmap			| map and verify user address
	jra	cpyoutcom		| common copy(i)out code

_copyiout:
_copyiou:
	movl	8(sp),a0		| user-address
	movl	12(sp),d0		| byte-count
	jbsr	umap			| map and verify user address
cpyoutcom:
	jcs	cperr			| error: return
	movl	a0,8(sp)		| replace logical addr w/physical
	jra	_copybuf		| use copybuf to do the copy!


