/*
 *  "dbug.h" -  Macros for debug output that won't clutter up code with #ifdefs
 *
 *	#define DEBUG n  to compile in the debugging statements [n = debug lvl]
 *	#define DEBUGS m  to define which sections you wish debug output from
 *		`m' should be each section or'd together.  
 *		dbALL is predefined to mean all sections.
 *		This macro package can handle up to 15 seperate sections.
 *		Each section should be defined as a bit in a short word, but
 *		each dbug() call checks for any or all the bits.  The high bit
 *		(0x8000) is reserved, and is always set.  The section name
 *		dbANY is predefined and will always produce output, depending
 *		on the level.
 *	#define DbugVar to be what you want your debug variable to be called
 *	#define DbugSS (debug single-step) to pause after each debug line
 */

#ifdef DEBUG
#  define dbANY	0x8000
#  define dbALL	0x7FFF
   short   DbugVar[2] = { DEBUG, (dbANY|DEBUGS) };
#  ifndef DbugQual
#    define DbugQual
#  endif /*DbugQual*/
#  define DbugOK(s,l)	(DbugQual ((DbugVar[1] & (s)) && (DbugVar[0] >= l)))
#  define  dbPause()	do { register int x;  x = spl7(); getchar(); splx(x); } while (0)
#  ifdef DbugSS
#    define  dbSSpause()  dbPause()
#  else /*DbugSS*/
#    define  dbSSpause()
#  endif /*DbugSS*/
#  define  dbug(s,l, f)		do { if (DbugOK(s,l)) { printf(f); dbSSpause(); } } while (0)
#  define  dbug1(s,l, f, a)	do { if (DbugOK(s,l)) { printf(f, a); dbSSpause(); } } while (0)
#  define  dbug2(s,l, f, a,b)	do { if (DbugOK(s,l)) { printf(f, a,b); dbSSpause(); } } while (0)
#  define  dbug3(s,l, f, a,b,c)	do { if (DbugOK(s,l)) { printf(f, a,b,c); dbSSpause(); } } while (0)
#  define  dbug4(s,l, f, a,b,c,d)  do { if (DbugOK(s,l)) { printf(f, a,b,c,d); dbSSpause(); } } while (0)
#  define  dbug5(s,l, f, a,b,c,d,e)  do { if (DbugOK(s,l)) { printf(f, a,b,c,d,e); dbSSpause(); } } while (0)
#  define  IFdbug(x)	x
#  define  ICEbrk()	asm("	movw	d0,/780F00")
#else /*DEBUG*/
#  define  dbug(s,l, f)
#  define  dbug1(s,l, f, a)
#  define  dbug2(s,l, f, a,b)
#  define  dbug3(s,l, f, a,b,c)
#  define  dbug4(s,l, f, a,b,c,d)
#  define  dbug5(s,l, f, a,b,c,d,e)
#  define  IFdbug(x)
#  define  ICEbrk()
#endif /*DEBUG*/
