/*
 *	%Z% %M% %I% %D% %Q%
 */




/*
 * THIS FILE CONTAINS CODE WHICH IS SPECIFIC TO THE
 * TRS-80 Model 16 COMPUTER. IT MAY REQUIRE CHANGING WHEN
 * MOVING XENIX TO DIFFERENT MACHINE ARCHITECTURES
 * AND CONFIGURATIONS.
 */



/*
 * The user structure.
 * One allocated per process.
 * Contains all per process data that doesn't need to be referenced
 * while the process is swapped.
 * The user block is USIZE*click bytes long;
 * contains the system stack per user; is cross referenced
 * with the proc structure for the same process.
 */

#ifndef X_MAGIC
#  include "a.out.h"
#endif
 
struct	user
{                               /* octal offset */
	label_t u_rsav;             /* 0  save info when exchanging stacks */
	label_t u_qsav;             /* 64 label variable for quits and interrupts */
	label_t u_ssav;             /*150 label variable for swapping */
	char    u_segflg;           /*234 IO flag: 0:user D; 1:system; 2:user I */
	char    u_error;            /*235 return error code */
	ushort  u_uid;              /*236 effective user id */
	ushort  u_gid;              /*240 effective group id */
	ushort  u_ruid;             /*242 real user id */
	ushort  u_rgid;             /*244 real group id */
	struct proc *u_procp;       /*246 pointer to proc structure */
	int     *u_ap;              /*252 pointer to arglist */
	union {                     /*256 syscall return values */
		struct	{
			int     r_val1;  /* 256*/
			int     r_val2;  /* 262*/
		}r_reg;
		long	r_long;
		off_t	r_off;
		time_t	r_time;
	} u_r;
	caddr_t u_base;             /*266 base address for IO */
	unsigned u_count;           /*272 bytes remaining for IO */
	off_t   u_offset;           /*276 offset in file for IO */
	short   u_fmode;            /*302 file mode for IO */
	short   u_errcnt;           /*304 syscall error count */
	struct inode *u_cdir;       /*306 pointer to inode of current directory */
	struct inode *u_rdir;       /*312 root directory of current process */
	caddr_t u_dirp;             /*316 pathname pointer */
	struct direct u_dent;       /*322 current directory entry */
	struct inode *u_pdir;       /*342 inode of parent directory of dirp */
	struct file *u_ofile[NOFILE];/*346 pointers to file structures of open files */
	char    u_pofile[NOFILE];   /*352 per-process flags of open files */
	int     u_arg[5];           /*372 arguments to current system call */
	unsigned u_tsize;           /*416 text size (clicks) */
	unsigned u_dsize;           /*422 data size (clicks) */
	unsigned u_ssize;           /*426 stack size (clicks) */
	int     u_signal[NSIG];     /*316 disposition of signals */
	time_t  u_utime;            /*366 this process user time */
	time_t  u_stime;            /*372 this process system time */
	time_t  u_cutime;           /*376 sum of childs' utimes */
	time_t  u_cstime;           /*402 sum of childs' stimes */
	int     *u_ar0;             /*406 address of users saved R0 */
	struct {                    /*410 profile arguments */
		short   *pr_base;   /* buffer base */
		unsigned pr_size;   /* buffer size */
		unsigned pr_off;    /* pc offset */
		unsigned pr_scale;  /* pc scaling */
	} u_prof;
	char    u_intflg;           /*420 catch intr from sys */
	char    u_sep;              /*421 flag for I and D separation */
	struct tty *u_ttyp;         /*422 controlling tty pointer */
	dev_t   u_ttyd;             /*424 controlling tty dev */
	struct xexec u_exdata;      /*426 header of executable file */
	struct xext     u_exext;    /*466 extension to header */
	unsigned        u_hdrsiz;   /*512 total size of header on file */
	char    u_comm[DIRSIZ];     /*514*/
	time_t  u_start;            /*532*/
	time_t  u_ticks;            /*536*/
	long    u_mem;              /*542*/
	long    u_ior;              /*546*/
	long    u_iow;              /*550*/
	long    u_iosw;             /*554*/
	long    u_ioch;             /*560*/
	char    u_acflag;           /*564*/
	short   u_cmask;            /*566 mask for file creation */
	daddr_t u_limit;            /*570 maximum write address */
	int     u_t[32];            /*574 **super dirty** (see iupdat) */
	int     u_sdsaved;          /* we have saved the shared data */
	int     u_sdcnt;            /* we have attached to shared data */
	struct sd {                 /* per proc shared data table */
	    struct inode *sd_inode; /* pointer to inode for segment */
	    char    *sd_addr;       /* address in this procs data space */
	    int     sd_vnum;        /* version # for local copy */
	    char    sd_flags;       /* describing state of this proc */
	} u_sdtab[NSDSEGS];
/* machine dependent fields */
	int     u_fper;             /*674 FP error register */
	char    u_fpsaved;          /*676 FP regs saved for this proc */
	char    u_fpused;           /*677 FP used for this process    */
	struct {                    /*700 8087 save status */
		int dummy;          /*700 to be added */
	} u_fps;
	unsigned short  u_lxrw;     /*702 for mmumap */

	caddr_t	u_addr;		    /* address of UPAGE */
	caddr_t	u_stktop;	    /* End of stack */
	int     u_stack[1];         /*706*/
					/* kernel stack per user
					 * extends from u + USIZE*64
					 * backward not to reach here
					 */
};
extern struct user u;

#define	u_rval1	u_r.r_reg.r_val1
#define	u_rval2	u_r.r_reg.r_val2
#define	u_roff	u_r.r_off
#define	u_rtime	u_r.r_time
#define u_rlong u_r.r_long

/* u_error codes */
#define	EPERM	1
#define	ENOENT	2
#define	ESRCH	3
#define	EINTR	4
#define	EIO	5
#define	ENXIO	6
#define	E2BIG	7
#define	ENOEXEC	8
#define	EBADF	9
#define	ECHILD	10
#define	EAGAIN	11
#define	ENOMEM	12
#define	EACCES	13
#define	EFAULT	14
#define	ENOTBLK	15
#define	EBUSY	16
#define	EEXIST	17
#define	EXDEV	18
#define	ENODEV	19
#define	ENOTDIR	20
#define	EISDIR	21
#define	EINVAL	22
#define	ENFILE	23
#define	EMFILE	24
#define	ENOTTY	25
#define	ETXTBSY	26
#define	EFBIG	27
#define	ENOSPC	28
#define	ESPIPE	29
#define	EROFS	30
#define	EMLINK	31
#define	EPIPE	32
#define	EDOM	33
#define	ERANGE	34
#define	EUCLEAN	35
#define EDEADLOCK 36
#define ENOTNAM 37
#define ENAVAIL 38
#define EISNAM  39

/* ioflag values: Read/Write, User/Kernel, Ins/Data */
#define	U_WUD	0
#define	U_RUD	1
#define	U_WKD	2
#define	U_RKD	3
#define	U_WUI	4
#define	U_RUI	5

#define	EXCLOSE	01
