/*****************************************

z80attn.h -- Interface to z80 attention queue

Copyright (C) Tandy Corporation, 1983.

This file must be kept consistent with z80/ctl/68kdefs.h

******************************************

****/

/* Description of an entry in the attention queue */

struct atn_q_entry {
	unsigned char
		dev,	/* High 4 bits are the major dev, low = minor # */
		cmd;	/* Tells z80 what kind of request is being made */
};


/*********************************/
/* The attention queue structure */
/*********************************/

/*
 *  WARNING!	The fifo size MUST be a power of two, so the Z80
 *		and 68k code can do fast, easy modulo arithmetic
 */
#define REQFIFOSIZE 32

 /*
  *  The request queue has an array of entries and two event counts
  *  (for the z80's benefit).  rear is the item-has-been-added event.
  *  front is the item-has-been-removed event.  An event is signalled
  *  simply by incrementing the event counters.
  *
  *  Note:  The rear  pointer for the queue =   rear % REQFIFOSIZE
  *         the front pointer for the queue =  front % REQFIFOSIZE
  */

struct req_fifo {
	unsigned char
		rear,		/* item-has-been-added event counter */
		front;		/* item-has-been-removed event counter */
	struct atn_q_entry
		q[REQFIFOSIZE];	/* Array containing queue entries themselves */
};


/* Device definitions for req_fifo.q[].dev */
#define	FDdev	(unsigned char)0x00	/* Request for the Floppy Disk */
#define	HDdev	(unsigned char)0x10	/* Request for the Hard Disk */
#define	SERdev	(unsigned char)0x20	/* Request for Serial I/O */
#define	SCSIdev	(unsigned char)0x30	/* Request for the IOmega */
#define	GRdev	(unsigned char)0x40	/* Graphics I/O request */
#define	CPUdev	(unsigned char)0x50	/* HALT/REBOOT request */

/* Request definitions for req_fifo.q[].cmd */
#define	RqStart		(unsigned char)0x00	/* I/O request */
#define	RqSTwrite	(unsigned char)0x01	/* Status change request */
#define	RqSTread	(unsigned char)0x02	/* Status update request */
#define	RqOVFLOW	(unsigned char)0x80	/* Request queue overflowed */

/* The attention queue itself */
struct req_fifo	z80req_fifo;
