/*
 *	TRS-16 line-printer driver
 *	Hacked from TRS-16 serial driver and PDP-11 LP driver.
 *	February 28, 1983
 */

/*
 * Modifications
 *	M000	26 Dec 82	JJD
 *		- added msgbuf support
 *	M001	07 Jan 83	JJD
 *		- line printer hacks
 *		- fixes to line printer
 *		- form-feed on close of line printer should be processed
 *		  by lpoutput
 *		- form-feed on close if not col 0, even if line 0
 *	M002	14 Jan 83	JJD
 *		- only super-user can change printer parameters
 *	M003	[was a mod to serial driver]
 *	M004	21 Jan 83	JJD
 *		- new line printer bit, suppresses auto form-feed on
 *		  close, also remembers current line number across opens
 *		  For lack of a better name, this is the weird bit.
 *		- miscellaneous changes to save text
 *	M005	12 May 83	jeIV  (Tandy)
 *		- kludged accesses from z80lp[] back to z80ser[3]  [ *sigh* ]
 *		  WARNING!!!  If you change NZ80LP > 1,  you will
 *			      have to move the lpr iopb's and put
 *			      them after the serial iopbs and put
 *			      the array z80lp[] back in!!!!!!!!!!
 *	M006	24 Jun 83	jeIV  (Tandy)
 *		- added calls to z80attn() to notify z80 of changes to mboxes
 *	M007	20 Sept 83	jeIV  (Tandy)
 *		- added a check to lpintr() to see 
 *		  if interrupt was really for us
 *	M008	29 Nov 83	jeIV  (Tandy)
 *		- Anyone may change the printer parms if they have it open
 *		  unless compiled with conditional ONLYROOT defined
 *
 * $Log:	lp.c,v $
 * Revision 6.0  87/10/12  17:04:47  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.2  86/08/29  15:08:30  iv
 * 	- Changed debugging statements over to new dbug.h macro package
 * 
 * Revision 5.1  85/12/04  10:03:44  iv
 * 	- Added debugging messages
 * 	- Put interrupt fix int (it was possible for the 68000 code to see
 * 	  TXINT on, and set up the next I/O *before* the Z80 had reset the
 * 	  GO bit, and then output would die, because GO got reset.  *sigh*)
 * 
 * Revision 5.0  85/08/23  10:30:25  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.1  85/03/29  10:23:15  iv
 * 	- Removed Z80 interrupt array code; now use iopb & TXINT bit
 * 
 * Revision 4.0  85/03/13  14:43:44  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:48:14  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:57:47  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.5  84/12/19  10:30:16  iv
 * 	- Put serial channel kludge back in.  *sigh*  z80ctl can't cope,
 * 	  yet, with pulling out the lpr iopb from the serial array.  *sigh*
 * 
 * Revision 1.4  84/07/19  15:55:39  iv
 * 	- Structure of a CBLOCK changed, make appropriate data-grabbing
 * 	  changes where we are handing z80 phys addr of output list.
 * 
 * Revision 1.3  84/04/16  17:00:32  iv
 * 	- Converted misc defines and types from v7 to sys3 
 * 
 * Revision 1.2  84/04/12  17:01:41  iv
 * 	- Removed lpr iopb from serial array.
 * 	  This also affects ../h/z80ser.h, which has the array z80lp[] again.
 * 
 * Revision 1.1  84/04/11  14:00:54  iv
 * Initial checkin of Tandy parallel printer driver for TRS-XENIX 2.0.
 * Hacked from original Tandy serial drivers.
 *
 * 
 * Revision 2.0  84/03/23  17:43:13  iv
 * Checkin for v2.0  (thus locking in TRS-XENIX 1.3.2 at v1.x)
 * 
 * Revision 1.3  84/03/23  17:18:41  iv
 * Changed interrupt handling to new system with byte array
 * 
 * Revision 1.2  84/03/23  13:57:51  iv       
 * Checkin for TRS-XENIX 1.3.2 source.  Many hacks and changes
 * 
 */

/*#define DEBUG	7		/* Enable lp debugging messages */
#define DEBUGS dbALL		/* Output from all debug sections */
#define	DbugVar	lp_debug

#include "../h/param.h"
#include "../h/systm.h"		/*M000*/
#include "../h/conf.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/tty.h"
#include "../h/z80ser.h"
#include "../h/z80attn.h"	/*M006*/
#include "../h/dbug.h"

#define	SPL	spl4
#define	LPPRI	(PZERO+8)
#define	LPLOWAT	CLSIZE
#define	LPHIWAT	(2*CLSIZE)

#define	lineno(dev)	((dev) & 0x1f)
#define	isprint(dev)	(((dev) & 0x80) != 0)	/* special printer processing */
#define	hasmemory(dev)	(((dev) & 0x40) != 0)	/* remember line & col across */
						/*	open/close */
#define	noblank(dev)	(((dev) & 0x20) != 0)	/* form-feeds do nothing at   */
						/*	top of page */

struct zs_iopb *z80lp = &z80ser[3];

struct lp {
	struct clist l_outq;		/* output queue */
	unsigned short l_state;		/* state */
	caddr_t	l_addr;			/* device address */
	struct lpiocb {
	    unsigned char ioc_lflag;	/* flags: minor(dev) */
	    unsigned char ioc_ind;	/* indent -- unused */
	    unsigned short ioc_col;	/* nr of columns per page -- unused */
	    unsigned short ioc_line;	/* number of lines per page */
	} iocb;
	unsigned short line;		/* current line number on printer */
	unsigned short col;		/* current column */
} lp[NZ80LP];

/*
 * Open a line printer
 */
/*ARGSUSED*/
lpopen(dev, flag)
register dev_t dev;					/*M004*/
int  flag;
{
	register d;
	register struct lp *lpp;

	static unsigned char once = 0;

	dbug1(dbANY,1,"[open(%d)]", minor(dev));
	if (once == 0) {
		once++;
		lpp = &lp[0];
#if NZ80LP != 1
		while (lpp < &lp[NZ80LP]) {
#endif
			lpp->iocb.ioc_col = 132;
			lpp->iocb.ioc_line = 66;
#if NZ80LP != 1
			lpp++;
		}
#endif
	}

	d = lineno(dev);
	if ((unsigned) d >= NZ80LP) {
		dbug(dbANY,1,"\nlpopen:  return (ENXIO)\n");
		u.u_error = ENXIO;
		return;
	}
#if NZ80LP != 1
	lpp = &lp[d];
#else
	lpp = &lp[0];
#endif
	if (lpp->l_state&XCLUDE) {
		dbug(dbANY,1,"\nlpopen:  return (EBUSY)\n");
		u.u_error = EBUSY;
		return;
	}
	lpp->l_addr = (caddr_t) &z80lp[d];
	lpp->l_state |= XCLUDE|CARR_ON;
	if (isprint(dev) && !hasmemory(dev)) {
		lpp->line = 0;
		lpp->col = 0;
	}
	lpp->iocb.ioc_lflag = dev;
}

/*
 * Close a line-printer.
 */
/*ARGSUSED*/
lpclose(dev, flag)
register dev_t dev;
int  flag;
{
#if NZ80LP != 1
	register struct lp *lpp = &lp[lineno(dev)];
#else
	register struct lp *lpp = &lp[0];
#endif

	dbug1(dbANY,1,"[close(%d)]", minor(dev));
	if (isprint(dev) && !hasmemory(dev) && (lpp->line != 0 || lpp->col != 0))
		lpoutput('\f', lpp);
	lpp->l_state &= ~(XCLUDE|CARR_ON);
	lpstart(lpp);
}

/*
 * write on a line-printer
 */
lpwrite(dev)
{
	register struct lp *lpp;
	register int c, s;

#if NZ80LP != 1
	lpp = &lp[lineno(dev)];
#else
	lpp = &lp[0];
#endif

	while (u.u_count) {
		s = SPL();
		while (lpp->l_outq.c_cc > LPHIWAT) {
			lpstart(lpp);
			lpp->l_state |= OASLP;
			sleep((caddr_t)&lpp->l_outq, LPPRI);
		}
		splx(s);
		if ((c = cpass()) < 0)
			break;
		lpoutput(c, lpp);
	}
	lpstart(lpp);
}

/*
 * line-printer interrupt.
 */
lpintr()
{
	register struct lp *lpp;
	register struct zs_iopb *addr;
#if NZ80LP != 1
	register i;
#endif

	lpp = &lp[0];
	addr = &z80lp[0];

	if ((addr->zs_csr & (TXGO|TXINT)) == TXINT)
		addr->zs_csr &= ~TXINT;
	else
		return;

	dbug(dbANY,3,"[intr]");
	/*
	 * Transmitter handling.
	 * Restart each line which used to be active but has
	 * terminated transmission since the last interrupt.
	 */

#if NZ80LP != 1
	for (i = NZ80LP; i; i--) {
#endif
		if ((lpp->l_state&BUSY) && (addr->zs_csr&TXGO) == 0) {
			ndflush(&lpp->l_outq, (int) addr->zs_count);
			lpp->l_state &= ~BUSY;
			lpstart(lpp);
		}
#if NZ80LP != 1
		lpp++;
		addr++;
	}
#endif
}

#define LIOCGETP	(('l'<<8)+0)			/*M001*/
#define LIOCSETP	(('l'<<8)+1)			/*M001*/

/*
 * stty/gtty for line-printer.
 */
/*ARGSUSED*/
lpioctl(dev, cmd, addr, flag)
caddr_t addr;
{
#if NZ80LP != 1
	register struct lp *lpp = &lp[lineno(dev)];
#else
	register struct lp *lpp = &lp[0];
#endif

	dbug(dbANY,1,"[ioctl]");
	switch(cmd) {
		case LIOCGETP:
			if (copyout((caddr_t) &lpp->iocb, addr,
			    sizeof lpp->iocb))
				u.u_error = EFAULT;
			break;
		case LIOCSETP:
#ifdef ONLYROOT
			if (!suser())			/*M002*//*M008*/
				return;			/*M002*//*M008*/
#endif ONLYROOT
			if (copyin(addr, (caddr_t) &lpp->iocb,
			    sizeof lpp->iocb))
				u.u_error = EFAULT;
			else {
				lpp->line = 0;
				lpp->col = 0;
			}
			break;
		default:
			u.u_error = ENOTTY;
	}
}

/*
 * Start (restart) transmission on the given line-printer.
 */
lpstart(lpp)
register struct lp *lpp;
{
	register struct zs_iopb *addr;
	register nch;
	register int s;

	/*
	 * If it's currently active, or delaying,
	 * no need to do anything, except maybe clear stop bit.
	 */
	s = SPL();
	if (lpp->l_state&BUSY)
		goto out;

	addr = (struct zs_iopb *)lpp->l_addr;

	/*
	 * If the writer was sleeping on output overflow,
	 * wake him when low tide is reached.
	 */
	if (lpp->l_state&OASLP && lpp->l_outq.c_cc<=LPLOWAT) {
		lpp->l_state &= ~OASLP;
		wakeup((caddr_t)&lpp->l_outq);
	}

	if (lpp->l_outq.c_cc == 0)
		goto out;

	/*
	 * Find number of characters to transfer.
	 */
	nch = ndqb(&lpp->l_outq, 0);

	/*
	 * If any characters were set up, start transmission;
	 */
	if (nch) {
		register 	int	d;
		register struct cblock	*cbp;

		cbp = lpp->l_outq.c_cf;
		addr->zs_addr = (caddr_t)&cbp->c_data[cbp->c_first];
		addr->zs_count = nch;
		addr->zs_csr = TXIENAB | TXGO;

		dbug1(dbANY,2,"[start %d]", nch);
#if NZ80LP != 1
		d = addr - z80lp;				/*M006*/
		z80attn(LPdev, (unsigned char)d, RqStart);	/*M006*/
#else NZ80LP
		d = addr - z80ser;
		z80attn(SERdev, (unsigned char)d, RqStart);	/*M006*/
#endif NZ80LP
		lpp->l_state |= BUSY;
	}
    out:
	splx(s);
}


lpoutput(c, lpp)
register c;
register struct lp *lpp;
{
	register int i;

	if (!isprint(lpp->iocb.ioc_lflag))
		putc(c, &lpp->l_outq);
	else switch(c) {
		case '\t':
			i = (8 - (lpp->col & 7));
			c = ' ';
			lpp->col += i;
			while (i--)
				putc(c, &lpp->l_outq);
			break;
		case '\n':
		case '\r':
			c = (lpp->col ? '\r' : '\n');
			if (++lpp->line >= lpp->iocb.ioc_line)
				lpp->line -= lpp->iocb.ioc_line;
			lpp->col = 0;
			putc(c, &lpp->l_outq);
			break;
		case '\f':
			if (lpp->col == 0 && lpp->line == 0 &&
				noblank(lpp->iocb.ioc_lflag))
					return;
			i = (lpp->iocb.ioc_line - lpp->line);
			c = (lpp->col ? '\r' : '\n');
			lpp->col = 0;
			lpp->line = 0;
			putc(c, &lpp->l_outq);
			c = '\n';
			while (--i)
				putc(c, &lpp->l_outq);
			break;
		case '\b':
			if (lpp->col)
				lpp->col--;
			putc(c, &lpp->l_outq);
			break;
		default:
			lpp->col++;
			putc(c, &lpp->l_outq);
			break;
	}
}
