/*
 *	"debug.c" -  MC60000 debug module for XENIX
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: debug.c,v 6.0 87/10/12 17:07:12 iv Locked $
 *
 * $Log:	debug.c,v $
 * Revision 6.0  87/10/12  17:07:12  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.3  87/10/12  16:36:59  iv
 * 	- Changed "unsigned short" to "ushort" everywhere
 * 	- Changed "trace" support to <trace> <n> for stepping <n> instructions
 * 	- Prettied up misc messages
 * 	- Added inode table dump
 * 	- Added file table dump
 * 	- Added open-file table dump (from U)
 * 	- Added new support subrs  getint() and getkey()
 * 	- Added new <c>hange <r>egister function
 * 	- Added new support routine getupage()
 * 	- Changed apropriate routines to use new getupage()
 * 
 * Revision 5.2  86/11/12  11:00:50  iv
 * 	- Some minor cosmetic changes
 * 	- Added a few fields to the Upage dump
 * 
 * Revision 5.1  86/08/28  17:31:40  iv
 * 	-  Changed MMU register dump routine to access new, longer, MMU regs
 * 
 * Revision 5.0  85/12/03  17:08:46  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:44:58  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.1  85/01/22  14:05:43  iv
 * 	- Added code to doCtrace() to check to make sure that
 * 	  the process is swapped IN
 * 	- Added code to doCtrace() to 'POP' stuff off the stack
 * 	- Fixed the bit instruction disassembly 
 * 	  (they were clobbering an argument declaration & disassembling
 * 	   trash off of the stack)
 * 	- Made misc formatting changes in the disassembler
 * 
 * Revision 3.0  85/01/17  15:46:10  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.1  85/01/14  18:26:04  iv
 * 	- Got rid of debugging statement in doCtrace()
 * 
 * Revision 2.0  85/01/10  13:49:54  iv
 * 	- Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.2  84/04/20  17:46:35  iv
 * 	- Removed some Lisa specific code from debugger
 * 
 * Revision 1.1  84/04/10  17:35:30  iv
 *	- Checkin for initial version of TRS-XENIX 2.0
 */


#include "../h/param.h"
#include "../h/mmu.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/reg.h"
#include "../h/proc.h"
#include "../h/mount.h"
#include "../h/var.h"
#include "../h/buf.h"
#include "../h/iobuf.h"
#include "../h/file.h"
#include "../h/inode.h"

/*
 *      return 1 if valid brk pt (RETURN TO PROGRAM (pc-2))
 *      return 0 otherwise
 */
#define NBK     10              /* max number of brk pts */
struct {
	ushort *badr;              /* address */
	ushort oldinst;            /* old instruction */
} brks[NBK];

/*
 *  A *symbol table* !!   Hope it works...
 */
#include "../conf/debug_sym.h"

#define	NOFF	3
int	offset[NOFF];		/* Offset registers for address calcs */

char	gethex(), getint(), getkey();
int	supertrace = 0;		/* trace supervisor mode execution */
int	strflag = 0;
int	*ss;			/* Supervisor register save area pointer */

#define  ESC	'\033'


char Asrtstr[] = "Invalid assertion line %d of %s :%s\n";


debugsub(addr)
  ushort *addr;
{
	register ushort i, *ip;
	register char c;
	int tmp;
	int ps;

	ps = spl7();

	if (!strflag && addr != (ushort *)0) {
		ip = addr - 1;
		for (i=0; i<NBK; i++) if (brks[i].badr == ip) break;
		if (i < NBK) {
			brks[i].badr = 0;
			bugput(ip, brks[i].oldinst);
			brks[i].oldinst = 0;
		}
	}

	if (strflag && supertrace) {
		strflag = 0;
		ip = (ushort *)ss[UPC];
		if (--supertrace > 0) {
			printf("STEP  ");
			disasm(ip);	/* Show what we're about to step */

			splx(ps);
			clkstart();
			return (1);
		}
		ss[SR] &= ~SR_TRACE;
	}

	if (addr != (ushort *)0) {
		printf("\nBRK at  ");
		disasm(ip);
	} else
		printf("\nKernel DEBUG\n");

	for ( ; ; ) {
		printf("\ndbug> ");
		c = dgetc();
		switch (c) {
			default:
				printf(" ??   [Type ? for help]\n");
				c = '\n';
				break;
			case '\n':
				break;
			case 't':
				printf("race ");
				c = dotrace(ip);
				if (!supertrace)
					break;
				/* Fall through */
			case 'r':
				if (c == 'r')
					printf("un\n");
				splx(ps);
				clkstart();
				return (1);
			case 'R':
				printf("un AFTER\7\7\7\n");
				splx(ps);
				clkstart();
				return (0);
			case 'b':
				printf("rk pt = ");
				c = getbrk();
				break;
			case 'd':
				printf("ump ");
				c = dodump();
				break;
			case 'D':
				printf("isassemble ");
				c = dodisasm();
				break;
			case 'p':
				printf("rint brk pts\n");
				c = doprt();
				break;
			case 'c':
				printf("hange ");
				c = dochg();
				break;
			case 'u':
				printf("n-brk pt\n");
				c = doun();
				break;
			case 's':
				printf("how User registers\n");
				c = dostate(u.u_ar0);
				break;
			case 'S':
				printf("how Supervisor registers\n");
				c = dostate(ss);
				break;
			case 'P':
				printf("roc table slot = ");
				c = doproc();
				break;
			case 'U':
				printf("page dump\n");
				c = doupage();
				break;
			case 'm':
				printf("mu register dump\n");
				c = dommu();
				break;
#ifdef NEVER
			case 'e':
				printf("vent trace\n");
				c = doevent();
				break;
#endif /*NEVER*/
			case 'h':
				printf("ex search value (word) = ");
				if((c=gethex(&tmp)) == ESC) break;
				c = dohex(tmp);
				break;
			case 'x':
				printf("-sum memory region\n");
				c = dosum();
				break;
			case 'l':
				printf("ook up what symbol value? ");
				c = dolook();
				break;
			case 'C':
				printf(" stack trace\n");
				c = doCtrace();
				break;
			case '?':
				putchar('\n');
				printf("\tr)un\tR)un AFTER\n");
				printf("\tb)reak pt\tu)n-break pt");
				printf("\tp)rint break pts\n");
				printf("\ts)how user registers");
				printf("\tS)how supervisor registers\n");
				printf("\tm)mu register dump\n");
				printf("\tt)race (single-step)\n");
				printf("\tl)ook up a symbol name\n");
			/*	printf("\te)vent trace display\n");	*/
				printf("\th)ex value search");
				printf("\tx)-sum memory region\n");
				printf("\td)ump something");
				printf("\tc)hange something\n");
				printf("\tD)isassemble memory\n");
				printf("\tP)roc slot display\n");
				printf("\tU)page dump\n");
				printf("\tC) stack trace\n");
				c = '\n';
				break;
		}
		if(c==ESC) {
			putchar('\n');
			c='\n';
		}
		if(c != '\n')
			while(dgetc() != '\n');
	}
}

getbrk()
{
	register int i, j;
	register ushort *ip;
	int tmp;

	for(i=0;i<NBK;i++) if(brks[i].badr == (ushort *)0) break;
	if(i>=NBK) {
		printf("brk table full\n");
		return ('\n');
	}
	j = gethex(&tmp);
	if (j == ESC)
		return (j);

	if (tmp == 0 || tmp & 1) {
		if (j != '\n')
			putchar('\n');
		printf("Can't brk 0 or odd addr.\n");
		return ('\n');
	}
	brks[i].badr = (ushort *)tmp;
	ip = brks[i].badr;
	brks[i].oldinst = buget(ip);
	bugput(ip,0x4e4f);	/* TRAP 15 - BREAKpoint */
	return (j);
}

char fixchr(c)
unsigned c;
{
	if (c < ' ' || c >= 0177)
		return('.');
	return((char) c);
}

dodump()
{
	char	c;

	c = getkey("mdfio?\033");
	if (c == ESC)
		return (c);
	
	if (c == 'm') {
		printf("memory ");
		return (domemdump());
	}
	if (c == 'd') {
		printf("disk structs\n");
		return (dodkdump());
	}
	if (c == 'f') {
		printf("file tbl slot=");
		return (doftdump());
	}
	if (c == 'i') {
		printf("inode tbl slot=");
		return (doitdump());
	}
	if (c == 'o') {
		putchar(c);
		c = getkey("fp\033");
		if (c == ESC)
			return (c);
		if (c == 'p') {
			printf("pen file tbl\n");
			return (doOfdump());
		}
		printf("ffset registers\n");
		return (doshowoffset());
	}

	printf("?\nDump:\n");
	printf("  d)isk structs\n");
	printf("  f)ile tbl\n");
	printf("  i)node tbl\n");
	printf("  of)fset registers\n");
	printf("  op)en file tbl\n");
	return ('\n');
}

domemdump()
{
	register ushort	*ip, *jp, k;
		 unsigned int	i, j;

	j = 0;
	k = gethex(&i);
	if (k == ESC)
		return (k);
	if (k != '\n')
		k = gethex(&j);
	if (k != '\n')
		putchar('\n');

	if (i & 1) {
		printf("WARNING: Odd address %X", i);
		i &= ~1;
		printf(" rounded to %X\n", i);
	}
	if (j <= i)
		j = i+2;
	if (j - i > 0x4000) {
		printf("about to dump %X bytes, continue? (y/n)?", j - i);
		if (dgetc() != 'y') {
			goto out;
		}
	}
	ip = (ushort *)i;
	for(;;) {
		while(ip < (ushort *)j) {
			putchar('\r');
			printf("%X  ",ip);
			jp = ip;
			if(buget(ip)== -1)	printf("* ");
			else			printf("  ");
			for(i=0;i<8;i++) printf("%x ",buget(ip++));
			for(i=0;i<8;i++) {
				k = buget(jp++);
				putchar(fixchr((k>>8)&0x00FF));
				putchar(fixchr(k&0x00FF));
			}
			printf("\n");
		}
		if (dgetc() == ESC) break;
		j = (int)(ip + 1);
	}
out:
	return('\n');
}

char gethex(x)
int *x;
{
	register int i;
	register char c;

restart:
	i = 0;
	for ( ; ; ) {
		if ((c = getchar()) == ESC) {
			break;
		}
		if (c == 8) {			/* backspace */
			putchar(c);
			putchar(' ');
			putchar(c);
			i >>= 4;
			continue;
		}
		if (c >= '0' && c <= '9') {
			putchar(c);
			i = (i << 4) | (c - '0');
		} else if (c >= 'a' && c <= 'f') {
			putchar(c);
			i = (i << 4) | (c - 'a' + 10);
		} else if (c >= 'A' && c <= 'F') {
			putchar(c);
			i = (i << 4) | (c - 'A' + 10);
		} else if (c == '@') {
			putchar(c);
			c = dgetc();
			if (c == 8) {
				putchar(' ');
				putchar(c);
				continue;
			}
			putchar('\n');
			if (c < '0' || c > ('0' + NOFF) ) {
				printf("Bad offset register");
				if (c >= ' ' && c <= '~') {
					putchar(' ');
					putchar(c);
				}
				printf(".  Redo.\n");
				c = ESC;
				break;
			}
			i += offset[c - '0'];
			c = '\n';
			break;
		} else if (c == '/') {
			char	symbol[15];
			register char	*symptr = symbol;

			putchar(c);
			do {
				c = getchar();
				if (c == '\r')  c = '\n';

				if (c == 8) {
					printf("\b \b");
					if (symptr == symbol)
						break;
					if (symptr > symbol)
						--symptr;
				}
				else if (c != '\n' && (c < ' ' || c > '~') )
					putchar('\007');
				else if (c != '\n') {
					*symptr++ = c;
					putchar(c);
				}
			} while (c != '\n');

			if (c == '\b')
				continue;

			*symptr = '\0';
			for (i = 0; dsymtab[i].name; ++i)
				if (strcmp(dsymtab[i].name, symbol) == 0)
					break;
			
			if (!dsymtab[i].name) {
				printf("   Symbol not found.\n");
				c = ESC;
				break;
			}

			i = dsymtab[i].addr;
			printf(" = 0x%X\n", i);
			c = '\n';
			break;
		} else if (c == ':') {
			int	*rp;

			putchar(c);
			c = getregp(&rp);
			if (c == '\b')
				printf("\b \b");
			if (c == ESC || c == '\b')
				break;
			i = *rp;
			printf(" = 0x%X\n", i);
			c = '\n';
			break;
		}
		else {
			if (c == ' ' || c == '\t' || c == '\n') {
				putchar(c);
				break;
			} else
				putchar('\007');
		}
	}
	*x = i;
	return (c);
}

char getint(x)
int *x;
{
	register int i;
	register char c;

	i = 0;
	for ( ; ; ) {
		if ((c = getchar()) == ESC) {
			break;
		}
		if (c == 8) {			/* backspace */
			putchar(c);
			putchar(' ');
			putchar(c);
			i /= 10;
			continue;
		}
		if (c >= '0' && c <= '9') {
			putchar(c);
			i = (i * 10) + (c - '0');
		}
		else {
			if (c == ' ' || c == '\t' || c == '\n') {
				putchar(c);
				break;
			} else
				putchar('\007');
		}
	}
	*x = i;
	return (c);
}

char
getkey(list)
  char	*list;
{
	register char	*sp;
	char	c;

	for ( ; ; ) {
		c = getchar();
		if (c >= 'A' && c <= 'Z')  c |= 0x20;	/* tolower() */
		for (sp = list; *sp; ++sp) {
			if (c == *sp)
				return (c);
		}
	}
}

dgetc()
{
	register char	c;

	c = getchar();
	if (c=='\r') c='\n';
	putchar(c);
	return (c);
}

doprt()
{
	register int i;

	for (i=0;i<NBK;i++) if (brks[i].badr)
		printf("%X %x\n", brks[i].badr, brks[i].oldinst);
	return ('\n');
}

doun()
{
	register ushort *ip;
	register i,c;
	int val;

	for(i=0;i<NBK;i++) if(brks[i].badr) break;
	if(i==NBK) {
		printf("no brk pts\n");
		return('\n');
	}
	printf("Address? ");
	for(;;) {
		c = gethex(&val);
		if (c==ESC) break;
		for(i=0;i<NBK;i++) if(brks[i].badr == (ushort *)val) {
			ip = brks[i].badr;
			*ip = brks[i].oldinst;
			brks[i].badr = 0;
		}
		if (c=='\n') break;
	}
	printf("\nNew breakpoint list:\n");
	doprt();
	return ('\n');
}

dochg()
{
	char	c;

	c = getkey("rmo?\033");
	if (c == ESC)
		return (c);
	if (c == 'm') {
		printf("memory ");
		return (domemchg());
	}
	if (c == 'o') {
		printf("offset register #? ");
		return (dochgoffset());
	}
	if (c == 'r') {
		printf("register ");
		return (doregchg());
	}

	printf("?\nChange:\n");
	printf("  m)emory\n");
	printf("  o)ffset registers\n");
	printf("  r)egisters\n");
	return ('\n');
}

domemchg()
{
	register ushort *ip, *jp, kv;
	ushort	i;
	int	j, k;
	char	c;

	c = gethex(&j);
	if (c == ESC)
		return (c);
	if (j & 1) {
		printf("WARNING: Odd address %X", j);
		j &= ~1;
		printf(" rounded to %X\n", j);
	}
	for(;;) {
		ip = (ushort *)j;
		if(c != '\n') putchar('\n');
		putchar('\n');
		for (k=0;k<8;k++)  for (;;) {
			ip = (ushort *)j;
			jp = ip;
			putchar('\r');
			printf("%X  ",ip);
			if(buget(ip)== -1)	printf("* ");
			else			printf("  ");
			jp = ip;
			for(i=0;i<8;i++) printf("%x ",buget(ip++));
			for(i=0;i<8;i++) {
				kv = buget(jp++);
				putchar(fixchr((kv>>8)&0x00FF));
				putchar(fixchr(kv&0x00FF));
			}
			ip = (ushort *)j;
			putchar('\r');
			printf("%X  ",ip);
			if(buget(ip)== -1)	printf("* ");
			else			printf("  ");
			for(i=0;i<=k;i++)
				printf("    \b\b\b\b%x ",buget(ip++));
			printf("\b\b");
			i = buget(--ip);
			c = dgetc();
			if(c>='0' && c<='9') i = (i<<4) | (c - '0');
			else if(c>='a' && c<='f') i = (i<<4) | (c - 'a' + 10);
			else if(c>='A' && c<='F') i = (i<<4) | (c - 'a' + 10);
			else if(c=='\n' || c=='\33') goto out;
			else if(c==8) {
				k -= 1;
				if(k<0) k = 0;
			} else break;
			if(bugput(ip,i) == -1) printf("\7");
		}
	    out:
		j += 16;
		if(c=='\33') break;
	}
	if(c=='\33') printf("\n");
	return('\n');
}

getregp(ipp)
  int	**ipp;
{
	char	c, d;

	c = getkey("adpsu\033\b");

	if (c == 'a' || c == 'd') {
		putchar(c);
		do {
			d = getchar();
		} while (d < '0' || d > '7');
		putchar(d);

		d -= '0';
		if (c == 'd')
			*ipp = &ss[D0] + d;
		else
			*ipp = &ss[A0] + d;
	}
	else if (c == 'p') {
		printf("pc");
		*ipp = &ss[UPC];
	}
	else if (c == 's') {
		printf("sr");
		*ipp = &ss[SR];
	}
	else if (c == 'u') {
		printf("usp");
		*ipp = &ss[USP];
	}

	return (c);
}

doregchg()
{
	char	c;
	int	val;
	int	*cp;

	c = getregp(&cp);
	if (c == ESC)
		return (c);

	printf(" to? ");
	c = gethex(&val);
	if (c != ESC)
		*cp = val;
	return (c);
}

dostate(ar0)
int *ar0;
{
	extern int end;
	struct proc	*pp;

	pp = u.u_procp;
	printf("upage=%x &d0=%X sp=%X ",
		(pp >= proc && pp < v.ve_proc )? pp->p_addr.p_caddr : -1,
		ar0,
		ar0+19);
	printf("sr=%X pc=%X usp=%X\n",
		ar0[SR],
		ar0[UPC],
		ar0[USP]);
		
	printf("d0-d7 %X %X %X %X  %X %X %X %X\n",
		ar0[D0],
		ar0[D1],
		ar0[D2],
		ar0[D3],
		ar0[D4],
		ar0[D5],
		ar0[D6],
		ar0[D7]);
	printf("a0-a7 %X %X %X %X  %X %X %X %X\n",
		ar0[A0],
		ar0[A1],
		ar0[A2],
		ar0[A3],
		ar0[A4],
		ar0[A5],
		ar0[A6],
		ar0[A7]);
	return ('\n');
}

buget(adr)
ushort *adr;
{
	return (get_w(adr));
}

bugput(adr,val)
ushort *adr;
ushort val;
{
	return (put_w(adr, val));
}

doproc()
{
	register struct proc *pp;
	register int c;
	int slot;
	int newslot;

	if ((c = getint(&slot)) == ESC)
		return;

	for (;;) {
		printf("Proc[%d] %X\n", slot, pp = &proc[slot]);

		printf("stat flag pri  time cpu  nice sig  uid  pgrp pid  ppid addr size ssize\n");
		printf("%x %x %x %x %x %x %x %x %x %x %x %x %x %x\n",
			pp->p_stat, pp->p_flag, pp->p_pri, pp->p_time,
			pp->p_cpu, pp->p_nice, pp->p_sig, pp->p_uid,
			pp->p_pgrp, pp->p_pid, pp->p_ppid, pp->p_addr.p_caddr,
			pp->p_size, pp->p_ssize);
		printf("  pnum wchan    textp    link     clktim\n");
		printf("  %x %X %X %X %x\n",
			pp->p_pnum, pp->p_wchan, pp->p_textp, pp->p_link,
			pp->p_clktim);
		if ((c = getint(&newslot)) == ESC) {
			break;
		}
		if (newslot != 0 || c == '\n') {
			slot = newslot;		/* new slot was specified */
		} else {
			slot++;			/* else dump next slot */
		}
	}
	return ('\n');
}

char
getupage(upp)
  struct user	**upp;
{
	ushort	*caddr;
	struct proc	*addr;
	int	scr;
	char	c;

	printf("Proc slot? ");
	c = getint(&scr);
	if (c == ESC)
		return (c);
	if (c != '\n')
		putchar('\n');
	
	addr = &proc[scr];
	if (!(addr->p_stat & SLOAD)) {
		printf("Proc swapped out");
		return (ESC);
	}

	
	caddr = (ushort *)(addr->p_addr.p_caddr << LMMPGSZ);
	if (addr == u.u_procp)
		*upp = &u;
	else
		*upp = (struct user *)caddr;
	
	return ('\n');
}

doupage()
{
	struct user	*up;
	struct foo {
		char	fname[DIRSIZ + 2];
	} *fp1, *fp2;
	char	name[DIRSIZ + 2];
	char	c;

	c = getupage(&up);
	if (c == ESC)
		return (c);

	printf("segflg procp    #    base     count    offset   error\n");
	printf("  %x %X %x %X %X %X %X\n",
			up->u_segflg,
			up->u_procp,
			up->u_procp - &proc[0],
			up->u_base,
			up->u_count,
			up->u_offset,
			up->u_error);
	printf("limit    ap       ar0      rval1    rval2    &ofile[0] &signal[0]\n");
	printf("%X %X %X %X %X %X  %X\n",
			up->u_limit,
			up->u_ap,
			up->u_ar0,
			up->u_rval1,
			up->u_rval2,
			&up->u_ofile[0],
			&up->u_signal[0]);

	/* Copy accounting command name & make sure NULL terminated */
	fp1 = (struct foo *)name;
	fp2 = (struct foo *)up->u_comm;
	*fp1 = *fp2;
	name[DIRSIZ] = '\0';
	printf("tsize   dsize   ssize   name\n");
	printf(" %x    %x    %x   %s\n",
			up->u_tsize,
			up->u_dsize,
			up->u_ssize,
			up->u_comm);
	return ('\n');
}

#ifdef NEVER
doevent()
{
	extern	long	*tracep, traceb[], tracee[];
	register long	*p;
	char	*addr;
	int	count = 1;
	int	code;

	p = tracep;
	do {
		addr = (char *) (*p & 0xffffff);
		if (addr != 0) {
			if (count == 0) {
				switch(getchar()) {
					case '\33':
						return('\33');
					case ' ':
						putchar('\n');
						count = 21;
						break;
					default:
						count = 1;
						break;
				}
			}
			count--;
			switch(code = ((*p >> 24) & 0xff)) {
				case 0:
				case 1:
				case 5:
				case 6:
				case 7:
					printf("spl%o", code);
					break;
				case 0xc:
					printf("idle");
					break;
				case 0x1c:
					printf("serial");
					break;
				case 0x1d:
					printf("disk");
					break;
				case 0x1e:
					printf("clock");
					break;
				case 0xf0:
					printf("splx");
					break;
				default:
					printf("trap %d", code);
					break;
			}
			printf("\t%X\n",addr);
		}
		if (++p >= tracee)
			p = traceb;
	} while (p != tracep);
	return('\33');
}
#endif /*NEVER*/

dommu()
{
	extern  ushort	mmuLoLim, mmuLoOff, mmuHiLim, mmuHiOff;

	printf("mmu LO offset is 0x%X\n", mmuLoOff << LMMPGSZ);
	printf("mmu LO limit  is 0x%X\n", mmuLoLim << LMMPGSZ);
	printf("mmu HI offset is 0x%X\n", mmuHiOff << LMMPGSZ);
	printf("mmu HI limit  is 0x%X\n", mmuHiLim << LMMPGSZ);

	return('\n');
}

dotrace(ip)
  ushort  *ip;
{
	int	c;

	c = getint(&supertrace);
	if (c == ESC)
		return (c);
	else if (c != '\n')
		putchar('\n');
	if (supertrace < 1)
		supertrace = 1;

	ss[SR] |= SR_TRACE;
	printf("STEP  ");
	disasm(ip);
	return('\n');
}

dohex(val)
register int val;
{
	register ushort *p;
	register int i;

	if (val < 0 || val > 0xffff) {
		printf("value out of range\n");
		return('\n');
	}
	p = (ushort *) 0;
	for (;;) {
		i = get_w(p++);
		if (i < 0)
			break;
		if (i == val) {
			printf("%X: %x ", p - 1, val);
			if (dgetc() == ESC)
				break;
		}
	}
	return(ESC);
}

dosum()
{
	register int c;
	int first, last;
	register ushort *fp, *lp;
	register ushort checksum = 0;

	printf("starting addr = ");
	if((c = gethex(&first)) == ESC)
		return(c);
	printf("ending addr = ");
	if((c = gethex(&last)) == ESC)
		return(c);
	fp = (ushort *) first;
	lp = (ushort *) last;
	if (get_w(fp) == -1 || get_w(lp) == -1)
		printf("invalid address\n");
	else {
		while (fp < lp)
			checksum += *fp++;
		printf("checksum = %x\n", checksum);
	}
	return('\n');
}

doshowoffset()
{
	register int	i = 0;

	printf("Offset registers:\n");
	for (i = 0; i <= NOFF; ++i)
		printf("  [%d]  0x%X\n", i, offset[i]);

	return('\n');
}

dochgoffset()
{
	register char	c, i;
		 int	addr;

	c = dgetc();
	putchar('\n');
	if (c < '0' || c > ('0'+NOFF) ) {
		if (c < ' ' || c > '~')  
			c == ' ';
		printf("Bad offset register %c\n", c);
		return('\n');
	}

	i = c - '0';
	printf("Change to? ");
	if ((c = gethex(&addr)) == ESC)
		return (c);
	offset[i] = addr;
	return (c);
}

dolook()
{
	register int	i;
	unsigned int	addr;

	i = gethex(&addr);
	if (i == ESC)
		return (i);
	if (i != '\n')
		putchar('\n');

	for (i = 0; dsymtab[i].name; ++i)
		if (dsymtab[i].addr > addr)
			break;

	if (i > 0)  --i;
	printf("%s = 0x%X\n", dsymtab[i].name, dsymtab[i].addr);

	return ('\n');
}

dodkdump()
{
	register char	c, g;
	register struct	iobuf	*ioptr;
	register struct	buf	*bptr;
	unsigned int	addr;

	printf("Dump: i)obuf or b)uf? ");
	c = getkey("ib\033");
	if (c != ESC) putchar(c);
	putchar('\n');
	if (c == ESC)
		return (c);

	printf("==> Address? ");
	g = gethex(&addr);
	if (g == ESC)
		return (g);
	if (g != '\n')
		putchar('\n');
	if (addr & 1) {
		addr &= ~1;
		printf("WARNING:  Odd address rounded to 0x%X\n", addr);
	}

	if (c == 'i') {
		ioptr = (struct iobuf *)addr;
		printf("\nAddress = 0x%X;  dev = %d/%d\n",
			addr, major(ioptr->b_dev), minor(ioptr->b_dev));
		printf("b_flags= %s|%s|%s\n",
			(ioptr->b_flags&B_ONCE)?   "ONCE" : "",
			(ioptr->b_flags&B_TAPE)?   "TAPE" : "",
			(ioptr->b_flags&B_TIME)?   "TIME" : "");
		printf("b_active = %d;  b_errcnt = %d\n",
			ioptr->b_active, ioptr->b_errcnt);
		printf("io_addr = 0x%X;  io_s1 = %d;  io_s2 = %d\n",
			ioptr->io_addr, ioptr->io_s1, ioptr->io_s2);
		printf("b_forw => 0x%X;  b_back => 0x%X\n",
			ioptr->b_forw, ioptr->b_back);
		printf("b_actf => 0x%X;  b_actl => 0x%X\n",
			ioptr->b_actf, ioptr->b_actl);
	}
	else {
		bptr = (struct buf *)addr;
		printf("\nAddress = 0x%X;  dev = %d/%d\n",
			addr, major(bptr->b_dev), minor(bptr->b_dev));
		printf("b_flags= %s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s\n",
			(bptr->b_flags&B_READ)?   "READ" : "WRITE",
			(bptr->b_flags&B_DONE)?   "DONE" : "",
			(bptr->b_flags&B_ERROR)?  "ERROR" : "",
			(bptr->b_flags&B_BUSY)?   "BUSY" : "",
			(bptr->b_flags&B_PHYS)?   "PHYS" : "",
			(bptr->b_flags&B_MAP)?    "MAP"  : "",
			(bptr->b_flags&B_WANTED)? "WANTED" : "",
			(bptr->b_flags&B_AGE)?    "AGE"    : "",
			(bptr->b_flags&B_ASYNC)?  "ASYNC"  : "",
			(bptr->b_flags&B_DELWRI)? "DELWRI" : "",
			(bptr->b_flags&B_OPEN)?   "OPEN"  : "",
			(bptr->b_flags&B_STALE)?  "STALE" : "",
			(bptr->b_flags&B_UAREA)?  "UAREA" : "",
			(bptr->b_flags&B_FLUSH)?  "FLUSH" : "");
		printf("b_count = %u;  b_paddr = 0x%X;  b_blkno = %d\n",
			bptr->b_bcount, bptr->b_paddr, bptr->b_blkno);
		printf("b_cylin = %u;  b_error = %d;  b_resid = %u\n",
			bptr->b_cylin, bptr->b_error, bptr->b_resid);
		printf("b_forw => 0x%X;  b_back => 0x%X\n",
			bptr->b_forw, bptr->b_back);
		printf("b_actf => 0x%X;  b_actl => 0x%X\n",
			bptr->av_forw, bptr->av_back);
	}

	return ('\n');
}

doftdump()
{
	struct file	*fp;
	int	c, slot, new_slot;

	for ( ; ; ) {
		c = getint(&new_slot);
		if (c == ESC)
			break;
		if (c == '\n')
			slot = new_slot;
		else if (&file[++slot] >= v.ve_file) {
			c = '\n';
			break;
		}

		fp = &file[slot];
		if (fp < file || fp >= v.ve_file) {
			printf("\nBad file slot\n");
			c = '\n';
			break;
		}

		putchar('\r');
		dumpfp(fp);
	}

	return (c);
}

dumpfp(fp)
  register struct file	*fp;
{
	printf("[%d]: flag = %x  count = %d  inode = ", 
				fp - file,
				fp->f_flag,
				fp->f_count);
	if (fp->f_inode < inode || fp->f_inode >= v.ve_inode)
		printf("%X\n", fp->f_inode);
	else
		printf("&inode[%d]\n", fp->f_inode - inode);
}

doOfdump()
{
	register struct file	**fpp;
	struct user	*up;
	int		i;
	char		c;

	c = getupage(&up);
	if (c == ESC)
		return (c);
	
	for (fpp = up->u_ofile, i = 0; fpp < &up->u_ofile[NOFILE]; ++fpp, ++i) {
		if (*fpp != (struct file *)0) {
			printf("(%d) ==> ", i);
			dumpfp(*fpp);
		}
	}

	return ('\n');
}

struct cpystuff {
	char	stuff[8];
} sp = {
       /*0123456*/
	"       "
};
char *
domode(mode)
{
	static char	rval[8];
	register char	*rp;

	mode &= 0177777;
	*((struct cpystuff *)rval) = sp;
	rp = &rval[8];
	do {
		*--rp = "01234567"[mode & 7];
		mode >>= 3;
	} while (mode > 0);

	return (rval);
}

doitdump()
{
	struct inode	*ip;
	int		c, slot, new_slot, pass = 0;
	ushort		fmt;

	for ( ; ; ) {
		c = getint(&new_slot);
		if (c == ESC)
			break;
		if (c == '\n')
			slot = new_slot;
		else if (&inode[++slot] >= v.ve_inode) {
			c = '\n';
			break;
		}

		ip = &inode[slot];
		if (ip < inode || ip >= v.ve_inode) {
			printf("\nBad inode slot\n");
			break;
		}

		putchar('\r');
		if (c == '\n' || ++pass == 22) {
			pass = 0;
			printf("\tFlags  Cnt Device Inode   Mode  Nlink  UID  GID  Size/Dev\n");
		}

		printf("[%d]:\t%c%c%c%c%c%c%c %d   %x   %x%s  %x %x %x  ",
				slot,
				(ip->i_flag & ILOCK)? 'L' : ' ',
				(ip->i_flag & IUPD)?  'U' : ' ',
				(ip->i_flag & IACC)?  'A' : ' ',
				(ip->i_flag & IMOUNT)? 'M' : ' ',
				(ip->i_flag & IWANT)? 'W' : ' ',
				(ip->i_flag & ITEXT)? 'T' : ' ',
				(ip->i_flag & ICHG)?  'C' : ' ',
				ip->i_count,
				ip->i_dev,
				ip->i_number,
				domode(ip->i_mode),
				ip->i_nlink,
				ip->i_uid,
				ip->i_gid);
		fmt = (ip->i_mode & IFMT);
		if (fmt == IFBLK || fmt == IFCHR)
			printf("%d/%d\n", major(ip->i_rdev), minor(ip->i_rdev));
		else
			printf("%ld\n", ip->i_size);
	}

	return (c);
}

unsigned chkget();
extern long	maxaddr;
doCtrace()
{
	register ushort	*caddr, *frame, *nframe, *stack;
	register struct proc	*addr;
	register char	c;
	struct user	*up;
	int		running, scr;

	c = getupage(&up);
	if (c == ESC)
		return (c);
	
	running = (up == &u);
	if (running) {
		frame = (ushort *)ss[A6];
		stack = (ushort *)ss[19];
	}
	else {
		frame = (ushort *)up->u_rsav[10];
		stack = (ushort *)up->u_rsav[11];
	}


	printf("[Press <I> to trace the next frame as an interrupt frame,\n");
	printf(" or <R> to merely RETurn to the address @SP]\n");

	/* Show where we are now (where PC is) */
	printf("  --------:  %X %X --> ",
		running? frame : (frame - (ushort *)&u) + caddr,
		running? ss[UPC] : up->u_rsav[12]);
	if (!running)
		psymoff(up->u_rsav[12], 0, "\n");
	else
		psymoff(ss[UPC], 0, "\n");

	for ( ; ; ) {
		c = getchar();
		if (c == ESC)
			return (c);
		
		if (c >= 'A' || c <= 'Z')
			c |= 0x20;

		if (!running)
			frame = (frame - (ushort *)&u) + caddr;
		if ((int)frame > maxaddr || (int)frame & 1) {
			printf("**** Trace ERROR:  ");
			printf("BAD address obtained = 0x%X\n", frame);
			return ('\n');
		}

		if (c == 'r') {
			nframe = frame;
			frame = stack;
			scr = (int)chkget(frame, 0, 4);
			stack += 2;	/* "POP" RETurn value */
		}
		else {
			/* Get next frame pointer */
			nframe = (ushort *)chkget(frame, 0, 4);
			/* Get the return address */
			scr = (int)chkget(frame + 2, 0, 4);
			stack = frame + 4;	/* 1/2 UNLK A6; RTS */
		}
		printf("  %X:  %X %X --> ", frame, nframe, scr);
		psymoff(scr, 0, "\n");
		frame = nframe;
	}
}


dodisasm()
{
	register ushort *ip;
	unsigned int	i, j;
	char		c;

	j = 0;
	c = gethex(&i);
	if (c == ESC)
		return (c);
	if (c != '\n')
		c = gethex(&j);
	if (c != '\n')
		putchar('\n');
	if (i & 1) {
		printf("WARNING: Odd address %X", i);
		i &= ~1;
		printf(" rounded to %X\n", i);
	}
	if (j <= i)
	    j = i+2;
	if (j - i > 0x4000) {
	    printf("about to disassemble %X bytes, continue? (y/n)?", j - i);
	    if (dgetc() != 'y') {
	    	goto out;
	    }
	}
	ip = (ushort *)i;
	for(;;) {
		while(ip < (ushort *)j) {
			putchar('\r');
			/* Disassemble at ip */
			ip += disasm(ip);
		}
		if(dgetc() == ESC) break;
		j = (int)(ip + 1);
	}
out:
	return('\n');
}

disasm(ip)
  register ushort *ip;
{
	extern	 int	dotinc;
	register int	inc, cnt;

	printf("%X  ",ip);
	if (buget(ip) == -1) {
		printf("**** **** **** **** ****\t???");
		inc = 1;
	}
	else {
		printins(0, 0, ip);
		dotinc = dotinc / sizeof(short);
		cnt = 0;
		while (cnt < dotinc)
			printf("%x ", buget(ip + cnt++));
		cnt = 5 - dotinc;
		if (cnt > 0)
			while (cnt--)
				printf("---- ");
		printins(0, 1, ip);
		inc = dotinc / sizeof(short);
	}
	printf("\n");
	return(inc);
}



/*
 * =========================================================================
 *  68000 instruction disassembler.  Stolen from the halls of src/cmd/adb
 * =========================================================================
 */

/* M000	-Hans.  May 10, 1982
 *	Changed to do assembly syntax dissassembly
 * M001 -Hans	May 12, 1982
 *	Change use of get()
 * M002 -Hans	Oct 7, 1982
 *	Made it understand dbra and the 'false' condition.
 *	Made it handle cc and sr in immediate instructions
 * M003 -Hans Oct 7, 1982
 *	rewrote mapsize to use a much better algorithm.
 * M004 -Hans Dec 7, 1982
 *	made n(pc,ixreg) addressing mode work
 * M005 -Hans Dec 7, 1982
 *	made it understand movep
 */
#
/****************************************************************************

	DEBUGGER - 68000 instruction printout

****************************************************************************/

/* #include "defs.h" */
#define	ISYM		2
#define dprintf		if (prt_on) printf
#define	leng(a)		((long)((unsigned)(a)))
#define inkdot(incr)	((int)dot + (incr))

ushort		*dot;
int		dotinc;
short		prt_on = 0;
unsigned	space;

char *badop = "\t???";
char *IMDF;				/* immediate data format */

char *bname[16] = { "ra", "sr", "hi", "ls", "cc", "cs", "ne",
		    "eq", "vc", "vs", "pl", "mi", "ge", "lt", "gt", "le" };

char *shro[4] = { "as", "ls", "rox", "ro" };

char *bit[4] = { "btst", "bchg", "bclr", "bset" };

int omove(), obranch(), oimmed(), oprint(), oneop(), soneop(), oreg(), ochk();
int olink(), omovem(), oquick(), omoveq(), otrap(), oscc(), opmode(), shroi();
int extend(), biti(),  omovep();

struct opdesc
{			
	ushort mask, match;
	int (*opfun)();
	char *farg;
} opdecode[] =
{					/* order is important below */
  0xF000, 0x1000, omove, "b",		/* move instructions */
  0xF000, 0x2000, omove, "l",
  0xF000, 0x3000, omove, "w",
  0xF000, 0x6000, obranch, 0,		/* branches */
  0xF038, 0x0008, omovep, 0,		/*M005*/
  0xFF00, 0x0000, oimmed, "or",		/* op class 0  */
  0xFF00, 0x0200, oimmed, "and",
  0xFF00, 0x0400, oimmed, "sub",
  0xFF00, 0x0600, oimmed, "add",
  0xFF00, 0x0A00, oimmed, "eor",
  0xFF00, 0x0C00, oimmed, "cmp",
  0xF100, 0x0100, biti, 0,
  0xF800, 0x0800, biti, 0,
  0xFFC0, 0x40C0, oneop, "move_from_sr", /* op class 4 */
  0xFF00, 0x4000, soneop, "negx",
  0xFF00, 0x4200, soneop, "clr",
  0xFFC0, 0x44C0, oneop, "move_to_ccr",
  0xFF00, 0x4400, soneop, "neg",
  0xFFC0, 0x46C0, oneop, "move_to_sr",
  0xFF00, 0x4600, soneop, "not",
  0xFFC0, 0x4800, oneop, "nbcd",
  0xFFF8, 0x4840, oreg, "\tswap\td%D",
  0xFFC0, 0x4840, oneop, "pea",
  0xFFF8, 0x4880, oreg, "\textw\td%D",
  0xFFF8, 0x48C0, oreg, "\textl\td%D",
  0xFB80, 0x4880, omovem, 0,
  0xFFC0, 0x4AC0, oneop, "tas",
  0xFF00, 0x4A00, soneop, "tst",
  0xFFF0, 0x4E40, otrap, 0,
  0xFFF8, 0x4E50, olink, 0,
  0xFFF8, 0x4880, oreg, "\textw\td%D",
  0xFFF8, 0x48C0, oreg, "\textl\td%D",
  0xFFF8, 0x4E58, oreg, "\tunlk\ta%D",
  0xFFF8, 0x4E60, oreg, "\tmove\ta%D,usp",
  0xFFF8, 0x4E68, oreg, "\tmove\tusp,a%D",
  0xFFFF, 0x4E70, oprint, "reset",
  0xFFFF, 0x4E71, oprint, "nop",
  0xFFFF, 0x4E72, oprint, "stop",
  0xFFFF, 0x4E73, oprint, "rte",
  0xFFFF, 0x4E75, oprint, "rts",
  0xFFFF, 0x4E76, oprint, "trapv",
  0xFFFF, 0x4E77, oprint, "rtr",
  0xFFC0, 0x4E80, oneop, "jsr",
  0xFFC0, 0x4EC0, oneop, "jmp",
  0xF1C0, 0x4180, ochk, "chk",
  0xF1C0, 0x41C0, ochk, "lea",
  0xF0C0, 0x50C0, oscc, 0,
  0xF100, 0x5000, oquick, "addq",
  0xF100, 0x5100, oquick, "subq",
  0xF000, 0x7000, omoveq, 0,
  0xF1C0, 0x80C0, ochk, "divu",
  0xF1C0, 0x81C0, ochk, "divs",
  0xF1F0, 0x8100, extend, "sbcd",
  0xF000, 0x8000, opmode, "or",
  0xF1C0, 0x91C0, opmode, "sub",
  0xF130, 0x9100, extend, "subx",
  0xF000, 0x9000, opmode, "sub",
  0xF1C0, 0xB1C0, opmode, "cmp",
  0xF138, 0xB108, extend, "cmpm",
  0xF100, 0xB000, opmode, "cmp",
  0xF100, 0xB100, opmode, "eor",
  0xF1C0, 0xC0C0, ochk, "mulu",
  0xF1C0, 0xC1C0, ochk, "muls",
  0xF1F8, 0xC188, extend, "exg",
  0xF1F8, 0xC148, extend, "exg",
  0xF1F8, 0xC140, extend, "exg",
  0xF1F0, 0xC100, extend, "abcd",
  0xF000, 0xC000, opmode, "and",
  0xF1C0, 0xD1C0, opmode, "add",
  0xF130, 0xD100, extend, "addx",
  0xF000, 0xD000, opmode, "add",
  0xF100, 0xE000, shroi, "r",
  0xF100, 0xE100, shroi, "l",
  0, 0, 0, 0
};

printins(f, dspflg, addr)
ushort *addr;
{
	register struct opdesc *p;
	ushort	inst;

	dot = addr;
	inst = buget(addr);
	prt_on = dspflg;
	space = 0; dotinc = 2;
	if (f) IMDF = "#%D."; else IMDF = "#%X";
	for (p = opdecode; p->mask; p++)
		if ((inst & p->mask) == p->match) break;
	if (p->mask != 0) (*p->opfun)(inst, p->farg);
	else dprintf(badop);
}

long
instfetch(size)
int size;
{
	long l1;
	/* M001 */
	l1 = leng(chkget(inkdot(dotinc), space, (size==4)?4:2));
	if (size!=4)
	{
		if (l1 & 0x8000) l1 |= 0xFFFF0000;
		else l1 &= 0xFFFF;
		dotinc +=2;
	}
	else	dotinc +=4;
	return (l1);
}

printea(mode,reg,size)
long mode, reg;
int size;
{
	long index,disp;

	switch ((int)(mode)) {
	  case 0:	dprintf("d%D",reg);
			break;

	  case 1:	dprintf("a%D",reg);
			break;

	  case 2:	dprintf("(a%D)",reg);	/* M000 */
			break;

	  case 3:	dprintf("(a%D)+",reg);	/* M000 */
			break;

	  case 4:	dprintf("-(a%D)",reg);	/* M000 */
			break;

	  case 5:	index = instfetch(2);
	  		dprintf("%D.(a%D)", index, reg);	/* M000 */
			break;

	  case 6:	index = instfetch(2);
			disp = (char)(index&0377);
			dprintf("%d.(a%d,%c%D%c)", disp, reg,	/* M000 */
			  (index&0100000)?'a':'d',(index>>12)&07,
			  (index&04000)?'l':'w');
			break;

	  case 7:	switch ((int)(reg))
			{
			  case 0:	index = instfetch(2);
					dprintf("%x:w",index);
					break;

			  case 1:	index = instfetch(4);
					psymoff(index, ISYM, "");
					break;

			  case 2:	index = instfetch(2);	/* M000 */
					dprintf("%D.(pc)",index);
					break;

			  case 3:	index = instfetch(2);
					disp = (char)(index&0377);
						/* M000 *//*M004*/
					dprintf("%D.(pc,%c%D:%c)", disp,
        			        (index&0100000)?'a':'d',(index>>12)&07,
					(index&04000)?'l':'w');
					break;

			  case 4:	index = instfetch(size==4?4:2);
					dprintf(IMDF, index);
					break;

			  default:	dprintf("???");
					break;
			}
			break;

	  default:	dprintf("???");
	}
}

printEA(ea,size)
long ea;
int size;
{
	printea((ea>>3)&07,ea&07,size);
}

/*
 * turn the size field into a byte count.  this used to be done
 * about as inefficiently as possible.  inefficiency is ok when
 * it improves simplicity or readability but this was disgusting
 * in all accounts.  -Hans Oct 7, 82
 */
mapsize(inst)
long inst;
{
	static int sizemap[] = {1, 2, 4, -1};	/* M003 */
	return(sizemap[(inst>>6)&03]);		/* M003 */
}

char suffix(size)
register int size;
{
	return((size==1) ? 'b' : (size==2) ? 'w' : (size==4) ? 'l' : '?');
}

omove(inst, s)
long inst;
char *s;
{
	int size;

	dprintf("\tmov%c\t",*s);
	size = ((*s == 'b') ? 1 : (*s == 'w') ? 2 : 4);
	printea((inst>>3)&07,inst&07,size);
	printc(',');
	printea((inst>>6)&07,(inst>>9)&07,size);
}

obranch(inst,dummy)
long inst;
{
	long disp = inst & 0377;
	char *s; 

	s = "s ";
	if (disp > 127) disp |= ~0377;
	else if (disp == 0) { s = " "; disp = instfetch(2); }
	dprintf("\tb%s%s\t", bname[(int)((inst>>8)&017)], s);
	psymoff(disp+inkdot(2), ISYM, "");
}

oscc(inst,dummy)
long inst;
{
	/* M002  Hacked to understand dbcc and the 'false' condition */
	register int cc;
	cc=(inst>>8)&017;
	if ((inst&070) == 010)
	{
		dprintf("\tdb%s\td%d,", cc==1 ? "ra" : bname[cc], inst&07);
		cc = instfetch(2);
		psymoff(cc+inkdot(2), ISYM, "");
	}
	else
	{
		dprintf("\ts%s\t", cc==1 ? "f" : bname[cc]);
		printea((inst>>3)&07,inst&07,1);
	}
}

biti(inst, dummy)
long inst;
{
	int	immd;

	dprintf("\t%s\t", bit[(int)((inst>>6)&03)]);
	if (inst&0x0100) {
		dprintf("d%D,", inst>>9);
	}
	else {
		immd = instfetch(2);
		dprintf("#%D.", immd); printc(','); 
	}
	printEA(inst);
}

/* movep instruction.   M005 */
omovep(inst, dummy)
{
	int dreg, areg, imm;
	dreg = (inst>>9) & 7;
	areg = inst & 7;
	dprintf("\tmovep%c\t", "wl"[(inst>>6) &1]);
	imm = instfetch(2);
	if (inst&0x80)
	{
		dprintf("d%d,", dreg);
		dprintf(IMDF, imm);
		dprintf("(a%d)", areg);
	}
	else
	{
		dprintf(IMDF, imm);
		dprintf("(a%d),d%d", areg, dreg);
	}
}

opmode(inst,opcode)
long inst;
{
	register int opmode = (int)((inst>>6) & 07);
	register int reg = (int)((inst>>9) & 07);
	int size;

	size = (opmode==0 || opmode==4) ?
		1 : (opmode==1 || opmode==3 || opmode==5) ? 2 : 4;
	dprintf("\t%s%c\t", opcode, suffix(size));
	if (opmode>=4 && opmode<=6)
	{
		dprintf("d%d,",reg);
		printea((inst>>3)&07,inst&07, size);
	}
	else
	{
		printea((inst>>3)&07,inst&07, size);
		dprintf(",%c%d",(opmode<=2)?'d':'a',reg);
	}
}

shroi(inst,ds)
long inst;
char *ds;
{
	int rx, ry;
	char *opcode;
	if ((inst & 0xC0) == 0xC0)
	{
		opcode = shro[(int)((inst>>9)&03)];
		dprintf("\t%s%s\t", opcode, ds);
		printEA(inst);
	}
	else
	{
		opcode = shro[(int)((inst>>3)&03)];
		dprintf("\t%s%s%c\t", opcode, ds, suffix(mapsize(inst)));
		rx = (int)((inst>>9)&07); ry = (int)(inst&07);
		if ((inst>>5)&01) {
			dprintf("d%d,d%d", rx, ry);
		}
		else {
			dprintf(IMDF, (rx ? rx : 8));
			dprintf(",d%d", ry);
		}
	}
}		

/*
 * immediates, of the form addw	#nnn,EA
 */
oimmed(inst,opcode) 
long inst;
register char *opcode;
{
	register int size = mapsize(inst);
	long const;

	if (size > 0)
	{
		dprintf("\t%s%c\t", opcode, suffix(size));
		const = instfetch(size==4?4:2);
		dprintf(IMDF, const); printc(',');
		if ((inst&074) == 074) {	/* M002 */
			dprintf("%s", size==1? "cc" : "sr");
		} else
			printEA(inst,size);
	}
	else dprintf(badop);
}

oreg(inst,opcode)
long inst;
register char *opcode;
{
	dprintf(opcode, (inst & 07));
}

extend(inst, opcode)
long	inst;
char	*opcode;
{
	register int size = mapsize(inst);
	int ry = (inst&07), rx = ((inst>>9)&07);
	char c;

	c = ((inst & 0x1000) ? suffix(size) : ' ');
	dprintf("\t%s%c\t", opcode, c);
	if (*opcode == 'e')
	{
		if (inst & 0x0080) {dprintf("d%D,a%D", rx, ry);}
		else if (inst & 0x0008) {dprintf("a%D,a%D", rx, ry);}
		else dprintf("d%D,d%D", rx, ry);
	}
	else if ((inst & 0xF000) == 0xB000) {dprintf("a%D@+,a%D@+", ry, rx);}
	else if (inst & 0x8) {dprintf("a%D@-,a%D@-", ry, rx);}
	else dprintf("d%D,d%D", ry, rx);
}

olink(inst,dummy)
long inst;
{
	int	size;

	dprintf("\tlink\ta%D,", inst&07);
	size = instfetch(2);
	dprintf(IMDF, size);
}

otrap(inst,dummy)
long inst;
{
	dprintf("\ttrap\t");
	dprintf("#%D", inst&017);
}

oneop(inst,opcode)
long inst;
register char *opcode;
{
	dprintf("\t%s\t",opcode);
	printEA(inst);
}

pregmask(mask)
register int mask;
{
	register int i;
	register int flag = 0;

	dprintf("#<");
	for (i=0; i<16; i++)
	{
		if (mask&1)
		{
			if (flag) printc(','); else flag++;
			dprintf("%c%d",(i<8)?'d':'a',i&07);
		}
		mask >>= 1;
	}
	dprintf(">");
}

omovem(inst,dummy)
long inst;
{
	register int i, list = 0, mask = 0100000;
	register int reglist = (int)(instfetch(2));

	if ((inst & 070) == 040)	/* predecrement */
	{
		for(i = 15; i > 0; i -= 2)
		{ list |= ((mask & reglist) >> i); mask >>= 1; }
		for(i = 1; i < 16; i += 2)
		{ list |= ((mask & reglist) << i); mask >>= 1; }
		reglist = list;
	}
	dprintf("\tmovem%c\t",(inst&100)?'l':'w');
	if (inst&02000)
	{
		printEA(inst);
		printc(',');
		pregmask(reglist);
	}
	else
	{
		pregmask(reglist);
		printc(',');
		printEA(inst);
	}
}

ochk(inst,opcode)
long inst;
register char *opcode;
{
	dprintf("\t%s\t",opcode);
	printEA(inst);
	dprintf(",%c%D",(*opcode=='l')?'a':'d',(inst>>9)&07);
}

soneop(inst,opcode)
long inst;
register char *opcode;
{
	register int size = mapsize(inst);

	if (size > 0)
	{
		dprintf("\t%s%c\t",opcode,suffix(size));
		printEA(inst);
	}
	else dprintf(badop);
}

oquick(inst,opcode)
long inst;
register char *opcode;
{
	register int size = mapsize(inst);
	register int data = (int)((inst>>9) & 07);

	if (data == 0) data = 8;
	if (size > 0)
	{
		dprintf("\t%s%c\t", opcode, suffix(size));
		dprintf("#%D.", data); printc(',');
		printEA(inst);
	}
	else dprintf(badop);
}

omoveq(inst,dummy)
long inst;
{
	register int data = (int)(inst & 0377);

	if (data > 127) data |= ~0377;
	dprintf("\tmoveq\t"); dprintf("#%x", data);
	dprintf(",d%D", (inst>>9)&07);
}

oprint(inst,opcode)
long inst;
register char *opcode;
{
	dprintf("\t%s",opcode);
}

/* =========================================================================
     Follows is a debugger quickie version of chkget
   ========================================================================= */
unsigned chkget(addr, space, size)
  ushort *addr;
{
	ushort	*maddr;
	unsigned long 	value;

	switch (size) {
	    case 1:
		maddr = (ushort *)((long)addr & ~1);
		value = (unsigned long)buget(maddr);
		if ((long)addr & 1)
			value &= 0xff;
		else
			value = value >> 8;
		return (value);

	    default:
	    case 2:
		return ((unsigned)buget(addr));

	    case 4:
		value = ((unsigned)buget(addr)) << 16;
		value |= (ushort)buget(addr+1);
		return (value);
	}
}

printc(c)
  register char	c;
{
	if (prt_on)
		putchar(c);
}

psymoff(v,x,s)
  register unsigned int	v;
  int		x;
  char		*s;
{
	register int	i;

	if (x == ISYM && !prt_on)
		return;

	for (i = 0; dsymtab[i].name; ++i)
		if (dsymtab[i].addr >= v)  break;

	if (dsymtab[i].name  &&  dsymtab[i].addr == v) {
		printf("%s", dsymtab[i].name);
	}
	else {
		if (x != ISYM && dsymtab[i].name) {
			--i;
			printf("%s+%d", dsymtab[i].name, v - dsymtab[i].addr);
		} else
			printf("0x%X", v);
	}

	printf(s);
}
