/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: exec.c,v 6.0 87/10/12 17:07:24 iv Exp $
 *
 * $Log:	exec.c,v $
 * Revision 6.0  87/10/12  17:07:24  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.3  87/05/06  14:02:14  iv
 * 	- Fixed degenerate signed-char comparison
 * 
 * Revision 5.2  86/08/28  14:37:44  iv
 * 	- Correct typos & illegal name in "68[kK]MODEL"  (now "m68kMODEL")
 * 
 * Revision 5.1  86/08/28  13:58:29  iv
 * 	- Switched debugging over to new dbug.h package.
 * 	- General readability overhaul.  No actual code changes.
 * 
 * Revision 5.0  85/12/03  17:09:25  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:45:31  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:46:28  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:09  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.2  84/12/19  11:19:09  iv
 * 	- Corrected stack-size checking code for better FIXEDSTACK operation
 * 	  (we now just set u.u_exext.xe_stksize to SSIZE if it is 0)
 * 	  ((or if the XE_FS bit is NOT on))
 * 	- Added a missing FIXEDSTACK conditional to initialize u.u_stktop
 * 	- Corrected a minor nit lint was complaining about
 * 	  [wrong type argument to closef()]
 * 
 * Revision 1.1  84/04/10  17:36:10  iv
 * Checkin for initial version of TRS-XENIX 2.0
 */


#include "../h/param.h"
#include "../h/inode.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/proc.h"
#include "../h/reg.h"
#include "../h/mmu.h"
#include "../h/systm.h"
#include "../h/acct.h"


/*#define DEBUG	7		/* Include debugging statements */
#ifdef DEBUG
#  define DbugVar  ex_debug
#  define dbTX	0x0001			/* Debug shared text */
#  define dbPR	0x0002			/* Debug preadi() code */
#endif /*DEBUG*/
#define DEBUGS	dbALL
#include "../h/dbug.h"


/* lookup the file, check access & read in the header */
/* sets up u.u_exdata, returns 0 on error */
/* NON PORTABLE only in that it must support different kind of old headers */
struct inode *gethead()
{
	register struct inode *ip;
	struct aexec oldhead;

	if ((ip = namei(uchar, 0)) == NULL)
		return (NULL);
	if (access(ip, IEXEC) ||
	   (ip->i_mode & IFMT) != IFREG ||
	   (ip->i_mode & (IEXEC|(IEXEC>>3)|(IEXEC>>6))) == 0) 
	{
		u.u_error = EACCES;
		goto bad;
	}
	/*
	 * read in first few bytes of file for segment sizes
	 */

	u.u_base = (caddr_t)&u.u_exdata;
	u.u_count = sizeof(u.u_exdata);
	u.u_offset = 0;
	u.u_segflg = 1;
	readi(ip);
	u.u_segflg = 0;
	if (u.u_error)
		goto bad;
	if (u.u_count!=0) {
		goto enoexe;
	}

#define m68kMODEL	(XE_EXEC|XE_LTEXT|XE_LDATA)
	if (u.u_exdata.x_magic != X_MAGIC
	    || (u.u_exdata.x_renv & m68kMODEL) != m68kMODEL
	    || (uchar_t)u.u_exdata.x_cpu != (uchar_t)(XC_68K | XC_BSWAP))
	{
		goto enoexe;
	}

	clearbuf((caddr_t) &u.u_exext, sizeof u.u_exext);
	if (u.u_exdata.x_ext) {
		u.u_base = (caddr_t)&u.u_exext;
		u.u_count = min(sizeof(u.u_exext), u.u_exdata.x_ext);
		u.u_offset = sizeof(u.u_exdata);
		u.u_segflg = 1;
		readi(ip);
		u.u_segflg = 0;
		if (u.u_error)
			goto bad;
	}
	return (ip);
enoexe:
	u.u_error=ENOEXEC;
bad:
	iput(ip);
	return (0);
}


/*
 * Read in and set up memory for executed file.
 * Zero return is normal;
 * non-zero means only the text is being replaced
 */
getxfile(ip, nargc)
register struct inode *ip;
{
	register msize_t ds;
	register sep;
	register msize_t ts, ss;
	register overlay;
	register struct proc *pp;
	long lsize;


	sep = 0;
	overlay = 0;
	switch (u.u_exdata.x_renv&(XE_SEP|XE_PURE|XE_OVER)) {
		case 0:			/* plain executable */
			lsize = (long)u.u_exdata.x_data + u.u_exdata.x_text;
			if (u.u_exdata.x_data == 0)
				lsize += TEXTPAD;
			u.u_exdata.x_data = lsize;
			if (lsize != u.u_exdata.x_data) {/* check ovflow */
				u.u_error = ENOMEM;
				goto out;
			}
			u.u_exdata.x_text = 0;
			break;
		case XE_SEP|XE_PURE:	/* separated I & D */
			sep++;
			break;
		case XE_OVER|XE_PURE:	/* text overlay */
			overlay++;
			break;
		case XE_PURE:		/* read-only text */
			break;
		default:
			goto enoexe;
	}


	if (u.u_exdata.x_text!=0 && (ip->i_flag&ITEXT)==0 && ip->i_count!=1) {
		u.u_error = ETXTBSY;
		goto out;
	}

	/*
	 * find text and data sizes
	 * try them out for possible
	 * overflow of max sizes
	 */
	if (u.u_exdata.x_text)
		ts = btoms(u.u_exdata.x_text + TEXTPAD);
	else
		ts = 0;
	lsize = (long)u.u_exdata.x_data + u.u_exdata.x_bss;
	if (lsize != (unsigned)lsize) {
		u.u_error = ENOMEM;
		goto out;
	}
#ifdef FIXEDSTACK
	if ((u.u_exdata.x_renv & XE_FS)==0 || u.u_exext.xe_stksize == 0)
		u.u_exext.xe_stksize = SSIZE;

	ds = btoms((unsigned) lsize + nargc + u.u_exext.xe_stksize);
	ss = 0;
#else /*FIXEDSTACK*/
	ds = btoms((unsigned) lsize);
	ss = btoms((unsigned) (nargc + SSIZE));
#endif /*FIXEDSTACK*/


#define	ENDOFFSET(component) \
	(int) &(((struct xext *) 0)->component) \
			+ sizeof (((struct xext *) 0)->component)

	if (ts) {
		if (u.u_exdata.x_ext >= ENDOFFSET(xe_tbase))
			if (u.u_exext.xe_tbase != (long) TEXTSTART)
				goto enoexe;
		if (u.u_exdata.x_ext >= ENDOFFSET(xe_dbase))
			if (u.u_exext.xe_dbase != (long) DATASTART(ts, sep))
				goto enoexe;
	}
	else if (u.u_exdata.x_ext >= ENDOFFSET(xe_tbase)) {
		if (u.u_exext.xe_tbase != (long) DATASTART(ts, sep))
			goto enoexe;
	}


	if (overlay) {
		if (u.u_sep==0 && ts != u.u_tsize || nargc) {
			u.u_error = ENOMEM;
			goto out;
		}
		ds = u.u_dsize;
		ss = u.u_ssize;
		sep = u.u_sep;
		xfree();
		xalloc(ip);
		u.u_ar0[UPC] = u.u_exdata.x_entry & ~01;
	} else {
		/*
		 * We cannot fiddle with u.u_?size until
		 * we after we know that we can do the exec
		 * (after call on mmuchk).
		 */
		if (mmuchk(ts, ds, ss, sep))
			goto out;
		/*
		 * allocate and clear core
		 * at this point, committed
		 * to the new image
		 */
		u.u_prof.pr_scale = 0;
		xfree();
		pp = u.u_procp;
		pp->p_ssize = 0;
		expand(USIZE);  /* free data & stack space */
		u.u_dsize = u.u_ssize = 0;
		expand(USIZE + ds + ss);
		pp->p_ssize = ss;
#ifdef FIXEDSTACK
		u.u_stktop = (caddr_t)(lsize + nargc + u.u_exext.xe_stksize);
#endif /*FIXEDSTACK*/
		u.u_sep = sep;
		xalloc(ip);  /* sets u.u_tsize */

		/*
		 * read in data segment
		 */
		u.u_tsize = 0;
		u.u_dsize = ds;
		u.u_ssize = 0;
		u.u_sep = 0;
		mmuset(RO);
		u.u_base = DATASTART(u.u_tsize, u.u_sep);
		u.u_offset = 
			(sizeof u.u_exdata)+u.u_exdata.x_ext+u.u_exdata.x_text;
		u.u_count = u.u_exdata.x_data;

		dbug3(dbTX,2, "data read: u_base=%X u_offset=%X u_count=%X\n",
						    u.u_base,
						    u.u_offset,
						    u.u_count);
		u.u_procp->p_flag |= SLOCK;	/* Lock in core for the read */
		readi(ip);
		u.u_procp->p_flag &= ~SLOCK;
		/*
		 * set SUID/SGID protections, if no tracing
		 */
		if ((pp->p_flag&STRC)==0) {
			if (ip->i_mode&ISUID)
				if (u.u_uid != 0) {
					u.u_uid = ip->i_uid;
					pp->p_uid = ip->i_uid;
				}
			if (ip->i_mode&ISGID)
				u.u_gid = ip->i_gid;
		} else
			psignal(pp, SIGTRAP);
	}
	u.u_tsize = ts;
	u.u_dsize = ds;
	u.u_ssize = ss;
	u.u_sep = sep;
	mmuset(RO);
	dbug(dbTX,3, "Returning from getxfile with shared text\n");
	goto out;

enoexe:
	u.u_error = ENOEXEC;
out:
	return (overlay);
}

/*
 * Clear registers on exec
 */
setregs()
{
	register int *rp;
	register char *cp;
	register i;

	for (rp = &u.u_signal[0]; rp < &u.u_signal[NSIG]; rp++)
		if((*rp & 1) == 0)
			*rp = 0;

	for (cp = &regloc[0]; cp < &regloc[ 15 ];)        /* BUG PORTABILITY */
		u.u_ar0[*cp++] = 0;

#ifdef FIXEDSTACK
	u.u_ar0[A0] = (int)u.u_stktop;
#endif /*FIXEDSTACK*/
	u.u_ar0[UPC] = u.u_exdata.x_entry & ~01;

#ifdef NEVER
	for(rp = (int *)&u.u_fps; rp < (int *)&u.u_fps.u_fpregs[8];)
							/* BUG PORTABILITY */
		*rp++ = 0;
	/*
	 * The code below brings up the floating point package with
	 * no traps enabled.  It is up to the startup to enable the
	 * appropriate traps.
	 */
	u.u_fps.u_fpsr = u.u_fps.u_fpctrl = 0;
#endif /*NEVER*/

	for (i=0; i<NOFILE; i++) {
		if (u.u_pofile[i]&EXCLOSE) {
			closef((struct file *)u.u_ofile[i]);
			u.u_ofile[i] = NULL;
			u.u_pofile[i] &= ~EXCLOSE;
		}
	}
	/*
	 * Remember file name for accounting.
	 */
	u.u_acflag &= ~AFORK;
	bcopy((caddr_t)u.u_dent.d_name, (caddr_t)u.u_comm, DIRSIZ);
}
