/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: local.c,v 6.0 87/10/12 17:07:28 iv Exp $
 *
 * $Log:	local.c,v $
 * Revision 6.0  87/10/12  17:07:28  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/12/03  17:09:34  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:45:35  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:46:33  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:13  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.1  84/04/10  17:36:22  iv
 * Checkin for initial version of TRS-XENIX 2.0
 * 
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/mount.h"
#include "../h/filsys.h"
#include "../h/buf.h"
#include "../h/dir.h"
#include "../h/a.out.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/var.h"
#include "../h/sites.h"
#include "../h/utsname.h"

/*      local.c - Custom Local System Call(s)
 *
 *      the 'clocal' system call is processed here.
 *      This system call is used for all additions to the kernel
 *      made by oems or source licensees.  Programs making use of
 *      clocal calls will not be portable to other versions of Xenix.
 *
 *      Microsoft's policy is to minimize the incompatibilities between
 *      different versions of XENIX, whether they be versions for different
 *      CPUs, different OEMS, or whatever.  For this reason, every effort
 *      is made to satisfy a customers unique requirements in a fashion
 *      cross compatible with other XENIX systems.  Customers requiring
 *      some extension are requested to consult with Microsoft to see
 *      if their needs can be met by some generally acceptable system
 *      extension, which would then be included in future general releases.
 *
 *      This system call is intended for features which cannot be
 *      made a part of standard XENIX.  Microsoft feels that this is
 *      acceptable only for features that in some way depend upon an OEM's
 *      special hardware or proprietary products, and thus cannot in
 *      any case be made portable.
 *
 *
 *      USE:
 *
 *      The clocal call always takes 5 arguments.  One of these arguments
 *      has the site(oem) designation and the sub-function.  This was
 *      done so that the use of a clocal function on a machine which
 *      does not implement it (or implements some other set of clocal
 *      functions) can produce a meaningful diagnostic.  The individual
 *      subfunctions make use of the other 4 arguments.
 *
 *	Each system call has its user entry point through a similarly named
 *	function in /lib/libc.a.   This function is usually written in 
 *	assembler and sets up the argument passing sequence to the system call.
 *      The argument passing sequence is machine dependent.  See the example
 *      local0.s below for details
 *
 *	NOTE:  At present in this distribution clocal calls are not available
 *		To create your clocal calls remove the dummy clocal, and
 *		the #ifdef and #endif statements, around the real clocal()
 */



/*
 * initialize structure with general system/OEM info.
 */

struct utsname utsname = {
	SYSNAME,
	NODE,
	RELEASE,
	VERSION,
	ORIGIN,
	OEMNUMBER,
	SERIAL
};



	/* 
	 *
	 * Dummy clocal system call, until you add your own
	 *
	 */

clocal()
{
	u.u_error = EINVAL;
}




#ifdef NEVER
	/*************  ACTUAL CLOCAL CODE  ***************/
	/*
	 * Use this code as an example for writing system calls
	 * Change the system call names to your own and remove the
	 * other unnecessary calls
	 */


/*
 *      Custom Local System Call
 */

extern  int     nosys();
extern  int     local0();
extern  int     local1();
extern  int     local2();


int (*claddr[])() = {
	local0,			/* 0 - local call 0 */
	local1,			/* 1 - local call 1 */
	local2,			/* 2 - local call 2 */
	};

#define CLMAX 3                /* number of subfunctions */


/*
 *      clocal - custom function dispatcher
 */

clocal()
{
	register struct a {
		unsigned int arg[5];
	} *uap;
	register subfunc;
	register unsigned oem;

	uap = (struct a *)u.u_ap;
	oem = (uap->arg[4]>>16) &0xFFFF;
	if( oem !=  utsname.sysoem ) {
		u.u_error = EINVAL;
		return;
	}
	subfunc = (uap->arg[4]>>8) & 0xFF;
	if (subfunc > CLMAX) {
		u.u_error = EINVAL;
		return;
	}
	(*claddr[subfunc])();
}
#endif NEVER



#ifdef EXAMPLE
/*	Examples of system calls - and argument passing
 *	***** This code is not meant to be compiled *****
 * 
 *
 * local system call 0
 *
 *	passes two arguments, returns a long
 *	in this example argument 1 must only have the value 1 or 2
 *
 * C definition 
 *
 *	long local0(arg1, arg2)
 *	int arg1;
 *	long arg2;
 *
 */

local0()
{
	register l1;
	long l2;
	register struct a {
		int	arg1;
		long	arg2;
	} *uap;

	uap = (struct a *)u.u_ap;
	l1 = uap->arg1;			/* sys call argument 1 */
	l2 = uap->arg2;			/* sys call argument 2 */
	if( l1 !=1 && l1 != 2 ) {
		u.u_error = EINVAL;	/* pass pack error number */
		return;
	}
	/*
		.
		.
	 your code
		.
		.
		.
	u.u_rlong =  whatever value	/* return long value */
	
}


/*
 * local system call 1
 *
 *	passes two arguments, and returns an integer
 *
 * C definition
 *
 *	local1(arg1,arg2)
 *	char *arg1;
 *	int arg2;
 */
local1()
{
	register struct a {
		char	*arg1;
		int	arg2;
	} *uap;

	uap = (struct a *)u.u_ap;
	/*	.
	 *	.
	 *   your code
	 *	.
	 *	.
	 *
	 */
	u.u_rval1 = whatever value
}

/*
 * local system call 2
 *
 * two arguments, and passing back a structure
 *
 * C definition
 * 
 *	struct {
 *		int arga;
 *		long argb;
 *		} arg1;
 *
 *
 *	local2(arg1,arg2)
 *	struct arg1 arg1;
 *	int arg2;
 *
 */
local2()
{
	struct arg1 locarg1;
	struct arg1 *locarg2;
	register struct a {
		struct arg1 *ptrarg1;
		int arg2;
	} *uap;

	uap = (struct a *)u.u_ap;	/* get at arguments */
	locarg2 = uap->ptrarg1;
	/*	.
	 *	.
	 * put system info., into locarg1
	 *    (ie. locarg1.arga = system value
	 *	 locarg1.argb = another system value)
	 *	.
	 *	.
	 *	.
	 *
	 * 
	 *
	 * copy info, from locarg1 back to original structure pointed to by
	 * locarg2.  If there is an error, setup the error number and return
 	 *
	 */
	if (copyout((caddr_t)&locarg1, (caddr_t)locarg2, sizeof(locarg1)) < 0)
		u.u_error = EFAULT;
}
#endif  EXAMPLE



#ifdef LIB_EXAMPLE

|	
|	The following code is an example of how to setup a library entry
|	point to a custom system call.
|
|	All that is necessary, is to replace LOCAL0 by the
|	appropriate custom system call number.
|
|	Note that regester d0 is set up as follows:
|
|		byte 1			call number for clocal
|		byte 2			actual for the custom call
|		bytes 3+4		oemnumber
|
|
|***    C library -- local0
|
|       val = local0(arg1,arg2);
|	int arg1;	/* argument 1 */
|	long arg2;	/* long argument 2 */
|	long  val;	/* return value */
|

#include 	<sys/syscall.h>
#include	<sys/sites.h>

#define LOCAL0		0
#define	SUBFUNCTION	LOCAL0
	.globl	_local0

	.text
_local0:
	movl	#OEMNUMBER,d0		| move in OEMNUMBER
	asll	#8,d0
	addl	#SUBFUNCTION,d0		| get particular custom subfunction
	asll	#8,d0
	addl	#TCLOCAL,d0		| specify that it is a custom call
	jra	call
	rts

	.end

#endif LIB_EXAMPLE
