/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 Model 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/text.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/reg.h"
#include "../h/map.h"
#include "../h/mmu.h"
#include "../h/assert.h"


/* Turn DEBUG on in mmu.h so everyone who uses MMU macros in non-DEBUG code
 *  will be recompiled to use real routines for MMU debugging */
#define DbugVar	mm_debug
#ifdef MMDEBUG
#  define DEBUG	MMDEBUG			/* Use debug level set in mmu.h */
#  define db
#endif /*MMDEBUG*/
#define DEBUGS	dbALL
#include "../h/dbug.h"


/*
 *   Machine dependent MMU routines for TRS-80.
 *
 *	This code is not portable and must be rewritten for new configurations
 *	or kernel enhancement. Once this code is debugged, all routines except
 *	mmufree, mmusub, and mmuchk get replaced with macros defined in mmu.h,
 *	which is why MMDEBUG is defined there.
 */

/*  These had better match the declarations in the include files!
 *  [ie, this code assumes that an mloc_t and an msize_t are shorts]
 *
 *	typedef	unsigned short	mloc_t;
 *	typedef	unsigned short	msize_t;
 */


/*
 * The assembly language support needs to know many bytes are
 * in a struct u.
 */
int dustack = sizeof (struct user);


#ifdef DEBUG
/*
 * MMUINIT
 *	Initialize software and hardware tables.
 */
mmuinit()
{
}

/*
 * MMUPGET
 *	Allocates memory for given process.  Returns MMUERR if unable to
 *	allocate memory.  Return memory location if memory allocated.
 */
mloc_t 
mmupget(p)
  register struct proc *p;
{
	dbug1(dbANY,1,"*** mmupget %X\n", p);
	return (mmuget(p->p_size));
}

/*
 * MMUPFREE
 *	Free all memory resources associated with proc.
 */
mmupfree(p)
  register struct proc *p;
{
	dbug1(dbANY,1,"*** mmupfree %X\n", p);
	mmufree(p->p_addr.p_caddr, p->p_size);
}

/*
 * MMUPCOPY
 *	Copy all of one proc to another.  From proc is active one.
 */
mmupcopy(from, to)
  register struct proc *from, *to;
{
	dbug2(dbANY,1,"*** mmupcopy from %X to %X\n", from, to);

	/*
	 * Copy data + U-page.
	 */
	copybuf(mltoa(from->p_addr.p_caddr), mltoa(to->p_addr.p_caddr),
		mstob(from->p_size));
}


/*
 * MMUCLEAR
 *	Clear part of a memory region.  The caller is assumed to
 *	have checked that the bytes desired for clearing must all
 *	live in the memory region.
 */
mmuclear(mloc, msize)
  mloc_t mloc;
  msize_t msize;
{
	dbug2(dbANY,1,"*** mmuclear mloc=%x msize=%x\n", mloc, msize);
	clearbuf(mltoa(mloc), mstob(msize));
}

/*
 * MMUCOPY
 *	Copy part of a memory region.  The code must be
 *	able to handle ripple-copying correctly.
 */
mmucopy(source, dest, size)
  mloc_t source, dest;
  msize_t size;
{
	dbug3(dbANY,1,"*** mmucopy source=%x dest=%x size=%x\n",
					source,
					dest,
					size);
	copybuf(mltoa(source), mltoa(dest), mstob(size));
}

/*
 * GIVEMAP
 *	Give mapping resources from one process to another.
 *
 *	p1 - donor process
 *	p2 - receiver process
 */
/*ARGSUSED*/
givemap(p1,p2)
struct proc *p1, *p2;
{
}

/*
 * MMUGET
 *      Allocate memory for region of size msize, returns
 *      description of new location or MMUERR on failure
 */
mloc_t 
mmuget(msize)
  register msize_t	msize;
{
	register ret;

	dbug1(dbANY,1,"*** mmuget msize=%x", msize);

	ret = malloc(coremap, (int) msize);
	if (ret == 0) {
		dbug(dbANY,1,"\tfailure\n");
		return (MMUERR);
	}
	dbug1(dbANY,1,"\tsuccess %x\n",ret);
	return (ret);
}
/*
 * MMUGETAT
 *      Allocate memory for region of size msize at location loc, returns
 *      description of new location or MMUERR on failure
 */
mloc_t 
mmugetat(msize, loc)
  register msize_t	msize;
{
	register ret;

	dbug1(dbANY,1,"*** mmuget msize=%x", msize);

	ret = mallocat(coremap, (int) msize, (int) loc);
	if (ret == 0) {
		dbug(dbANY,1,"\tfailure\n");
		return (MMUERR);
	}
	dbug1(dbANY,1,"\tsuccess %x\n",ret);
	return (ret);
}
#endif /*DEBUG*/

/*
 * MMUFREE
 *	Free region of allocated memory of size msize
 *	at location mloc.
 */
mmufree(mloc, msize)
  mloc_t mloc;
  msize_t msize;
{
	dbug2(dbANY,1,"*** mmufree mloc = %x msize = %x", mloc, msize);

	mfree(coremap, (int)msize, (int)mloc);
	if (runin) {
		runin = 0;
		wakeup((caddr_t)&runin); /* Wake swapper when freeing core */
	}
	dbug(dbANY,1,"\tdone\n");
}


/*
 * Definitions:
 *	mmumap() is called when the process's required mapping
 *		is known not to have changed since the last call on mmumap
 *		or mmuset for the same process.
 *	mmuset(rwflag) is called when the required mapping is new or has
 *		changed.
 *
 *	On some systems, mmumap() will be synonymous with mmuset(u.u_lxrw).
 */		

#ifdef DEBUG
/*
 *	mmumap()
 *
 *		Called to reestablish the mapping previously held by a process.
 *		The process must have memory and must have a u page.
 */
mmumap()
{
	dbug1(dbANY,2,"*** mmumap p=%X\n", u.u_procp);
	mmusub(u.u_procp, (int)u.u_lxrw);
}

/*
 *	mmuset(rwflag)
 *
 *		Establish new mapping for a process, with text (if any)
 *		having the protection specified by rwflag.
 */
mmuset(rwflag)
  register int rwflag;
{
	dbug1(dbANY,2,"*** mmuset p=%X\n", u.u_procp);
	u.u_lxrw = rwflag;
	mmusub(u.u_procp, rwflag);
}

/* return maximum dsize */
long 
mmu_ulimit()
{
	return (mstob((long)maxmem - USIZE - u.u_ssize - u.u_tsize));
}
#endif /*DEBUG*/


/*
 *	MMUSUB(p, rwflag)
 *		Set up the mapping of process p, with text protected
 *		according to rwflag.
 */
mmusub(p, rwflag)
  register struct proc *p;
  register int rwflag;
{
	register struct text *xp;
	register unsigned char	*hp, *lp;
	register msize_t ts = u.u_tsize;
	register msize_t ds = u.u_dsize;
	extern unsigned short	mmuHiLim, 
				mmuHiOff,
				mmuLoLim,
				mmuLoOff;	/* software copy of MMU */

	dbug2(dbANY,2,"*** mmusub ts=%x ds=%x\n\t\t", ts, ds);

	xrwflag = rwflag;

	if (mmuchk(ts, ds, 0, u.u_sep))
		panic("mmusub: chk");

	if (ts && (xp = p->p_textp)) {
		mmuHiOff = xp->x_caddr;
		mmuHiLim = xp->x_caddr + ts - 1;
	}
	else {
		mmuHiOff = 0xFF;
		mmuHiLim = 0;
	}

	mmuLoOff = _paddr(p) + USIZE;
	mmuLoLim = _paddr(p) + USIZE - 1 + ds;

	dbug4(dbANY,3,"\tmmusub: Hi O/L=%x/%x, Lo O/L=%x/%x",
				mmuHiOff,
				mmuHiLim,
				mmuLoOff,
				mmuLoLim);

/* WARNING:  This section of code *depends* on and uses the fact
 *	     that the word _mmuHiOff follows _mmuHiLim and that
 *	     the word _mmuLoOff follows _mmuLoLim.
 *	     [See the declarations and comment in mch.S]
 */
	/* Program lower 8 bits of HI segment */
	hp = (unsigned char *)&mmuHiLim;
	MMU_HI = MMU_MkLow(hp[3], hp[1]);

	/* Program lower 8 bits of LO segment */
	lp = (unsigned char *)&mmuLoLim;
	MMU_LO = MMU_MkLow(lp[3], lp[1]);

	/* Program upper 3 bits of everybody */
	MMU_REST = MMU_MkRest(hp[2], hp[0], lp[2], lp[0]);

	dbug(dbANY,2,"\tdone\n");
}

/*
 * MMUCHK
 *	Check validity for this system of proposed sizes.  
 *	Totally system dependent and full of magic numbers.
 */
mmuchk(ts, ds, ss, sep)
  msize_t	ts, ds, ss;
  int		sep;
{
	dbug4(dbANY,2,"*** mmuchk ts=%x ds=%x ss=%x sep=%x\n", ts, ds, ss, sep);

	if (sep || ss)
		goto err;
	if (ts + ds + USIZE <= maxmem)
		return (0);
err:
	u.u_error = ENOMEM;
	dbug(dbANY,1,"\tmmuchk error return\n");
	return (-1);
}
