/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: physio.c,v 6.0 87/10/12 17:07:44 iv Exp $
 *
 * $Log:	physio.c,v $
 * Revision 6.0  87/10/12  17:07:44  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.1  87/08/26  12:15:30  iv
 * 	- Converted debugging to new dbug macro package.  MUCH more readable.
 * 
 * Revision 5.0  85/12/03  17:10:13  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:46:01  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:46:53  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:31  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.2  84/04/18  14:49:35  iv
 * 	- Got rid of "extern pagoff" declaration, which was unused
 * 
 * Revision 1.1  84/04/10  17:39:35  iv
 * -	Initial checkin for TRS-XENIX 2.0
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/text.h"
#include "../h/buf.h"
#include "../h/iobuf.h"
#include "../h/conf.h"
#include "../h/mmu.h"
#include "../h/map.h"

/*#define DEBUG	3	/* set debug level */
#ifdef DEBUG
#  define pPH	0x0001		/* Debug output from physio() */
#  define pSW	0x0002		/* Debug output from swap() */
#  define pSI	0x0004		/* Debug output from swapin() */
#  define pSO	0x0008		/* Debug output from xswap() */
#  define DbugVar  sw_debug
#  define DEBUGS (pSW|pSI|pSO)
#endif DEBUG
#include "../h/dbug.h"

/*
 *      Machine Dependent I/O   (Not drivers)
 *
 *      physio          perform raw i/o to user or system space
 *      swap            read or write to swap device
 *      swapin          swap a process in
 *      xswap           swap a process out
 */

/*
 * Modifications
 *	JJD	20 Dec 81	M000
 *		- Incoporated mods promoting greater machine independence.
 *		  Xswap and swapin should now be mmu independent.  Mostly
 *		  unnoted.
 *	JJD	20 Sep 82	M001
 *		- Mods for contiguous memory allocation
 *		  Code assumes that msize_t's can be added to each other
 *		  and to mloc_t's.
 *		- Conditional code on FIXEDSTACK
 *	JJD	29 Oct 82	M002
 *		- new physio, with improvements for portability and clarity
 *	JJD	24 Nov 82	M003
 *		- removed reference to proc.h SESTAB bit
 *	vwh	15 Aug 83	M004
 *		- laddr was set too early in physio.  On rare occasions
 *		  we got swapped out between setting laddr and locking
 *		  ourselves in memory; set the core position dependent
 *		  part AFTER we have control of the buffer.
 */

/*
 * Raw I/O. The arguments are
 *	The strategy routine for the device
 *	A buffer, which will always be a special buffer
 *	  header owned exclusively by the device for this purpose
 *	The device number
 *	Read/write flag
 * Essentially all the work is computing physical addresses and
 * validating them.
 *
 *      We will not do raw i/o into text or into unallocated user memory.
 *
 *      normal,pure:    TTTTDDDDXXXXSSSS
 *                      <----base---->
 *
 *      seperate:       DDDDDDDDXXXXSSSS
 *                      <----base---->
 *
 */


/* Begin M002 .... */
physio(strat, bp, dev, flag)
register struct buf *bp;
int (*strat)();
int flag;       /* r/w and B_TAPE options */
{
	/* the declarations depend upon sizeof(unsigned) == sizeof(caddr_t) */

	register unsigned base;		/* addr of xfer in user address space */
	register unsigned limit;	/* end of xfer in user address space */
	register unsigned origin;
	register unsigned endpoint;
	register long laddr;		/* long address of start of transfer */
	register unsigned dsize;	/* size of process's data space */

	base = (unsigned)u.u_base;
	limit = base + u.u_count;

	/* Check odd base, odd count, and address wraparound */
	/* On disk, check that file offset and count are BSIZE multiples */

	/* NOTE: variable record size device drivers MUST set the B_TAPE */
	/* bit of their physical buffers.  If they don't, we will check  */
	/* to make sure that they are asking for a whole number of       */
	/* blocks, and will not attempt to transfer a non-standard record*/
	/* size.							 */

	if (((base|u.u_count)&01) || base >= limit)
		goto bad;
	if((flag&B_TAPE)!=0 && ((u.u_count|u.u_offset)&BMASK)!=0)
		goto bad;
	flag &= ~B_TAPE;

	laddr = 0;
	dsize = (unsigned) mstob(u.u_dsize);

	/*
	 * Set origin and endpoint to user-space start and end of data space.
	 */
	origin = (unsigned) DATASTART(u.u_tsize, u.u_sep);
	endpoint = origin + dsize;

	if (base >= origin && base < endpoint) {
		if (limit > endpoint)
			goto bad;

	} else {
		/*
		 * Transfer must be in stack space; set
		 * origin and endpoint to describe user stack space;
		 * adjust long physical address for bytes of data space.
		 */
		origin = (unsigned) STACKSTART(u.u_ssize);
		endpoint = origin + mstob(u.u_ssize);
		laddr += dsize;

		if (base < origin || base >= endpoint || limit > endpoint)
			goto bad;
	}
	laddr += base - origin;

	/* wait for buffer */

	spl6();
	while (bp->b_flags&B_BUSY) {
		bp->b_flags |= B_WANTED;
		sleep((caddr_t)bp, PRIBIO+1);
	}
	bp->b_flags = B_BUSY | B_PHYS | flag;
	bp->b_dev = dev;

	/* get physical address as xmem and addr */

	laddr += (long) mltoa(u.u_procp->p_addr.p_caddr) + mstob(USIZE);/*M004*/
	bp->b_paddr = (paddr_t) laddr;
	bp->b_blkno     = u.u_offset >> BSHIFT;
	bp->b_bcount = u.u_count;

	dbug5(pPH,2,"physio base=%X count=%d poff=%X laddr=%X addr=%X\n",
			base,
			bp->b_bcount,
			laddr - (long)mltoa(u.u_procp->p_addr.p_caddr),
			laddr,
			bp->b_paddr);
	dbug3(pPH,2,"\tu.u_tsize=%d u.u_dsize=%d u.u_ssize=%d\n",
			u.u_tsize,
			u.u_dsize,
			u.u_ssize);
	IFdbug( /*debug(0);*/	)

	bp->b_error = 0;
	u.u_procp->p_flag |= SLOCK;
	(*strat)(bp);
	spl6();
	while ((bp->b_flags&B_DONE) == 0)
		sleep((caddr_t)bp, PRIBIO);
	u.u_procp->p_flag &= ~SLOCK;
	if (bp->b_flags&B_WANTED)
		wakeup((caddr_t)bp);
	spl0();
	bp->b_flags &= ~(B_BUSY|B_WANTED);
	u.u_count = bp->b_resid;
	geterror(bp);
	return;
    bad:
	dbug2(pPH,0,"*** physio bad: base=%X u.u_count=%X", base, u.u_count);
	dbug3(pPH,2," tsize=%x dsize=%x ssize=%x",
			u.u_tsize,
			u.u_dsize,
			u.u_ssize);
	dbug2(pPH,1," origin=%X endpoint=%X", origin, endpoint);
	dbug(pPH,0,"\n");
	IFdbug( /*debug(0);*/	)

	u.u_error = EFAULT;
}

/* ... End M002 */

/* PAGE  */

/*
 * swap IO headers.
 * they are filled in to point
 * at the desired IO operation.
 */
struct	buf	swbuf1;
struct	buf	swbuf2;

/* you can't wait on swbuf[12] for scheduling buffer use because */
/* they are used some number of times for each swap */

int     swbf1   = 0;            /* used to schedule the buffers */
int     swbf2   = 0;            /* used to schedule the buffers */
int     swbwant = 0;            /* used to schedule the buffers */

/*
 * swap(blkno, mloc, msize, rdflg)	- swap a region to/from a device
 */
swap(blkno, mloc, msize, rdflg)
	unsigned  blkno;
	register mloc_t mloc;
	register msize_t msize;
	int rdflg;
{
	register struct buf *bp;

	dbug4(pSW,2,"swap: %s blkno=%d addr=%X bytes=%d\n",
			(rdflg==B_WRITE)? "WRITE" : "READ",
			blkno,
			mltoa(mloc),
			mstob(msize));
	/*
	 * M000
	 * We can now be called with a zero length memory region.
	 * Check for it.
	 */
	if (msize == 0)
		return;

	/* wait for a i/o buffer to do swapping with */
	spl6();

	while (swbf1 && swbf2) {
		swbwant = 1;
		sleep((caddr_t)&swbwant, PSWP+1);     /* sleep if both busy */
	}
	/* select a free buffer */
	if (swbf1==0) { bp = &swbuf1; swbf1++; }
	else          { bp = &swbuf2; swbf2++; }

	spl0();

	/* calculate physical address from page number */
	bp->b_paddr = (paddr_t)mltoa(mloc);

	/* do the i/o */
	bp->b_bcount = mstob(msize);            /* all at once */
	bp->b_blkno = swplo + blkno;            /* disk loc for this page */
	bp->b_flags = B_BUSY | B_PHYS | rdflg;
	bp->b_dev = swapdev;

	dbug4(pSW,3,"swap bp=%X addr=%X blk=%d cnt=%d\n",
			bp,
			bp->b_paddr,
			(int)bp->b_blkno,
			bp->b_bcount);
	IFdbug( /*debug(0);*/	)
	(*bdevsw[major(swapdev)].d_strategy)(bp);

	/* Wait for I/O completion */
	spl6();
	while((bp->b_flags&B_DONE)==0)
		sleep((caddr_t)bp, PSWP);

	if (bp->b_flags & B_ERROR)
		panic("IO err in swap");	/* Oops... */

	/* set flags to indicate free buffer and wake everyone waiting */
	if (bp == &swbuf1)
		swbf1 = 0;
	else
		swbf2 = 0;

	if (swbwant) {
		swbwant = 0;
		wakeup((caddr_t)&swbwant);
	}

	bp->b_flags &= ~(B_BUSY|B_WANTED);
	spl0();
}

/*
 * Swap a process in.
 * Allocate data and possible text separately.
 * It would be better to do largest first.
 */
swapin(p)
register struct proc *p;
{
	register struct text *xp;
	register unsigned short daddr;
	register mloc_t caddr;

	dbug1(pSI,1,"SWAPIN pid %d\n", p->p_pid);

	/* try to get main mem for the process */
	if ((caddr = mmupget(p)) == MMUERR) {
		dbug1(pSI,0,"swapin %d: mmupget FAIL\n", p->p_pid);
		return (0);
	}
	daddr = p->p_addr.p_daddr;
	p->p_addr.p_caddr = caddr;

	/* if the process has shared text try to attach it and possibly  */
	/* swap in the shared text, which may fail for not enough memory */

	if (xp = p->p_textp) {  /* has shared text segment */
		xlock(xp);      /* lock it */
		if (xp->x_ccount==0) {  /* shared text is on swap */
		      /* try to allocate memory for shared text */
			dbug1(pSI,2,"swapin: alloc %x text mem\n", xp->x_size);
			if ((caddr = mmuget(xp->x_size)) == MMUERR) {
			    /* not enough core for text; free memory & return */
				dbug1(pSI,0,"swapin %d: text mmuget FAIL\n",
						p->p_pid);
				xunlock(xp);    /* unlock text */
				mmupfree(p);	/* free allocated memory */
				p->p_addr.p_daddr = daddr;
				return (0);
			}
			xp->x_caddr = caddr;
			if ((xp->x_flag&XLOAD)==0) {
				dbug(pSI,2,"swapin: swap in text\n");
				swap( (unsigned)xp->x_daddr,
						xp->x_caddr,
						xp->x_size,
						B_READ);	/*M001*/
			}
		}
		xp->x_ccount++;
		xunlock(xp);
	}

	/* read in the process */
	dbug(pSI,2,"swapin: swap in data\n");
	swap(daddr, p->p_addr.p_caddr, p->p_size, B_READ);	/*M001*/
	mfree(swapmap, (int)mstod(p->p_size), (int)daddr);	/*M001*/

	p->p_flag |= SLOAD;					/*M003*/
	p->p_time = 0;
	return (1);
}

/*
 * Swap out process p.
 * The ff flag causes its core to be freed--
 * it may be off when called to create an image for a
 * child process in newproc.
 * Osize is the old size of the process and is supplied on calls from
 * expand() during core expansion swaps from sbreak and grow.
 * When doing an expansion swap, we swap out the process in two
 * pieces so that the "new" data is between the current data and
 * stack, which is right where expand() wants it.
 *
 * NOTE CHANGE: Osize should indicate size 0
 * only if desired size is 0.
 *
 * panic: out of swap space
 */
xswap(p, ff, osize)				/*M001*/
register struct proc *p;
msize_t osize;					/*M001*/
{
	register unsigned a;
	register msize_t ssize;
	register msize_t x;

	if (osize == 0)
		osize = p->p_size;

	dbug4(pSO,1,"XSWAP pid %d: %s; osize=%x, size=%x\n",
			p->p_pid,
			ff? "SWAPOUT" : "EXPAND",
			osize,
			p->p_size);

	a = malloc(swapmap, (int) mstod(p->p_size));
	if (a == NULL) {
		dbug2(pSO,0,"xswap %d: malloc(swapmap, %d) FAIL\n",
				p->p_pid,
				mstod(p->p_size));
		panic("out of swap space");
	}
	ssize = p->p_ssize;
	p->p_flag |= SLOCK;
	xccdec(p->p_textp);

	/*
	 * If either the data or stack is expanding, the process is
	 * swapped out with "holes" in it for the new regions.
	 */
	swap(a, p->p_addr.p_caddr, osize - ssize, B_WRITE);	/*M001*/
	swap(a + mstod(p->p_size - ssize),
		p->p_addr.p_caddr + osize - ssize, ssize, B_WRITE);	/*M001*/


	/* When called from a core expansion, p->p_size holds the new  */
	/* size, and osize holds the old ones.  We have allocated      */
	/* enough room for the new size, but have only swapped out     */
	/* the old size.  We now free main memory using the old size.  */
	if (ff) {
		dbug(pSO,3,"xswap: freeing proc's mem\n");
		x = p->p_size;			/* remember size */	/*M001*/
		p->p_size = osize;		/* the actual size */	/*M001*/
		mmupfree(p);                    /* free the procs memory */
		p->p_size = x;			/* new size */		/*M001*/
	}

	p->p_addr.p_daddr = a;
	p->p_flag &= ~(SLOAD|SLOCK);
	p->p_time = 0;
	if (runout) {
		runout = 0;
		wakeup((caddr_t)&runout);
	}
}
