/*
 *	%Z% %M% %I% %D% %Q%
 */

/*
 * THIS FILE CONTAINS CODE SPECIFIC TO THE Tandy TRS-80 MODEL 16
 * COMPUTER. IT MAY REQUIRE MODIFICATION WHEN MOVING
 * XENIX TO DIFFERENT MACHINE ARCHITECTURES OR CONFIGURATIONS.
 *
 *
 * $Header: /usr/sys/mdep/RCS/sbreak.c,v 6.0 87/10/12 17:07:55 iv Exp $
 *
 * $Log:	sbreak.c,v $
 * Revision 6.0  87/10/12  17:07:55  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/12/03  17:10:29  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:46:13  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:47:03  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:39  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.1  84/04/10  17:40:07  iv
 * -	Initial checkin for TRS-XENIX 2.0
 * 
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/text.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/mmu.h"
#include "../h/assert.h"


/*
 * Modifications
 *	JJD	24 Feb 83	M000
 *		- check for overflow in sbreak.  (This code superfluous
 *		  on some machines.)
 */

/*
 *      Machine dependent process size changing.
 *
 *      expand          change the size of the data + stack of a process
 *      grow            extend the stack
 *      sbreak          "break" system call
 */

/*
 * Change the size of the data+stack regions of the current process.
 *
 * This routine has built into it certain assumptions about its
 * callers; in particular, it "knows" that if asked to shrink
 * a process, it is either throwing way the end of data space,
 * or else all of data and stack space.
 *
 * o Called when setting up process 1, in which case current values
 *   of p_size and p_ssize are 0.
 * o Called during the an exec either to free all of data + stack,
 *   or else to allocate data and stack with current p_size = USIZE,
 *   and p_ssize = 0.
 * o Called by sbreak() when enlarging or shrinking data.
 * o Called by grow() when enlarging stack.
 *
 * Upon return, p_size has been updated, and data and stack have
 * been appropriately shuffled in the memory space.
 */
#ifdef DEBUG
int	sb_debug = 0;
#endif
int	smart_expand = 0;		/* turned off by default */
expand(newsize)
	register msize_t newsize;
{
	register struct	proc *p;
	register msize_t ssize;
	register msize_t oldsize;
	register int diff;
	register mloc_t oldloc;
	register mloc_t newloc;

	p = u.u_procp;
	ASSERT(newsize >= p->p_ssize);
	oldloc = p->p_addr.p_caddr;
	ssize = p->p_ssize;
	oldsize = p->p_size;
	p->p_size = newsize;

	diff = (int) newsize - (int) oldsize;

#ifdef DEBUG
	if (sb_debug) {
		printf("expand: oldsize=%d newsize=%d ssize=%d diff=%d\n",
			oldsize, newsize, ssize, diff);
		if (sb_debug > 2)
			debug(0);
	}
#endif
	if (diff == 0)
		return;
	
	if (diff < 0) {
		if(newsize >= ssize)    /* if not removing stack also */
		    mmucopy(oldloc+oldsize-ssize, oldloc+newsize-ssize, ssize);
		mmufree(oldloc + newsize, (msize_t) -diff);
		return;
	}

	/*
	 * Note that even if we do not force a swap on this expand,
	 * we still do a save and resume to continue execution.
	 * This is necessary because when we resume execution,
	 * our u is in a different place.
	 */
	if (save(u.u_ssav)) {
		mmumap();
#ifdef DEBUG
		if (sb_debug) {
			printf("%d: expand growth return\n", p->p_pid);
			if (sb_debug > 1)
				debug(0);
		}
#endif
		/*
		 * The new space already lies between data and stack.
		 * Either we explicitly did the in-core copying to set
		 * this up (below) or else swapin cleverly swapped us
		 * in in two pieces, with a gap between.
		 */
		mmuclear(p->p_addr.p_caddr + oldsize - ssize, (msize_t) diff); 
		return;
	}
	/* If we can allocate the memory just after the current allocation,
	 * then just use it, (provided smart_expand is turned on).
	 */
	if (smart_expand == 0 || ssize != 0 || 
		mmugetat((msize_t) diff, oldloc+oldsize) == MMUERR)
	{
	  if ((newloc = mmuget(newsize)) == MMUERR) {
#ifdef DEBUG
		if (sb_debug) {
			printf("%d: expand force swap\n", p->p_pid);
			if (sb_debug > 1)
				debug(0);
		}
#endif
		xswap(p, 1, oldsize);
		p->p_flag |= SSWAP;
		qswtch();
		/* no return from qswtch() */
	  }
	  p->p_addr.p_caddr = newloc;
	/*
	 * Special hack to minimize fragmentation:
	 *	If the memory which we found follows immediately
	 *	after the current memory region, just use the
	 *	first few pages of the new region, and free
	 *	the rest.
	 * This could be improved upon with the use of some
	 *	sort of realloc routine, which tries for adjacent
	 *	memory pages.
	 */
	  if (oldloc + oldsize == newloc) {
#ifdef DEBUG
		if (sb_debug)
			printf("%d: clever expand: oldloc=%x newloc=%x\n",
				p->p_pid, oldloc, newloc);
#endif
		p->p_addr.p_caddr = oldloc;
		mmucopy(oldloc + oldsize - ssize,
			oldloc + newsize - ssize,
			ssize);
		mmufree(oldloc + newsize, oldsize);
	  } else {
		mmucopy(oldloc, newloc, (msize_t) (oldsize - ssize));
		mmucopy(oldloc + oldsize - ssize,
			newloc + newsize - ssize,
			ssize);
		mmufree(oldloc, oldsize);

	  }
	}
	resume(p->p_addr.p_caddr, u.u_ssav);
}
/* PAGE  */
/*
 * grow the stack to include the SP
 * true return if successful.
 */
grow(sp)
unsigned sp;
{
#ifdef FIXEDSTACK
	return(0);
#else
	register struct proc *p;
	register msize_t newssize;

	p = u.u_procp;
	if( (unsigned)sp >= (unsigned)(USRSTACK-mstob(u.u_ssize)) )
		return(0);
	newssize = btoms((unsigned)USRSTACK - sp + SINCR);
#ifdef DEBUG
	if (sb_debug) {
		printf("grow %d->%d\n",u.u_ssize,newssize);
		if (sb_debug > 2)
			debug(0);
	}
#endif
	if(mmuchk(u.u_tsize, u.u_dsize, newssize, u.u_sep))
		return(0);
	expand(USIZE + u.u_dsize + newssize);
	p->p_ssize = u.u_ssize = newssize;
	mmuset(RO);
	return(1);
#endif
}

/* PAGE  */
/*
 * break system call.
 *  -- bad planning: "break" is a dirty word in C.
 */
sbreak()
{
	struct a {
		char	*nsiz;
	};
	register msize_t newdsize;
	register unsigned bytes = ((struct a *)u.u_ap)->nsiz -
				DATASTART(u.u_tsize, u.u_sep); /*M000*/

	/* set newdsize to requested data size */

	newdsize = btoms(bytes);				/*M000*/

	/* Make sure that we haven't had an overflow problem */	/*M000*/

	if (mstob(newdsize) - bytes >= mstob(1)) {		/*M000*/
		u.u_error = ENOMEM;				/*M000*/
		return;						/*M000*/
	}							/*M000*/

	/* if no change in size return */

	if (newdsize == u.u_dsize)
		return;

#ifdef DEBUG
	if (sb_debug) {
		printf("sbrk nsiz=%X dsize=%d newdsize=%d\n",
		     (int)((struct a *)u.u_ap)->nsiz, u.u_dsize, newdsize);
		if (sb_debug > 2)
			debug(0);
	}
#endif
	/*
	 * Check if it's possible.
	 */
	if (mmuchk(u.u_tsize, newdsize, u.u_ssize, u.u_sep)) {
#ifdef DEBUG
		if (sb_debug)
			printf("sbrk error return\n");
#endif
		return;
	}
	expand(newdsize + u.u_ssize + USIZE);
	u.u_dsize = newdsize;
	mmuset(RO);
	return;
}
