/*
 *  Shared Data functions.
 *	Paul Butzi	5/25/83
 *
 *	Shared data segments are inodes of the type IFNAM.  Thus, they
 *	live in the same name space as semaphores.  The per segment data
 *	lives in the inode for that segment; the per process data
 *	lives in the u area in the sdtab. (see inode.h and user.h)
 *
 *  Modification History:
 *
 *    M000	3/27/84		fiasco!buckm
 *		numerous unmarked changes made while bringing up
 *		shared data for the SCO IBM XENIX project.
 *		sdget() was especially modified.
 *    M001	3/31/84		bobt
 *		change cast to long in btoms usage to unsigned
 *
 *  $Log:	sd.c,v $
 * Revision 6.0  87/10/12  17:07:59  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.2  86/10/01  17:48:37  iv
 * 	-  Changed debugging stuff to new dbug.h macro package
 * 	   (allows sectional debug)
 * 
 * Revision 5.1  86/08/28  18:09:31  iv
 * 	-  Changed sdget() to return the new size of the process
 * 	   with a prefix byte of `0x0A' (so that the library can
 * 	   tell if it's running on the new kernel or not)
 * 
 * Revision 5.0  85/12/03  17:10:35  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:46:18  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.1  85/01/22  15:11:38  iv
 * 	- Changed copyio() call back to the original argument order
 * 	  and swapped the last arguments.   Believe it or not, 
 * 	  "U_WUD" means we want to READ UserData and "U_RUD" means 
 * 	  we want to WRITE UserData.  *sigh*
 * 
 * Revision 3.0  85/01/17  15:47:06  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:50:43  iv
 *	- Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.5  85/01/10  12:48:08  iv
 * 	- Defined ROUTINE for "sdcopy"
 * 
 * Revision 1.4  84/12/21  14:30:09  iv
 * 	- New SCO sd module.  Work in Progress...
 */

/* #define DEBUG   			/* Turns on debugging messages */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/file.h"
#include "../h/dir.h"
#include "../h/inode.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/mmu.h"
#include "../h/sd.h"
#include "../h/dbug.h"

#define	SDWRITE		0001
#define	SDCREAT		0002
#define	SDUNLOCK	0004
#define	SDNOWAIT	0010
#define SDLOCKED	0020
#define SDNTFY		0040
#define SDBTWN		0100

#define I_SD i_fdep.i_namef.i_ndata.i_sd

/* 
 * search for matching slot in per proc sdtab.
 * by convention, addr == 0 means slot is empty.
 */
struct sd *sdsrch( addr )
char *addr;
{
	register int i;

	for ( i = 0; i < NSDSEGS; i++)
		if ( addr == u.u_sdtab[i].sd_addr )
			return &u.u_sdtab[i];
	u.u_error = ENXIO;
	return NULL;
}

/* 
 *	Sdget system call
 */
#define ROUTINE "sdget"
sdget()
{
	struct a {
		char *path;
		int flags;
		int size;
		int mode;
	} *uap;
	register struct sd *sdp;
	register struct inode *ip;
	register msize_t msize;
	mloc_t buf = NULL;
	mloc_t save;
	unsigned size;
	extern struct inode *maknode();

	uap = (struct a *)u.u_ap;

	/* find free slot in sdtab. */
	if ( (sdp = sdsrch((char *)0)) == 0 )
	{
		dbug(dbANY,0, "No space in sdtab; return (EMFILE)...\n");
		u.u_error = EMFILE;
		return;
	}
		
	/* get an inode */
	u.u_dirp = uap->path;
	if ( uap->flags & SD_CREAT )
	{
		ip = namei(uchar, 1);
		if ( ip == NULL )
		{
			if ( u.u_error )
				return;
			ip = maknode(IFNAM | ( uap->mode & 0777));
			if ( ip == NULL )
				return;
			ip->i_fdep.i_namef.i_type = IFSHD;
		}
	}
	else
	{
		ip = namei(uchar, 0);
		if ( ip == NULL )
			return;
	}

	/* make sure inode is type SD */
	if ( ((ip->i_mode & IFMT) != IFNAM) ||
		(ip->i_fdep.i_namef.i_type != IFSHD) )
	{
#ifdef DEBUG
		if ((ip->i_mode & IFMT) != IFNAM)
		   printf("IFMT != IFNAM; i_mode = 0%o\n", ip->i_mode);
		else if (ip->i_fdep.i_namef.i_type != IFSHD)
		   printf("i_namef.i_type != IFSHD; i_type = 0x%x\n",
						    ip->i_fdep.i_namef.i_type);
		else
		   printf("HELP:  i_mode / i_type test FAILED!\n");
#endif DEBUG
		dbug(dbANY,0,"return (ENOTNAM)...\n");
		u.u_error = ENOTNAM;
		iput(ip);
		return;
	}

	/* check permissions */
	access(ip, IREAD);
	if ( uap->flags & SD_WRITE )
		access(ip, IWRITE);
	if ( u.u_error )
	{
		dbug(dbANY,0,"access(ip, IWRITE) failed...\n");
		iput(ip);
		return;
	}

	/*
	 * SD segment needs to be created, and the SD inode fields
	 *	initialized, if we are 1st to access
	 */
	if ( ip->i_count == 1 )
	{
		if ( (uap->flags & SD_CREAT) == 0 )
		{
			dbug(dbANY,0,"!SD_CREAT and ip->i_count == 0;");
			dbug(dbANY,0," return (ENOENT)...\n");
			u.u_error = ENOENT;
			iput(ip);
			return;
		}
		size  = uap->size;
		msize = btoms(size);		/* M001 */
		/* allocate a buffer. */
		if ((buf = mmuget(msize)) == MMUERR)
		{
			dbug1(dbANY,0,"mmuget(%d) gives MMUERR;", msize);
			dbug(dbANY,0," return (ENOMEM)...\n");
			u.u_error = ENOMEM;
			iput(ip);
			return;
		}
		/* put the stuff in the inode */
		ip->I_SD.i_buf   = buf;
		ip->I_SD.i_len   = size;
		ip->I_SD.i_vnum  = 0;
		ip->I_SD.i_snum  = 0;
		ip->I_SD.i_flags = uap->flags & SDUNLOCK;
		ip->i_flag |= IACC;
	}

	/* Make sure we can expand the process */
	size  = ip->I_SD.i_len;
	msize = btoms(size);		/* M001 */
	if (mmuchk(u.u_tsize, u.u_dsize + msize, u.u_ssize, u.u_sep))
	{
		dbug(dbANY,0,"mmuchk() says no room; return (ENOMEM)...\n");
		u.u_error = ENOMEM;
		iput(ip);
		if (buf)
			mmufree(buf, msize);
		return;
	}

	/* expand the process */
	save = u.u_dsize + u.u_ssize;		/* BUG  Machine dependant */
	expand((msize_t)(u.u_dsize + msize + u.u_ssize + USIZE));
	u.u_dsize += msize;
	mmuset(RO);

	/* release the inode */
	prele(ip);

	/* fill the per process data with right stuff. */
	sdp->sd_addr  = (char *)mstob(save);
	sdp->sd_inode = ip;
	sdp->sd_vnum  = 0;
	sdp->sd_flags = uap->flags & SDWRITE;
	u.u_sdsaved   = 0;
	u.u_sdcnt    += 1;

	/* set contents of user proc if necessary */
	sdcopy(sdp, 1);

	/* set rval */
	u.u_rval1 = (int)sdp->sd_addr;
	u.u_rval2 = (int)(0x0A000000 | size);
	return;
}
 

/*
 *	Sdfree system call
 */
sdfree()
{
	struct a {
		char *addr;
	} *uap;
	register struct sd *sdp;

	uap = (struct a *)u.u_ap;
	/* find the slot in sdtab. */
	if ( (sdp = sdsrch(uap->addr)) == NULL)
		return;

	sdfrcm(sdp);
}

sdfrcm(sdp)
register struct sd *sdp;
{
	register struct inode *ip;
	msize_t msize;

	ip = sdp->sd_inode;

	/* force an sdleave if necessary. */
	if ( sdp->sd_flags & SDBTWN )
		sdlvcm(sdp);

	/* copy if necessary. */
	sdcopy(sdp, 0);

	/* if this is the last process free the kernel copy buffer. */
	msize = btoms((unsigned)ip->I_SD.i_len);	/* M001 */
	if ( ip->i_count == 1 )
		mmufree(ip->I_SD.i_buf, msize);

	/* put the inode. */
	ip->i_flag |= IACC;
	iput(ip);

	/* zero out the per proc stuff */
	sdp->sd_addr = 0;
	u.u_sdcnt -= 1;
}

/*
 *	Sdenter system call
 */
sdenter()
{
	struct a {
		char *addr;
		int  flags;
	} *uap;
	register struct sd *sdp;
	register int *flg;

	uap = (struct a *)u.u_ap;

	/* find the right slot */
	if ( (sdp = sdsrch(uap->addr)) == NULL )
		return;

	flg = &(sdp->sd_inode)->I_SD.i_flags;
	if ( *flg & SDLOCKED )
		if ( uap->flags & SDNOWAIT )
		{
			u.u_error = ENXIO;
			return;
		} else {
			while ( *flg & SDLOCKED )
			{
				*flg |= SDNTFY;
				sleep((caddr_t)sdp->sd_inode, PSLEP);
			}
		}
	if ( (*flg & SDUNLOCK) == 0 )
		*flg |= SDLOCKED;
	sdp->sd_flags |= SDBTWN;

	if ( (sdp->sd_flags & SDWRITE ) && (uap->flags & SD_WRITE))
	{
		sdp->sd_vnum += 1;
	}
}


/*
 *	Sdleave system call
 */
 sdleave()
 {
	struct a {
		char *addr;
	} *uap;
	struct sd *sdp;

	uap = (struct a *)u.u_ap;
	/* find the slot. */
	if ( (sdp = sdsrch(uap->addr)) == 0 )
		return;

	sdlvcm(sdp);

	return;
}

/*
 * common code for sdleave, sdfree
 */
sdlvcm(sdp)
register struct sd *sdp;
{
	register int *flg;

	/* unlock segment if of type LOCKED */
	flg = &((sdp->sd_inode)->I_SD.i_flags);
	if ( (*flg & SDUNLOCK) == 0 )
		*flg &= ~(SDLOCKED);

	/* reset BETWEEN */
	sdp->sd_flags &= ~(SDBTWN);

	/* increment snum */
	sdp->sd_inode->I_SD.i_snum += 1;

	/* wakeup if necessary */
	if ( *flg & SDNTFY )
	{
		*flg &= ~(SDNTFY);
		wakeup((caddr_t)sdp->sd_inode);
	}

	return;
}


/*
 *	Sdgetv system call
 */
sdgetv()
{
	struct a {
		char *addr;
	} *uap;
	struct sd *sdp;

	uap = (struct a *)u.u_ap;
	if ( (sdp = sdsrch(uap->addr)) == 0)
		return;

	u.u_rval1 = (sdp->sd_inode)->I_SD.i_snum;
}

/*
 *	Sdwaitv system call
 */
sdwaitv()
{
	struct a {
		char *addr;
		int	num;
	} * uap;
	register struct sd *sdp;

	uap = (struct a *)u.u_ap;
	if ( (sdp = sdsrch(uap->addr)) == 0)
		return;

	while ( (sdp->sd_inode)->I_SD.i_snum == uap->num)
	{
		sdp->sd_inode->I_SD.i_flags |= SDNTFY;
		sleep((caddr_t)sdp->sd_inode, PSLEP);
	}

	u.u_rval1 = (sdp->sd_inode)->I_SD.i_snum;
}

/*
 *	Sdswtch- called twice during process switch.
 *		arg == 0 when switching from this proc.
 *		arg == 1 when switching to this proc.
 */
#undef ROUTINE
#define ROUTINE "sdswtch"
sdswtch(d)
int d;
{
	register struct sd *sdp;

	for ( sdp = &u.u_sdtab[0]; sdp < &u.u_sdtab[NSDSEGS]; sdp++ )
		if ( sdp->sd_addr != NULL ) {
			dbug2(dbANY,0,"sdcopy(sd[%d], %s)\n",
					(sdp - u.u_sdtab),
					(d == 0)? "LEAVE" : "ENTER");
			sdcopy(sdp, d);
		}
}

/*
 *	Sdcopy- copy to/from to kernel copy if necessary
 *		arg 2 == 0 when copying back to kernel copy
 *		arg 2 == 1 when copying back to user copy
 */
sdcopy(sdp, d)
struct sd *sdp;
int d;
{
	if ( d == 0 )
	{	/* switching away from process */
		if ( (sdp->sd_vnum > sdp->sd_inode->I_SD.i_vnum) &&
			( ((sdp->sd_flags & SDBTWN) == 0) || 
				(sdp->sd_inode->I_SD.i_flags & SDUNLOCK) ) )
		{
				cputok(sdp);
				sdp->sd_inode->I_SD.i_vnum = sdp->sd_vnum;
		}
	} 
	else 
	{	/* switching to the process */
		if ( sdp->sd_inode->I_SD.i_vnum > sdp->sd_vnum )
		{
			cpktou(sdp);
			sdp->sd_vnum = sdp->sd_inode->I_SD.i_vnum;
		}
		if ( (sdp->sd_flags & SDBTWN) && 
				(sdp->sd_inode->I_SD.i_flags & SDUNLOCK) )
			sdp->sd_vnum += 1;
	}
}

cputok(sdp)
register struct sd *sdp;
{
	copyio(mltoa(sdp->sd_inode->I_SD.i_buf), sdp->sd_addr, 
		sdp->sd_inode->I_SD.i_len, U_WUD);
}

cpktou(sdp)
register struct sd *sdp;
{
	copyio(mltoa(sdp->sd_inode->I_SD.i_buf), sdp->sd_addr,
		sdp->sd_inode->I_SD.i_len, U_RUD);
}

/*
 *	Sdexit- called to free segments on proc exit.
 */
sdexit()
{
	struct sd *sdp;

	if (u.u_sdcnt <= 0)
		return;
	for ( sdp = &u.u_sdtab[0]; sdp < &u.u_sdtab[NSDSEGS]; sdp++)
		if ( sdp->sd_addr != NULL )
			sdfrcm(sdp);
}

/*
 *	Sdfork- called to adjust inode reference counts during fork
 */
sdfork()
{
	struct sd *sdp;

	if (u.u_sdcnt <= 0)
		return;
	if (u.u_sdsaved == 0)		/* child gets shared data */
	{
		u.u_sdsaved = 1;
		sdswtch(0);
	}
	for ( sdp = &u.u_sdtab[0]; sdp < &u.u_sdtab[NSDSEGS]; sdp++)
		if ( sdp->sd_addr != NULL )
			sdp->sd_inode->i_count += 1;
}
