/*
 *	@(#) sys/clock.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: clock.c,v 6.0 87/10/12 17:09:06 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 *
 *  MODIFICATION HISTORY
 *	M000	20 Feb 1984	blf
 *		- Added spl6/slpx to delay().
 *	M001	22 Feb 1984	blf
 *		- Made lticks global for use in sys/sys4.c:ftime(),
 *		  as per PC XENIX 3.0.
 *	M002	03 Oct 1984	jeIV (Tandy)
 *		- Changed references to HZ to Hz for configurability
 *		- Changed clock() to use new return value from clkreld(),
 *		  which now returns the number of ticks since the last time
 *
 * $Log:	clock.c,v $
 * Revision 6.0  87/10/12  17:09:06  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/08/23  10:34:03  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:47:37  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:42:51  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:22:07  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.2  84/10/03  18:52:43  iv
 * 	- Changed clock() to use the new return value from clkreld(), 
 * 	  which now returns the number of ticks since the last interrupt
 * 
 * Revision 1.1  84/05/17  16:05:03  iv
 * 	- Changed references to HZ to Hz for configurability
 * 
 * Revision 0.1  84/05/17  15:07:35  iv
 * 	- spl6() / splx()  added to delay().
 * 	- lticks was made global to use in sys/sys4.c @ ftime(),
 * 	  as per PC XENIX 3.0.
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/sysinfo.h"
#include "../h/callo.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/text.h"
#include "../h/var.h"

/*
 * clock is called straight from
 * the real time clock interrupt.
 *
 * Functions:
 *	reprime clock
 *	implement callouts
 *	maintain user/system times
 *	maintain date
 *	profile
 *	alarm clock signals
 *	jab the scheduler
 *
 *      For the most part, clock acts like a typical interrupt
 *      service routine.  However, once a second it does a
 *      'lightning bolt' sequence, wherin some possibly time-consuming
 *      table searches are done.  When processing 'lightning bolt'
 *      this routine dismisses the clock interrupt so that it runs
 *      'task time', and can be interrupted for I/O service.  clock()
 *      returns non-zero when it has dismissed its interrupt and changed
 *      the priority level, 0 otherwise.
 *
 *      Note that lightning bolt is posponed as necessary so that it
 *      never takes place when the clock interrupted a non-level 0
 *      sequence (interrupt service or spl-protected sequence)
 *      If we were to force 'task level' state during a clock interrupt
 *      which interrupted an spl?-splx interlock, we would 'break'
 *      the interlock by allowing such interrupts.  The 'CLKONLY' macro
 *      defines a way to tell if its safe to go to 'task time.'
 */

short   lblock = 0;     /* interlock against recursive lbolt */
short	lticks = 0;	/* 1000/Hz (millisec) counter	M001 */

clock(pc, ps)
caddr_t pc;
{
	register struct callo *p1, *p2;
	register struct proc *pp;
	register a, lboltinc;			/* M002 */
	int s;
/* M001	static short lticks;	***/
	extern caddr_t waitloc;

	/* restart clock -- see machdep.c */
	lboltinc = clkreld();			/* M002 */

	/*
	 * callouts
	 * if none, just continue
	 * else update first non-zero time
	 */

	if(callout[0].c_func == NULL)
		goto out;
	p2 = &callout[0];
	while(p2->c_time<=0 && p2->c_func!=NULL)
		p2++;
	p2->c_time--;

	/*
	 * if ps is high, just return
	 */

	if (!CLKONLY(ps) || lblock!=0)
		goto out;

	/*
	 * callout
	 */
	tasktime(1);
	s=spl5();
	if(callout[0].c_time <= 0) {
		p1 = &callout[0];
		while(p1->c_func != 0 && p1->c_time <= 0) {
			(*p1->c_func)(p1->c_arg);
			p1++;
		}
		p2 = &callout[0];
		while(p2->c_func = p1->c_func) {
			p2->c_time = p1->c_time;
			p2->c_arg = p1->c_arg;
			p1++;
			p2++;
		}
	}
	splx(s);

out:
	if (USERMODE(ps)) {
		a = CPU_USER;
		u.u_utime += lboltinc;			/* M002 */
		if(u.u_prof.pr_scale)
			addupc(pc, &u.u_prof, 1);
	} else {
		if (pc == waitloc) {
			a = CPU_IDLE;
			if (syswait.iowait)
				sysinfo.wait[W_IO] += lboltinc;	    /* M002 */
			if (syswait.swap)
				sysinfo.wait[W_SWAP] += lboltinc;   /* M002 */
			if (syswait.physio)
				sysinfo.wait[W_PIO] += lboltinc;    /* M002 */
		} else
			a = CPU_KERNEL;
		u.u_stime += lboltinc;			/* M002 */
	}
	sysinfo.cpu[a] += lboltinc;			/* M002 */
	pp = u.u_procp;
	if(pp->p_stat==SRUN) {
		u.u_mem += (unsigned)pp->p_size;
		if(pp->p_textp) {
			a = pp->p_textp->x_ccount;
			if(a==0)
				a++;
			u.u_mem += pp->p_textp->x_size/a;
		}
	}
	if(pp->p_cpu < 80)
		pp->p_cpu++;
	lbolt += lboltinc;	/* time in ticks */	/* M002 */
	lticks -= lboltinc;				/* M002 */
	if (lticks <= 0) {
		if (!CLKONLY(ps) || lblock!=0)
			return(0);
		lblock++;
		lticks += Hz;				/* M002 */
		++time;
		tasktime();
		runrun++;
		for(pp = &proc[0]; pp < (struct proc *)v.ve_proc; pp++)
		if (pp->p_stat) {
			if(pp->p_time != 127)
				pp->p_time++;
			if(pp->p_clktim)
				if(--pp->p_clktim == 0)
					psignal(pp, SIGALRM);
			pp->p_cpu >>= 1;
			if(pp->p_pri >= PUSER) {
				pp->p_pri = (pp->p_cpu>>1) + PUSER +
					pp->p_nice - NZERO;
			}
		}
		if(runin!=0) {
			runin = 0;
			setrun(&proc[0]);
		}
		lblock=0;
		return(1);
	}
	return(0);
}

/*
 * timeout is called to arrange that fun(arg) is called in tim/Hz seconds.
 * An entry is sorted into the callout structure.
 * The time in each structure entry is the number of Hz's more
 * than the previous entry. In this way, decrementing the
 * first entry has the effect of updating all entries.
 *
 * The panic is there because there is nothing
 * intelligent to be done if an entry won't fit.
 */
timeout(fun, arg, tim)
int (*fun)();
caddr_t arg;
{
	register struct callo *p1, *p2;
	register int t;
	int s;

	t = tim;
	p1 = &callout[0];
	s = spl7();
	while(p1->c_func != 0 && p1->c_time <= t) {
		t -= p1->c_time;
		p1++;
	}
	if (p1 >= &callout[v.v_call-1])
		panic("Timeout table overflow");
	p1->c_time -= t;
	p2 = p1;
	while(p2->c_func != 0)
		p2++;
	if (p2 >= &callout[v.v_call-1])
		panic("Timeout table overflow");
	while(p2 >= p1) {
		(p2+1)->c_time = p2->c_time;
		(p2+1)->c_func = p2->c_func;
		(p2+1)->c_arg = p2->c_arg;
		p2--;
	}
	p1->c_time = t;
	p1->c_func = fun;
	p1->c_arg = arg;
	splx(s);
}

#define	PDELAY	(PZERO-1)
delay(ticks)
{
	int s;
	extern wakeup();

	if (ticks<=0)
		return;

	s = spl6();						/* M000 */
		timeout(wakeup, (caddr_t)u.u_procp+1, ticks);
		sleep((caddr_t)u.u_procp+1, PDELAY);
	splx( s );						/* M000 */
}
