/*
 *	@(#) sys/fio.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: fio.c,v 6.0 87/10/12 17:09:12 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 *
 * $Log:	fio.c,v $
 * Revision 6.0  87/10/12  17:09:12  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.5  87/05/06  15:35:47  iv
 * 	- Added new dev_open() routine [code from closef()] for use
 * 	  by other routines which need to know if a device is open
 * 	- Changed closef() to call dev_open()
 * 
 * Revision 5.4  86/11/21  14:30:20  iv
 * 	- Forgot to iput() when returning EROFS.   Oops.
 * 
 * Revision 5.3  86/11/21  13:42:37  iv
 * 	- Fixed an OOOOOOOOOLD bug in the system where if more than one
 * 	  person had a file open on a read-only filesystem, one was allowed
 * 	  to chmod and chown it.   Ooops.  Added read-only check to owner().
 * 
 * Revision 5.2  86/11/19  10:33:19  iv
 * 	- Added some debugging stuff.  No code changes.
 * 
 * Revision 5.1  86/10/01  18:33:43  iv
 * 	- Fixed bug in closef() where if a mount table entry was in an
 * 	  indeterminate state (ie, in use but not mounted) the close
 * 	  routine would get called anyway.  Fixed the test.
 * 
 * Revision 5.0  85/08/23  10:34:18  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:47:47  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:43:03  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:22:14  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 0.1  84/05/17  15:09:44  iv
 * 	- No code changes.  Added RCS stuff.
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/filsys.h"
#include "../h/file.h"
#include "../h/conf.h"
#include "../h/inode.h"
#include "../h/mount.h"
#include "../h/var.h"
#include "../h/acct.h"

/*#define DEBUG	3	/* Compile in debug code */
#define DbugVar	fio_debug
#define DbugQual  (fio_pid == u.u_procp->p_pid) &&
#ifdef DEBUG
#  include "../h/proc.h"
#  define fioCF	0x0001		/* Debug closef() code */
#  define fioOI	0x0002		/* Debug openi() code */
  short	fio_pid = 0;			/* PID of process to watch */
#endif /*DEBUG*/
#define DEBUGS (fioCF|fioOI)		/* Sections I want to hear from */
#include "../h/dbug.h"


/*
 * Convert a user supplied file descriptor into a pointer
 * to a file structure.
 * Only task is to check range of the descriptor.
 */
struct file *
getf(f)
register int f;
{
	register struct file *fp;

	if(0 <= f && f < NOFILE) {
		fp = u.u_ofile[f];
		if(fp != NULL)
			return(fp);
	}
	u.u_error = EBADF;
	return(NULL);
}

/*
 * Internal form of close.
 * Decrement reference count on file structure.
 * Also make sure the pipe protocol does not constipate.
 *
 * Decrement reference count on the inode following
 * removal to the referencing file structure.
 * On the last close switch out to the device handler for
 * special files.  Note that the handler is called
 * on every open but only the last close.
 */
closef(fp)
register struct file *fp;
{
	register struct inode *ip;
	int flag, fmt;
	dev_t dev;
	register int (*cfunc)();

	if(fp == NULL)
		return;
	dbug2(fioCF,3,"%d: closef(&file[%d])\n", u.u_procp->p_pid, fp - file);
	unlckf(fp->f_inode);
	if ((unsigned)fp->f_count > 1) {
		fp->f_count--;
		dbug1(fioCF,3,"%d closef: f_count > 1\n", u.u_procp->p_pid);
		return;
	}
	ip = fp->f_inode;
	plock(ip);
	flag = fp->f_flag;
	dev = (dev_t)ip->i_rdev;
	fmt = ip->i_mode&IFMT;
	fp->f_count = 0;
	switch(fmt) {

	case IFNAM:
		if(ip->i_fdep.i_namef.i_type == IFSEM)
		    closesem(fp,ip);
		dbug1(fioCF,3,"%d closef: IFNAM\n", u.u_procp->p_pid);
		iput(ip);
		return;

	case IFCHR:
		dbug3(fioCF,3,"%d closef: IFCHR(%d/%d)\n",
				u.u_procp->p_pid,
				major(dev),
				minor(dev));
		cfunc = cdevsw[major(dev)].d_close;
		break;

	case IFBLK:
		dbug3(fioCF,3,"%d closef: IFBLK(%d/%d)\n",
				u.u_procp->p_pid,
				major(dev),
				minor(dev));
		cfunc = bdevsw[major(dev)].d_close;
		break;

	case IFIFO:
		dbug1(fioCF,3,"%d closef: IFIFO\n", u.u_procp->p_pid);
		closep(ip, flag);

	default:
		dbug1(fioCF,3,"%d closef: default\n", u.u_procp->p_pid);
		goto out;
	}

	if (dev_open(dev, fmt))
		goto out;

	if (fmt == IFBLK) {
		register struct mount *mp;

		dbug1(fioCF,3,"%d closef: IFBLK mntchk\n", u.u_procp->p_pid);
		for (mp = mount; mp < (struct mount *)v.ve_mount; mp++)
			if (mp->m_flags != MFREE && mp->m_dev == dev)
				goto out;
		bflush(dev);
		(*cfunc)(minor(dev), flag, 1);
		dbug3(fioCF,3,"%d closef: b close(%d/%d)\n", 
				u.u_procp->p_pid,
				major(dev),
				minor(dev));
		binval(dev);
	} else {
		prele(ip);
		dbug3(fioCF,3,"%d closef: c close(%d/%d)\n", 
				u.u_procp->p_pid,
				major(dev),
				minor(dev));
		(*cfunc)(minor(dev), flag, 0);
	}
out:
	iput(ip);
}

dev_open(dev, fmt)
  dev_t	dev;
  int	fmt;
{
	register struct file	*fp;

	/* Avoid violating open/close protocol:
	 * rootdev/pipedev/swapdev implicitly ALWAYS open */
	if (dev == rootdev || dev == pipedev || dev == swapdev)
		return (1);

	for (fp = file; fp < (struct file *)v.ve_file; fp++) {
		register struct inode *tip;

		if (fp->f_count) {
			tip = fp->f_inode;
			if (tip->i_rdev == dev && (tip->i_mode&IFMT) == fmt)
				return (1);
		}
	}

	return (0);
}

/*
 * openi called to allow handler of special files to initialize and
 * validate before actual IO.
 */
openi(ip, flag)
register struct inode *ip;
{
	dev_t dev;
	register unsigned int maj;

	dev = (dev_t)ip->i_rdev;
	maj = major(dev);
	switch(ip->i_mode&IFMT) {

	case IFCHR:
		if(maj >= cdevcnt)
			goto bad;
		if (u.u_ttyp == NULL)
			u.u_ttyd = dev;
		(*cdevsw[maj].d_open)(minor(dev), flag, 0);
		break;

	case IFBLK:
		if(maj >= bdevcnt)
			goto bad;
		(*bdevsw[maj].d_open)(minor(dev), flag, 1);
		break;

	case IFIFO:
		openp(ip, flag);
		break;
	}
	return;

bad:
	u.u_error = ENXIO;
}

/*
 * Check mode permission on inode pointer.
 * Mode is READ, WRITE or EXEC.
 * In the case of WRITE, the read-only status of the file
 * system is checked. Also in WRITE, prototype text
 * segments cannot be written.
 * The mode is shifted to select the owner/group/other fields.
 * The super user is granted all permissions.
 */
access(ip, mode)
register struct inode *ip;
{
	register m;

	m = mode;
	if(m == IWRITE) {
		if(getfs(ip->i_dev)->s_ronly != 0) {
			u.u_error = EROFS;
			return(1);
		}
		if (ip->i_flag&ITEXT)
			xrele(ip);
		if(ip->i_flag & ITEXT) {
			u.u_error = ETXTBSY;
			return(1);
		}
	}
	if(u.u_uid == 0)
		return(0);
	if(u.u_uid != ip->i_uid) {
		m >>= 3;
		if(u.u_gid != ip->i_gid)
			m >>= 3;
	}
	if((ip->i_mode&m) != 0)
		return(0);

	u.u_error = EACCES;
	return(1);
}

/*
 * Look up a pathname and test if the resultant inode is writable.
 * Then look to see if it is owned by the current user.
 * If not, try for super-user.
 * If writable and permission is granted, return inode pointer.
 */
struct inode *
owner()
{
	register struct inode *ip;

	ip = namei(uchar, 0);
	if(ip == NULL)
		return(NULL);
	if (getfs(ip->i_dev)->s_ronly != 0)
		u.u_error = EROFS;
	else {
		if(u.u_uid == ip->i_uid)
			return(ip);
		if(suser())
			return(ip);
	}
	iput(ip);
	return(NULL);
}

/*
 * Test if the current user is the super user.
 */
suser()
{

	if(u.u_uid == 0) {
		u.u_acflag |= ASU;
		return(1);
	}
	u.u_error = EPERM;
	return(0);
}

/*
 * Allocate a user file descriptor.
 */
ufalloc(i)
register i;
{

	for(; i<NOFILE; i++)
		if(u.u_ofile[i] == NULL) {
			u.u_rval1 = i;
			u.u_pofile[i] = 0;
			return(i);
		}
	u.u_error = EMFILE;
	return(-1);
}

/*
 * Allocate a user file descriptor and a file structure.
 * Initialize the descriptor to point at the file structure.
 *
 * no file -- if there are no available file structures.
 */
struct file *
falloc()
{
	register struct file *fp;
	register i;

	i = ufalloc(0);
	if(i < 0)
		return(NULL);
	for(fp = &file[0]; fp < (struct file *)v.ve_file; fp++)
		if(fp->f_count == 0) {
			u.u_ofile[i] = fp;
			fp->f_count++;
			fp->f_offset = 0;
			return(fp);
		}
	printf("no file\n");
	u.u_error = ENFILE;
	return(NULL);
}
