/*
 *	@(#) sys/iget.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: iget.c,v 6.0 87/10/12 17:09:17 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */

/*
 * MODIFICATION HISTORY:
 *
 *	21 Feb 1984	cliff	M000
 *		- add missing synchronous-write flag to writei() calls.
 *	21 March 1984	blf	M001
 *		- As per PC XENIX 3.0, change l3tol() to il3tol()
 *		  and ltol3() to iltol3() to avoid libc conflicts.
 *
 * $Log:	iget.c,v $
 * Revision 6.0  87/10/12  17:09:17  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/08/23  10:34:30  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:47:53  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:43:16  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:22:20  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 0.1  84/05/17  15:10:08  iv
 * 	- Missing synchronous-write argument was added to writei() calls
 * 	- As per PC XENIX 3.0, l3tol() was changed to il3tol(),
 * 	  and ltol3() was changed to iltol3() to aviod libc conflicts
 * 	  [which is funny, because we didn't have any problem linking
 * 	   the kernel using the 68000 sys3 libc.
 * 	   Oh, well.  Hope they know what they're doing...]
 * 
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/sysinfo.h"
#include "../h/mount.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/inode.h"
#include "../h/ino.h"
#include "../h/filsys.h"
#include "../h/buf.h"
#include "../h/var.h"

/*
 * Look up an inode by device,inumber.
 * If it is in core (in the inode structure), honor the locking protocol.
 * If it is not in core, read it in from the specified device.
 * If the inode is mounted on, perform the indicated indirection.
 * In all cases, a pointer to a locked inode structure is returned.
 *
 * printf warning: no inodes -- if the inode structure is full
 * panic: no imt -- if the mounted filesystem is not in the mount table.
 *	"cannot happen"
 */
struct inode *
iget(dev, ino)
dev_t dev;
ino_t ino;
{
	register i;
	register struct inode *ip;
	register struct inode *oip;
	register struct mount *mp;
	struct inode *iread();

	sysinfo.iget++;
loop:
	ip = &inode[0];
	oip = NULL;
	i = v.v_inode;
	do {
		if(ino == ip->i_number && dev == ip->i_dev) {
			goto found;
		}
		if(oip==NULL && ip->i_count==0)
			oip = ip;
		ip++;
	} while(--i);
	ip = oip;
	if(ip == NULL) {
		printf("Inode table overflow\n");
		u.u_error = ENFILE;
		return(NULL);
	}
	ip->i_dev = dev;
	ip->i_number = ino;
	ip->i_flag = ILOCK;
	ip->i_count++;
	ip->i_lastr = 0;
	return(iread(ip));
found:
	if((ip->i_flag&ILOCK) != 0) {
		ip->i_flag |= IWANT;
		sleep((caddr_t)ip, PINOD);
		goto loop;
	}
	if((ip->i_flag&IMOUNT) != 0) {
		for(mp = &mount[0]; mp < (struct mount *)v.ve_mount; mp++)
		if(mp->m_inodp == ip) {
			dev = mp->m_dev;
			ino = ROOTINO;
			goto loop;
		}
		panic("no imt");
	}
	ip->i_count++;
	ip->i_flag |= ILOCK;
	return(ip);
}

struct inode *
iread(ip)
register struct inode *ip;
{
	char *p1, *p2;
	struct buf *bp;
	unsigned i;

	bp = bread(ip->i_dev, itod(ip->i_number));
	if (u.u_error) {
		brelse(bp);
		iput(ip);
		return(NULL);
	}
	i = itoo(ip->i_number)*sizeof(struct dinode);
	copyio(paddr(bp)+i, &ip->i_mode,
	    sizeof(struct dinode)-3*sizeof(time_t), U_RKD);
	p1 = (char *)&ip->i_addr[NADDR];
	p2 = (char *)(&ip->i_addr[0]) + 3*NADDR;
	for(i=NADDR; i; i--) {
		il3tol(&p1,&p2);
	}
	brelse(bp);
	return(ip);
}

il3tol(rp1,rp2)			/* NON PORTABLE */
    char **rp1, **rp2;
{
    register char *p1, *p2;
	p1 = *rp1; p2 = *rp2;
#ifdef WORDSWAP			/* WORDSWAP code untested! */
#  ifdef BYTESWAP
	*--p1 = *--p2;		/* words & bytes swapped */
	*--p1 = 0;
	*--p1 = *--p2;
	*--p1 = *--p2;
#  else
	*--p1 = 0;		/* words swapped */
	*--p1 = *--p2;
	*--p1 = *--p2;
	*--p1 = *--p2;
#  endif
#else
#  ifdef BYTESWAP
	*--p1 = *--p2;		/* bytes swapped */
	*--p1 = *--p2;
	*--p1 = *--p2;
	*--p1 = 0;
#  else
	*--p1 = *--p2;		/* neither */
	*--p1 = *--p2;
	*--p1 = 0;
	*--p1 = *--p2;
#  endif
#endif
	*rp1=p1; *rp2=p2;
}

/*
 * Decrement reference count of an inode structure.
 * On the last reference, write the inode out and if necessary,
 * truncate and deallocate the file.
 */
iput(ip)
register struct inode *ip;
{

	if(ip->i_count == 1) {
		ip->i_flag |= ILOCK;
		if(ip->i_nlink <= 0) {
			itrunc(ip,0L);
			ip->i_mode = 0;
			ip->i_flag |= IUPD|ICHG;
			ifree(ip->i_dev, ip->i_number);
		}
		if(ip->i_flag&(IACC|IUPD|ICHG))
			iupdat(ip, &time, &time,0);
		prele(ip);
		ip->i_flag = 0;
		ip->i_number = 0;
		ip->i_count = 0;
		return;
	}
	ip->i_count--;
	prele(ip);
}

/*
 * Update the inode with the current time.
 */
iupdat(ip, ta, tm,syncf)
register struct inode *ip;
time_t *ta, *tm;
int syncf;
{
	register struct buf *bp;
	struct dinode *dp;
	char *p1, *p2;
	unsigned i;

	if(getfs(ip->i_dev)->s_ronly) {
		if(ip->i_flag&(IUPD|ICHG))
			u.u_error = EROFS;
		ip->i_flag &= ~(IACC|IUPD|ICHG);
		return;
	}
	bp = bread(ip->i_dev, itod(ip->i_number));
	if (bp->b_flags & B_ERROR) {
		brelse(bp);
		return;
	}
	dp = (struct dinode *)&u.u_t[0];
	dp->di_mode = ip->i_mode;
	dp->di_nlink = ip->i_nlink;
	dp->di_uid = ip->i_uid;
	dp->di_gid = ip->i_gid;
	dp->di_size = ip->i_size;
	p1 = (char *)dp->di_addr;
	p2 = (char *)ip->i_addr;
	if ((ip->i_mode&IFMT)==IFIFO) {
		for (i=0; i<NFADDR; i++) {
			iltol3(ip,&p1,&p2);
		}
		for (; i<NADDR; i++) {
			*p1++ = 0;
			*p1++ = 0;
			*p1++ = 0;
		}
	} else
		for(i=0; i<NADDR; i++) {
			iltol3(ip,&p1,&p2);
	}
	dp = &((struct dinode *)0)[itoo(ip->i_number)];
	copyio(paddr(bp)+(unsigned)dp, &u.u_t[0],
	    sizeof(struct dinode)-3*sizeof(time_t), U_WKD);
	if(ip->i_flag&IACC)
		biputl(bp, (int)&dp->di_atime, *ta);
	if(ip->i_flag&IUPD)
		biputl(bp, (int)&dp->di_mtime, *tm);
	if(ip->i_flag&ICHG)
		biputl(bp, (int)&dp->di_ctime, time);
	ip->i_flag &= ~(IACC|IUPD|ICHG);
	if(syncf) bswrite(bp);
	else bdwrite(bp);
}


iltol3(ip,rp1,rp2)                 /* NON PORTABLE */
    struct inode *ip;
    char **rp1, **rp2;
{
    register char *p1,*p2;
	p1 = *rp1; p2 = *rp2;
#ifdef WORDSWAP				/* WORDSWAP code untested! */
#  ifdef BYTESWAP			/* words & bytes swapped */
	*p1++ = *p2++;
	*p1++ = *p2++;
	if(*p2++ != 0 && (ip->i_mode&IFMT)!=IFNAM && (ip->i_nlink > 0))
		printf("iaddress > 2^24\n");
	*p1++ = *p2++;
#  else					/* words swapped */
	*p1++ = *p2++;
	*p1++ = *p2++;
	*p1++ = *p2++;
	if(*p2++ != 0 && (ip->i_mode&IFMT)!=IFNAM && (ip->i_nlink > 0))
		printf("iaddress > 2^24\n");
#  endif
#else
#  ifdef BYTESWAP			/* bytes swapped */
	if(*p2++ != 0 && (ip->i_mode&IFMT)!=IFNAM && (ip->i_nlink > 0))
		printf("iaddress > 2^24\n");
	*p1++ = *p2++;
	*p1++ = *p2++;
	*p1++ = *p2++;
#  else					/* neither */
	*p1++ = *p2++;
	if(*p2++ != 0 && (ip->i_mode&IFMT)!=IFNAM && (ip->i_nlink > 0))
		printf("iaddress > 2^24\n");
	*p1++ = *p2++;
	*p1++ = *p2++;
#  endif
#endif
	*rp1=p1; *rp2=p2;
}


/*
 * Free all the disk blocks associated with the specified inode structure.
 * The blocks of the file are removed in reverse order. This FILO
 * algorithm will tend to maintain
 * a contiguous free list much longer than FIFO.
 * Don't actually free a blk unless its
 * logical block number is >= the starting logical blk # of the truncation
 * Still scans the whole file even if truncation is not to zero.
 */
itrunc(ip,size)
register struct inode *ip;
off_t size;
{
	register i;
	dev_t dev;
	daddr_t bn;
	daddr_t slbn;  /* starting logical blk # of trunc*/
	daddr_t blbn;  /* begining logical blk # currently at*/

	slbn = (size+BMASK)>>BSHIFT;
	i = ip->i_mode & IFMT;
	if (i!=IFREG && i!=IFDIR)
		return;
	dev = ip->i_dev;
	for(i=NADDR-1; i>=0; i--) {
		bn = ip->i_addr[i];
		if(bn == (daddr_t)0)
			continue;
		switch(i) {

		default:
			blbn = i;
			if(slbn<=blbn)
			    free(dev, bn);
			break;

		case NADDR-3:
			blbn = NADDR-3;
			tloop(dev, bn, 0, 0, blbn, slbn);
			break;

		case NADDR-2:
			blbn = NADDR-3+NINDIR;
			tloop(dev, bn, 1, 0, blbn, slbn);
			break;

		case NADDR-1:
			blbn = NADDR-3+NINDIR*NINDIR;
			tloop(dev, bn, 1, 1, blbn, slbn);
		}
		if(slbn<=blbn)
		    ip->i_addr[i] = (daddr_t)0;
	}
	ip->i_size = size;
	ip->i_flag |= IUPD|ICHG;
}

tloop(dev, bn, f1, f2, iblbn, slbn)
dev_t dev;
daddr_t bn;
daddr_t iblbn, slbn;
{
	register i;
	register struct buf *bp;
	daddr_t nb;
	daddr_t blbn;

	blbn = iblbn;
	bp = NULL;
	for(i=NINDIR-1; i>=0; i--) {
		if(f2!=0)blbn=iblbn+i*NINDIR*NINDIR;
		else if(f1!=0)blbn=iblbn+i*NINDIR;
		else blbn=iblbn+i;

		if(bp == NULL) {
			bp = bread(dev, bn);
			if (bp->b_flags & B_ERROR) {
				brelse(bp);
				return;
			}
		}
		nb = bigetl(bp, i*sizeof(daddr_t));
		if(nb == (daddr_t)0)
			continue;
		/* zap indirect pointer if necessary */
		if(slbn<=blbn)
		    biputl(bp, i*sizeof(daddr_t),0L);

		if(f1) {
			brelse(bp);
			bp = NULL;
			tloop(dev, nb, f2, 0, blbn, slbn);
		} else
		    if(slbn<=blbn)
			free(dev, nb);
	}
	if(bp != NULL)
		brelse(bp);
	if(slbn<=iblbn)
	    free(dev, bn);
}

/*
 * Make a new file.
 */
struct inode *
maknode(mode)
register mode;
{
	register struct inode *ip;

	ip = ialloc(u.u_pdir->i_dev);
	if(ip == NULL) {
		iput(u.u_pdir);
		return(NULL);
	}
	ip->i_flag |= IACC|IUPD|ICHG;
	if((mode&IFMT) == 0)
		mode |= IFREG;
	ip->i_mode = mode & ~u.u_cmask;
	ip->i_nlink = 1;
	ip->i_uid = u.u_uid;
	ip->i_gid = u.u_gid;
	iupdat(ip, &time, &time,0);
	wdir(ip);
	return(ip);
}

/*
 * Write a directory entry with parameters left as side effects
 * to a call to namei.
 */
wdir(ip)
struct inode *ip;
{

	u.u_dent.d_ino = ip->i_number;
	u.u_count = sizeof(struct direct);
	u.u_segflg = 1;
	u.u_base = (caddr_t)&u.u_dent;
	writei(u.u_pdir, 0);				/* M000 non-sync */
	iput(u.u_pdir);
}
