/*
 *	@(#) sys/prf.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: /usr/sys/sys/RCS/prf.c,v 6.0 87/10/12 17:09:39 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */

/*
 * MODIFICATION HISTORY
 *
 *	M000	25 Jan 1984	cliff
 *		- prevent infinite recursive panic 
 *	M001	26 Jan 1984	cliff
 *		- skip the linecount bullshit for kernel printf
 *		deleted using #ifdef NEVER
 *	23 Feb 1984	blf	M002
 *		- Print out device status & command in deverr() in hex,
 *		  not (unsigned) decimal!
 *	21 March 1984	blf	M003
 *		- Added %l[dx], %c formats to printf(),
 *		  as per PC XENIX 3.0.
 *		- Added some lint(1) directives.
 *
 * $Log:	prf.c,v $
 * Revision 6.0  87/10/12  17:09:39  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/08/23  10:35:49  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:48:36  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:44:00  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:22:57  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 0.1  84/05/17  15:15:07  iv
 * 	- Code added to prevent infinite, recursive panic
 * 	- "#ifdef NEVER" linecount garbage for kernel printf
 * 	- Print out device status and command in HEX, not
 * 	  unsigned decimal at deverr()
 * 	- Formats %l[dx] were added to printf()
 * 	- Some lint directives were added
 * 
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/buf.h"
#include "../h/iobuf.h"

/*
 * In case console is off,
 * panicstr contains argument to last call to panic.
 */
char	*panicstr = (char *)0;				/*M000*/

/*
 * Scaled down version of C Library printf.
 * Only %c %s %u %d (==%u) %o %x %D %X are recognized.
 * Used to print diagnostic information
 * directly on console tty.
 * When called with `putchar' it is not interrupt driven,
 * all system activities are pretty much suspended.
 * Printf should not be used for chit-chat.
 *
 *  VARARGS1
 */
printf(fmt, x1)
  char *fmt;
  unsigned x1;
{
	extern int	putchar();

	_printf(fmt, &x1, putchar);
}
fprintf(putcfn, fmt, x1)
  int	(*putcfn)();
  char *fmt;
  unsigned x1;
{
	_printf(fmt, &x1, putcfn);
}
_printf(fmt, adx, putcfn)
  register char *fmt;
  register unsigned int *adx;
  register int	(*putcfn)();
{
	register c;
	char *s;
loop:
	while ((c = *fmt++) != '%') {
		if (c == '\0')
		    return;
		(*putcfn)(c);
	}
	if ((c = *fmt++) == 'l') {			/* M003 begin... */
		if ((c = *fmt++) == 'd' || c == 'x') {
			c += 'A' - 'a';
		}
	}						/* ...end M003 */
	if (c == 'd' || c == 'u' || c == 'o' || c == 'x')
		printn((long)*adx, c=='o'? 8: (c=='x'? 16:10), 0, putcfn);
	else if (c == 's') {
		s = (char *)*adx;
		while (c = *s++)
			(*putcfn)(c);
	} else if (c == 'D' || c == 'X') {
		printn(*(long *)adx, c=='D'?10:16, 1, putcfn);
		adx += (sizeof(long) / sizeof(int)) - 1;
	}
	else if (c == 'c')				/* M003 begin... */
		(*putcfn)(*adx & 0xFF);
	else {
		(*putcfn)(c);
		adx--;
	}						/* ...end M003 */
	adx++;
	goto loop;
}

/* print value "n" in base "b". If "longf" is 0, "n" is only a short. */
/* print leading zeros for octal & hex, but not for decimal. */
printn(n, b, longf, putcfn)
  long n;
  register b;
  int longf;
  register int	(*putcfn)();
{
	register i, nd, c;
	int	flag;
	int	plmax;
	char d[12];

	c = 1;
	flag = n < 0;
	if (flag)
		n = (-n);
	if (b==16) plmax = longf?8:4;
	else plmax = longf?11:6;
	if (flag && b==10) {
		flag = 0;
		(*putcfn)('-');
	}
	/* build digit values in an array */
	for (i=0;i<plmax;i++) {
		if(n==0 && b==10 && i>0) break;
		nd = n%b;
		if (flag) {
			nd = (b - 1) - nd + c;
			if (nd >= b) {
				nd -= b;
				c = 1;
			} else
				c = 0;
		}
		d[i] = nd;
		n = n/b;
	}
	/* now print all we generated out */
	for (n=i-1;n>=0;n--) {
		(*putcfn)("0123456789ABCDEF"[d[n]]);
	}
}

/*
 * Panic is called on unresolvable fatal errors.
 * It prints "panic: mesg", syncs, and then loops.
 */
panic(s)
char *s;
{
	if (panicstr)						/*M000*/
		printf("recursive panic: %s\n", s);		/*M000*/
	else {							/*M000*/
		panicstr = s;
		printf("panic: %s\n", s);
#ifdef DEBUG
		debug(0);
#endif
		update();
	}
	for(;;)
		haltcpu();					/*M000*/
}

/*
 * prdev prints a warning message of the
 * form "mesg on dev x/y".
 * x and y are the major and minor parts of
 * the device argument.
 */
prdev(str, dev)
char *str;
dev_t dev;
{
	printf("%s on dev %u/%u\n", str, major(dev), minor(dev));
}

/*
 * Deverr prints a diagnostic from a device driver.
 * It prints: error on device name (major/minor), block number,
 * and two arguments, usually error status.
 */
deverr(dp, o1, o2, dn)
    struct iobuf *dp;
    char *dn;		/* device name */
{
    register struct buf *bp;

	bp=dp->b_actf;
	printf("error on dev %s (%u/%u)",dn,major(bp->b_dev),minor(bp->b_dev));
	printf(", block=%D cmd=%x status=%x\n", bp->b_blkno, o1, o2);  /*M002*/
}
