/*
 *	@(#) sys/sys1.c 1.2 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: sys1.c,v 6.0 87/10/12 17:10:03 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */

/*
 * MODIFICATION HISTORY
 *
 *	M000	25 Jan 1984	cliff
 *		- In exit(), adjust u_ssize, u_dsize
 *		upon return from expand(USIZE), then do mmuset()
 *		to sync the process mapping.  This is necessary since
 *		expand() leaves this up to the caller, and xfree()
 *		may send us through sleep() & swtch() & mmuset(),
 *		which will remap the proc using the u_[sd]size values.
 *
 * $Log:	sys1.c,v $
 * Revision 6.0  87/10/12  17:10:03  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/08/23  10:36:46  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:49:36  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:44:36  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:23:29  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.4  84/12/17  13:45:15  iv
 * 	- Changed swapmap check from a check for "maxmem" swap blocks
 * 	  to "procp->p_size" blocks, as otherwise we cannot run on floppy
 * 
 * Revision 1.3  84/10/09  15:38:23  iv
 * 	- Removed SCO's M000 and the expand(USIZE) call before it.
 * 	  Historically, the exit sequence has always had an expand(USIZE) 
 * 	  to "shrink the process back to a naked U page", as the saying goes.
 * 	  However, when uSoft put in their MMU code they added an mmupfree
 * 	  call a few lines later, but left in the expand.  This was not
 * 	  a good move.  If the call chain xfree-iput-iupdat-bread-iowait-
 * 	  sleep-swtch-(after wakeup)-mmumap-setmapregs (!!!) occurs, the
 * 	  pagelist will be WAY TOO SHORT, and boom...
 * 
 * Revision 1.2  84/05/17  17:19:20  iv
 * 	- Got rid of PC XENIX only code
 * 
 * Revision 1.1  84/05/17  16:08:40  iv
 * 	- Added missing FIXEDSTACK code for setting up user stack
 * 	  using u.u_stktop
 * 
 * Revision 0.1  84/05/17  15:30:45  iv
 * 	- In exit(), u_ssize and u_dsize are adjusted
 * 	  upon return from expand(USIZE), then mmuset() is called 
 * 	  to sync the process mapping.  This is necessary since
 * 	  expand() leaves this up to the caller, and xfree() may
 * 	  send us through sleep(), swtch(), and mmuset(); which
 * 	  will remap the proc using the u_[sd]size values.
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/map.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/buf.h"
#include "../h/reg.h"
#include "../h/inode.h"
#include "../h/mmu.h"
#include "../h/acct.h"
#include "../h/var.h"
#include "../h/tty.h"

/*
 * exec system call, with and without environments.
 */

struct execa {
	char	*fname;
	char	**argp;
	char	**envp;
};

exec()
{
	((struct execa *)u.u_ap)->envp = NULL;
	exece();
}

#define	NCABLK	(NCARGS+BSIZE-1)/BSIZE
exece()
{
	register unsigned nc;
	register int cp;
	register struct buf *bp;
	register struct execa *uap;
	int na, ne, ucp, ap, c, sep, overlay;
	unsigned bno;
	struct inode *ip;
	extern struct inode *gethead();

	if ((ip = gethead()) == NULL)
		return;
	bp = 0;
	na = nc = ne = 0;
	uap = (struct execa *)u.u_ap;
	/* collect arglist */
	if ((bno = malloc(swapmap,NCABLK)) == 0)
		panic("Out of swap");
	if (uap->argp) for (;;) {
		ap = NULL;
		if (uap->argp) {
			ap = fuword((caddr_t)uap->argp);
			uap->argp++;
		}
		if (ap==NULL && uap->envp) {
			uap->argp = NULL;
			if ((ap = fuword((caddr_t)uap->envp)) == NULL)
				break;
			uap->envp++;
			ne++;
		}
		if (ap==NULL)
			break;
		na++;
		if (ap == -1)
			u.u_error = EFAULT;
		do {
			if (nc >= NCARGS-1)
				u.u_error = E2BIG;
			if ((c = fubyte((caddr_t)ap++)) < 0)
				u.u_error = EFAULT;
			if (u.u_error)
				goto bad;
			if ((nc&BMASK) == 0) {
				if (bp)
					bdwrite(bp);
				bp = getblk(swapdev, swplo+bno+(nc>>BSHIFT));
				cp = 0;
			}
			nc++;
			biputc(bp, cp++, c);
		} while (c>0);
	}
	if (bp)
		bdwrite(bp);
	bp = 0;
	nc = (nc + NBPW-1) & ~(NBPW-1);
	if(getxfile(ip, nc + sizeof(char *)*na)!=0)
		goto bad;       /* Not really bad, just an overlay */
	if (u.u_error) {
		psignal(u.u_procp, SIGKILL);
		goto bad;
	}

	/* copy back arglist */
	/* USRSTACK is the starting address of the users stack.  This */
	/* typically is some large constant, the end of the text segment */
	/* or zero depending on the machine.  The value URSBC is a fixup */
	/* for the # of bytes of registers that get pushed on the users */
	/* stack during an interrupt, currently used on the 8086 only */

#ifdef FIXEDSTACK
	ucp = (int) (u.u_stktop - nc - SHWDSZ);   /* address of 1st arg */
#else FIXEDSTACK
	ucp = (int) (USRSTACK - nc - SHWDSZ);   /* address of 1st arg */
#endif FIXEDSTACK
	ap = ucp - na*NBPW - 3*NBPW;    	/* address of argc */
	u.u_ar0[USP] = ap - URSBC;
	suword((caddr_t)ap, na-ne);
	nc = 0;
	for (;;) {
		ap += NBPW;
		if (na==ne) {
			ap += NBPW;     /* null terminate argv */
		}
		if (--na < 0)
			break;
		suword((caddr_t)ap, ucp);
		do {
			if ((nc&BMASK) == 0) {
				if (bp)
					brelse(bp);
				bp = bread(swapdev, swplo+bno+(nc>>BSHIFT));
				bp->b_flags |= B_AGE;
				bp->b_flags &= ~B_DELWRI;
				cp = 0;
			}
			subyte((caddr_t)ucp++, (c = bigetc(bp, cp++)));
			nc++;
		} while (c&0377);
	}
	if((ucp&1)!=0) ucp++;           /* pad to even boundary */
	cprite(ucp);
	setregs();
	if (bp)
		brelse(bp);
	iput(ip);
	mfree(swapmap, NCABLK, bno);
	return;
bad:
	if (bp)
		brelse(bp);
	iput(ip);
	for (nc = 0; nc < NCABLK; nc++) {
		bp = getblk(swapdev, swplo+bno+nc);
		bp->b_flags |= B_AGE;
		bp->b_flags &= ~B_DELWRI;
		brelse(bp);
	}
	mfree(swapmap, NCABLK, bno);
}

/*
 * exit system call:
 * pass back caller's arg
 */
rexit()
{
	register struct a {
		int	rval;
	} *uap;

	uap = (struct a *)u.u_ap;
	exit((uap->rval & 0377) << 8);
}

/*
 * Release resources.                           NON PORTABLE
 * Enter zombie state.
 * Wake up parent and init processes,
 * and dispose of children.
 */
exit(rv)
{
	register int i;
	register struct proc *p, *q;
	extern short maxttys;

	p = u.u_procp;
	p->p_flag &= ~(STRC);
	p->p_clktim = 0;
	for (i=0; i<NSIG; i++)
		u.u_signal[i] = 1;

	/* M000 was here, but is no more... */

	if ((p->p_pid == p->p_pgrp)
	 && (u.u_ttyp != NULL)
	 && (u.u_ttyp->t_pgrp == p->p_pgrp)) {
		u.u_ttyp->t_pgrp = 0;
		maxttys--;
		signal(p->p_pgrp, SIGHUP);
	}
	p->p_pgrp = 0;
	for (i=0; i<NOFILE; i++) {
		if (u.u_ofile[i] != NULL)
			closef(u.u_ofile[i]);
	}
	plock(u.u_cdir);
	iput(u.u_cdir);
	if (u.u_rdir) {
		plock(u.u_rdir);
		iput(u.u_rdir);
	}
	xfree();
	acct(rv);
	sdexit();	/* free up attached shared data segments */
	mmupfree(p);                                    /*NONP*/
	p->p_stat = SZOMB;
	((struct xproc *)p)->xp_xstat = rv;
	((struct xproc *)p)->xp_utime = u.u_cutime + u.u_utime;
	((struct xproc *)p)->xp_stime = u.u_cstime + u.u_stime;
	for (q = &proc[1]; q < (struct proc *)v.ve_proc; q++) {
		if (p->p_pid == q->p_ppid) {
			q->p_ppid = 1;
			if (q->p_stat == SZOMB)
				psignal(&proc[1], SIGCLD);
			if (q->p_stat == SSTOP)
				setrun(q);
		} else
		if (p->p_ppid == q->p_pid)
			psignal(q, SIGCLD);
		if (p->p_pid == q->p_pgrp)
			q->p_pgrp = 0;
	}
	resume(proc[0].p_addr.p_caddr, u.u_qsav);
	/* no deposit, no return */
}

/*
 * Wait system call.
 * Search for a terminated (zombie) child,
 * finally lay it to rest, and collect its status.
 * Look also for stopped (traced) children,
 * and pass back status from them.
 */
wait()
{
	register f;
	register struct proc *p;

loop:
	f = 0;
	for (p = &proc[1]; p < (struct proc *)v.ve_proc; p++)
	if (p->p_ppid == u.u_procp->p_pid) {
		f++;
		if (p->p_stat == SZOMB) {
			freeproc(p, 1);
			return;
		}
		if (p->p_stat == SSTOP) {
			if ((p->p_flag&SWTED) == 0) {
				p->p_flag |= SWTED;
				u.u_rval1 = p->p_pid;
				u.u_rval2 = (fsig(p)<<8) | 0177;
				return;
			}
			continue;
		}
	}
	if (f) {
		sleep((caddr_t)u.u_procp, PWAIT);
		goto loop;
	}
	u.u_error = ECHILD;
}

/*
 * Remove zombie children from the process table.
 */
freeproc(p, flag)
register struct proc *p;
{
	if (flag) {
		u.u_rval1 = p->p_pid;
		u.u_rval2 = ((struct xproc *)p)->xp_xstat;
	}
	u.u_cutime += ((struct xproc *)p)->xp_utime;
	u.u_cstime += ((struct xproc *)p)->xp_stime;
	p->p_stat = NULL;
	p->p_pid = 0;
	p->p_ppid = 0;
	p->p_sig = 0;
	p->p_flag = 0;
	p->p_wchan = 0;
}

/*
 * fork system call.  NON PORTABLE
 */
fork()
{
	register n;
	register struct proc *p1, *p2;
	register a;

	/*
	 * Make sure there's enough swap space for another core
	 * image our size, thus reducing chances of running out
	 */
	n = mstod(u.u_procp->p_size);
	if ((a = malloc(swapmap, n)) == 0) {
		u.u_error = ENOMEM;
		goto out;
	}
	mfree(swapmap, n, a);
	a = 0;
	p1 = &proc[0];
	p2 = NULL;
	n = v.v_proc;
	do {
		if (p1->p_stat==NULL && p2==NULL)
			p2 = p1;
		else {
			if (p1->p_uid==u.u_uid && p1->p_stat!=NULL)
				a++;
		}
		p1++;
	} while (--n);
	/*
	 * Disallow if
	 *  No processes at all;
	 *  not su and too many procs owned; or
	 *  not su and would take last slot.
	 */
	if (p2==NULL || (u.u_uid!=0 && (p2==&proc[v.v_proc-1] || a>v.v_maxup))) {
		u.u_error = EAGAIN;
		goto out;
	}
	if (newproc()) {
		u.u_rval1 = p2->p_ppid;
#ifndef OLDFORK
		u.u_rval2 = 0;
#endif
		u.u_start = time;
		u.u_ticks = lbolt;
		u.u_mem = p2->p_size;
		u.u_ior = u.u_iow = u.u_ioch = 0;
		u.u_cstime = 0;
		u.u_stime = 0;
		u.u_cutime = 0;
		u.u_utime = 0;
		u.u_acflag = AFORK;
		mmuset(RO);
		/* handle inheritance of semaphores WARNING - the child */
		/* process can return with out of file structures error */
		if (passem() == -1) goto out;
		return;
	}
	u.u_rval1 = p2->p_pid;

out:
#ifndef OLDFORK
	u.u_rval2 = 1;
#else
	u.u_ar0[UPC] += 2;
#endif
}

lock()
{
	register struct proc *p;
	register struct a {
		int	flag;
	} *uap;

	uap = (struct a *)u.u_ap;
	if(suser()) {
		p = u.u_procp;
		p->p_flag &= ~SSYS;
		if(uap->flag)
			p->p_flag |= SSYS;
	}
}
