/*
 *	@(#) sys/sys2.c 1.3 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: sys2.c,v 6.0 87/10/12 17:10:08 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */


/*
 *  MODIFICATION HISTORY
 *	21 March 1984	blf	M000
 *		- Pass only the minor dev num to driver routines,
 *		  as per PC XENIX 3.0.
 *
 * $Log:	sys2.c,v $
 * Revision 6.0  87/10/12  17:10:08  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.1  86/11/19  10:28:49  iv
 * 	- Found a bug in copen() where after falloc() was called to allocate
 * 	  a file entry, it was left unintialized and itrunc() was called,
 * 	  which can go to sleep!  Any process which wakes up later and examines
 * 	  the file table, eg via closef() will read the PREVIOUS contents of
 * 	  that file entry and act accordingly.  In one case a device driver
 * 	  close routine was failing to be called because closef() "saw" it
 * 	  in the open file table.   *SIGH*
 * 
 * Revision 5.0  85/08/23  10:36:57  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.1  85/03/24  13:37:30  iv
 * 	- Fixed bug in rdchk() where if a rdchk() on a pipe was done, and
 * 	  an EOF was lurking there rdchk() would incorrectly claim that
 * 	  the read would block, when it would actually return EOF.
 * 	- Added FNDELAY check to rdchk(), so that it knows about no-wait
 * 	  reads.
 * 
 * Revision 4.0  85/03/13  14:49:47  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:44:41  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:23:37  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 0.1  84/05/17  15:47:16  iv
 * 	- Pass only the minor dev number to driver routines, 
 * 	  (at cdevsw[]) as it should
 * 
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/file.h"
#include "../h/inode.h"
#include "../h/sysinfo.h"
#include "../h/conf.h"
#include "../h/ioctl.h"


/*
 * read system call
 */
read()
{
	rdwr(FREAD);
}

/*
 * write system call
 */
write()
{
	rdwr(FWRITE);
}

/*
 * common code for read and write calls:
 * check permissions, set base, count, and offset,
 * and switch out to readi or writei code.
 */
rdwr(mode)
register mode;
{
	register struct file *fp;
	register struct inode *ip;
	register struct a {
		int	fdes;
		char	*cbuf;
		unsigned count;
	} *uap;
	int type, lt;
	long count;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);
	if (fp == NULL)
		return;
	if ((fp->f_flag&mode) == 0) {
		u.u_error = EBADF;
		return;
	}
	u.u_base = (caddr_t)uap->cbuf;
	u.u_count = uap->count;
	u.u_segflg = 0;
	u.u_fmode = fp->f_flag;
	ip = fp->f_inode;
	type = ip->i_mode&IFMT;
	if (type==IFREG || type==IFDIR) {
		lt = (mode&FREAD) ? 1 : 0;
		if(ip->i_locklist &&
		   locked(lt,ip,fp->f_offset,fp->f_offset+u.u_count))
		     return;
		plock(ip);
		if ((u.u_fmode&FAPPEND) && (mode == FWRITE))
			fp->f_offset = ip->i_size;
	} else if (type == IFIFO) {
		plock(ip);
		fp->f_offset = 0;
	}
	u.u_offset = fp->f_offset;
	if (mode == FREAD){
		readi(ip);
	} else {
		/*
		 * Only return EFBIG if a regular file AND the entire transfer
		 * is over the limit.  Otherwise, if part of the transfer is
		 * under the limit, we request partial transfer to prevent low
		 * level routines from giving the error.
		 */
		if (type == IFREG && u.u_count != 0) {
			count = u.u_limit*512 - u.u_offset;  /* max allowed */
			if (count <= 0) {         /* SIGNED comparison */
				prele(ip);
				u.u_error = EFBIG;
				return;
			}
			if (count < u.u_count) {	/* partial transfer? */
				u.u_count = count;
			}
		}
		count = uap->count - u.u_count;		/* disallowed count */
		writei(ip, fp->f_flag & FSYNC);
		u.u_count += count;
	}
	if (type==IFREG || type==IFDIR || type==IFIFO)
		prele(ip);
	fp->f_offset += uap->count-u.u_count;
	u.u_rval1 = uap->count-u.u_count;
	u.u_ioch += u.u_rval1;
	if (mode == FREAD)
		sysinfo.readch += u.u_rval1;
	else
		sysinfo.writech += u.u_rval1;
}

/*
 * rdchk system call
 */
rdchk()
{
       register struct a {
		int fdes;
	} *uap;
	struct file *fp;
	struct inode *ip;
	dev_t dev;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);
	if(fp==NULL)
	    return;
	ip = fp->f_inode;
	u.u_rval1 = 0;

  	if((fp->f_flag&FREAD)==0) {
  	    u.u_error = EBADF;
  	    return;
  	}

	if ((ip->i_mode & IFMT) == IFIFO){
  	    if (ip->i_size > 0 || ip->i_count < 2 || u.u_fmode & FNDELAY)
		u.u_rval1 = 1;
	} else if ((ip->i_mode & IFMT) == IFCHR){
	    dev = (dev_t)ip->i_rdev;
	    (*cdevsw[major(dev)].d_ioctl)(minor(dev), FIORDCHK, 0, 0); /*M000*/
	} else if ((ip->i_mode&IFMT) == IFNAM){
	    u.u_error = EBADF;
	} else {
	    u.u_rval1 = 1;
	}
}
/*
 * open system call
 */
open()
{
	register struct a {
		char	*fname;
		int	mode;
		int	crtmode;
	} *uap;

	uap = (struct a *)u.u_ap;
	copen(uap->mode-FOPEN, uap->crtmode);
}

/*
 * creat system call
 */
creat()
{
	register struct a {
		char	*fname;
		int	fmode;
	} *uap;

	uap = (struct a *)u.u_ap;
	copen(FWRITE|FCREAT|FTRUNC, uap->fmode);
}

/*
 * common code for open and creat.
 * Check permissions, allocate an open file structure,
 * and call the device open routine if any.
 */
copen(mode, arg)
register mode;
{
	register struct inode *ip;
	register struct file *fp;
	int i;

	if ((mode&(FREAD|FWRITE)) == 0) {
		u.u_error = EINVAL;
		return;
	}
	if (mode&FCREAT) {
		ip = namei(uchar, 1);
		if (ip == NULL) {
			if (u.u_error)
				return;
			ip = maknode(arg&07777&(~ISVTX));
			if (ip == NULL)
				return;
			mode &= ~FTRUNC;
		} else {
			if (mode&FEXCL) {
				u.u_error = EEXIST;
				iput(ip);
				return;
			}
			mode &= ~FCREAT;
		}
	} else {
		ip = namei(uchar, 0);
		if (ip == NULL)
			return;
	}
	if (!(mode&FCREAT)) {
		if((ip->i_mode&IFMT)==IFNAM)
			u.u_error = EISNAM;
		if (mode&FREAD)
			access(ip, IREAD);
		if (mode&(FWRITE|FTRUNC)) {
			access(ip, IWRITE);
			if ((ip->i_mode&IFMT) == IFDIR)
				u.u_error = EISDIR;
		}
	}
	if (u.u_error || (fp = falloc()) == NULL) {
		iput(ip);
		return;
	}
	fp->f_flag = mode&FMASK;
	fp->f_inode = ip;
	if (mode&FTRUNC)
		itrunc(ip,0L);
	prele(ip);
	i = u.u_rval1;
	if (setjmp(u.u_qsav)) {	/* catch half-opens */
		if (u.u_error == 0)
			u.u_error = EINTR;
		u.u_ofile[i] = NULL;
		closef(fp);
	} else {
		openi(ip, mode);
		if (u.u_error == 0)
			return;
		u.u_ofile[i] = NULL;
		fp->f_count--;
		iput(ip);
	}
}

/*
 * close system call
 */
close()
{
	register struct file *fp;
	register struct a {
		int	fdes;
	} *uap;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);
	if (fp == NULL)
		return;
	u.u_ofile[uap->fdes] = NULL;
	closef(fp);
}

/*
 * seek system call
 */
seek()
{
	register struct file *fp;
	register struct inode *ip;
	register struct a {
		int	fdes;
		off_t	off;
		int	sbase;
	} *uap;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);
	if (fp == NULL)
		return;
	ip = fp->f_inode;
	if ((ip->i_mode&IFMT)==IFIFO) {
		u.u_error = ESPIPE;
		return;
	}
	if (uap->sbase == 1)
		uap->off += fp->f_offset;
	else if (uap->sbase == 2)
		uap->off += fp->f_inode->i_size;
	else if (uap->sbase != 0) {
		u.u_error = EINVAL;
		return;
	}
	if (uap->off < 0) {
		u.u_error = EINVAL;
		return;
	}
	fp->f_offset = uap->off;
	u.u_roff = uap->off;
}

/*
 * link system call
 */
link()
{
	register struct inode *ip, *xp;
	register struct a {
		char	*target;
		char	*linkname;
	} *uap;

	uap = (struct a *)u.u_ap;
	ip = namei(uchar, 0);
	if (ip == NULL)
		return;
	if (ip->i_nlink >= MAXLINK) {
		u.u_error = EMLINK;
		goto out;
	}
	if ((ip->i_mode&IFMT)==IFDIR && !suser())
		goto out;
	/*
	 * Unlock to avoid possibly hanging the namei.
	 * Sadly, this means races. (Suppose someone
	 * deletes the file in the meantime?)
	 * Nor can it be locked again later
	 * because then there will be deadly
	 * embraces.
	 */
	prele(ip);
	u.u_dirp = (caddr_t)uap->linkname;
	xp = namei(uchar, 1);
	if (xp != NULL) {
		u.u_error = EEXIST;
		iput(xp);
		goto out;
	}
	if (u.u_error)
		goto out;
	if (u.u_pdir->i_dev != ip->i_dev) {
		iput(u.u_pdir);
		u.u_error = EXDEV;
		goto out;
	}
	wdir(ip);
	if (u.u_error==0) {
		ip->i_nlink++;
		ip->i_flag |= ICHG;
	}

out:
	iput(ip);
}

/*
 * mknod system call
 */
mknod()
{
	register struct inode *ip;
	register struct a {
		char	*fname;
		int	fmode;
		int     dev;
	} *uap;

	uap = (struct a *)u.u_ap;
	if ((uap->fmode&IFMT)!=IFIFO && (uap->fmode&IFMT)!=IFNAM && !suser())
		return;
	ip = namei(uchar, 1);
	if (ip != NULL) {
		u.u_error = EEXIST;
		goto out;
	}
	if (u.u_error)
		return;
	ip = maknode(uap->fmode);
	if (ip == NULL)
		return;
	switch(ip->i_mode&IFMT) {
	case IFCHR:
	case IFBLK:
		ip->i_rdev = (dev_t)uap->dev;
		ip->i_flag |= ICHG;
	}

out:
	iput(ip);
}

/*
 * access system call
 */
saccess()
{
	register svuid, svgid;
	register struct inode *ip;
	register struct a {
		char	*fname;
		int	fmode;
	} *uap;

	uap = (struct a *)u.u_ap;
	svuid = u.u_uid;
	svgid = u.u_gid;
	u.u_uid = u.u_ruid;
	u.u_gid = u.u_rgid;
	ip = namei(uchar, 0);
	if (ip != NULL) {
		if (uap->fmode&(IREAD>>6))
			access(ip, IREAD);
		if (uap->fmode&(IWRITE>>6))
			access(ip, IWRITE);
		if (uap->fmode&(IEXEC>>6))
			access(ip, IEXEC);
		iput(ip);
	}
	u.u_uid = svuid;
	u.u_gid = svgid;
}
