/*
 *	@(#) sys/sys3.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */

/***************************************************************************
 * Modification log
 *	2/14/84, saf at sco, M000
 *		 check u.u_error after synchronous update() in smount to
 *		 catch a write protected file system & do the close().
 *	4/20/84, saf, M001,
 *	 	 change update call in mount() to write_sb; that's all we
 *		 really want to do, & above fix could get confused by other
 *		 update errors
 */


/*#define DEBUG	3	/* Compile in debug code */
#define DbugVar	s3_debug
#ifdef DEBUG
#  define s3M	0x0001		/* Debug mount code */
#  define s3UM	0x0002		/* Debug unmount code */
  char	*mnames[] = { "MFREE", "MINUSE", "MINTER" };
#endif /*DEBUG*/
#define DEBUGS (s3M|s3UM)		/* Sections I want to hear from */
#include "../h/dbug.h"


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/mount.h"
#include "../h/ino.h"
#include "../h/buf.h"
#include "../h/filsys.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/user.h"
#include "../h/inode.h"
#include "../h/file.h"
#include "../h/conf.h"
#include "../h/stat.h"
#include "../h/ioctl.h"
#include "../h/var.h"

/*
 * the fstat system call.
 */
fstat()
{
	register struct file *fp;
	register struct a {
		int	fdes;
		struct stat *sb;
	} *uap;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);
	if(fp == NULL)
		return;
	stat1(fp->f_inode, uap->sb);
}

/*
 * the stat system call.
 */
stat()
{
	register struct inode *ip;
	register struct a {
		char	*fname;
		struct stat *sb;
	} *uap;

	uap = (struct a *)u.u_ap;
	ip = namei(uchar, 0);
	if(ip == NULL)
		return;
	stat1(ip, uap->sb);
	iput(ip);
}

/*
 * The basic routine for fstat and stat:
 * get the inode and pass appropriate parts back.
 */
stat1(ip, ub)
register struct inode *ip;
struct stat *ub;
{
	register struct buf *bp;
	struct stat ds;

	if(ip->i_flag&(IACC|IUPD|ICHG))
		iupdat(ip, &time, &time,0);
	/*
	 * first copy from inode table
	 */
	ds.st_dev = ip->i_dev;
	ds.st_ino = ip->i_number;
	ds.st_mode = ip->i_mode;
	ds.st_nlink = ip->i_nlink;
	ds.st_uid = ip->i_uid;
	ds.st_gid = ip->i_gid;
	ds.st_rdev = (dev_t)ip->i_rdev;
	ds.st_size = ip->i_size;
	/*
	 * next the dates in the disk
	 */
	bp = bread(ip->i_dev, itod(ip->i_number));
	copyio(paddr(bp)+itoo(ip->i_number)*sizeof(struct dinode)
		+(long)(&((struct dinode *)0)->di_atime),
	    &ds.st_atime, 3*sizeof(time_t), U_RKD);
	brelse(bp);
	if (copyout((caddr_t)&ds, (caddr_t)ub, sizeof(ds)) < 0)
		u.u_error = EFAULT;
}

/*
 * the dup system call.
 */
dup()
{
	register struct file *fp;
	register i,m;
	register struct a {
		int	fdes;
		int     fdes2;
	} *uap;

	uap = (struct a *)u.u_ap;
	m = uap->fdes & ~077;
	uap->fdes &= 077;
	fp = getf(uap->fdes);
	if(fp == NULL)
		return;
	if ((m&0100) == 0) {
		if ((i = ufalloc(0)) < 0)
			return;
	} else {
		i = uap->fdes2;
		if (i<0 || i>=NOFILE) {
			u.u_error = EBADF;
			return;
		}
		u.u_rval1 = i;
	}
	if (i!=uap->fdes) {
		if (u.u_ofile[i]!=NULL)
			closef(u.u_ofile[i]);
		u.u_ofile[i] = fp;
		fp->f_count++;
	}
}

/*
 * the file control system call.
 */
fcntl()
{
	register struct file *fp;
	register struct a {
		int	fdes;
		int	cmd;
		int	arg;
	} *uap;
	register i;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);
	if (fp == NULL)
		return;
	switch(uap->cmd) {
	case 0:
		i = uap->arg;
		if (i < 0 || i > NOFILE) {
			u.u_error = EINVAL;
			return;
		}
		if ((i = ufalloc(i)) < 0)
			return;
		u.u_ofile[i] = fp;
		fp->f_count++;
		break;

	case 1:
		u.u_rval1 = u.u_pofile[uap->fdes];
		break;

	case 2:
		u.u_pofile[uap->fdes] = uap->arg;
		break;

	case 3:
		u.u_rval1 = fp->f_flag+FOPEN;
		break;

	case 4:
		fp->f_flag &= (FREAD|FWRITE);
		fp->f_flag |= (uap->arg-FOPEN) & ~(FREAD|FWRITE);
		break;

	default:
		u.u_error = EINVAL;
	}
}

/*
 * character special i/o control
 */
ioctl()
{
	register struct file *fp;
	register struct inode *ip;
	register struct a {
		int	fdes;
		int	cmd;
		int	arg;
	} *uap;
	register dev_t dev;

	uap = (struct a *)u.u_ap;
	if ((fp = getf(uap->fdes)) == NULL)
		return;
	u.u_rval1 = 0;
	if(uap->cmd == FIOCLEX){
		u.u_pofile[uap->fdes] |= EXCLOSE;
		return;
	} else if(uap->cmd == FIONCLEX){
		u.u_pofile[uap->fdes] &= ~EXCLOSE;
		return;
	}
	ip = fp->f_inode;
	if ((ip->i_mode&IFMT) != IFCHR) {
		u.u_error = ENOTTY;
		return;
	}
	dev = (dev_t)ip->i_rdev;
	(*cdevsw[major(dev)].d_ioctl)(minor(dev),uap->cmd,uap->arg,fp->f_flag);
}

/*
 * old stty and gtty
 */
stty()
{
	register struct a {
		int	fdes;
		int	arg;
		int	narg;
	} *uap;

	uap = (struct a *)u.u_ap;
	uap->narg = uap->arg;
	uap->arg = TIOCSETP;
	ioctl();
}

gtty()
{
	register struct a {
		int	fdes;
		int	arg;
		int	narg;
	} *uap;

	uap = (struct a *)u.u_ap;
	uap->narg = uap->arg;
	uap->arg = TIOCGETP;
	ioctl();
}

/*
 * the mount system call.
 */
smount()
{
	dev_t dev;
	register struct inode *ip;
	register struct mount *mp;
	struct mount *smp;
	register struct filsys *fp;
	struct buf *bp;
	register struct a {
		char	*fspec;
		char	*freg;
		int	ronly;
	} *uap;

	uap = (struct a *)u.u_ap;
	if(!suser())
		return;
	dev = getmdev();
	if(u.u_error)
		return;
	u.u_dirp = (caddr_t)uap->freg;
	ip = namei(uchar, 0);
	if(ip == NULL)
		return;
	if ((ip->i_mode&IFMT) != IFDIR) {
		u.u_error = ENOTDIR;
		goto out;
	}
	if (ip->i_count != 1)
		goto out;
	if (ip->i_number == ROOTINO)
		goto out;
	smp = NULL;
	dbug3(s3M,2,"%d: mount: looking for dev %d/%d\n", 
				u.u_procp->p_pid,
				major(dev),
				minor(dev));
	for(mp = &mount[0]; mp < (struct mount *)v.ve_mount; mp++) {
		if(mp->m_flags != MFREE) {
			if(dev == mp->m_dev) {
				dbug4(s3M,1,"%d: mount: found %d/%d => %s\n",
						u.u_procp->p_pid,
						major(dev),
						minor(dev),
						mnames[mp->m_flags]);
				goto out;
			}
		} else
		if(smp == NULL)
			smp = mp;
	}
	mp = smp;
	if(mp == NULL)
		goto out;
	mp->m_flags = MINTER;
	dbug4(s3M,1,"%d: Mount %d/%d at mount[%d]\n",
			u.u_procp->p_pid,
			major(dev),
			minor(dev),
			mp - mount);
	mp->m_dev = dev;
	dbug4(s3M,3,"%d: mount: d_open(%d/%d, %s, 1)\n",
			u.u_procp->p_pid,
			major(dev),
			minor(dev),
			uap->ronly? "RONLY" : "R/W");
	(*bdevsw[major(dev)].d_open)(minor(dev), uap->ronly? 0:FWRITE, 1);
	if(u.u_error)
		goto out1;
	bp = bread(dev, SUPERB);
	if(u.u_error) {
		goto out2;
	}
	mp->m_inodp = ip;
	mp->m_bufp = getablk(1);
	fp = (struct filsys *)paddr(mp->m_bufp);
	copyio(paddr(bp), fp, sizeof(struct filsys), U_RKD);
	/* check for v7 file system and adjust structure accordingly */
	if(fp->s_magic!=S_S3MAGIC)
	    cvtv7superb(fp);
	if (fp->s_clean != S_CLEAN) {
		u.u_error = EUCLEAN;    /* wasn't cleanly dismounted! */
		goto out3;
	}
	mp->m_flags = MINUSE;
	fp->s_ilock = 0;
	fp->s_flock = 0;
	fp->s_ninode = 0;
	fp->s_inode[0] = 0;
	fp->s_ronly = uap->ronly & 1;
	brelse(bp);
	if (fp->s_ronly == 0) {
		fp->s_clean = 0;        /* unclean */
		fp->s_fmod = 1;         /* flag modified */
		write_sb(fp, dev);	/* M001, was update()  */
	}

	if (u.u_error) { /* M000 */
		goto out3;
	}

	dbug4(s3M,1,"%d: Mounted %d/%d at mount[%d]\n",
			u.u_procp->p_pid,
			major(dev),
			minor(dev),
			mp - mount);

	ip->i_flag |= IMOUNT;
	prele(ip);
	return;
out3:
	brelse(mp->m_bufp);
out2:
	dbug3(s3M,1,"%d: mount: mdev_close(%d/%d)\n",
			u.u_procp->p_pid,
			major(dev),
			minor(dev));
	mdev_close(dev, !uap->ronly);
	brelse(bp);
out1:
	mp->m_flags = MFREE;
out:
	if (u.u_error == 0)
		u.u_error = EBUSY;
	dbug4(s3M,1,"%d: mount %d/%d error %d\n", 
			u.u_procp->p_pid,
			major(dev),
			minor(dev),
			u.u_error);
	iput(ip);
}


/* convert a v7 superblock to a system3 one */
cvtv7superb(fp)
    struct filsys *fp;
{
    char *dcp, *scp;

	dcp=(char *)&fp->s_clean;
	scp=dcp-2*sizeof(short);    /* sys3 has 2 extra shorts */
	while(scp >= (char *)&fp->s_dinfo[2])
	    *dcp-- = *scp--;
	fp->s_dinfo[2] = 0;
	fp->s_dinfo[3] = 0;
}

/* convert a s3 superblock to a v7 one */
cvts3superb(fp)
    struct filsys *fp;
{
    char *dcp, *scp;

	dcp=(char *)&fp->s_dinfo[2];
	scp=dcp+2*sizeof(short);    /* sys3 has 2 extra shorts */
	while(scp <= (char *)&fp->s_clean)
	    *dcp++ = *scp++;
}

/*
 * the umount system call.
 */
sumount()
{
	register struct inode *ip;
	register struct mount *mp;
	register dev_t dev;
	struct filsys *fp;
	register struct a {
		char	*fspec;
	};

	if(!suser())
		return;
	dev = getmdev();
	if(u.u_error)
		return;
	dbug3(s3UM,2,"%d: umount: looking for dev %d/%d\n", 
				u.u_procp->p_pid,
				major(dev),
				minor(dev));
	for(mp = &mount[0]; mp < (struct mount *)v.ve_mount; mp++)
		if(mp->m_flags == MINUSE && dev == mp->m_dev)
			goto found;
	dbug3(s3UM,1,"%d: umount %d/%d: return (EINVAL)\n",
				u.u_procp->p_pid,
				major(dev),
				minor(dev));
	u.u_error = EINVAL;
	return;

found:
	for(ip = &inode[0]; ip < (struct inode *)v.ve_inode; ip++)
		if(ip->i_number != 0 && dev == ip->i_dev) {
			dbug3(s3UM,1,"%d: umount %d/%d: return (EBUSY)\n",
						u.u_procp->p_pid,
						major(dev),
						minor(dev));
			u.u_error = EBUSY;
			return;
		}

	/* Prevent iput() or update() from allowing another proc
	 *   to umount out from under us while we're asleep by marking
	 *   the mount table entry BUSY.  This will also fix someone that
	 *   is stupid enough (Geoooooorge!) to try to mount it again while
	 *   the system is umounting it.  They'll get EBUSY while we're
	 *   busy writing and stuff.
	 */
	mp->m_flags = MINTER;

	xumount(dev);	/* remove unused sticky files from text table */

	ip = mp->m_inodp;
	ip->i_flag &= ~IMOUNT;
	plock(ip);
	iput(ip);

	update();
	bflush(dev);

	fp = (struct filsys *)paddr(mp->m_bufp);
	if (fp->s_ronly==0) {
		fp->s_clean = S_CLEAN;  /* flag properly dismounted */
		fp->s_fmod = 1;         /* flag modified */
		dbug3(s3UM,2,"%d: umount: write_sb(%d/%d)\n",
					u.u_procp->p_pid,
					major(dev),
					minor(dev));
		write_sb(fp, dev);
	}
	brelse(mp->m_bufp);

	dbug3(s3UM,1,"%d: umount: mdev_close(%d/%d)\n",
			u.u_procp->p_pid,
			major(dev),
			minor(dev));
	mdev_close(dev, 0);

	mp->m_flags = MFREE;
	mp->m_bufp = NULL;
}

/*
 * Common code for mount and umount.
 * Check that the user's argument is a reasonable
 * thing on which to mount, and return the device number if so.
 */
dev_t
getmdev()
{
	dev_t dev;
	register struct inode *ip;

	ip = namei(uchar, 0);
	if(ip == NULL)
		return(NODEV);
	if((ip->i_mode&IFMT) != IFBLK)
		u.u_error = ENOTBLK;
	dev = (dev_t)ip->i_rdev;
	if(major(dev) >= bdevcnt)
		u.u_error = ENXIO;
	iput(ip);
	return(dev);
}

/*
 * More common code for mount and umount.
 *  Check to see if there are any open references to a device;
 *  and if not, call its close routine.
 */
mdev_close(dev, flag)
  register dev_t  dev;
  int	flag;
{
	if (!dev_open(dev, IFBLK)) {
		/* Didn't find a reference, so close it */
		(*bdevsw[major(dev)].d_close)(minor(dev), flag, 1);
		binval(dev);
	}
}
