/*
 *	@(#) sys/sysx.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: sysx.c,v 6.0 87/10/12 17:10:24 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 *
 *  MODIFICATION HISTORY
 *	20 Feb 1984	blf	M000
 *		- Close all mounted filesystems, including the root,
 *		  on a shutdown.  Caution: the open/close protocol
 *		  is probably being violated (eg, more than 1 close).
 *	21 March 1984	blf	M001
 *		- Added some code from PC XENIX 3.0:
 *		  Consistency check (and obscure warning) on the
 *		  supplied superblock during a shutdn/up().
 *
 * $Log:	sysx.c,v $
 * Revision 6.0  87/10/12  17:10:24  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.5  87/05/06  15:40:16  iv
 * 	- Fixed bug in nap() where timeout could go off before we slept
 * 	- Fixed a couple of comments
 * 
 * Revision 5.4  86/10/16  15:44:59  iv
 * 	- Recoded nap() for speed/space
 * 	- Fixed hibernate() bug in nap() which was still there.
 * 
 * Revision 5.3  86/10/01  18:28:21  iv
 * 	- Fixed rounding error in nap() which sometimes caused it to 
 * 	  become hibernate() instead.
 * 
 * Revision 5.2  86/08/29  13:35:02  iv
 * 	- Fixed bug in shutup() where if update was already running
 * 	  when haltsys was run, some of the mounted superblocks might
 * 	  not get written out; esp. if some I/O was already queued to
 * 	  the disk driver [fix was to reset updlock; this can't hurt
 * 	  anything because Hogproc is set]
 * 
 * Revision 5.1  86/07/02  16:55:41  gordon
 * Nap should use time calculation that doesn't overflow on 1000 * time,
 * and use lticks for the milliseconds.
 * 
 * Revision 5.0  85/08/23  10:37:33  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:50:26  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:45:03  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:24:03  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.1  84/05/17  16:05:14  iv
 * 	- Changed references to HZ to Hz for configurability
 * 
 * Revision 0.1  84/05/17  15:53:38  iv
 * 	- Now close all mounted filesystems' device, including root,
 * 	  on a shutdown.  CAUTION:  the open/close protocol is probably
 * 	  being violated  (ie, more than one close).
 * 	- Now perform a consistency check on the supplied superblock
 * 	  and display an obscure warning if anything amiss during shutdn/up()
 * 
 */


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/mount.h"
#include "../h/filsys.h"
#include "../h/buf.h"
#include "../h/dir.h"
#include "../h/a.out.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/var.h"
#include "../h/conf.h"					/* M000 */

/*
 * Everything in this file is a routine implementing XENIX-special.
 *	system call.  In order to save space in the system call
 *	table, and to minimize conflicts with other unix systems,
 *	all custom XENIX calls are done via the cxenix call.
 *	There are five arguments.
 *	The last argument indicates the subfunction number,
 *	the other four arguments depend upon the particular
 *	subfunction.
 */

extern  int     nosys();
extern  int     shutup();
extern  int     locking();      /* file locking */
extern  int     rdchk();        /* read check */
#ifdef LARGECODE
extern	int	xptrace();	/* ptrace extension */
#endif
extern  int     creatsem();
extern  int     opensem();
extern  int     sigsem();
extern  int     waitsem();
extern  int     nbwaitsem();
extern  int     stkgrow();
extern  int     ftime();
extern  int     chsize();
extern  int     nap();
extern  int     sdget();
extern  int     sdfree();
extern  int     sdenter();
extern  int     sdleave();
extern  int     sdgetv();
extern  int     sdwaitv();

int (*cxaddr[])() = {
	shutup,                 /* 0 - shutup OS */
	locking,                /* 1 - record locking */
	creatsem,               /* 2 - create semaphore */
	opensem,                /* 3 - open semaphore */
	sigsem,                 /* 4 - signal semaphore */
	waitsem,                /* 5 - wait semaphore */
	nbwaitsem,              /* 6 - nonblocking waitsem */
	rdchk,                  /* 7 - blocking read check */
	stkgrow,		/* 8 - set stack limit */
#ifdef LARGECODE
	xptrace,		/* 9 - extended ptrace */
#else
	nosys,
#endif
	chsize,                 /* 10- change file size */
	ftime,                  /* 11- V7 ftime call */
	nap,                    /* 12 - sleep for short interval */
	sdget,                  /* 13 - attach to shared data */
	sdfree,                 /* 14 - release shared data */
	sdenter,                /* 15 - enter critical region */
	sdleave,                /* 16 - leave critical region */
	sdgetv,                 /* 17 - get SD version # */
	sdwaitv,                /* 18 - wait for new SD version */
	};
#define CXMAX 18                /* largest subfunction */


/*
 *      cxenix - XENIX custom function dispatcher
 */

cxenix()
{
	register struct a {
		unsigned int arg[5];
	} *uap;
	register subfunc;

	uap = (struct a *)u.u_ap;
	subfunc = (uap->arg[4]>>8) & 0xFF;
	if (subfunc > CXMAX) {
		u.u_error = EINVAL;
		return;
	}
	(*cxaddr[subfunc])();
}


/*      shutup - shut down the system
 *
 *	shutup is called to do an orderly system shutdown.  The procedure is:
 *
 *	1) Set Hogproc so that only this procedure gets service.
 *	2) Set all mounted structure superblocks to 'clean'.
 *	3) Sync so that all unlocked superblocks are written.
 *	           Note that if any are locked, they will not be written
 *	           out, and will thus be unclean.  Of course, we can't
 *	           wait for them to free up, since no one else can run.
 *	4) If specified, write a new user-supplied superblock to the
 *	   root structure.  This is a special hack to allow programs
 *	   like fsck to rebuild the root file structure and be sure of
 *	   overriding the invalid system copy of the superblock.
 *
 *	arg[0] = addr of final root superblock; if none, is 0.
 *	arg[1] = boolean complement of arg[1] if a final root superblock,
 *	         else 0.  This is just 'insurance' against a damaged
 *	         program accidentally trashing the root superblock.
 */

shutup()
{
	register struct mount *mp;
	register struct filsys *fp;
	struct buf *bp;
	register struct a {
		caddr_t rsb;            /* pointer to struct filsys */
		caddr_t invs;           /* inverse of 1st arg. */
	} *uap;
	long int i;
	int bad = 0;						/* M001 */

	if (!suser())                   /* must be super user */
		return;
	Hogproc = u.u_procp;            /* service only for me */

	for(mp = &mount[0]; mp < v.ve_mount; mp++)
		if(mp->m_flags!=MFREE) {
			fp = (struct filsys *)paddr(mp->m_bufp);
			fp->s_clean = S_CLEAN;
			fp->s_fmod = 1;
		}
	updlock = 0;
	update();                       /* sync them all */
	uap = (struct a *)u.u_ap;
	if ((uap->rsb != NULL) && ((int)uap->rsb == ~(int)uap->invs)) {
		/* we're going to rewrite the root superblock */
		bp = getablk(1);
		bp->b_dev = rootdev;
		bp->b_blkno = SUPERB;
							/* M001 begin... */
		if (copyin((caddr_t)uap->rsb, (caddr_t)bp->b_paddr, BSIZE)!=0)
			++bad;		/* Can't get new superblock	*/
		else {
			fp = (struct filsys *)paddr(bp);
			if (fp->s_isize > fp->s_fsize>>2
			 || fp->s_nfree > NICFREE
			 || fp->s_ninode > NICINOD)
				++bad;	/* Inconsistent new superblock	*/
			else
				bswrite(bp);
		}					/* ...end M001 */
	}

	for (mp = &mount[0] ; mp < v.ve_mount ; mp++)	/* M000 begin... */
		if (mp->m_flags != MFREE)
			(*bdevsw[major(mp->m_dev)].d_close)( mp->m_dev, 0, 1 );
	(*bdevsw[major(pipedev)].d_close)( pipedev, 0, 1 );
	(*bdevsw[major(swapdev)].d_close)( swapdev, 0, 1 );
	if (mount[0].m_flags == MFREE || mount[0].m_dev != rootdev)
		(*bdevsw[major(rootdev)].d_close)( rootdev, 0, 1 );

	for (i = 0 ; i < 100000L ; i++)
		;					/* ...end M000 */

	if (!bad) {					/* M001 begin... */
		printf("**  Normal System Shutdown  **\n");

		spl7();                         /* no interrupts */
		haltcpu();                      /* halt it */
	}

	printf( "**  WARNING: Abnormal System Shutdown  **\n");

	spl7();
	for (;;)
		haltcpu();				/* ...end M001 */
}

/*
 *	Grow the stack to include the address passed as an argument.
 *	Does not understand how to shrink the stack.
 *
 *	return value 1 = actual stack limit
 *	return value 2 = limit requested
 */
stkgrow()
{
	struct a {
		char	*stackbot;
	};
	register char *sp;

#ifdef	FIXEDSTACK
	u.u_error = ENOMEM;
#else
	sp = ((struct a *)u.u_ap)->stackbot;
	grow(sp);
	sp = (char *) ((unsigned)0 - mstob(u.u_ssize));
	if (u.u_error == 0 && sp > ((struct a *)u.u_ap)->stackbot)
		u.u_error = ENOMEM;
	if (u.u_error == 0)
		u.u_rval1 = (int) sp;
#endif  FIXEDSTACK
}

/*
 *	Nap for the specified number of milliseconds.
 */
nap()
{
	extern timeout(),wakeup(),sleep();
	extern long msec();
	register struct a {
		long msec;
	} *ap;
	long	start, limit;
	int	togo, s;
	caddr_t	s_chan;

	/* togo get time to nap in ticks */
	ap = (struct a *)u.u_ap;
	togo = (int)(((ap->msec+(1000/Hz)) * Hz) / 1000);
	if ( togo < 0 )
	{
		u.u_error = EINVAL;
		return;
	}

	/* limit gets time in msecs when we're done. */
	start = msec();
	limit = start + ap->msec;
	s_chan = (caddr_t)&(u.u_procp->p_wchan);

	s = spl7();
	timeout(wakeup, s_chan, togo);
	do {
		sleep(s_chan, PSLEP);
	} while ( u.u_error == 0 && (msec() < limit) );
	splx(s);

	/* type time_t is a long */
	u.u_r.r_time = msec() - start;
}

/*
 *	return time as number of millisecs.
 */
long msec()
{
	int f;
	time_t tsv;
	long lsv;
	extern short lticks;
	extern long bootime;

	f = spl7();
	tsv = time - bootime;
	lsv = Hz - lticks;
	splx(f);
	return (tsv * 1000) + (lsv * 1000) / Hz;
}
