/*
 *	Tandy 68000/XENIX serial driver
 *	Hacked from PDP-11 DH driver.
 */

/*
 * $Header: serial.c,v 6.0 87/10/12 17:04:55 iv Exp $
 *
 * $Log:	serial.c,v $
 * Revision 6.0  87/10/12  17:04:55  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.6  87/05/06  11:27:08  iv
 * 	- Got rid of unneeded initialization of serstat[] to 0 (bss is cleared
 * 	  by mch.s for us)
 * 	- Initialize carrier state on startup
 * 	- Optimized some code (used some common sequences, too)
 * 	- Changed close to not disable the chip until any xmits are finished
 * 	- Changed open code to re-assert DTR if off on entry (someone may
 * 	  have hung-up the line and kept it open)
 * 	- Changed carrier wait channel to &tp->t_state (rawq was in use!)
 * 
 * Revision 5.5  86/11/12  10:52:25  iv
 * 	- Changed debug stuff a little so that the table "Bauds[]" is not
 * 	  always compiled in.
 * 
 * Revision 5.4  86/10/16  15:41:28  iv
 * 	- Added serial stati dump debug define & code under LDEBUG
 * 	- Changed acuinit code to not also go through XCLUDE code in open
 * 	- Changed close code to first reset XCLUDE before calling ttclose
 * 
 * Revision 5.3  86/10/01  17:41:08  iv
 * 	- Redid code to not need ACUINIT; stole the bit for DOPARAM
 * 	- Changed serparam() to only run if there is not I/O on the
 * 	  channel.  We were locking up PCI chips when we reprogrammed
 * 	  them with pending output.
 * 	- Added a z80attn(RqSTread) call on TXSTOPs to help the Z80 code.
 * 
 * Revision 5.2  86/08/29  15:15:50  iv
 * 	- Changed debugging code over to new dbug.h macro package
 * 	- Improved serproc() debugging (now can dump everything)
 * 	- Re-arranged some code in open because of CLOCAL/carrier strangeness
 * 	- Removed brkflag[] from serintr(), now have a new status flag BRKRCVD
 * 	- Added z80msgbuf[] support; the 68000 now gets a copy of all z80
 * 	  error/warning messages for dmesg to copy
 * 	- Deleted all old status polling code; the z80 no longer updates the
 * 	  status register in the mailbox and so the polling code just won't
 * 	  work anymore.  Took all conditionals out.
 * 	- Reworked modemint() to tighten code
 * 	- Added new puts() call which putchar()s a string w/o logging it
 * 	  it in msgbuf[].  main() uses it to put the boring half of Legal[] out
 * 
 * Revision 5.1  85/12/04  11:01:33  iv
 * 	- Changed kernel debugger conditional to KDEBUG [compiles in getchar()
 * 	  and extra code for ^X and ^T in serintr()]
 * 	- Added more debugging messages - more complete now
 * 	- Moved some #defines to `serial.h', where they should have been
 * 	- Removed `dbug' macro definitions; now use "../h/dbug.h"
 * 	- Rewrote CLOCAL/DCD handling in modemint() & sermodpoll()
 * 	  to correct some minor bugs in logic
 * 
 * Revision 5.0  85/08/23  10:30:41  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.3  85/08/22  17:18:35  iv
 * 	- Added new flag to seropen(): ACUINIT  (tells seropen to just
 * 	  initialize for I/O)
 * 	- Added code to serclose() to turn off the XCLUDE bit
 * 
 * Revision 4.2  85/03/29  10:24:19  iv
 * 	- Removed the Z80 interrupt ID array code; it was not being used
 * 
 * Revision 4.1  85/03/26  15:19:01  iv
 * 	- Changed seropen() to allow the super-user to open a line
 * 	  with XCLUDE set if no one has it open
 * 	- Moved the modem(d, MODEMON) call to AFTER we have turned off
 * 	  the RXDISAB bit and thus enabled the line
 * 	- Added IXANY to console initialization
 * 	- Added a commented-out line of code to turn OFF the XCLUDE bit
 * 	  to serclose()
 * 
 * Revision 4.0  85/03/13  14:43:57  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.6  85/02/18  14:37:02  iv
 * 	- Turned WASCL bit off so we correctly turn it on (or leave it off)
 * 	  when setting remembered CLOCAL status
 * 
 * Revision 3.5  85/02/18  11:12:59  iv
 * 	- Fixed type problem ("cc warning")
 * 
 * Revision 3.4  85/02/18  10:42:52  iv
 * 	- Replaced status array "acustatus[]" with new array "serstat[]"
 * 
 * Revision 3.3  85/02/12  13:42:02  iv
 * 	- Added the rest of system III support:
 * 	  "stty cread clocal ignpar parmrk inpck ignbrk brkint cstopb cs[5-8]"
 * 	  now work correctly (as opposed to being ignored)
 * 
 * Revision 3.2  85/02/07  16:30:21  iv
 * 	- Added debug output on FRAMERR, PARERR, and OVRRUN
 * 	- Took out test for !(t_oflag&OPOST) when looking to set parity
 * 	  (The way the test was coded, you only got parity if OPOST was off!)
 * 
 * Revision 3.1  85/01/22  15:31:14  iv
 * 	- Moved the call to modem() to turn DTR off to AFTER we close
 * 	  the line so we don't hang waiting for the PCIs to drain
 * 	  their output queues
 * 	- Cleaned up some conditionally compiled stuff to make it
 * 	  easier to read.  No functional change.
 * 
 * Revision 3.0  85/01/17  15:48:24  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.1  85/01/17  14:18:35  iv
 * 	- Added some new debugging code to putchar() to aid printf's
 * 	  that occur during the middle of draining console's t_outq
 * 
 * Revision 2.0  85/01/10  13:57:55  iv
 *	- Checkin to save XENIX 3.0(22)
 * 
 * Revision 1.5  84/12/19  10:32:07  iv
 * 	- Replaced serstart(), serstop() & etc with serproc()
 * 	  and did misc sys III conversion stuff.
 * 
 * Revision 1.4  84/04/16  17:02:40  iv
 * 	- Debugged sys 3 conversion.  Many changes.
 * 	- Got rid of some anachronisms which are not supported and no longer
 * 	  work or even compile under sys 3 [so THATs what happened to mcstart!]
 * 
 * Revision 1.3  84/04/13  13:34:43  iv
 * 	- Changed z80 define "PARODD" to "zPARODD"
 * 	- Changed z80 define "OVERRUN" to "zOVRRUN"
 * 
 * Revision 1.2  84/04/12  17:02:55  iv
 * 	- Converted driver from v7 to sys III per Microsoft 3.0 conversion guide
 * 
 * Revision 1.1  84/04/11  14:03:43  iv
 * 	- Initial revision for TRS-XENIX 3.0 (aka XENIX 3.0)
 */

/*#define DEBUG	8		/* Enable serial debugging messages */
#define DbugVar serDbg
#ifdef DEBUG
#  define srI		0x0001		/* Debug Interrupt code */
#  define srO		0x0002		/* Debug Open code */
#  define srC		0x0004		/* Debug Carrier / modem control */
#  define srIO		0x0008		/* Debug IOctl code */
#  define srPA		0x0010		/* Debug PAram code */
#  define srPR		0x0020		/* Debug PRoc code */
#  define srRW		0x0040		/* Debug Read/Write code */
#endif /*DEBUG*/
#define DEBUGS		(srI)

/*#define KDEBUG   	/* Compile in code for the kernel debugger */
/*#define KPAUSE   	/* Only compile in ^T & getchar code (needs KDEBUG) */
/*#define LDEBUG   	/* Compile in line stati dump code (needs KDEBUG) */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/conf.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/tty.h"
#include "../h/file.h"
#include "../h/z80ser.h"
#include "../h/serial.h"
#include "../h/z80msgbuf.h"
#include "../h/z80attn.h"
#include "../h/dbug.h"


struct tty	ser_tty[NZ80SER];
u_char		serstat[NZ80SER];
static	 char	ser_xon = CSTART;
static	 char	ser_xoff = CSTOP;

#if DEBUG || LDEBUG
char	*Bauds[] = {
	"0", "50", "75", "110", "134", "150", "200", "300", "600",
	"1200", "1800", "2400", "4800", "9600", "EXTA", "EXTB"
};
#endif /*DEBUG*/

int	ttrstrt();
int	wakeup();
int	sertimeout();

#ifdef SM
    /*
     * Smflag values:			see also "sm.c"
     *		0 = shared tty line not open
     *		1 = open as sm driver
     *		2 = open as tty
     */
int Smflag;
#endif SM

/*
 * Open a serial line.
 */
/*ARGSUSED*/
seropen(d, flag)
  int  d;
  int  flag;
{
	register struct tty	*tp;
	register struct zs_iopb	*addr;
	short			x;
	u_char			*sp, sf;
	static	 u_char		once;
	int sermodpoll(), serproc();

	dbug1(dbANY,1,"*** seropen(%d)\n", d);
	if (once == 0) {
		once++;
		for (addr = z80ser, tp = ser_tty, x = 0;
		     x < NZ80SER;
		     ++addr, ++tp, ++x) 
		{
			if (addr->zs_rctl & DCDON)
				tp->t_state |= CARR_ON;
			addr->zs_mode |= RXDISAB;
			z80attn(SERdev, (u_char)x, RqSTwrite);
		}

		/* M009  We are enabling the RX fifo here, and since it
			 has no driver, as such, we need no z80attn() call */
		if ((z80rcv.zr_csr & RXIENAB) == 0)
			z80rcv.zr_csr |= RXIENAB;
	}
		
	if (d >= NZ80SER) {
		dbug(dbANY,1,"seropen: bad chan; return(ENXIO)\n");
		u.u_error = ENXIO;
		return;
	}

	sp = &serstat[d];
	tp = &ser_tty[d];
	addr = &z80ser[d];
	if (*sp & (ACUOPEN | LINEOPEN | ONLINE)) {
		/* Is the ACU itself asking for an open? */
		if (!(*sp & ACURQST)) {
			/* Nope, tell 'em forget it! */
			dbug(dbANY,1,"seropen: ACU busy; return(EBUSY)\n");
			u.u_error = EBUSY;
			return;
		}
		/* Yes, give it to him */
	}
	else if (tp->t_lflag & XCLUDE) {
		if (tp->t_state & ISOPEN || !suser()) {
			dbug2(dbANY,1,"seropen: XCLUDE set/ uid = %d, tp->t_state & ISOPEN = 0x%x\n", u.u_uid, (tp->t_state & ISOPEN));
			u.u_error = EBUSY;
			return;
		}
	}

#ifdef SM
	if (d == SM) {
		if (Smflag == 1) {
			u.u_error = EBUSY;
			return;
		} else
			Smflag = 2;
	}
#endif SM

	if ((tp->t_state&ISOPEN) == 0) {
		if (addr->zs_mode & RXDISAB) {
			addr->zs_mode &= ~RXDISAB;
			z80attn(SERdev, (u_char)d, RqSTwrite);
		}

		tp->t_proc = serproc;
	}
	if (!(addr->zs_wctl & DTRON))
		modem(d, MODEMON);

	/*
	 *  If the ACU dev asked for an open,
	 *  he just wanted the line setup for I/O
	 */
	sf = *sp;
	*sp &= ~ACURQST;
	if ((sf & ACURQST) && !(sf & (LINEOPEN|ONLINE))) {
		dbug(srO,2,"seropen: ACUINIT set, calling ttinit()\n");
		ttinit(tp);
		return;
	}

	x = SPL();
	if (!(tp->t_state & ISOPEN)) {
		/*
		 *  Special hack because console has no "del" key,
		 *  and the "break" key sends a control-C; and it is
		 *  nice to enable cannonical processing CRT-style
		 */
		dbug(srO,3,"seropen: Initializing channel...\n");
		ttinit(tp);
		if (d == CONSOLE) {
			tp->t_state |= CARR_ON;
			tp->t_cc[VINTR] = '\03';
			tp->t_oflag = OPOST|ONLCR;
			tp->t_iflag = ICRNL|ISTRIP|IXON|IXANY;
			tp->t_lflag = ECHO|ECHOE|ICANON|ISIG;
		}

		serparam(d);	/* Initializes carrier if not set */
	}

	if (!(flag & FNDELAY)) {
		dbug(srC,2,"seropen: sleeping on carrier...\n");
		while ((tp->t_state&CARR_ON)==0) {
			tp->t_state |= WOPEN;
			sleep((caddr_t)&tp->t_state, TTIPRI);
		}
		dbug(srC,2,"seropen: Awake, got carrier...\n");
	}

	(*linesw[tp->t_line].l_open)(tp);
	splx(x);
}

/*
 * Close a serial line.
 */
/*ARGSUSED*/
serclose(d, flag)
  register int	 d;
  int  flag;
{
	register struct tty *tp;
	register struct zs_iopb *addr;
	u_char			*sp;
	int			s;

	tp = &ser_tty[d];
	addr = &z80ser[d];
	sp = &serstat[d];

	tp->t_lflag &= ~XCLUDE;		/* ttclose should be doing this */

	dbug1(dbANY,1,"*** serclose(%d)\n", d);
	(*linesw[tp->t_line].l_close)(tp);
	
	if (tp->t_cflag & HUPCL)
		modem(d, MODEMOFF);		/* close modem control */

	s = SPL();
	if (addr->zs_csr & TXGO)
		*sp |= DOPARAM;
	else {
		addr->zs_mode |= RXDISAB;
		z80attn(SERdev, (u_char)d, RqSTwrite);
	}
	splx(s);

#ifdef SM
	if (d == SM)
		Smflag = 0;		/* shared line not open any more */
#endif SM
	
	dbug(dbANY,1,"*** serclose DONE\n");
}

/*
 * Read from a serial line.
 */
serread(d)
{
	struct tty *tp;
	
	tp = &ser_tty[d];
	dbug1(srRW,8,"*** serread(%d)\n", d);
	(*linesw[tp->t_line].l_read)(tp);
}

/*
 * write on a serial line
 */
serwrite(d)
{
	struct tty *tp;

	tp = &ser_tty[d];
	dbug1(srRW,8,"*** serwrite(%d)\n", d);
	(*linesw[tp->t_line].l_write)(tp);
}

/*
 * serial interrupt.
 */
serintr()
{
	/*
	 * Receiver handling.
	 * Get character from input FIFO.
	 */
	{
		register struct tty	*tp;
		register struct zs_fifo	*entry;
		register u_int	c;
		register u_char	s;
		register u_char	*sp;
		u_int		line;

		dbug(srI,9,"\n*** serRXintr\n");
		while (z80rcv.zr_inptr != z80rcv.zr_outptr) {
			entry = &z80fifo[z80rcv.zr_outptr];
			s = entry->zf_s;
			c = entry->zf_ch;
			entry->zf_ch = 0;	/* Mess up /dev/kmem spy pgms */
			z80rcv.zr_outptr = (z80rcv.zr_outptr + 1) % FIFOSIZE;
			line = (s & LINENO);

			if (line != CONSOLE)
				dbug1(srI,8," I[s/c=%x]", (s<<8)|c);
#ifdef KDEBUG
			if (line == CONSOLE)
#  ifndef LDEBUG
#    ifndef KPAUSE
				if (c == ('X' & 0x1f)) {	/* ^X */
					printf("^X\n");
					debug(0);
					continue;
				} else
#    endif /*KPAUSE*/
				if (c == ('T' & 0x1f)) {	/* ^T */
					printf("^T\n");
					getchar();
					continue;
				}
#  else /*LDEBUG*/
				if (c == ('T' & 0x1f)) {	/* ^T */
					printf("^T\nShow tty=..\b\b");
					do
						c = getchar();
					while (c < '0' || c > '9');
					putchar(c);
					line = c - '0';
					do {
						c = getchar();
						if (c == '\n')
							break;
					} while (c < '0' || c > '9');
					if (c != '\n') {
						putchar(c);
						line = (line * 10) + (c - '0');
					}
					putchar('\n');
					
					if (line < NZ80SER)
						dumpstati(line);
					else
						printf("Bad line %d\n", line);
					continue;
				}
#  endif /*LDEBUG*/
#endif /*KDEBUG*/
#ifdef SM
			if (Smflag == 1) {
				if (s & zOVRRUN)
					smovint();
				else if (line == SM) {
					if (s & (PARERR|FRAMERR))
						smerint();
					else
						smrxint((int) c);
					continue;
				}
			}
#endif SM
					
			if (line > NZ80SER)
				continue;

			tp = &ser_tty[line];
			sp = &serstat[line];

			if (s & MCTLNTFY) {
				modemint((u_short)line, (u_short)c);
				continue;
			}
			if ((tp->t_state&ISOPEN)==0 || !(tp->t_cflag&CREAD)) {
				if (!(*sp & ACUOPEN))
					continue;
			}
			if (s&zOVRRUN) {
				dbug1(srI,1,"serintr:  OVERRUN on chan %d\n", line);
				c |= OVERRUN;
			}
			if (s&PARERR) {
				dbug1(srI,1,"serintr:  PARERR on chan %d\n", line);
				c |= PERROR;
			}
			if (s&FRAMERR) {			/* break */
				dbug1(srI,1,"serintr:  FRAMERR on chan %d\n", line);
				*sp |= BRKRCVD;
				c = FRERROR;	/* zero received garbage char */
			}
			else if (*sp & BRKRCVD) {
				*sp &= ~BRKRCVD;
				if (c == 0)
					continue;
			}
			(*linesw[tp->t_line].l_input)(tp, c, 0);
		}
		z80rcv.zr_csr = RXIENAB;
	}

	/*
	 * Transmitter handling.
	 * Restart each line which used to be active but has
	 * terminated transmission since the last interrupt.
	 */
	{
		register struct tty *tp;
		register struct zs_iopb *addr;
		register i;

		dbug(srI,9,"\n*** serTXintr\n");
		tp = &ser_tty[NZ80SER - 1];
		addr = &z80ser[NZ80SER - 1];
		for (i = NZ80SER; i; --i, --tp, --addr) {
#ifdef SM
			if (Smflag == 1 && i - 1 == SM) {
				if ((addr->zs_csr & (TXGO|TXINT)) == TXINT)
					smtxint();
			} else
#endif SM
			if ((tp->t_state&BUSY) && (addr->zs_csr&TXGO) == 0) {
				ndflush(&q3, (int)addr->zs_count);
			        tp->t_state &= ~BUSY;
				serproc(tp, T_OUTPUT);
			}
		}
	}

	/*
	 *  If the z80 has put out any messages since we last looked,
	 *  update the master buffer that `dmesg' looks at
	 */
	{
		register struct zm_iopb *zmp;

		zmp = &z80msgbuf;
		if (zmp->zm_back != zmp->zm_front) 
			putchar('\0');
	}

	dbug(srI,9,"\n*** Exit serintr\n");
}

#ifdef LDEBUG
dumpstati(line)
{
	register struct tty	*tp;
	register struct zs_iopb	*iop;
	register u_char		*sp;

	tp = &ser_tty[line];
	iop = &z80ser[line];
	sp = &serstat[line];

	printf("    z80ser[%d]:\n", line);
	printf("\taddr = 0x%X\n", iop->zs_addr);
	printf("\tcsr = %x (%s|%s|%s)\n",
			iop->zs_csr,
			(iop->zs_csr & TXGO)? "TXGO" : "",
			(iop->zs_csr & TXIENAB)? "TXIENAB" : "",
			(iop->zs_csr & TXINT)? "TXINT" : "");
	printf("\tcount = %d\n", iop->zs_count);
	printf("\tbits = %d  /  baud = %s\n", 
			(iop->zs_baud >> 4) + 5,
			Bauds[iop->zs_baud & 0x0F]);
	printf("\tmode = %x (%s|%s)\n",
			iop->zs_mode,
			(iop->zs_mode & RXDISAB)? "RXDISAB" : "",
			(iop->zs_mode & TXSTOP)? "TXSTOP" : "");
	printf("\twctl = %x (%s|%s|%s)\n",
			iop->zs_wctl,
			(iop->zs_wctl & RTSON)? "RTSON" : "RTSOFF",
			(iop->zs_wctl & DTRON)? "DTRON" : "DTROFF",
			(iop->zs_wctl & MINTENAB)? "MINTENAB" : "MINTDSAB");
	printf("\trctl = %x (%s|%s|%s)\n\n",
			iop->zs_rctl,
			(iop->zs_rctl & DCDON)? "DCDON" : "DCDOFF",
			(iop->zs_rctl & CTSON)? "CTSON" : "CTSOFF",
			(iop->zs_rctl & DSRON)? "DSRON" : "DSROFF");

	printf("    serstat[%d]:\n", line);
	printf("\t = %x (%s|%s|%s|%s|%s|%s|%s|%s)\n",
			*sp,
			(*sp & ACUOPEN)? "ACUOPEN" : "",
			(*sp & ONLINE)? "ONLINE" : "",
			(*sp & ACURQST)? "ACURQST" : "",
			(*sp & LINEOPEN)? "LINEOPEN" : "",
			(*sp & WASCL)? "WASCL" : "",
			(*sp & INITTED)? "INITTED" : "",
			(*sp & DOPARAM)? "DOPARAM" : "",
			(*sp & BRKRCVD)? "BRKRCVD" : "");

	printf("    ser_tty[%d]:\n", line);
	printf("\tiflag = %x (%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s)\n",
			tp->t_iflag,
			(tp->t_iflag & IGNBRK)? "IGNBRK" : "",
			(tp->t_iflag & BRKINT)? "BRKINT" : "",
			(tp->t_iflag & IGNPAR)? "IGNPAR" : "",
			(tp->t_iflag & PARMRK)? "PARMRK" : "",
			(tp->t_iflag & INPCK)? "INPCK" : "",
			(tp->t_iflag & ISTRIP)? "ISTRIP" : "",
			(tp->t_iflag & INLCR)? "INLCR" : "",
			(tp->t_iflag & IGNCR)? "IGNCR" : "",
			(tp->t_iflag & ICRNL)? "ICRNL" : "",
			(tp->t_iflag & IUCLC)? "IUCLC" : "",
			(tp->t_iflag & IXON)? "IXON" : "",
			(tp->t_iflag & IXANY)? "IXANY" : "",
			(tp->t_iflag & IXOFF)? "IXOFF" : "");
	printf("\toflag = %x (%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s)\n",
			tp->t_oflag,
			(tp->t_oflag & OPOST)? "OPOST" : "",
			(tp->t_oflag & OLCUC)? "OLCUC" : "",
			(tp->t_oflag & ONLCR)? "ONLCR" : "",
			(tp->t_oflag & OCRNL)? "OCRNL" : "",
			(tp->t_oflag & ONOCR)? "ONOCR" : "",
			(tp->t_oflag & ONLRET)? "ONLRET" : "",
			(tp->t_oflag & OFILL)? "OFILL" : "",
			(tp->t_oflag & OFDEL)? "OFDEL" : "",
			(tp->t_oflag & NLDLY)? "NL1" : "NL0",
			(tp->t_oflag & CRDLY) == CR3? "CR3" 
				: (tp->t_oflag & CRDLY) == CR2? "CR2"
				: (tp->t_oflag & CRDLY) == CR1? "CR1" : "CR0",
			(tp->t_oflag & TABDLY) == TAB3? "TAB3"
				: (tp->t_oflag & TABDLY) == TAB2? "TAB2"
				: (tp->t_oflag & TABDLY) == TAB1? "TAB1":"TAB0",
			(tp->t_oflag & BSDLY)? "BS1" : "BS0",
			(tp->t_oflag & VTDLY)? "VT1" : "VT0",
			(tp->t_oflag & FFDLY)? "FF1" : "FF0");
	printf("\tcflag = %x (%s %d bits / [%s|%s|%s|%s|%s|%s])\n",
			tp->t_cflag,
			Bauds[tp->t_cflag & CBAUD],
			((tp->t_cflag & CSIZE) >> 4) + 5,
			(tp->t_cflag & CSTOPB)? "CSTOPB" : "",
			(tp->t_cflag & CREAD)? "CREAD" : "",
			(tp->t_cflag & PARENB)? "PARENB" : "",
			(tp->t_cflag & PARODD)? "PARODD" : "",
			(tp->t_cflag & HUPCL)? "HUPCL" : "",
			(tp->t_cflag & CLOCAL)? "CLOCAL" : "");
	printf("\tlflag = %x (%s|%s|%s|%s|%s|%s|%s|%s|%s)\n",
			tp->t_lflag,
			(tp->t_lflag & ISIG)? "ISIG" : "",
			(tp->t_lflag & ICANON)? "ICANON" : "",
			(tp->t_lflag & XCASE)? "XCASE" : "",
			(tp->t_lflag & ECHO)? "ECHO" : "",
			(tp->t_lflag & ECHOE)? "ECHOE" : "",
			(tp->t_lflag & ECHOK)? "ECHOK" : "",
			(tp->t_lflag & ECHONL)? "ECHONL" : "",
			(tp->t_lflag & NOFLSH)? "NOFLSH" : "",
			(tp->t_lflag & XCLUDE)? "XCLUDE" : "");
	printf("\tstate = %x (%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s)\n",
			tp->t_state,
			(tp->t_state & TIMEOUT)? "TIMEOUT" : "",
			(tp->t_state & WOPEN)? "WOPEN" : "",
			(tp->t_state & ISOPEN)? "ISOPEN" : "",
			(tp->t_state & TBLOCK)? "TBLOCK" : "",
			(tp->t_state & CARR_ON)? "CARR_ON" : "",
			(tp->t_state & BUSY)? "BUSY" : "",
			(tp->t_state & OASLP)? "OASLP" : "",
			(tp->t_state & IASLP)? "IASLP" : "",
			(tp->t_state & TTSTOP)? "TTSTOP" : "",
			(tp->t_state & EXTPROC)? "EXTPROC" : "",
			(tp->t_state & TACT)? "TACT" : "",
			(tp->t_state & ESC)? "ESC" : "",
			(tp->t_state & RTO)? "RTO" : "",
			(tp->t_state & TTIOW)? "TTIOW" : "",
			(tp->t_state & TTXON)? "TTXON" : "",
			(tp->t_state & TTXOFF)? "TTXOFF" : "");
	printf("\tpgrp = %d\n", tp->t_pgrp);
	printf("\tline = %d\n", tp->t_line);
	printf("\trawq has %d chars\n", tp->t_rawq.c_cc);
	printf("\tcanq has %d chars\n", tp->t_canq.c_cc);
	printf("\toutq has %d chars\n", tp->t_outq.c_cc);
}
#endif /*LDEBUG*/

/*
 * ioctl for serial line
 */
/*ARGSUSED*/
serioctl(d, cmd, addr, flag)
{
	register struct tty *tp;
	register status;

	tp = &ser_tty[d];
	dbug2(srIO,1,"*** serioctl(%d, 0x%x)\n", d, cmd);
	status = ttiocom(tp, cmd, addr, flag);
	if (status == 1) 
		serparam(d);
	else if (status != 0) {
		dbug(srIO,1,"serioctl: bad sts; return (ENOTTY)\n");
		u.u_error = ENOTTY;
	}
}

/*
 * Set parameters from open or stty into the serial hardware
 * registers.
 */
serparam(d)
  register d;
{
	register struct tty	*tp;
	register struct zs_iopb	*addr;
	register int	s, wordlen, baud;
	register u_char	*sp, thisClocal, lastClocal;
	u_char		chan;

	dbug1(srPA,1,"*** serparam(%d):", d);
	tp = &ser_tty[d];
	addr = &z80ser[d];
	sp = &serstat[d];
	chan = d;

	s = SPL();

	/* Is the chip busy? */
	if (addr->zs_csr & TXGO) {
		*sp |= DOPARAM;
		goto out;
	}

	/*
	 * Hang up line?
	 */
	baud = (tp->t_cflag & CBAUD);
	if (baud == B0) {
		dbug(srPA,2,"  HUP request\n");
		tp->t_cflag |= HUPCL;
		modem(d, MODEMOFF);
		goto out;
	}
	else if (!(addr->zs_wctl & DTRON))
		modem(d, MODEMON);

	thisClocal = ((tp->t_cflag & CLOCAL) == CLOCAL)? WASCL : 0;
	lastClocal = (*sp & WASCL);
	*sp &= ~WASCL;

	if (!(*sp & INITTED)) {
		*sp |= INITTED;
		dbug(srPA,4,"  INIT");
	}
	else {
		if (thisClocal && !lastClocal) {
			dbug(srPA,4,"  CLOCAL>on");
			tp->t_state |= CARR_ON;		/* Fake Carrier ON */
			wakeup((caddr_t)&tp->t_state);
		}
		else if (!thisClocal && lastClocal) {
			dbug(srPA,4,"  CLOCAL>off");
			modemint((u_short)d, (u_short)-1); /* Get REAL status */
		}
	}
	*sp |= thisClocal;		/* save current CLOCAL status */

	d = addr->zs_mode;
	d &= ~(RXDISAB|PARENAB|zPARODD|STOP2);

	switch (tp->t_cflag & CSIZE) {
	    case CS5:
		wordlen = WORD5;
		break;
	    case CS6:
		wordlen = WORD6;
		break;
	    case CS7:
		wordlen = WORD7;
		break;
	    default:
	    case CS8:
		wordlen = WORD8;
		break;
	}

	dbug2(srPA,4,"  %d bit @ %s baud", (wordlen>>4) + 5, Bauds[baud]);

	if (tp->t_cflag & CSTOPB)
		d |= STOP2;
	dbug1(srPA,4,"  %s stop", (d & STOP2)? "2" : "1");

	/* no support for 134.5 baud */
	if (tp->t_cflag & PARENB) {
		d |= PARENAB;
		if (tp->t_cflag & PARODD)
			d |= zPARODD;
	}
	dbug1(srPA,4,"  %s parity",  (d & PARENAB)? 
				  (d & zPARODD)? "Odd" : "Even" 
				: "No");
	addr->zs_baud = wordlen | baud;
	addr->zs_mode = d;
	z80attn(SERdev, chan, RqSTwrite);
   out:
	splx(s);

	dbug(srPA,1,"  done\n");
}

serproc(tp, cmd)
  register struct tty *tp;
{
	register struct	zs_iopb	*addr;
	register u_char		*sp;
	register nch;
	register int	s, d;
	static	 short	breaking = FALSE;

	d = tp - ser_tty;
	addr = &z80ser[d];
	sp = &serstat[d];

	switch (cmd) {
	    case T_BREAK:
		dbug(srPR,2,"serproc: T_BREAK requested\n");
		addr->zs_mode |= TXBREAK;
		z80attn(SERdev, (u_char)d, RqSTwrite);

		tp->t_state |= TIMEOUT;
		breaking = TRUE;
		timeout(ttrstrt, (caddr_t)tp, Hz / 4);
		break;

	    case T_SUSPEND:
		s = SPL();
		tp->t_state |= TTSTOP;
		if (tp->t_state & BUSY) {
			dbug(srPR,2,"\nserproc: T_SUSPEND requested\n");
			addr->zs_mode |= TXSTOP;
			/* RqSTread is interpreted to mean TXSTOP change */
			z80attn(SERdev, (u_char)d, RqSTread);
		}
		splx(s);
		break;

	    case T_WFLUSH:
		dbug(srPR,2,"serproc: T_WFLUSH requested\n");
	    case T_RESUME:
		dbug(srPR,2,"serproc: T_RESUME requested\n");
		tp->t_state &= ~TTSTOP;
		goto start;

	    case T_TIME:
		dbug(srPR,2,"serproc: T_TIME requested\n");
		tp->t_state &= ~TIMEOUT;
		if (breaking) {
			breaking = FALSE;
			addr->zs_mode &= ~TXBREAK;
			z80attn(SERdev, (u_char)d, RqSTwrite);
		}
		/* Fall thru intentional, we wish to restart transmission... */

	    case T_OUTPUT:
		/*
		 *  Start (or restart) transmission on a line
		 */
		dbug(srPR,9,"\nserproc: T_OUTPUT requested\n");

	     start:
		/*
		 * If it's currently active, or delaying,
		 * no need to do anything, except maybe clear stop bit.
		 */
		s = SPL();
		if (tp->t_state&(TIMEOUT|BUSY|TTSTOP))
			goto out;

		if (addr->zs_csr & TXGO) {
			/* When this happens, it usually means that the kernel
			   has recently called printf thus using the iopb, and
			   so the Z80 is still busy outputting the last char.
			   The reason we just don't go to sleep waiting for the
			   iopb to become free is that we might have been called
			   at a hi pri or from an interrupt routine, and thus
			   we cannot call sleep.  *sigh*  */
			dbug(srPR,1,"serproc @ T_OUTPUT: iopb BUSY\n");
			timeout(ttrstrt, (caddr_t)tp, Hz / 8);
			goto out;
		}

		if (*sp & DOPARAM) {
			/* serparam() needs to be called first,
			 *   or the line is closing */
			*sp &= ~DOPARAM;
			if (tp->t_state & ISOPEN)
				serparam(d);
			else {
				addr->zs_mode |= RXDISAB;
				z80attn(SERdev, (u_char)d, RqSTwrite);
			}
		}

		/*
		 *  If we have received a flow control request,
		 *  then honor it by sending out ^S/^Q as apropriate
		 */
		if (tp->t_state & (TTXON|TTXOFF)) {
			dbug1(srPR,2,"\nserproc(T_OUTPUT): TTX%s requested\n",
				(tp->t_state & TTXOFF)? "ON" : "OFF");

			addr->zs_addr =
				(tp->t_state & TTXOFF)? &ser_xoff : &ser_xon;
			tp->t_state &= ~(TTXON|TTXOFF);

			addr->zs_count = 1;
			addr->zs_csr = TXIENAB | TXGO;
			tp->t_state |= BUSY;
			z80attn(SERdev, (u_char)d, RqStart);
			goto out;
		}

		/*
		 * If the writer was sleeping on output overflow,
		 * wake him when low tide is reached.
		 */
		if (tp->t_state & OASLP
			  && tp->t_outq.c_cc <= ttlowat[tp->t_cflag & CBAUD]) {
			tp->t_state &= ~OASLP;
			wakeup((caddr_t)&tp->t_outq);
		}

		if (tp->t_state&TTIOW && tp->t_outq.c_cc==0) {
			tp->t_state &= ~TTIOW;
			wakeup((caddr_t)&tp->t_oflag);
		}

		if (tp->t_outq.c_cc == 0)
			goto out;
		/*
		 * Find number of characters to transfer.
		 */
		if (tp->t_oflag&OPOST) {
			nch = ndqb(&tp->t_outq, 0x80);
			if (nch == 0) {
				register struct cblock	*cbp;
				register u_char	next;

				cbp = tp->t_outq.c_cf;
				next = cbp->c_data[cbp->c_first];
				if (next == 0x80) {
				    next = getc(&tp->t_outq);
				    if (tp->t_outq.c_cc > 0) {
					cbp = tp->t_outq.c_cf;
					next = cbp->c_data[cbp->c_first];
					if (next > 0x80) {
					    next = getc(&tp->t_outq);
					    tp->t_state |= TIMEOUT;
					    timeout(ttrstrt, tp, (next&0x7f)+6);
					    goto out;
					} else
					    nch = 1;
				    } else
					nch = 0;
				} else
				    nch = 1;
			}
		} else
			nch = ndqb(&tp->t_outq, 0);
		
		/*
		 * If any characters were set up, start transmission;
		 */
		if (nch) {
			register struct cblock	*cbp;

			cbp = tp->t_outq.c_cf;
			addr->zs_addr = (caddr_t)&cbp->c_data[cbp->c_first];
			addr->zs_count = nch;
			addr->zs_mode &= ~TXSTOP;
			addr->zs_csr = TXIENAB | TXGO;
			tp->t_state |= BUSY;
			z80attn(SERdev, (u_char)d, RqStart);
		}
	     out:
		splx(s);
		break;

	    case T_BLOCK:
		dbug(srPR,2,"serproc: T_BLOCK requested\n");
		tp->t_state = (tp->t_state | (TBLOCK|TTXOFF)) & ~TTXON;
		goto start;

	    case T_RFLUSH:
		dbug(srPR,2,"serproc: T_RFLUSH requested\n");
		if (!(tp->t_state & TBLOCK))
			break;
		/* Fall thru intentional, we wish to restart transmission... */

	    case T_UNBLOCK:
		dbug(srPR,2,"serproc: T_UNBLOCK requested\n");
		tp->t_state = (tp->t_state | TTXON) & ~(TTXOFF|TBLOCK);
		goto start;

	} /*end switch*/
}

/* Begin M007 .... */
#define MSERON  DTRON|RTSON|MINTENAB
#define MSEROFF MINTENAB
modem(d, flag)
  register int d;
{
	register struct zs_iopb *addr = &z80ser[d];
	register struct tty *tp = &ser_tty[d];

	dbug2(srC,1,"*** modem(%d, MODEM%s)\n", d, (flag==MODEMOFF)?"OFF":"ON");
	if (flag != MODEMOFF)
		addr->zs_wctl = MSERON;
	else
		addr->zs_wctl = MSEROFF;

	z80attn(SERdev, (u_char)d, RqSTwrite);
}

modemint(line, status)
  u_short	 line;
  register u_short	status;
{
	register struct tty	*tp;
	register u_char	*cp;
	static u_char	last_status[NZ80SER];

	tp = &ser_tty[line];
	cp = &last_status[line];

	if ((short)status == -1)	/* serparam() wants current status */
		status = *cp;

	dbug4(srC,1,"\n*** modemint(%d) = <DCD%s CTS%s DSR%s>\n",
			line,
			(status & DCDON)? "ON, " : "OFF,",
			(status & CTSON)? "ON, " : "OFF,",
			(status & DSRON)? "ON" : "OFF");

	if (status & DCDON) {
		tp->t_state |= CARR_ON;
		wakeup((caddr_t)&tp->t_state);
	}
	else {
		if (!(tp->t_cflag & CLOCAL)) {
			if (tp->t_state & (ISOPEN|WOPEN)) {
				signal(tp->t_pgrp, SIGHUP);
				ttyflush(tp, (FREAD|FWRITE));
			}
			tp->t_state &= ~CARR_ON;
		}
	}

	*cp = status;
}
/* ... End M007 */


/*
 *  This is a special version of putchar that is for the kernel only.
 *  It outputs a character to the console (at spl7 priority [!]).
 *  It also saves the character outputted into the system msgbuf for
 *    later examination by the cron daemon dmesg.
 *  This is not to be used for chit-chat.  Everything STOPS when it is called.
 *  Any kernel routine that uses printf is using this routine.
 */
putchar(c)
  char c;
{
	register struct zm_iopb *zmp;
	register int	zc;

	zmp = &z80msgbuf;
	/* Copy any Z80 messages to our buffer */
	while (zmp->zm_back != zmp->zm_front) {
		zc = zmp->zm_z80[zmp->zm_back++];
		zmp->zm_back &= ZMQSIZE-1;	/* Cheap %= ZMQSIZE */
		if (zc != '\0' && zc != '\r') {
			zmp->zm_68k[zmp->zm_6front++] = zc;
			if (zmp->zm_6front >= MSGBUFS)
				zmp->zm_6front = 0;
		}
	}

	if (c != '\0' && c != '\r' && c != 0177) {
		/* Finally, store OUR character */
		zmp->zm_68k[zmp->zm_6front++] = c;
		zmp->zm_6front %= MSGBUFS;
	}

	if (c == '\0')
		return;			/* Hook for z80 queue copy-in */

	_putchar(c);
}
_putchar(c)
  char c;
{
	register int x;
	IFdbug( register int old_cnt;	)
	register struct zs_iopb *addr;
	static char ch;

	if (c == '\n')
		_putchar('\r');

	addr = &z80ser[CONSOLE];

	x = spl7();
	while (addr->zs_csr & TXGO)
		;

	IFdbug( old_cnt = addr->zs_count;	)

	ch = c;
	addr->zs_addr = &ch;
	addr->zs_count = 1;
	addr->zs_mode &= ~TXSTOP;
	addr->zs_csr = (addr->zs_csr | TXGO) & ~TXINT;
	z80attn(SERdev, (u_char)CONSOLE, RqStart);

	IFdbug( while(addr->zs_csr & TXGO); addr->zs_count = old_cnt;	)
	splx(x);
}

puts(s)
  register char	*s;
{
	while (*s)
		_putchar(*s++);
}

#ifdef KDEBUG
getchar()
{
	register u_char	c;
	register int	l;
	register int	x;

	x = spl7();
	z80rcv.zr_csr &= ~RXIENAB;
	for ( ; ; ) {
		while (z80rcv.zr_inptr == z80rcv.zr_outptr)
			;
		c = z80fifo[z80rcv.zr_outptr].zf_ch;
		l = z80fifo[z80rcv.zr_outptr].zf_s;
		z80rcv.zr_outptr = (z80rcv.zr_outptr + 1) % FIFOSIZE;
		if ((l & LINENO) == CONSOLE) {
			c &= 0x7f;
			if (c == '\r') 
				c = '\n';
			z80rcv.zr_csr = RXIENAB;
			splx(x);
			return (c);
		}
	}
}
#endif /*KDEBUG*/
